<?php

/**
 * SHD CABLE API
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
abstract class cable_api_model_abstract implements isys_api_model_interface
{

    /**
     * Name of this module.
     *
     * @var string
     */
    protected $module_name;

    /**
     * Name of this category.
     *
     * @var string
     */
    protected $category_name;
    
    /**
     * Modules DAO.
     *
     * @var isys_module_dao
     */
    protected $dao;

    /**
     * Indicating whether authentication system should be used during api calls.
     *
     * @var boolean
     */
    protected $useAuth = false;

    /**
     * Constructor.
     */
    function __construct()
    {
        global $g_config;

        // Determine status of authentication system
        $this->useAuth = $g_config['use_auth'];
    }

    /**
     * Read objects.
     *
     * @param array $p_params
     *
     * $p_param structure:
     * array (
     *     'filter' => array|null,
     * )
     *
     * @return array
     * @throws isys_exception_auth
     * @throws isys_exception_general
     */
    public function read($p_params)
    {
        $this->check_permission_of_category(isys_auth::VIEW);

        try {
            if (isset($p_params["filter"])) {
                // get filtered objects
                return $this->dao->filter_objects($p_params["filter"]);
            }
            if (isset($p_params["id"])) {
                // get object
                return $this->dao->get_object('id', $p_params["id"]);
            }
            return $this->dao->get_data();
        } catch (Exception $ex) {
            return $this->return_error($this->module_name, 'read', $ex->getMessage());
        }
    }
    
    /**
     * Create object.
     *
     * @param  array $p_params
     * @return array
     */
    abstract public function create($p_params);

    /**
     * Create a new object.
     *
     * @param array $p_params
     * @param array $required_parameters
     * @param array $optional_params
     * @return array
     * @throws isys_exception_auth
     * @throws isys_exception_general
     */
    protected function _create($p_params, $required_parameters, $optional_params = [])
    {
        $this->check_permission_of_category(isys_auth::CREATE, $p_params['obj_id'] ?: null);

        try {
            $this->check_required_parameters($required_parameters, $p_params);
        } catch (isys_exception_validation $ex) {
            return $this->return_error($this->module_name, 'create', 'could not create object - ' . $ex->getMessage());
        }
        
        $params = [];
        foreach ($required_parameters as $required_parameter) {
            $params[$required_parameter] = $p_params[$required_parameter];
        }
        
        $optParams = [];
        foreach ($optional_params as $optionalParam) {
            $val = isset($p_params[$optionalParam]) ? $p_params[$optionalParam] : null;
            $optParams[$optionalParam] = $val;
        }
        
        $l_return_id = $this->dao->create_object($params, $optParams);
        if ($l_return_id) {
            return $this->return_success($this->module_name, 'create', $l_return_id);
        } else {
            return $this->return_error($this->module_name, 'create', 'could not create object');
        }
    }
    
    /**
     * Update object.
     *
     * @param  array $p_params
     * @return array
     */
    abstract public function update($p_params);

    /**
     * Update an existing object.
     *
     * @param array $p_params
     * @param int $id
     * @param array $parameters
     * @param array $ref_parameters
     * @return array
     * @throws isys_exception_auth
     * @throws isys_exception_general
     */
    protected function _update($p_params, $id, $parameters = [], $ref_parameters = [])
    {
        // check if object exists
        $object = $this->dao->get_object('id', $id);
        if (empty($object)) {
            return $this->return_error($this->module_name, 'update', 'could not find object');
        }

        $this->check_permission_of_category(isys_auth::EDIT, $object['obj_id'] ?: null);
        
        $params = $this->_get_update_params_from_request($p_params, $parameters, $ref_parameters);
        $success = $this->dao->update_object($id, [], $params);
        if ($success) {
            return $this->return_success($this->module_name, 'update', $id);
        } else {
            return $this->return_error($this->module_name, 'update', 'could not update object');
        }
    }
    
    /**
     * Get update parameters from request.
     *
     * @param  array $p_params
     * @param  array $parameters
     * @param  array $ref_parameters
     * @return array
     */
    protected function _get_update_params_from_request($p_params, $parameters = [], $ref_parameters = [])
    {
        $params = [];
        foreach ($parameters as $parameter) {
            if (! isset($p_params[$parameter])) {
                continue;
            }
            $value = $p_params[$parameter];
            if ($value == "") {
                $value = null;
            }
            $params[$parameter] = $value;
        }
        
        foreach ($ref_parameters as $ref_parameter) {
            if (! isset($p_params[$ref_parameter])) {
                // parameter not set
                continue;
            }
            $value = $p_params[$ref_parameter];
            if ($value == "0") {
                $value = null;
            }
            $params[$ref_parameter] = $value;
        }
        
        return $params;
    }
    
    /**
     * Delete object.
     *
     * @param  array $p_params
     * @return boolean
     */
    abstract public function delete($p_params);

    /**
     * Delete an object by ID-parameter.
     *
     * @param string $idParameterName
     * @param array $p_params
     * @return array
     * @throws isys_exception_auth
     * @throws isys_exception_general
     */
    protected function _delete($idParameterName, $p_params)
    {
        $required_parameters = array (
            $idParameterName,
        );
        try {
            $this->check_required_parameters($required_parameters, $p_params);
        } catch (isys_exception_validation $ex) {
            return $this->return_error($this->module_name, 'delete', 'could not delete object - ' . $ex->getMessage());
        }
        
        // check if object exists
        $object = $this->dao->get_object($idParameterName, $p_params[$idParameterName]);
        if (empty($object)) {
            return $this->return_error($this->module_name, 'delete', 'could not find object');
        }

        $this->check_permission_of_category(isys_auth::DELETE, $object['obj_id'] ?: null);
        
        $success = $this->dao->delete_object($p_params[$idParameterName]);
        if ($success) {
            return $this->return_success($this->module_name, 'delete', null);
        } else {
            return $this->return_error($this->module_name, 'delete', 'could not delete object');
        }
    }
    
    /**
     * Return success message.
     *
     * @param  string   $module
     * @param  string   $action
     * @param  int|null $returned_id
     * @return array
     */
    protected function return_success($module, $action, $returned_id)
    {
        return array(
            'message'      => "$module.$action.success",
            'id'           => $returned_id,
            'current_date' => date('Y-m-d'),
        );
    }
    
    /**
     * Return error message.
     *
     * @param  string $module
     * @param  string $action
     * @param  string $message
     * @return array
     */
    protected function return_error($module, $action, $message)
    {
        return array(
            'message'      => "$module.$action.error",
            'error'        => $message,
            'current_date' => date('Y-m-d'),
        );
    }

    /**
     * Check if request contains required parameters.
     *
     * @param  array $required_parameters
     * @param  array $request
     * @throws isys_exception_validation
     */
    protected function check_required_parameters($required_parameters, $request)
    {
        $missing_parameters = [];
        foreach ($required_parameters as $param) {
            if (! isset($request[$param]) || $request[$param] == '') {
                $missing_parameters[] = $param;
            }
        }
        
        if (! empty($missing_parameters)) {
            $params = implode(',', $missing_parameters);
            $message = "following parameters are missing: $params";
            throw new isys_exception_validation($message, $missing_parameters);
        }
    }

    /**
     * Check permission of category.
     *
     * @param int $permission
     * @param int|null $objId
     * @return bool
     * @throws isys_exception_auth
     * @throws isys_exception_general
     */
    protected function check_permission_of_category($permission, $objId = null)
    {
        if (! $this->useAuth || empty($this->category_name)) {
            return true;
        }

        if (empty($objId)) {
            return isys_auth_cmdb::instance()->category($permission, strtolower($this->category_name));
        } else {
            return isys_auth_cmdb::instance()->check_rights_obj_and_category($permission, $objId, $this->category_name);
        }
    }
}
