<?php

/**
 * SHD SM-VIEW Cable Tracing API
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class cable_api_model_cable_tracing
{

    /**
     * Name of this module.
     *
     * @var string
     */
    protected $module_name = 'cable_tracing';

    /**
     * Indicating whether authentication system should be used during api calls.
     *
     * @var boolean
     */
    protected $useAuth = false;

    /**
     * Constructor.
     */
    function __construct()
    {
        global $g_config;

        // Determine status of authentication system
        $this->useAuth = $g_config['use_auth'];
    }
    
    /**
     * Get connected cables of an object.
     *
     * @param  array $p_params
     * @return array
     */
    public function get_connected_cables($p_params)
    {
        if ($this->useAuth) {
            isys_auth_auth::instance()->module(isys_auth::VIEW, strtolower("C__MODULE__SMVIEW_CABLE"));
        }

        $required_parameters = array (
            'obj_id',
        );
        
        try {
            $this->check_required_parameters($required_parameters, $p_params);
        } catch (\isys_exception_validation $ex) {
            return $this->return_error($this->module_name, 'get_connected_cables', 'error: ' . $ex->getMessage());
        }
        
        $obj_id = $p_params['obj_id'];
        $tracing = new \idoit\Module\Cable\CableTracing($obj_id);
        return $tracing->get_connected_cables();
    }
    
    /**
     * Get unpatched connections of an object.
     *
     * @param  array $p_params
     * @return array
     */
    public function get_unpatched_connections($p_params)
    {
        if ($this->useAuth) {
            isys_auth_auth::instance()->module(isys_auth::VIEW, strtolower("C__MODULE__SMVIEW_CABLE"));
        }

        $required_parameters = array (
            'obj_id',
        );
        
        try {
            $this->check_required_parameters($required_parameters, $p_params);
        } catch (\isys_exception_validation $ex) {
            return $this->return_error($this->module_name, 'get_unpatched_connections', 'error: ' . $ex->getMessage());
        }
        
        $obj_id = $p_params['obj_id'];
        $ignored_wire_id = isset($p_params['ignored_wire_id']) ? $p_params['ignored_wire_id'] : null;
        $tracing = new \idoit\Module\Cable\CableTracing($obj_id);
        return $tracing->get_unpatched_connections($ignored_wire_id, $ignored_wire_id);
    }
    
    /**
     * Return success message.
     *
     * @param  string   $module
     * @param  string   $action
     * @param  int|null $returned_id
     * @return array
     */
    protected function return_success($module, $action, $returned_id)
    {
        return array(
            'message'      => "$module.$action.success",
            'id'           => $returned_id,
            'current_date' => date('Y-m-d'),
        );
    }
    
    /**
     * Return error message.
     *
     * @param  string $module
     * @param  string $action
     * @param  string $message
     * @return array
     */
    protected function return_error($module, $action, $message)
    {
        return array(
            'message'      => "$module.$action.error",
            'error'        => $message,
            'current_date' => date('Y-m-d'),
        );
    }

    /**
     * Check if request contains required parameters.
     *
     * @param  array $required_parameters
     * @param  array $request
     * @throws \isys_exception_validation
     */
    protected function check_required_parameters($required_parameters, $request)
    {
        $missing_parameters = [];
        foreach ($required_parameters as $param) {
            if (! isset($request[$param]) || $request[$param] == '') {
                $missing_parameters[] = $param;
            }
        }
        
        if (! empty($missing_parameters)) {
            $params = implode(',', $missing_parameters);
            $message = "following parameters are missing: $params";
            throw new \isys_exception_validation($message, $missing_parameters);
        }
    }
}
