<?php

/**
 * SHD SM-VIEW Cable Pin API
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class cable_api_model_pin extends cable_api_model_abstract
{

    /**
     * Name of this module.
     *
     * @var string
     */
    protected $module_name = 'pin';

    /**
     * Name of this category.
     *
     * @var string
     */
    protected $category_name = 'C__CATG__SMVIEW_CABLE_PIN';
    
    /**
     * Constructor.
     */
    function __construct()
    {
        parent::__construct();
        $this->dao = isys_cmdb_dao_smview_cable_pin::instance(isys_application::instance()->database);
    }

    /**
     * Create pin.
     *
     * $p_param structure:
     * array (
     *     'obj_id'                => int,
     *     'description'           => string,
     *     'number'                => int,
     *     'type'                  => string,
     *     'internal_pin_id'       => int|null,
     *     'patch_pin_id'          => int|null,
     *     'connected_wire_pin_id' => int|null,
     *     'connector_id'          => int|null,
     *     'classification'        => string|null,
     * )
     *
     * @return array
     */
    public function create($p_params_all)
    {
        if ($this->useAuth && ! empty($this->category_name)) {
            isys_auth_cmdb::instance()->category(isys_auth::CREATE, strtolower($this->category_name));
        }

        $p_params = $p_params_all['data'];
        $required_parameters = array (
            'obj_id',
            'description',
            'type',
        );
        try {
            $this->check_required_parameters($required_parameters, $p_params);
        } catch (\isys_exception_validation $ex) {
            return $this->return_error($this->module_name, 'create', 'could not create object - ' . $ex->getMessage());
        }
        
        $type = isys_cmdb_dao_smview_cable_pin::PIN_TYPE_INPUT;
        if ($p_params['type'] == isys_cmdb_dao_smview_cable_pin::PIN_TYPE_OUTPUT) {
            $type = isys_cmdb_dao_smview_cable_pin::PIN_TYPE_OUTPUT;
        }
        $number = isset($p_params['number']) ? $p_params['number'] : null;
        $internal_pin_id = isset($p_params['internal_pin_id']) ? $p_params['internal_pin_id'] : null;
        $patch_pin_id = isset($p_params['patch_pin_id']) ? $p_params['patch_pin_id'] : null;
        $connector_id = isset($p_params['connector_id']) ? $p_params['connector_id'] : null;
        $classification = isset($p_params['classification']) ? $p_params['classification'] : null;
        $l_return_id = $this->dao->create_pin($p_params['obj_id'], $p_params['description'], $number, $type, $internal_pin_id, $patch_pin_id, $connector_id, $classification);
        
        $wire_pin_id = isset($p_params['connected_wire_pin_id']) ? $p_params['connected_wire_pin_id'] : null;
        $this->_handle_wire_pin_connection($l_return_id, $wire_pin_id);

        if ($l_return_id) {
            return $this->return_success($this->module_name, 'create', $l_return_id);
        } else {
            $this->return_error($this->module_name, 'create', 'could not create object');
        }
    }

    /**
     * Update pin.
     *
     * @param  array $p_params
     *
     * $p_param structure:
     * array (
     *     'id'                    => int,
     *     'description'           => string,
     *     'number'                => int,
     *     'type'                  => string,
     *     'internal_pin_id'       => int|null,
     *     'patch_pin_id'          => int|null,
     *     'connected_wire_pin_id' => int|null,
     *     'connector_id'          => int|null,
     *     'classification'        => string|null,
     * )
     * @return array
     */
    public function update($p_params)
    {
        if ($this->useAuth && ! empty($this->category_name)) {
            isys_auth_cmdb::instance()->category(isys_auth::EDIT, strtolower($this->category_name));
        }

        $parameters = array (
            'description',
            'number',
            'type',
            'classification',
        );
        $ref_params = array(
            'internal_pin_id',
            'patch_pin_id',
            'connected_wire_pin_id',
            'connector_id',
        );
        
        $id = $p_params['id'];
        $data = $p_params['data'];
        
        $params = $this->_get_update_params_from_request($data, $parameters, $ref_params);
        
        $pin = $this->dao->get_object('id', $id);
        $description = array_key_exists('description', $params) ? $params['description'] : $pin['description'];
        $number = array_key_exists('number', $params) ? $params['number'] : $pin['number'];
        $type = array_key_exists('type', $params) ? $params['type'] : $pin['type'];
        $internal_pin_id = array_key_exists('internal_pin_id', $params) ? $params['internal_pin_id'] : $pin['internal_pin_id'];
        $patch_pin_id = array_key_exists('patch_pin_id', $params) ? $params['patch_pin_id'] : $pin['patch_pin_id'];
        $connector_id = isset($params['connector_id']) ? $params['connector_id'] : null;
        $classification = array_key_exists('classification', $params) ? $params['classification'] : $pin['classification'];
        $success = $this->dao->update_pin($id, $description, $number, $type, $internal_pin_id, $patch_pin_id, $connector_id, $classification);

        if (isset($data['connected_wire_pin_id'])) {
            $wire_pin_id = $data['connected_wire_pin_id'];
            try {
                $this->_handle_wire_pin_connection($id, $wire_pin_id, true);
            } catch (Exception $e) {
                return $this->return_error($this->module_name, 'update', $e->getMessage());
            }
        }
        
        return $this->return_success($this->module_name, 'update', $id);
    }

    /**
     * Update connector id of a pin.
     *
     * @param  array $p_params
     *
     * $p_param structure:
     * array (
     *     'id'           => int,
     *     'connector_id' => int,
     * )
     * @return array
     * @throws isys_exception_database
     */
    public function update_connector_id($p_params)
    {
        if ($this->useAuth && ! empty($this->category_name)) {
            isys_auth_cmdb::instance()->category(isys_auth::EDIT, strtolower($this->category_name));
        }

        $id = $p_params['id'];
        $pin = $this->dao->get_object('id', $id);
        if (empty($pin)) {
            $message = "could not find pin with ID: $id";
            return $this->return_error($this->module_name, 'update_connector_id', "could not update connector id - $message");
        }

        $data = $p_params['data'];
        if (empty($data['connector_id'])) {
            $message = "following parameters are missing: connector_id";
            return $this->return_error($this->module_name, 'update_connector_id', "could not update connector id - $message");
        }

        $this->dao->update_pin(
            $id,
            $pin['description'],
            $pin['number'],
            $pin['type'],
            $pin['internal_pin_id'],
            $pin['patch_pin_id'],
            $data['connector_id'],
            $pin['classification']
        );
        return $this->return_success($this->module_name, 'update_connector_id', $id);
    }
    
    /**
     * Handle connection to wire pin.
     *
     * @param  int $pin_id
     * @param  int $wire_pin_id
     * @throws Exception
     */
    private function _handle_wire_pin_connection($pin_id, $wire_pin_id, $is_update = false)
    {
        $pin_dao = isys_cmdb_dao_category_g_smview_cable_pin::instance(isys_application::instance()->database);
        $wire_pin_dao = isys_cmdb_dao_smview_cable_wire_pin::instance(isys_application::instance()->database);
        
        // check if a wire is in use by a section or a route
        if ($is_update && $wire_pin_id) {
            if (! $pin_dao->is_wire_pin_updateable($pin_id)) {
                throw new Exception('can not update pin because it is in use by a section or a route');
            }
        }
        
        $wire_pin = $wire_pin_dao->get_object('connected_pin_id', $pin_id);
        if ($wire_pin) {
            // remove current connection
            $wire_pin_dao->update_wire_pin($wire_pin['id'], $wire_pin['description'], $wire_pin['status_id']);
        }
        
        // add new connection
        $wpin = $wire_pin_dao->get_object('id', $wire_pin_id);
        $wire_pin_dao->update_wire_pin($wpin['id'], $wpin['description'], $wpin['status_id'], null, $pin_id);
    }

    /**
     * Find usages of an element.
     *
     * @param $p_params
     *
     * $p_param structure:
     * array (
     *     'id' => int
     * )
     *
     * @return array
     * @throws isys_exception_database
     */
    public function find_usages($p_params)
    {
        if ($this->useAuth && ! empty($this->category_name)) {
            isys_auth_cmdb::instance()->category(isys_auth::VIEW, strtolower($this->category_name));
        }

        $pin_catg = new isys_cmdb_dao_category_g_smview_cable_pin(isys_application::instance()->database);
        return $pin_catg->find_usages($p_params['id']);
    }

    /**
     * Delete pin.
     *
     * @param array $p_params
     *
     * $p_param structure:
     * array (
     *     'id' => int
     * )
     *
     * @return array
     */
    public function delete($p_params)
    {
        // check if a wire is in use by a section or a route
        $pin_dao = isys_cmdb_dao_category_g_smview_cable_pin::instance(isys_application::instance()->database);
        if (! $pin_dao->is_wire_pin_updateable($p_params['id'])) {
            $message = 'can not update pin because it is in use by a section or a route';
            return $this->return_error($this->module_name, 'delete', $message);
        }
        return $this->_delete('id', $p_params);
    }

    public function properties($p_params)
    {
        if ($this->useAuth && ! empty($this->category_name)) {
            isys_auth_cmdb::instance()->category(isys_auth::VIEW, strtolower($this->category_name));
        }

        $dao = isys_cmdb_dao_smview_cable_properties::instance(isys_application::instance()->database);
        $properties = [
            "p_catg_id" => ["C__CATG__SMVIEW_CABLE_PIN", "C__CATG__SMVIEW_CABLE_ROUTE"]
        ];
        return $dao->query($properties);
    }
}
