<?php

use idoit\Module\Cable\CableTracing;
use idoit\Module\Cable\SectionTracing;

/**
 * SHD SM-VIEW Cable Route API
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class cable_api_model_route extends cable_api_model_abstract
{

    /**
     * Name of this module.
     *
     * @var string
     */
    protected $module_name = 'route';

    /**
     * Name of this category.
     *
     * @var string
     */
    protected $category_name = 'C__CATG__SMVIEW_CABLE_ROUTE';
    
    /**
     * Constructor.
     */
    function __construct()
    {
        parent::__construct();
        $this->dao = isys_cmdb_dao_smview_cable_route::instance(isys_application::instance()->database);
    }

    /**
     * Create route.
     *
     * $p_param structure:
     * array (
     *     'obj_id'                   => int,
     *     'description'              => string,
     *     'connection_count'         => int,
     *     'current_connection_count' => int,
     *     'permission'               => string,
     *     'location'                 => string,
     *     'room'                     => string,
     *     'device_type'              => string,
     *     'device_number'            => string,
     *     'phone_number'             => string,
     *     'route_usage_id'           => int,
     *     'phone_subscriber'         => string,
     *     'hint'                     => string,
     *     'measured_length'          => int,
     *     'measure_date'             => string,
     *     'start_obj_id'             => int,
     *     'end_obj_id'               => int,
     *     'tka_participant'          => string,
     *     'is_external'              => boolean,
     *     'damping_left1'            => string,
     *     'damping_left2'            => string,
     *     'damping_right1'           => string,
     *     'damping_right2'           => string,
     *     'associated_route'         => int,
     *     'redundant_routes'         => string,
     * )
     *
     * @return array
     */
    public function create($p_params)
    {
        $required_parameters = array (
            'obj_id',
        );
        $optional_params = array(
            'description',
            'connection_count',
            'current_connection_count',
            'permission',
            'location',
            'room',
            'device_type',
            'device_number',
            'phone_number',
            'route_usage_id',
            'phone_subscriber',
            'hint',
            'measured_length',
            'measure_date',
            'start_obj_id',
            'end_obj_id',
            'tka_participant',
            'is_external',
            'damping_left1',
            'damping_left2',
            'damping_right1',
            'damping_right2',
            'associated_route',
        );
        $data = $p_params['data'];
        $result = $this->_create($data, $required_parameters, $optional_params);

        $dao = new isys_cmdb_dao_category_g_smview_cable_route(isys_application::instance()->database);
        if (! empty($data['redundant_routes'])) {
            $redundantRoutes = explode(',', $data['redundant_routes']);
            $dao->update_redundant_routes($data['obj_id'], $redundantRoutes);
        }

        if (! empty($data['associated_route'])) {
            $dao->update_associated_route($data['obj_id'], $data['associated_route']);
        }

        return $result;
    }

    /**
     * Update route.
     *
     * @param  array $p_params
     *
     * $p_param structure:
     * array (
     *     'id'                       => int,
     *     'description'              => string,
     *     'connection_count'         => int,
     *     'current_connection_count' => int,
     *     'permission'               => string,
     *     'location'                 => string,
     *     'room'                     => string,
     *     'device_type'              => string,
     *     'device_number'            => string,
     *     'phone_number'             => string,
     *     'route_usage_id'           => int,
     *     'phone_subscriber'         => string,
     *     'hint'                     => string,
     *     'measured_length'          => int,
     *     'measure_date'             => string,
     *     'start_obj_id'             => int,
     *     'end_obj_id'               => int,
     *     'tka_participant'          => string,
     *     'is_external'              => boolean,
     *     'damping_left1'            => string,
     *     'damping_left2'            => string,
     *     'damping_right1'           => string,
     *     'damping_right2'           => string,
     *     'associated_route'         => int,
     *     'redundant_routes'         => string,
     * )
     * @return array
     */
    public function update($p_params)
    {
        $parameters = array (
            'description',
            'connection_count',
            'current_connection_count',
            'permission',
            'location',
            'room',
            'device_type',
            'device_number',
            'phone_number',
            'phone_subscriber',
            'hint',
            'measured_length',
            'measure_date',
            'start_obj_id',
            'end_obj_id',
            'tka_participant',
            'is_external',
            'damping_left1',
            'damping_left2',
            'damping_right1',
            'damping_right2',
            'associated_route',
        );
        
        $id = $p_params['id'];
        $data = $p_params['data'];
        $result = $this->_update($data, $id, $parameters, ['route_usage_id']);

        $dao = new isys_cmdb_dao_category_g_smview_cable_route(isys_application::instance()->database);
        if (! empty($data['redundant_routes'])) {
            $redundantRoutes = explode(',', $data['redundant_routes']);
            $route = $this->dao->get_object('id', $id);
            $dao->update_redundant_routes($route['obj_id'], $redundantRoutes);
        }

        if (! empty($data['associated_route'])) {
            $route = $this->dao->get_object('id', $id);
            $dao->update_associated_route($route['obj_id'], $data['associated_route']);
        }

        return $result;
    }

    /**
     * Get connected objects.
     *
     * @param  array $p_params
     * @return array
     */
    public function get_connected_objects($p_params)
    {
        if (empty($p_params['obj_id'])) {
            return $this->return_error($this->module_name, 'get_connected_objects', 'following parameters are missing: obj_id');
        }

        $this->check_permission_of_category(isys_auth::VIEW, $p_params['obj_id']);

        $obj_id = $p_params['obj_id'];
        $cable_tracing = new CableTracing($obj_id);
        $section_tracing = new SectionTracing($obj_id);
        $ignored_wire_id = ! empty($p_params['wire_id']) ? $p_params['wire_id'] : null;
        $ignored_section_id = ! empty($p_params['section_id']) ? $p_params['section_id'] : null;
        $associated_route = ! empty($p_params['associated_route']) ? $p_params['associated_route'] : null;

        try {
            $connections = $cable_tracing->get_unpatched_connections($ignored_wire_id, $ignored_wire_id, $associated_route);
            $sections = $section_tracing->get_unpatched_sections($ignored_section_id, $ignored_wire_id, $associated_route);
        } catch (isys_exception_database $e) {
            return $this->return_error($this->module_name, 'get_connected_objects', 'could not get connected objects: ' . $e->getMessage());
        }
        $merged_elements = array_merge($connections['objects'], $sections['objects']);
        $connections['objects'] = $merged_elements;
        return $connections;
    }

    /**
     * Get route elements.
     *
     * @param  array $p_params
     * @return array
     */
    public function get_route_elements($p_params)
    {
        if (empty($p_params['obj_id'])) {
            return $this->return_error($this->module_name, 'get_route_elements', 'following parameters are missing: obj_id');
        }

        $this->check_permission_of_category(isys_auth::VIEW, $p_params['obj_id']);

        $obj_id = $p_params['obj_id'];
        $dao = new \isys_cmdb_dao_category_g_smview_cable_route(\isys_application::instance()->database);
        $data = $dao->get_data(null, $obj_id)->get_row();
        if (empty($data)) {
            return [];
        }

        $route_id = $data['isys_catg_smview_cable_route_list__id'];
        $add_connection_information = isset($p_params['add_connection_information']) ? $p_params['add_connection_information'] : true;
        $dao = new \isys_cmdb_dao_category_g_smview_cable_route_element(\isys_application::instance()->database);
        return $dao->get_elements_by_route_id($route_id, false, $add_connection_information);
    }

    /**
     * Get calculated damping of a route.
     *
     * @param  array $p_params
     * @return array
     * @throws isys_exception_auth
     * @throws isys_exception_database
     * @throws isys_exception_general
     */
    public function get_calculated_damping($p_params)
    {
        if (empty($p_params['obj_id'])) {
            return $this->return_error($this->module_name, 'get_calculated_damping', 'following parameters are missing: obj_id');
        }

        $this->check_permission_of_category(isys_auth::VIEW, $p_params['obj_id']);

        $obj_id = $p_params['obj_id'];
        $damping_1310_factor = isys_settings::get(isys_module_smview_cable::DAMPING_1310, 0);
        $damping_1550_factor = isys_settings::get(isys_module_smview_cable::DAMPING_1550, 0);

        $route_dao = new isys_cmdb_dao_category_g_smview_cable_route(isys_application::instance()->database);
        $route = $route_dao->get_data(null, $obj_id)->get_row();
        $dao = new isys_cmdb_dao_category_g_smview_cable_route_element(isys_application::instance()->database);
        $elements = $dao->get_elements_by_route_id($route['isys_catg_smview_cable_route_list__id'], false, false);

        $section_dao = new isys_cmdb_dao_category_g_smview_cable_section(isys_application::instance()->database);
        return $section_dao->calculate_damping($elements, $damping_1310_factor, $damping_1550_factor);
    }

    /**
     * Get redundant routes.
     *
     * @param  array $p_params
     * @return array
     * @throws isys_exception_auth
     * @throws isys_exception_database
     * @throws isys_exception_general
     */
    public function get_redundant_routes($p_params)
    {
        if (empty($p_params['obj_id'])) {
            return $this->return_error($this->module_name, 'get_redundant_routes', 'following parameters are missing: obj_id');
        }

        $this->check_permission_of_category(isys_auth::VIEW, $p_params['obj_id']);

        $dao = new isys_cmdb_dao_category_g_smview_cable_route(isys_application::instance()->database);
        $res = $dao->get_assigned_redundant_routes($p_params['obj_id']);
        $routes = [];
        while ($row = $res->get_row()) {
            $routes[] = [
                'id'    => $row['isys_obj__id'],
                'title' => $row['isys_obj__title'],
            ];
        }
        return $routes;
    }

    /**
     * Delete route.
     *
     * @param array $p_params
     *
     * $p_param structure:
     * array (
     *     'id' => int,
     * )
     *
     * @return array
     */
    public function delete($p_params)
    {
        return $this->_delete('id', $p_params);
    }
}
