<?php

use idoit\Module\Cable\CableTracing;

/**
 * SHD SM-VIEW Cable Section API
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class cable_api_model_section extends cable_api_model_abstract
{

    /**
     * Name of this module.
     *
     * @var string
     */
    protected $module_name = 'section';

    /**
     * Name of this category.
     *
     * @var string
     */
    protected $category_name = 'C__CATG__SMVIEW_CABLE_SECTION';
    
    /**
     * Constructor.
     */
    function __construct()
    {
        parent::__construct();
        $this->dao = isys_cmdb_dao_smview_cable_section::instance(isys_application::instance()->database);
    }

    /**
     * Create section.
     *
     * $p_param structure:
     * array (
     *     'obj_id'                => int,
     *     'description'           => string,
     *     'user_state_id'         => int,
     *     'channel_count'         => int,
     *     'current_channel_count' => int,
     *     'start_obj_id'          => int,
     *     'end_obj_id'            => int,
     * )
     *
     * @return array
     */
    public function create($p_params)
    {
        $required_parameters = array (
            'obj_id',
        );
        $optional_params = array(
            'description',
            'user_state_id',
            'channel_count',
            'current_channel_count',
            'start_obj_id',
            'end_obj_id',
        );
        $data = $p_params['data'];
        return $this->_create($data, $required_parameters, $optional_params);
    }

    /**
     * Update section.
     *
     * @param  array $p_params
     *
     * $p_param structure:
     * array (
     *     'id'                    => int,
     *     'description'           => string,
     *     'user_state_id'         => int,
     *     'channel_count'         => int,
     *     'current_channel_count' => int,
     *     'start_obj_id'          => int,
     *     'end_obj_id'            => int,
     * )
     * @return array
     */
    public function update($p_params)
    {
        $parameters = array (
            'description',
            'channel_count',
            'current_channel_count',
            'start_obj_id',
            'end_obj_id',
        );
        $ref_params = array(
            'user_state_id',
        );
        
        $id = $p_params['id'];
        $data = $p_params['data'];
        return $this->_update($data, $id, $parameters, $ref_params);
    }

    /**
     * Find usages of an element.
     *
     * @param $p_params
     *
     * $p_param structure:
     * array (
     *     'id' => int
     * )
     *
     * @return array
     * @throws isys_exception_database
     */
    public function find_usages($p_params)
    {
        $this->check_permission_of_category(isys_auth::VIEW, $p_params['id']);

        $section_catg = new isys_cmdb_dao_category_g_smview_cable_section(isys_application::instance()->database);
        return $section_catg->find_usages($p_params['id']);
    }

    /**
     * Get connected objects.
     *
     * @param  array $p_params
     * @return array
     */
    public function get_connected_objects($p_params)
    {
        if (empty($p_params['obj_id'])) {
            return $this->return_error($this->module_name, 'get_connected_objects', 'following parameters are missing: obj_id');
        }

        $this->check_permission_of_category(isys_auth::VIEW, $p_params['obj_id']);

        $obj_id = $p_params['obj_id'];
        $tracing = new CableTracing($obj_id);
        $ignored_wire_id = ! empty($p_params['wire_id']) ? $p_params['wire_id'] : null;
        try {
            return $tracing->get_unpatched_connections($ignored_wire_id, $ignored_wire_id);
        } catch (isys_exception_database $e) {
            return $this->return_error($this->module_name, 'get_connected_objects', 'could not get connected objects: ' . $e->getMessage());
        }
    }

    /**
     * Get section elements.
     *
     * @param  array $p_params
     * @return array
     * @throws isys_exception_auth
     * @throws isys_exception_database
     * @throws isys_exception_general
     */
    public function get_section_elements($p_params)
    {
        if (! empty($p_params['section_id'])) {
            $section_id = $p_params['section_id'];

            $dao = new isys_cmdb_dao_smview_cable_section(isys_application::instance()->database);
            $section = $dao->get_object('id', $section_id);
            $this->check_permission_of_category(isys_auth::VIEW, $section['obj_id']);
        } else {
            if (empty($p_params['obj_id'])) {
                return $this->return_error($this->module_name, 'get_section_elements', 'following parameters are missing: obj_id');
            }

            $this->check_permission_of_category(isys_auth::VIEW, $p_params['obj_id']);

            $obj_id = $p_params['obj_id'];
            $dao = new isys_cmdb_dao_category_g_smview_cable_section(isys_application::instance()->database);
            $data = $dao->get_data(null, $obj_id)->get_row();
            if (empty($data)) {
                return [];
            }
            $section_id = $data['isys_catg_smview_cable_section_list__id'];
        }
        $add_connection_information = isset($p_params['add_connection_information']) ? $p_params['add_connection_information'] : true;
        $add_capacity_information = isset($p_params['add_capacity_information']) ? $p_params['add_capacity_information'] : true;
        $dao = new isys_cmdb_dao_category_g_smview_cable_section_element(isys_application::instance()->database);
        return $dao->get_elements_by_section_id($section_id, false, $add_connection_information, $add_capacity_information);
    }

    /**
     * Get calculated damping of a section.
     *
     * @param  array $p_params
     * @return array
     * @throws isys_exception_auth
     * @throws isys_exception_database
     * @throws isys_exception_general
     */
    public function get_calculated_damping($p_params)
    {
        if (empty($p_params['obj_id'])) {
            return $this->return_error($this->module_name, 'get_calculated_damping', 'following parameters are missing: obj_id');
        }

        $this->check_permission_of_category(isys_auth::VIEW, $p_params['obj_id']);

        $obj_id = $p_params['obj_id'];
        $damping_1310_factor = isys_settings::get(isys_module_smview_cable::DAMPING_1310, 0);
        $damping_1550_factor = isys_settings::get(isys_module_smview_cable::DAMPING_1550, 0);

        $section_dao = new isys_cmdb_dao_category_g_smview_cable_section(isys_application::instance()->database);
        $section = $section_dao->get_data(null, $obj_id)->get_row();
        $dao = new isys_cmdb_dao_category_g_smview_cable_section_element(isys_application::instance()->database);
        $elements = $dao->get_elements_by_section_id($section['isys_catg_smview_cable_section_list__id'], false, false, false);
        return $section_dao->calculate_damping($elements, $damping_1310_factor, $damping_1550_factor);
    }

    /**
     * Delete section.
     *
     * @param array $p_params
     *
     * $p_param structure:
     * array (
     *     'id' => int,
     * )
     *
     * @return array
     * @throws isys_exception_dao
     * @throws isys_exception_database
     */
    public function delete($p_params)
    {
        // remove section elements first
        $sectionElementDao = isys_cmdb_dao_category_g_smview_cable_section_element::instance(isys_application::instance()->database);
        $sectionId = $p_params['id'];

        // first get current elements (to determine patches to be deleted later on)
        $currentElements = $sectionElementDao->get_elements_by_section_id($sectionId, null, false, false);

        // remove all section elements
        $elementDao = isys_cmdb_dao_smview_cable_section_element::instance(isys_application::instance()->database);
        $sectionElements = $elementDao->filter_objects(array('section_id' => $sectionId));
        foreach ($sectionElements as $current_element) {
            $elementDao->delete_object($current_element['id']);
        }

        // remove obsolete patches
        $sectionElementDao->patch_section_elements($sectionId, $currentElements);

        // delete section
        return $this->_delete('id', $p_params);
    }
}
