<?php

/**
 * SHD SM-VIEW Cable Section Element API
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class cable_api_model_section_element extends cable_api_model_abstract
{

    /**
     * Name of this module.
     *
     * @var string
     */
    protected $module_name = 'section_element';

    /**
     * Name of this category.
     *
     * @var string
     */
    protected $category_name = 'C__CATG__SMVIEW_CABLE_SECTION_ELEMENT';
    
    /**
     * Constructor.
     */
    function __construct()
    {
        parent::__construct();
        $this->dao = isys_cmdb_dao_smview_cable_section_element::instance(isys_application::instance()->database);
    }

    /**
     * Create section element.
     *
     * $p_param structure:
     * array (
     *     'section_id' => int,
     *     'wire_id'    => int,
     *     'index'      => int,
     * )
     *
     * @return array
     */
    public function create($p_params)
    {
        $required_parameters = array (
            'section_id',
            'wire_id',
            'index',
        );
        $data = $p_params['data'];

        // check permission and disable auth after it to disable default behaviour
        $dao = new isys_cmdb_dao_smview_cable_section(isys_application::instance()->database);
        $section = $dao->get_object('id', $data['section_id']);
        $this->check_permission_of_category(isys_auth::CREATE, $section['obj_id']);
        $this->useAuth = false;

        $result = $this->_validate_index($data, $data['section_id']);
        if (! empty($result)) {
            return $result;
        }

        // first get current elements (to determine patches to be deleted later on)
        $section_element_dao = new \isys_cmdb_dao_category_g_smview_cable_section_element(\isys_application::instance()->database);
        $current_elements = $section_element_dao->get_elements_by_section_id($data['section_id'], null, false, false);

        $result = $this->_create($data, $required_parameters);
        $this->_update_section_endpoints($data['section_id']);

        // create new patches and remove obsolete ones
        $section_element_dao->patch_section_elements($data['section_id'], $current_elements);
        return $result;
    }

    /**
     * Update section element.
     *
     * @param  array $p_params
     *
     * $p_param structure:
     * array (
     *     'id'      => int,
     *     'wire_id' => int,
     *     'index'   => int,
     * )
     * @return array
     */
    public function update($p_params)
    {
        $parameters = array (
            'index',
        );
        $ref_parameters = array (
            'wire_id',
        );

        $id = $p_params['id'];
        $data = $p_params['data'];
        $section_element = $this->dao->get_object('id', $id);
        if (empty($section_element)) {
            return $this->return_error($this->module_name, 'update', 'could not find object');
        }

        $section_id = $section_element['section_id'];
        $result = $this->_validate_index($data, $section_id, $id);
        if (! empty($result)) {
            return $result;
        }

        // check permission and disable auth after it to disable default behaviour
        $dao = new isys_cmdb_dao_smview_cable_section(isys_application::instance()->database);
        $section = $dao->get_object('id', $section_id);
        $this->check_permission_of_category(isys_auth::EDIT, $section['obj_id']);
        $this->useAuth = false;

        // first get current elements (to determine patches to be deleted later on)
        $section_element_dao = new isys_cmdb_dao_category_g_smview_cable_section_element(isys_application::instance()->database);
        $current_elements = $section_element_dao->get_elements_by_section_id($section_id, null, false, false);

        $response = $this->_update($data, $id, $parameters, $ref_parameters);
        $this->_update_section_endpoints($section_id);

        // create new patches and remove obsolete ones
        $section_element_dao->patch_section_elements($section_id, $current_elements);
        return $response;
    }

    /**
     * Validate elements index.
     *
     * @param array $data
     * @param int|null $section_id
     * @param int|null $id
     * @return array|null
     */
    private function _validate_index($data, $section_id = null, $id = null)
    {
        if (! isset($data['index'])) {
            return null;
        }

        $index = $data['index'];
        if (! isset($section_id)) {
            return $this->return_error($this->module_name, 'update', 'following parameters are missing: section_id');
        }

        $section_elements = $this->dao->filter_objects(array('section_id' => $section_id));
        foreach ($section_elements as $element) {
            if ($element['index'] == $index) {
                if (isset($id) && $element['id'] == $id) {
                    continue;
                }
                return $this->return_error($this->module_name, 'update', 'index already used by another element');
            }
        }
    }

    /**
     * Delete section element.
     *
     * @param array $p_params
     *
     * $p_param structure:
     * array (
     *     'id' => int,
     * )
     *
     * @return array
     */
    public function delete($p_params)
    {
        // get elements section
        $element = $this->dao->get_object('id', $p_params['id']);
        if (empty($element)) {
            return $this->return_error($this->module_name, 'delete', 'could not find object');
        }

        $section_id = $element['section_id'];
        $section_dao = isys_cmdb_dao_smview_cable_section::instance(isys_application::instance()->database);
        $element_dao = \isys_cmdb_dao_category_g_smview_cable_section_element::instance(\isys_application::instance()->database);
        $elements = $element_dao->get_elements_by_section_id($section_id, null, false, false);

        // first get current elements (to determine patches to be deleted later on)
        $current_elements = $elements;
        $section = $section_dao->get_object('id', $section_id);

        // check permission and disable auth after it to disable default behaviour
        $this->check_permission_of_category(isys_auth::DELETE, $section['obj_id']);
        $this->useAuth = false;

        // check, if element is the last one
        $first_element = array_shift($elements);
        $last_element = count($elements) > 0 ? array_pop($elements) : $first_element;
        if ($last_element['id'] != $p_params['id'] && $first_element['id'] != $p_params['id']) {
            return $this->return_error($this->module_name, 'delete', 'can only delete first or last element of a section');
        }

        $start_obj_id = null;
        $end_obj_id = null;
        if ($first_element['id'] != $p_params['id']) {
            $start_obj_id = $section['start_obj_id'];
        }
        if ($last_element['id'] != $p_params['id']) {
            $end_obj_id = $section['end_obj_id'];
        }

        $result = $this->_delete('id', $p_params);
        $this->_update_section_endpoints($section_id, $start_obj_id, $end_obj_id);

        if ($first_element['id'] == $p_params['id']) {
            $element_dao->re_order_indices($section_id);
        }

        // create new patches and remove obsolete ones
        $element_dao->patch_section_elements($section_id, $current_elements);
        return $result;
    }

    /**
     * Update section endpoints.
     *
     * @param  int $section_id
     * @param  int|null $start_obj_id
     * @param  int|null $end_obj_id
     * @throws isys_exception_dao
     */
    private function _update_section_endpoints($section_id, $start_obj_id = null, $end_obj_id = null)
    {
        $section_dao = isys_cmdb_dao_smview_cable_section::instance(isys_application::instance()->database);
        $element_dao = isys_cmdb_dao_category_g_smview_cable_section_element::instance(isys_application::instance()->database);
        $elements = $element_dao->get_elements_by_section_id($section_id, true, false, false);
        $this->update_endpoints($section_dao, $section_id, $elements, $start_obj_id, $end_obj_id);
    }

    /**
     * Update section endpoints.
     *
     * @param  isys_cmdb_dao_smview_cable_abstract $dao
     * @param  int $id
     * @param  array $elements
     * @param  int|null $start_obj_id
     * @param  int|null $end_obj_id
     * @throws isys_exception_dao
     */
    public static function update_endpoints($dao, $id, $elements, $start_obj_id = null, $end_obj_id = null)
    {
        if (empty($elements)) {
            // cleanup start- and end-objects
            $params = array(
                'start_obj_id' => null,
                'end_obj_id'   => null,
            );
            $dao->update_object($id, [], $params);
            return;
        }

        $firstElement = array_shift($elements);
        $lastElement = count($elements) > 0 ? array_pop($elements) : $firstElement;

        $params = array(
            'start_obj_id' => $firstElement['obj_out_id'],
            'end_obj_id'   => $lastElement['obj_in_id'],
        );

        if ($start_obj_id) {
            if (count($elements) == 0 && $params['start_obj_id'] != $start_obj_id) {
                // flip start and end if only one element remains
                $params['end_obj_id'] = $params['start_obj_id'];
            }
            $params['start_obj_id'] = $start_obj_id;
        }

        if ($end_obj_id) {
            if (count($elements) == 0 && $params['end_obj_id'] != $end_obj_id) {
                // flip start and end if only one element remains
                $params['start_obj_id'] = $params['end_obj_id'];
            }
            $params['end_obj_id'] = $end_obj_id;
        }
        $dao->update_object($id, $params);
    }
}
