<?php

/**
 * SHD SM-VIEW Cable Station API
 *
 * @package    idoit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  2023 SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license MIT
 */
class cable_api_model_station extends cable_api_model_abstract
{

    /**
     * Name of this module.
     *
     * @var string
     */
    protected $module_name = 'station';

    /**
     * Read method.
     *
     * @param  array $p_params
     * @return array
     */
    public function read($p_params)
    {
        return array_values($this->_getStations());
    }

    /**
     * Get all stations with assigned category "extended location".
     *
     * @return array
     */
    private function _getStations(): array
    {
        $dao = new isys_cmdb_dao_category_g_global(isys_application::instance()->database);
        $assignedStationsDao = new isys_cmdb_dao_category_g_smview_cable_assigned_stations(isys_application::instance()->database);
        $extendedLocationDao = new isys_cmdb_dao_category_g_smview_cable_extended_location(isys_application::instance()->database);
        $lineObjTypeID = defined_or_default('C__SHD__LINE', 0);
        $stationObjTypeID = defined_or_default('C__SHD__STATION', 0);
        $sql = ' AND isys_obj_type__id = ' . $stationObjTypeID;
        $sql .= ' AND isys_obj__status = ' . defined_or_default('C__RECORD_STATUS__NORMAL');
        $res = $dao->get_data(null, null, $sql);

        // get all stations
        $stations = [];
        while ($stationRow = $res->get_row()) {
            $sql = ' AND isys_catg_smview_cable_assigned_stations_list__station_id = ' . $stationRow['isys_obj__id'];
            $sql .= ' AND mainObject.isys_obj__status = ' . defined_or_default('C__RECORD_STATUS__NORMAL');
            $combinedStationRow = $assignedStationsDao->get_data(null, null, $sql)->get_row();
            $centralStationRow = null;
            if (! empty($combinedStationRow)) {
                $sql = ' AND isys_obj__status = ' . defined_or_default('C__RECORD_STATUS__NORMAL');
                $centralStationRow = $dao->get_data(null, $combinedStationRow['isys_catg_smview_cable_assigned_stations_list__isys_obj__id'], $sql)->get_row();
            }

            $sql = ' AND mainObject.isys_obj__status = ' . defined_or_default('C__RECORD_STATUS__NORMAL');
            $extendedLocation = $extendedLocationDao->get_data(null, $stationRow['isys_obj__id'], $sql)->get_row();
            $stationData = [
                'obj_id'        => $stationRow['isys_obj__id'],
                'obj_title'     => $stationRow['isys_obj__title'],
                'station_id'    => isset($centralStationRow) ? $centralStationRow['isys_obj__id'] : null,
                'station_title' => isset($centralStationRow) ? $centralStationRow['isys_obj__title'] : null,
                'position'      => ! empty($extendedLocation) ? $extendedLocation['isys_catg_smview_cable_extended_location_list__position'] : null,
                'line_id'       => null,
                'line_title'    => null,
            ];

            // get associated line by assigned stations
            $lineRow = $this->_getMatchingParentByObjectType($stationRow['isys_obj__id'], $lineObjTypeID);
            if (! empty($lineRow)) {
                $stationData['line_id'] = $lineRow['isys_obj__id'];
                $stationData['line_title'] = $lineRow['isys_obj__title'];
            }
            $stations[$stationRow['isys_obj__id']] = $stationData;
        }
        return $stations;
    }

    /**
     * Try to get matching parent by its object type.
     *
     * @param int|null $parentObjId
     * @param int      $objTypeId
     *
     * @return array|null
     */
    private function _getMatchingParentByObjectType(?int $parentObjId, int $objTypeId): ?array
    {
        if (empty($parentObjId) || $parentObjId < 1) {
            return null;
        }

        $locationDao = new isys_cmdb_dao_category_g_location(isys_application::instance()->database);
        $sql = ' AND isys_obj__status = ' . defined_or_default('C__RECORD_STATUS__NORMAL');
        $row = $locationDao->get_data(null, $parentObjId, $sql)->get_row();
        if (! $row) {
            return null;
        }

        if ($row['isys_obj_type__id'] != $objTypeId) {
            return $this->_getMatchingParentByObjectType($row['isys_catg_location_list__parentid'], $objTypeId);
        } else {
            return $row;
        }
    }

    /**
     * Get all maps, containing a requested station.
     *
     * @param  array $p_params
     * @return array
     */
    public function get_maps(array $p_params): array
    {
        if (! class_exists('smview_cmdb_dao_map')) {
            return $this->return_error($this->module_name, __FUNCTION__, "add-on 'SM-VIEW Client' is not installed");
        }

        if (empty($p_params['station'])) {
            return $this->return_error($this->module_name, __FUNCTION__, "parameter 'station' is missing");
        }

        $filteredStations = $this->_filterStations([$p_params['station']]);
        if (empty($filteredStations)) {
            return $this->return_error($this->module_name, __FUNCTION__, "could not find any station with name: " . $p_params['station']);
        }

        $dao = new smview_cmdb_dao_map(isys_application::instance()->database);
        try {
            $maps = $dao->getMaps();
        } catch (isys_exception_database $e) {
            $maps = [];
        }

        $matchingMaps = [];
        foreach ($maps as $map) {
            $mapId = $map['map_id'];
            try {
                $mapObjects = $dao->getMapObjectByMap($mapId);
            } catch (Exception $e) {
                $mapObjects = [];
            }

            foreach ($mapObjects as $mapObject) {
                $mapObjId = $mapObject['isys_obj__id'];
                if (isset($filteredStations[$mapObjId])) {
                    $matchingMaps[$mapId] = $map;
                }
            }
        }
        return array_values($matchingMaps);
    }

    /**
     * Get cables between stations.
     *
     * @param  array $p_params
     * @return array
     */
    public function get_cables_between_stations(array $p_params): array
    {
        if (empty($p_params['stations'])) {
            return $this->return_error($this->module_name, __FUNCTION__, "parameter 'stations' is missing");
        }

        if (! is_array($p_params['stations'])) {
            return $this->return_error($this->module_name, __FUNCTION__, "parameter 'stations' has to be an array");
        }

        if (count($p_params['stations']) < 2) {
            return $this->return_error($this->module_name, __FUNCTION__, "at least 2 stations must be specified");
        }

        $filteredStations = $this->_filterStations($p_params['stations']);
        if (count($filteredStations) < 2) {
            return [];
        }

        // get all section cables
        $cableDao = new isys_cmdb_dao_smview_cable(isys_application::instance()->database);
        $cableTypeDao = new isys_cmdb_dao_smview_cable_type(isys_application::instance()->database);
        $objectDao = new isys_cmdb_dao_category_g_global(isys_application::instance()->database);
        $extLocationDao = new isys_cmdb_dao_category_g_smview_cable_extended_location(isys_application::instance()->database);
        try {
            $cables = $cableDao->filter_objects(['is_section_cable' => 1]);
        } catch (isys_exception_database $e) {
            $cables = [];
        }
        $sectionCables = [];
        $cableGroupIds = ! empty($p_params['cable_group_ids']) ? $p_params['cable_group_ids'] : null;

        $cableCapacityObj = null;
        if (class_exists('\idoit\Module\Cable\CableCapacity')) {
            $cableCapacityObj = \idoit\Module\Cable\CableCapacity::instance();
        }

        foreach ($cables as $cable) {
            if (! empty($cableGroupIds) && ! in_array($cable['group_id'], $cableGroupIds)) {
                // filter cables by group ID if provided
                continue;
            }

            // try to get corresponding stations of cable endings
            $stationIn = $this->_getMatchingParentInList($cable['location_in_id'], $filteredStations);
            $stationOut = $this->_getMatchingParentInList($cable['location_out_id'], $filteredStations);

            if ($stationIn && $stationOut) {
                // this is a section cable between two requested stations
                $sql = ' AND isys_obj__status = ' . defined_or_default('C__RECORD_STATUS__NORMAL');
                $object = $objectDao->get_data(null, $cable['obj_id'], $sql)->get_row();

                // get cable capacity
                $patched_wires = null;
                $sum_wires = null;
                if (isset($cableCapacityObj)) {
                    $cableCapacity = $cableCapacityObj->get_capacity_by_cable_id($cable['obj_id']);
                    $sum_wires = ((int) $cableCapacity['free_wires']) + ((int) $cableCapacity['patched_wires']);
                    $patched_wires = $cableCapacity['patched_wires'];
                }

                // get cable type
                if (! empty($cable['type_id'])) {
                    $row = $cableTypeDao->get_object('id', $cable['type_id']);
                    $cableTypeTitle = $row['description'];
                }

                // try to get rail location (Gleisseite)
                $sql = ' AND isys_obj__status = ' . defined_or_default('C__RECORD_STATUS__NORMAL');
                $row = $extLocationDao->get_data(null, $cable['obj_id'], $sql)->get_row();
                $sectionCables[$cable['obj_id']] = [
                    'obj_id'           => $cable['obj_id'],
                    'obj_title'        => $object['isys_obj__title'],
                    'line_id'          => $stationIn['line_id'] ?: $stationOut['line_id'],
                    'line_title'       => $stationIn['line_title'] ?: $stationOut['line_title'],
                    'rail_id'          => $row && $row['isys_shd_extended_location__id'] ? $row['isys_shd_extended_location__id'] : null,
                    'rail_title'       => $row && $row['isys_shd_extended_location__title'] ? $row['isys_shd_extended_location__title'] : null,
                    'all_wires'        => $sum_wires,
                    'patched_wires'    => $patched_wires,
                    'cable_group_id'   => $cable['group_id'],
                    'cable_type_id'    => $cable['type_id'],
                    'cable_type_title' => $cableTypeTitle ?? null,
                    'station_in'       => $stationIn,
                    'station_out'      => $stationOut,
                ];
            }
        }
        return array_values($sectionCables);
    }

    /**
     * Filter requested stations.
     *
     * @param  array $requestedStations
     * @return array
     */
    private function _filterStations(array $requestedStations): array
    {
        // get all stations
        $stations = $this->_getStations();

        // filter stations by requested station- or combined-station-ID
        $filteredStations = [];
        foreach ($requestedStations as $requestedStationId) {
            foreach ($stations as $id => $station) {
                // try to get normal station
                if ($id == $requestedStationId) {
                    $filteredStations[$id] = $station;
                }

                // try to get combined station
                if ($station['station_id'] == $requestedStationId) {
                    $filteredStations[$id] = $station;
                }
            }
        }
        return $filteredStations;
    }

    /**
     * Try to get a matching parent within listing.
     *
     * @param  int|null $parentObjId
     * @param  array    $listing
     * @return array|null
     */
    private function _getMatchingParentInList(?int $parentObjId, array $listing): ?array
    {
        if (empty($parentObjId) || $parentObjId < 1) {
            return null;
        }

        $locationDao = new isys_cmdb_dao_category_g_location(isys_application::instance()->database);
        $sql = ' AND isys_obj__status = ' . defined_or_default('C__RECORD_STATUS__NORMAL');
        $row = $locationDao->get_data(null, $parentObjId, $sql)->get_row();
        if (! $row) {
            return null;
        }

        if (! isset($listing[$row['isys_obj__id']])) {
            return $this->_getMatchingParentInList($row['isys_catg_location_list__parentid'], $listing);
        } else {
            return $listing[$row['isys_obj__id']];
        }
    }

    /**
     * Create method.
     *
     * @param  array $p_params
     * @return array
     */
    public function create($p_params)
    {
        return $this->return_error($this->module_name, 'create', 'method not implemented');
    }

    /**
     * Update method.
     *
     * @param  array $p_params
     * @return array
     */
    public function update($p_params)
    {
        return $this->return_error($this->module_name, 'create', 'method not implemented');
    }

    /**
     * Delete method.
     *
     * @param array $p_params
     * @return array
     */
    public function delete($p_params)
    {
        return $this->return_error($this->module_name, 'create', 'method not implemented');
    }
}
