<?php

use idoit\Module\Cable\LocationTracing;
use idoit\Module\Cable\Tracing;

/**
 * SHD SM-VIEW Cable Tracing API
 *
 * @package    idoit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  2019 SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license MIT
 */
class cable_api_model_tracing extends cable_api_model_abstract
{

    /**
     * Name of this module.
     *
     * @var string
     */
    protected $module_name = 'tracing';

    /**
     * Read method.
     *
     * @param array $p_params
     *
     * $p_param structure:
     * array (
     *     'location_start' => int,
     *     'location_end'   => int,
     * )
     *
     * @return array
     */
    public function read($p_params)
    {
        return $this->between_locations($p_params);
    }

    /**
     * Get nodes between locations.
     *
     * @param array $p_params
     *
     * $p_param structure:
     * array (
     *     'location_start' => int,
     *     'location_end'   => int,
     * )
     *
     * @return array
     */
    public function between_locations($p_params)
    {
        if ($this->useAuth) {
            isys_auth_auth::instance()->module(isys_auth::VIEW, strtolower("C__MODULE__SMVIEW_CABLE"));
        }

        $required_parameters = array (
            'location_start',
            'location_end',
        );
        $data = $p_params['data'];

        try {
            $this->check_required_parameters($required_parameters, $data);
        } catch (isys_exception_validation $ex) {
            return $this->return_error($this->module_name, 'create', 'could not create object - ' . $ex->getMessage());
        }

        $location_start = $data['location_start'];
        $location_end = $data['location_end'];

        $tracing = new LocationTracing();
        $data = $tracing->get_nodes_between_locations($location_start, $location_end);
        return $data;
    }

    /**
     * Get nodes between objects.
     *
     * @param array $p_params
     *
     * $p_param structure:
     * array (
     *     'object_start' => int,
     *     'object_end'   => int,
     * )
     *
     * @return array
     */
    public function between_objects($p_params)
    {
        if ($this->useAuth) {
            isys_auth_auth::instance()->module(isys_auth::VIEW, strtolower("C__MODULE__SMVIEW_CABLE"));
        }

        $required_parameters = array (
            'object_start',
            'object_end',
        );
        $data = $p_params['data'];

        try {
            $this->check_required_parameters($required_parameters, $data);
        } catch (isys_exception_validation $ex) {
            return $this->return_error($this->module_name, 'create', 'could not create object - ' . $ex->getMessage());
        }

        $object_start = $data['object_start'];
        $object_end = $data['object_end'];

        $tracing = new LocationTracing();
        $connections = $tracing->get_connections($object_start);
        $data = $tracing->get_nodes_between_objects($object_start, $object_end, $connections);
        return $data;
    }

    /**
     * Get nodes between objects.
     *
     * @param array $p_params
     *
     * $p_param structure:
     * array (
     *     'object_start' => int,
     *     'object_end'   => int,
     * )
     *
     * @return array
     */
    public function shortest_path($p_params)
    {
        if ($this->useAuth) {
            isys_auth_auth::instance()->module(isys_auth::VIEW, strtolower("C__MODULE__SMVIEW_CABLE"));
        }

        $required_parameters = array (
            'object_start',
            'object_end',
        );
        $data = $p_params['data'];

        try {
            $this->check_required_parameters($required_parameters, $data);
        } catch (isys_exception_validation $ex) {
            return $this->return_error($this->module_name, 'create', 'could not create object - ' . $ex->getMessage());
        }

        $object_start = $data['object_start'];
        $object_end = $data['object_end'];
        $include_sections = empty($data['exclude_sections']);
        $use_cache = empty($data['do_not_use_cache']);

        $tracing = new Tracing($object_start, $include_sections, $use_cache);
        $data = $tracing->get_shortest_path($object_end);
        return $data;
    }

    /**
     * Get nodes between objects.
     *
     * @param array $p_params
     *
     * $p_param structure:
     * array (
     *     'object_start' => int,
     *     'object_end'   => int,
     * )
     *
     * @return array
     */
    public function least_number_of_hops($p_params)
    {
        if ($this->useAuth) {
            isys_auth_auth::instance()->module(isys_auth::VIEW, strtolower("C__MODULE__SMVIEW_CABLE"));
        }

        $required_parameters = array (
            'object_start',
            'object_end',
        );
        $data = $p_params['data'];

        try {
            $this->check_required_parameters($required_parameters, $data);
        } catch (isys_exception_validation $ex) {
            return $this->return_error($this->module_name, 'create', 'could not create object - ' . $ex->getMessage());
        }

        $object_start = $data['object_start'];
        $object_end = $data['object_end'];
        $include_sections = empty($data['exclude_sections']);
        $use_cache = empty($data['do_not_use_cache']);

        $tracing = new Tracing($object_start, $include_sections, $use_cache);
        $data = $tracing->get_least_number_of_hops($object_end);
        return $data;
    }

    /**
     * Create method.
     *
     * @param  array $p_params
     * @return array
     */
    public function create($p_params)
    {
        return $this->return_error($this->module_name, 'create', 'method not implemented');
    }

    /**
     * Update method.
     *
     * @param  array $p_params
     * @return array
     */
    public function update($p_params)
    {
        return $this->return_error($this->module_name, 'create', 'method not implemented');
    }

    /**
     * Delete method.
     *
     * @param array $p_params
     * @return array
     */
    public function delete($p_params)
    {
        return $this->return_error($this->module_name, 'create', 'method not implemented');
    }
}
