<?php

/**
 * SHD SM-VIEW Cable Route Element API
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class cable_api_model_virtual_route_element extends cable_api_model_abstract
{

    /**
     * Name of this module.
     *
     * @var string
     */
    protected $module_name = 'virtual_route_element';

    /**
     * Name of this category.
     *
     * @var string
     */
    protected $category_name = 'C__CATG__SMVIEW_CABLE_VIRTUAL_ROUTE_ELEMENT';

    /**
     * Constructor.
     */
    function __construct()
    {
        parent::__construct();
        $this->dao = isys_cmdb_dao_smview_cable_virtual_route_element::instance(isys_application::instance()->database);
    }

    /**
     * Create route element.
     *
     * $p_param structure:
     * array (
     *     'route_id'   => int,
     *     'index'      => int,
     *     'wire_id'    => int,
     *     'section_id' => int,
     * )
     *
     * @return array
     */
    public function create($p_params)
    {
        $required_parameters = array (
            'route_id',
            'index',
        );
        $optional_params = array(
            'wire_id',
            'section_id',
        );
        $data = $p_params['data'];

        // check permission and disable auth after it to disable default behaviour
        $dao = new isys_cmdb_dao_smview_cable_virtual_route(isys_application::instance()->database);
        $route = $dao->get_object('id', $data['route_id']);
        $this->check_permission_of_category(isys_auth::CREATE, $route['obj_id']);
        $this->useAuth = false;

        $result = $this->_validate_index($data, $data['route_id']);
        if (! empty($result)) {
            return $result;
        }
        $update_endpoints = empty($data['exclude_endpoint_update']);

        // first get current elements (to determine patches to be deleted later on)
        $route_element_dao = new \isys_cmdb_dao_category_g_smview_cable_virtual_route_element(\isys_application::instance()->database);
        $current_elements = $route_element_dao->get_elements_by_route_id($data['route_id'], false, false);

        $result = $this->_create($data, $required_parameters, $optional_params);
        if ($update_endpoints) {
            $this->_update_route_endpoints($data['route_id']);
        }

        // create new patches and remove obsolete ones
        $route_element_dao->patch_route_elements($data['route_id'], $current_elements);
        return $result;
    }

    /**
     * Update route element.
     *
     * @param  array $p_params
     *
     * $p_param structure:
     * array (
     *     'id'    => int,
     *     'index' => int,
     * )
     * @return array
     */
    public function update($p_params)
    {
        $parameters = array (
            'index',
        );
        $ref_parameters = array (
            'wire_id',
            'section_id',
        );

        $id = $p_params['id'];
        $data = $p_params['data'];
        $route_element = $this->dao->get_object('id', $id);
        if (empty($route_element)) {
            return $this->return_error($this->module_name, 'update', 'could not find object');
        }

        $route_id = $route_element['route_id'];
        $result = $this->_validate_index($data, $route_id, $id);
        if (! empty($result)) {
            return $result;
        }

        // check permission and disable auth after it to disable default behaviour
        $dao = new isys_cmdb_dao_smview_cable_virtual_route(isys_application::instance()->database);
        $route = $dao->get_object('id', $route_id);
        $this->check_permission_of_category(isys_auth::EDIT, $route['obj_id']);
        $this->useAuth = false;

        // first get current elements (to determine patches to be deleted later on)
        $route_element_dao = new isys_cmdb_dao_category_g_smview_cable_virtual_route_element(isys_application::instance()->database);
        $current_elements = $route_element_dao->get_elements_by_route_id($route_id, false, false);

        $response = $this->_update($data, $id, $parameters, $ref_parameters);
        $this->_update_route_endpoints($route_id);

        // create new patches and remove obsolete ones
        $route_element_dao->patch_route_elements($route_id, $current_elements);
        return $response;
    }

    /**
     * Validate elements index.
     *
     * @param array $data
     * @param int|null $route_id
     * @param int|null $id
     * @return array|null
     */
    private function _validate_index($data, $route_id = null, $id = null)
    {
        if (! isset($data['index'])) {
            return null;
        }

        $index = $data['index'];
        if (! isset($route_id)) {
            return $this->return_error($this->module_name, 'update', 'following parameters are missing: route_id');
        }

        $route_elements = $this->dao->filter_objects(array('route_id' => $route_id));
        foreach ($route_elements as $element) {
            if ($element['index'] == $index) {
                if (isset($id) && $element['id'] == $id) {
                    continue;
                }
                return $this->return_error($this->module_name, 'update', 'index already used by another element');
            }
        }
    }

    /**
     * Delete route element.
     *
     * @param array $p_params
     *
     * $p_param structure:
     * array (
     *     'id' => int,
     * )
     *
     * @return array
     */
    public function delete($p_params)
    {
        // get elements route
        $element = $this->dao->get_object('id', $p_params['id']);
        if (empty($element)) {
            return $this->return_error($this->module_name, 'delete', 'could not find object');
        }

        $route_id = $element['route_id'];
        $route_dao = isys_cmdb_dao_smview_cable_virtual_route::instance(isys_application::instance()->database);
        $element_dao = \isys_cmdb_dao_category_g_smview_cable_virtual_route_element::instance(\isys_application::instance()->database);
        $elements = $element_dao->get_elements_by_route_id($route_id, false, false);

        // first get current elements (to determine patches to be deleted later on)
        $current_elements = $elements;
        $route = $route_dao->get_object('id', $route_id);

        // check permission and disable auth after it to disable default behaviour
        $this->check_permission_of_category(isys_auth::DELETE, $route['obj_id']);
        $this->useAuth = false;

        // check, if element is the last one
        $first_element = array_shift($elements);
        $last_element = count($elements) > 0 ? array_pop($elements) : $first_element;
        if ($last_element['id'] != $p_params['id'] && $first_element['id'] != $p_params['id']) {
            return $this->return_error($this->module_name, 'delete', 'can only delete first or last element of a route');
        }

        $start_obj_id = null;
        $end_obj_id = null;
        if ($first_element['id'] != $p_params['id']) {
            $start_obj_id = $route['start_obj_id'];
        }
        if ($last_element['id'] != $p_params['id']) {
            $end_obj_id = $route['end_obj_id'];
        }

        $result = $this->_delete('id', $p_params);
        $this->_update_route_endpoints($element['route_id'], $start_obj_id, $end_obj_id);

        if ($first_element['id'] == $p_params['id']) {
            $element_dao->re_order_indices($route_id);
        }

        // create new patches and remove obsolete ones
        $element_dao->patch_route_elements($route_id, $current_elements);
        return $result;
    }

    /**
     * Update route endpoints.
     *
     * @param  int $route_id
     * @param  int|null $start_obj_id
     * @param  int|null $end_obj_id
     * @throws isys_exception_dao
     */
    private function _update_route_endpoints($route_id, $start_obj_id = null, $end_obj_id = null)
    {
        $route_dao = isys_cmdb_dao_smview_cable_virtual_route::instance(isys_application::instance()->database);
        $element_dao = isys_cmdb_dao_category_g_smview_cable_virtual_route_element::instance(isys_application::instance()->database);
        $elements = $element_dao->get_elements_by_route_id($route_id, true, false);
        cable_api_model_section_element::update_endpoints($route_dao, $route_id, $elements, $start_obj_id, $end_obj_id);
    }
}
