<?php

/**
 * SHD SM-VIEW Cable Wire API
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class cable_api_model_wire extends cable_api_model_abstract
{

    /**
     * Name of this module.
     *
     * @var string
     */
    protected $module_name = 'wire';

    /**
     * Name of this category.
     *
     * @var string
     */
    protected $category_name = 'C__CATG__SMVIEW_CABLE';
    
    /**
     * Constructor.
     */
    function __construct()
    {
        parent::__construct();
        $this->dao = isys_cmdb_dao_smview_cable_wire::instance(isys_application::instance()->database);
    }

    /**
     * Create wire.
     *
     * $p_param structure:
     * array (
     *     'obj_id'             => int,
     *     'description'        => string,
     *     'color'              => string,
     *     'scope'              => string,
     *     'scope_index'        => int,
     *     'reserved'           => int,
     *     'user_state_id'      => int,
     *     'user_state'         => string,
     *     'connection_type_id' => int,
     *     'damping'            => string,
     *     'notice'             => string,
     * )
     *
     * @return array
     */
    public function create($p_params)
    {
        if ($this->useAuth && ! empty($this->category_name)) {
            isys_auth_cmdb::instance()->category(isys_auth::CREATE, strtolower($this->category_name));
        }

        $data = $p_params['data'];
        $required_parameters = array (
            'obj_id',
            'description',
        );
        
        try {
            $this->check_required_parameters($required_parameters, $data);
        } catch (\isys_exception_validation $ex) {
            return $this->return_error($this->module_name, 'create', 'could not create object - ' . $ex->getMessage());
        }
        
        $color = isset($data['color']) ? $data['color'] : null;
        $scope = isset($data['scope']) ? $data['scope'] : null;
        $scope_index = isset($data['scope_index']) ? $data['scope_index'] : null;
        $reserved = isset($data['reserved']) ? $data['reserved'] : null;
        $damping = isset($data['damping']) ? $data['damping'] : null;
        $notice = isset($data['notice']) ? $data['notice'] : null;
        $user_state_id = isset($data['user_state_id']) ? $data['user_state_id'] : null;
        $user_state = isset($data['user_state']) ? $data['user_state'] : null;
        $connection_type_id = isset($data['connection_type_id']) ? $data['connection_type_id'] : null;
        $l_return_id = $this->dao->create_wire($data['obj_id'], $data['description'], $color, $scope, $reserved, $scope_index, $user_state_id, $user_state, $connection_type_id, $damping, $notice);
        
        if ($l_return_id) {
            return $this->return_success($this->module_name, 'create', $l_return_id);
        } else {
            return $this->return_error($this->module_name, 'create', 'could not create object');
        }
    }

    /**
     * Update wire.
     *
     * @param  array $p_params
     *
     * $p_param structure:
     * array (
     *     'id'                 => int,
     *     'obj_id'             => int,
     *     'description'        => string,
     *     'color'              => string,
     *     'scope'              => string,
     *     'scope_index'        => int,
     *     'reserved'           => int,
     *     'user_state_id'      => int,
     *     'user_state'         => string,
     *     'connection_type_id' => int,
     *     'damping'            => string,
     *     'notice'             => string,
     * )
     * @return array
     */
    public function update($p_params)
    {
        $parameters = array (
            'description',
            'color',
            'scope',
            'scope_index',
            'reserved',
            'damping',
            'user_state',
            'notice',
        );
        $ref_params = array(
            'user_state_id',
            'connection_type_id',
        );
        
        $id = $p_params['id'];
        $data = $p_params['data'];
        return $this->_update($data, $id, $parameters, $ref_params);
    }
    
    /**
     * Generate wires.
     *
     * @param  array $p_params
     *
     * $p_param structure:
     * array (
     *     'id'            => int,
     *     'generator_id'  => int,
     *     'start_pattern' => string,
     *     'end_pattern'   => string,
     * )
     * @return array
     */
    public function generate_wires($p_params)
    {
        if ($this->useAuth && ! empty($this->category_name)) {
            isys_auth_cmdb::instance()->category(isys_auth::CREATE, strtolower($this->category_name));
        }

        try {
            $this->check_required_parameters(array('id'), $p_params);
        } catch (isys_exception_validation $ex) {
            return $this->return_error($this->module_name, 'generate_wires', 'could not generate wires - ' . $ex->getMessage());
        }
        
        $data = $p_params['data'];
        $required_parameters = array (
            'generator_id',
            'start_pattern',
            'end_pattern',
        );
        try {
            $this->check_required_parameters($required_parameters, $data);
        } catch (isys_exception_validation $ex) {
            return $this->return_error($this->module_name, 'generate_wires', 'could not generate wires - ' . $ex->getMessage());
        }
        
        $generator_id = $data['generator_id'];
        $start_pattern = $data['start_pattern'];
        $end_pattern = $data['end_pattern'];
        
        // check if object exists
        $dao = isys_cmdb_dao_category_g_global::instance(isys_application::instance()->database);
        $object = $dao->get_object_by_id($p_params['id']);
        if (empty($object)) {
            return $this->return_error($this->module_name, 'generate_wires', 'could not find object');
        }
        
        $cable_catg = new isys_cmdb_dao_category_g_smview_cable(isys_application::instance()->database);
        try {
            $cable_catg->generate_wires($p_params['id'], $generator_id, $start_pattern, $end_pattern);
        } catch (InvalidArgumentException $e) {
            return $this->return_error($this->module_name, 'generate_wires', 'could not generate wires - ' . $e->getMessage());
        }
        return $this->return_success($this->module_name, 'generate_wires', $p_params['id']);
    }

    /**
     * Find usages of an element.
     *
     * @param $p_params
     *
     * $p_param structure:
     * array (
     *     'id' => int
     * )
     *
     * @return array
     * @throws isys_exception_database
     */
    public function find_usages($p_params)
    {
        if ($this->useAuth && ! empty($this->category_name)) {
            isys_auth_cmdb::instance()->category(isys_auth::VIEW, strtolower($this->category_name));
        }

        $wire_catg = new isys_cmdb_dao_category_g_smview_cable_wire(isys_application::instance()->database);
        return $wire_catg->find_usages($p_params['id']);
    }

    /**
     * Delete wire.
     *
     * @param array $p_params
     *
     * $p_param structure:
     * array (
     *     'id' => int
     * )
     *
     * @return array
     */
    public function delete($p_params)
    {
        // check if a wire is in use by a section or a route
        $wire_catg = new isys_cmdb_dao_category_g_smview_cable_wire(isys_application::instance()->database);
        $language = isys_application::instance()->container->get('language');
        $in_use = $wire_catg->get_in_use($p_params['id']);
        if (! empty($in_use) && $in_use == $language->get('LC__CMDB__CATG__SMVIEW_CABLE__WIRE_IS_WIRED')) {
            return $this->return_error($this->module_name, 'delete', 'can not delete wire because it is in use by a section or a route');
        }
        return $this->_delete('id', $p_params);
    }
    
    /**
     * Delete all wires of an object.
     *
     * @param array $p_params
     *
     * $p_param structure:
     * array (
     *     'id' => int
     * )
     *
     * @return array
     */
    public function delete_all($p_params)
    {
        if ($this->useAuth && ! empty($this->category_name)) {
            isys_auth_cmdb::instance()->category(isys_auth::DELETE, strtolower($this->category_name));
        }

        $required_parameters = array (
            'id',
        );
        try {
            $this->check_required_parameters($required_parameters, $p_params);
        } catch (isys_exception_validation $ex) {
            return $this->return_error($this->module_name, 'delete_all', 'could not delete all - ' . $ex->getMessage());
        }
        
        // check if object exists
        $objects = $this->dao->filter_objects(array('obj_id' => $p_params['id']));
        if (empty($objects)) {
            return $this->return_error($this->module_name, 'delete_all', 'no wires assigned to this object');
        }
        
        $wire_catg = new isys_cmdb_dao_category_g_smview_cable_wire(isys_application::instance()->database);
        $language = isys_application::instance()->container->get('language');
        foreach ($objects as $object) {
            // check if a wire is in use by a section or a route
            $in_use = $wire_catg->get_in_use($object['id']);
            if (! empty($in_use) && $in_use == $language->get('LC__CMDB__CATG__SMVIEW_CABLE__WIRE_IS_WIRED')) {
                return $this->return_error($this->module_name, 'delete_all', 'can not delete wire because it is in use by a section or a route');
            }
            $this->dao->delete_object($object['id']);
        }
        return $this->return_success($this->module_name, 'delete_all', null);
    }

    public function properties($p_params)
    {
        if ($this->useAuth && ! empty($this->category_name)) {
            isys_auth_cmdb::instance()->category(isys_auth::VIEW, strtolower($this->category_name));
        }

        $dao = isys_cmdb_dao_smview_cable_properties::instance(isys_application::instance()->database);
        $properties = [
            "p_catg_id" => ["C__CATG__SMVIEW_CABLE_WIRE"]
        ];
        return $dao->query($properties);
    }
}
