import * as React from 'react'

interface AlgorithmItem {
    index: number,
    pos: number,
    length: number,
    type: string,
    description: string,
    elements: string,
}

interface Generator {
    id?: number,
    title: string,
    description: string,
    mask: string,
    counter_min: number,
    counter_max: number,
    algorithm_items: AlgorithmItem[],
}

export interface AlgorithmEditorComposeProps {
    language: {[key: string]: string},
    generator: Generator,
    is_editable: boolean,
}

interface AlgorithmEditorComposeState {
    generator: Generator,
}

export class AlgorithmEditorCompose extends React.Component<AlgorithmEditorComposeProps, AlgorithmEditorComposeState> {
    constructor (props: AlgorithmEditorComposeProps) {
        super(props)

        const generator = this.props.generator && Object.keys(this.props.generator).length > 0 ? this.props.generator : {
            title: '',
            description: '',
            mask: '',
            counter_min: 1,
            counter_max: 1,
            algorithm_items: [],
        }
        this.state = {
            generator: generator,
        }

        this.handleGeneratorChange = this.handleGeneratorChange.bind(this)
        this.handleAlgorithmItemChange = this.handleAlgorithmItemChange.bind(this)
        this._saveGenerator = this._saveGenerator.bind(this)
    }

    setElementsByMask(mask: string) {
        let count = (mask.match(/#/g) || []).length
        const elements = this.state.generator.algorithm_items
        let new_elements: AlgorithmItem[] = []
        for (let i = count - 1; i >= 0; i--) {
            let pos = count - i
            if (elements[i]) {
                // update element position
                let e = elements[i]
                e.pos = pos;
                new_elements.push(e)
            } else {
                // add new element
                new_elements.push({
                    index: i + 1,
                    pos: pos,
                    type: 'n',
                    length: 2,
                    description: '',
                    elements: ''
                })
            }
        }

        let generator = this.state.generator
        generator.algorithm_items = new_elements
        this._saveGenerator(generator)
    }

    handlePosUpdate(index: number, newPos1: string) {
        const newPos = parseInt(newPos1)
        if (isNaN(newPos)) {
            // value is not a valid integer
            return
        }

        let generator = this.state.generator
        let elementCount = generator.algorithm_items.length
        if (newPos < 1 || newPos > elementCount) {
            // value does not fit into range of positions
            return
        }

        // get current position of this element
        const elementsPosCurrent = generator.algorithm_items.filter(e => e.index == index)
        const oldPos = elementsPosCurrent ? elementsPosCurrent[0].pos : null

        // search current element on new position
        const elementsPosNew = generator.algorithm_items.filter(e => e.pos == newPos)
        const indexOfNewPos = elementsPosNew ? elementsPosNew[0].index : null

        // update positions
        let items: AlgorithmItem[] = []
        generator.algorithm_items.forEach((item: AlgorithmItem) => {
            if (item.index == index) {
                item.pos = newPos
            }
            if (item.index == indexOfNewPos) {
                item.pos = oldPos
            }
            items.push(item)
        })

        generator.algorithm_items = items
        this._saveGenerator(generator)
    }

    handleGeneratorChange (e: React.ChangeEvent<HTMLInputElement>) {
        let id = e.currentTarget.id
        let generator = this.state.generator
        if (id == 'mask') {
            this.setElementsByMask(e.currentTarget.value)
        }
        Object.assign(generator, { [id]: e.currentTarget.value })
        this._saveGenerator(generator)
    }

    handleAlgorithmItemChange (e: React.ChangeEvent<HTMLInputElement|HTMLSelectElement>) {
        let generator = this.state.generator
        let parts = e.currentTarget.id.split('_')
        const index = parseInt(parts[1])
        const property = parts[0]

        let updatedItems: AlgorithmItem[] = []
        generator.algorithm_items.forEach((item: AlgorithmItem) => {
            if (item.index == index) {
                if (property == 'pos') {
                    this.handlePosUpdate(index, e.currentTarget.value)
                } else {
                    Object.assign(item, { [property]: e.currentTarget.value })
                }
            }
            updatedItems.push(item)
        })

        generator.algorithm_items = updatedItems
        this._saveGenerator(generator)
    }

    _saveGenerator (generator: Generator) {
        this.setState((prevState) => ({
            ...prevState,
            generator,
        }), () => {
            // save json to DOM
            const json = JSON.stringify(this.state.generator)
            document.getElementById('C__CATG__SMVIEW_CABLE_GENERATOR').setAttribute('value', json)
        });
    }

    renderInputRow (identifier: string, label: string, value: string) {
        return <tr>
            <td className="key">{label}</td>
            <td className="value">
                <div className="ml20">
                    <input id={identifier} className="input" value={value} disabled={! this.props.is_editable} onChange={this.handleGeneratorChange} />
                </div>
            </td>
        </tr>
    }

    renderAlgorithmItemRow (i: AlgorithmItem) {
        return <tr id={i.index.toString()} key={i.index.toString()}>
            <td>
                {i.index}
            </td>
            <td>
                <input id={'pos_' + i.index}
                       className="input"
                       value={i.pos}
                       disabled={! this.props.is_editable}
                       onChange={this.handleAlgorithmItemChange}
                       type="number"
                       min={1}
                       max={Object.keys(this.state.generator.algorithm_items).length} />
            </td>
            <td>
                <select id={'type_' + i.index}
                        className="input input-small"
                        value={i.type}
                        disabled={! this.props.is_editable}
                        onChange={this.handleAlgorithmItemChange}>
                    <option value="n">n</option>
                    <option value="e">e</option>
                </select>
            </td>
            <td>
                <input id={'description_' + i.index}
                       className="input"
                       value={i.description}
                       disabled={! this.props.is_editable}
                       onChange={this.handleAlgorithmItemChange} />
            </td>
            <td>
                <input id={'length_' + i.index}
                       className="input"
                       value={i.length}
                       disabled={! this.props.is_editable}
                       onChange={this.handleAlgorithmItemChange}
                       type="number"
                       min={1}
                       max={8} />
            </td>
            <td>
                <input id={'elements_' + i.index}
                       className="input"
                       value={i.elements}
                       disabled={! this.props.is_editable}
                       onChange={this.handleAlgorithmItemChange} />
            </td>
        </tr>
    }

    render () {
        const { language } = this.props
        const { generator } = this.state

        let rows: JSX.Element[] = []
        if (generator.algorithm_items) {
            // sort by index
            const items = generator.algorithm_items.sort((a, b) => a.index > b.index ? 1 : -1)
            items.forEach(i => {
                rows.push(this.renderAlgorithmItemRow(i))
            })
        }

        return <table className="contentTable" id="algorithm_generator">
            <tbody>
                {this.renderInputRow('title', language.LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_TITLE, generator.title)}
                {this.renderInputRow('description', language.LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_DESCRIPTION, generator.description)}
                {this.renderInputRow('mask', language.LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_MASK, generator.mask)}
                {this.renderInputRow('counter_min', language.LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_COUNTER_MIN, generator.counter_min.toString())}
                {this.renderInputRow('counter_max', language.LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_COUNTER_MAX, generator.counter_max.toString())}
                <tr>
                    <td className="key">{language.LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_ALGORITHMS}</td>
                    <td className="value">
                        <div id="algorithms">
                            <table id='algorithms_tab'>
                                <thead>
                                    <tr>
                                        <th>{language.LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_ALGORITHM_INDEX}</th>
                                        <th>{language.LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_ALGORITHM_MASK_POSITION}</th>
                                        <th>{language.LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_ALGORITHM_TYPE}</th>
                                        <th>{language.LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_DESCRIPTION}</th>
                                        <th>{language.LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_ALGORITHM_LENGTH}</th>
                                        <th>{language.LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_ALGORITHM_ELEMENTS}</th>
                                    </tr>
                                </thead>
                                <tbody>
                                    {rows}
                                </tbody>
                            </table>
                        </div>
                    </td>
                </tr>
            </tbody>
        </table>
    }
}
