import * as React from 'react'
import { BayItemCell } from './BayItemCell'
import { BayItemProperties } from './BayItemProperties'

interface Bay {
  id: string,
  title: string,
  he: number,
  use_half_he?: boolean,
  items?: BayItem[],
}

interface WireBundle {
  cable_name: string,
  cable_url: string,
  wire_start: number,
  wire_end: number,
  wire_state: string,
  device_colors: {[key: number]: string},
  sibling_object_title: string,
  sibling_pin_scope: string,
  sibling_object_url: string,
}

export interface BayItem {
  id: string,
  title: string,
  position?: number,
  color?: number,
  short_name?: string,
  he?: number,
  project_identifier?: string,
  comment?: string,
  status?: string,
  wire_bundles?: WireBundle[],
}

export interface ColorLayout {
  id: number,
  title: string,
  layout: number,
  color_left_top: string,
  color_left_bottom?: string,
  color_right_top?: string,
  color_right_bottom?: string,
}

export interface BayEditorComposeProps {
  language: {[key: string]: string},
  is_editable: boolean,
  bays: Bay[],
  color_layouts: ColorLayout[],
  bay?: Bay,
}

interface BayEditorComposeState {
  available_bay_items: BayItem[],
  assigned_bay_items: BayItem[],
  selected_bay: Bay,
  selected_bay_item: BayItem,
  rack_he: number,
}

export class BayEditorCompose extends React.Component<BayEditorComposeProps, BayEditorComposeState> {
  constructor (props: BayEditorComposeProps) {
    super(props)

    let selected_bay = this.props.bays ? this.props.bays[0] : null
    if (this.props.bay) {
      selected_bay = this.props.bays.filter(b => b.id == this.props.bay.id)[0]
    }
    let he = this.props.bay && this.props.bay.he ? this.props.bay.he : null
    if (!he) {
      he = selected_bay ? selected_bay.items.length : 0
    }
    this.state = {
      available_bay_items: selected_bay ? selected_bay.items : [],
      assigned_bay_items: this.props.bay && this.props.bay.items ? this.props.bay.items : [],
      selected_bay: selected_bay,
      selected_bay_item: null,
      rack_he: he
    }
    this.getUnassignedBayItems = this.getUnassignedBayItems.bind(this)
    this.handleDataChange = this.handleDataChange.bind(this)
    this.handleBaySelection = this.handleBaySelection.bind(this)
    this.handleRackHeChange = this.handleRackHeChange.bind(this)
    this.handleAutoAssignment = this.handleAutoAssignment.bind(this)
    this.handleAllowDrop = this.handleAllowDrop.bind(this)
    this.handleDrop = this.handleDrop.bind(this)
    this.handleBayItemSelection = this.handleBayItemSelection.bind(this)
    this.handleBayItemRemove = this.handleBayItemRemove.bind(this)
    this.handleChangedBayItem = this.handleChangedBayItem.bind(this)
    this.handleAddPlannedItem = this.handleAddPlannedItem.bind(this)
    this.updateState = this.updateState.bind(this)
    this.saveToDom = this.saveToDom.bind(this)
  }

  getUnassignedBayItems () {
    const available_bay_items: BayItem[] = []
    this.state.available_bay_items.forEach(b => {
      const assigned_bay_items = this.state.assigned_bay_items.filter(a => a.id == b.id)
      if (assigned_bay_items.length == 0) {
        available_bay_items.push(b)
      }
    })
    return available_bay_items
  }

  handleDataChange (property_name: string, value: string) {
    this.setState((prevState) => ({
      ...prevState,
      [property_name]: value
    }), () => {
      this.saveToDom()
    })
  }

  handleBaySelection (e: React.FormEvent<HTMLSelectElement>) {
    const bayId = e.currentTarget.value
    const selected_bay = this.props.bays.filter(b => b.id == bayId)[0]

    this.setState((prevState) => ({
      ...prevState,
      available_bay_items: selected_bay.items,
      assigned_bay_items: [],
      selected_bay: selected_bay,
      selected_bay_item: null,
      rack_he: selected_bay.items.length
    }), () => {
      this.saveToDom()
    })
  }

  handleRackHeChange (e: React.FormEvent<HTMLInputElement>) {
    const he = parseInt(e.currentTarget.value)
    if (he > 0) {
      this.handleDataChange('rack_he', e.currentTarget.value)
    }
  }

  handleAutoAssignment () {
    const rack_height = this.state.rack_he
    const assigned_bay_items: BayItem[] = []
    for (let i = 1; i <= rack_height; i++) {
      const assigned_bays = this.state.assigned_bay_items.filter(b => b.position == i)
      if (assigned_bays.length > 0) {
        // this position is already assigned
        const assigned_bay = assigned_bays[0]
        assigned_bay_items.push(assigned_bay)
        continue
      }

      const available_bays = this.getUnassignedBayItems().filter(b => {
        const parts = b.id.match(/\d+/g)
        const pos = parseInt(parts[0])
        return pos == i
      })

      if (available_bays.length > 0) {
        const available_bay = available_bays[0]
        available_bay.position = i
        assigned_bay_items.push(available_bay)
      }
    }

    this.updateState(assigned_bay_items, this.state.selected_bay_item)
  }

  handleStartDrag (e: React.DragEvent<HTMLDivElement>) {
    e.dataTransfer.setData('text', e.currentTarget.id)
  }

  handleAllowDrop (e: React.DragEvent<HTMLDivElement>) {
    if (this.props.is_editable) {
      e.preventDefault()
    }
  }

  handleDrop (e: React.DragEvent<HTMLDivElement>) {
    if (!this.props.is_editable) {
      return
    }
    e.preventDefault()
    const bayId = e.dataTransfer.getData('text')

    const available_bay_items = this.state.available_bay_items.filter(b => b.id == bayId)
    if (available_bay_items.length == 0) {
      return
    }

    let bay = available_bay_items[0]
    const assigned_items = this.state.assigned_bay_items.filter(b => b.id == bayId)
    if (assigned_items.length > 0) {
      // item is already assigned
      bay = assigned_items[0]
    }

    // get position string ("row_1") from parent tr-element
    const posString = e.currentTarget.parentElement.parentElement.id
    // remove leading string part "row_"
    const pos = parseInt(posString.substring(4))
    bay.position = pos

    // get all assigned items except the current bay item
    let assigned_bay_items = this.state.assigned_bay_items.filter(b => b.id != bayId)

    // try to get a planned element at this position
    const planned_items = this.state.assigned_bay_items.filter(b => b.position == pos)
    if (planned_items.length > 0) {
      // merge with planned item
      const planned_item = planned_items[0]
      bay.project_identifier = planned_item.project_identifier
      bay.color = planned_item.color
      bay.comment = planned_item.comment

      // get all assigned items except the planned item at this position
      assigned_bay_items = assigned_bay_items.filter(b => b.position != pos)
    }
    assigned_bay_items.push(bay)
    this.updateState(assigned_bay_items, this.state.selected_bay_item)
  }

  handleBayItemSelection (bayId: string) {
    let bay: BayItem = null
    if (!this.state.selected_bay_item || this.state.selected_bay_item.id != bayId) {
      const bays = this.state.assigned_bay_items.filter(b => b.id == bayId)
      bay = bays.length > 0 ? bays[0] : null
    }
    this.setState((prevState) => ({
      ...prevState,
      selected_bay_item: bay
    }))
  }

  handleBayItemRemove (e: React.MouseEvent<HTMLButtonElement>) {
    const id = e.currentTarget.id
    const bayItemId = id.substring(7)
    const assigned_bay_items = this.state.assigned_bay_items.filter(b => b.id != bayItemId)
    this.updateState(assigned_bay_items, null)
  }

  handleChangedBayItem (bayItem: BayItem) {
    const assigned_bay_items: BayItem[] = []
    this.state.assigned_bay_items.forEach(b => {
      if (b.id == bayItem.id) {
        assigned_bay_items.push(bayItem)
        return
      }
      assigned_bay_items.push(b)
    })

    this.updateState(assigned_bay_items, bayItem)
  }

  handleAddPlannedItem (e: React.MouseEvent<HTMLButtonElement>) {
    const assigned_bay_items = this.state.assigned_bay_items
    const pos = parseInt(e.currentTarget.value)
    const prefix = '_planned_'

    // get next planning ID
    let nextId = null
    for (let i = 1; i <= 100; i++) {
      const itemId = prefix + i
      const assigned_items = assigned_bay_items.filter(b => b.id == itemId)
      if (assigned_items.length == 0) {
        nextId = itemId
        break
      }
    }

    const bayItem: BayItem = {
      id: nextId,
      title: this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_BAY_PLANNED_ITEM,
      position: pos,
      he: 1,
      project_identifier: '',
      comment: '',
      status: 'planning'
    }

    assigned_bay_items.push(bayItem)
    this.updateState(assigned_bay_items, bayItem)
  }

  updateState (assigned_bay_items: BayItem[], bayItem: BayItem) {
    this.setState((prevState) => ({
      ...prevState,
      assigned_bay_items: assigned_bay_items,
      selected_bay_item: bayItem
    }), () => {
      this.saveToDom()
    })
  }

  saveToDom () {
    // save json to DOM
    const json = JSON.stringify({
      bay_id: this.state.selected_bay ? this.state.selected_bay.id : null,
      assigned_bay_items: this.state.assigned_bay_items,
      he: this.state.rack_he
    })
    document.getElementById('C__CATG__SMVIEW_CABLE_ETSI_BAY_ITEMS').setAttribute('value', json)
  }

  render () {
    let bay_selector = null
    let items = null
    if (this.props.is_editable) {
      const options: JSX.Element[] = []
      this.props.bays.forEach(b => {
        options.push(<option key={'bay_' + b.id} value={b.id}>{b.title}</option>)
      })

      let baySelect = <select onChange={this.handleBaySelection} defaultValue={this.state.selected_bay ? this.state.selected_bay.id : null} className="input">
        {options}
      </select>

      if (this.props.bay) {
        baySelect = <span style={{ paddingLeft: '16px' }}>{this.props.bay.id}</span>
      }

      bay_selector = <div id='bay_editor_selector' key='bay_editor_selector'>
        <table className="contentTable">
          <tbody>
            <tr>
              <td className="key">{this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_BAY}</td>
              <td className="value">
                {baySelect}
              </td>
            </tr>
          </tbody>
        </table>
      </div>

      const available_bay_items: JSX.Element[] = []
      this.getUnassignedBayItems().forEach(b => {
        available_bay_items.push(<div draggable={true}
          onDragStart={this.handleStartDrag}
          id={b.id}
          key={'available_' + b.id}
          className='bay-editor-item'>{b.title}</div>)
      })

      const assign_button = available_bay_items.length == 0 ? null : <button type='button' className="btn" onClick={this.handleAutoAssignment}>
        {this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_AUTO_ASSIGNMENT}
      </button>
      items = <div id='bay_editor_items' key='bay_editor_items'>
        <div>
          {this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_AVAILABLE_BAY_ITEMS}:<br /><br />
          {available_bay_items}
          <br />
          {assign_button}
        </div>
      </div>
    }

    const rack_height = this.state.rack_he
    const assigned_rows = []
    const assigned_hes = []
    for (let i = 1; i <= rack_height; i++) {
      let heName: string = i.toString()
      const bays = this.state.assigned_bay_items.filter(b => b.position == i)
      const assigned_bay = bays.length > 0 ? bays[0] : null

      let div = <div className='rack-slot' id={'pos_' + i} onDrop={this.handleDrop} onDragOver={this.handleAllowDrop}>&nbsp;</div>
      let classPlanning = ''
      let classSelected = ''
      let mark = null
      let heightHe = 1
      if (assigned_bay) {
        classSelected = this.state.selected_bay_item && assigned_bay.id == this.state.selected_bay_item.id ? 'selected' : ''
        if (assigned_bay.project_identifier) {
          classPlanning = 'planning'
          heName = assigned_bay.project_identifier
        }

        heightHe = assigned_bay.he
        mark = assigned_bay.comment && assigned_bay.status != 'planning' ? <span className='exclamation-mark' title={assigned_bay.comment}>!</span> : null
        assigned_hes.push(i)
        div = <BayItemCell language={this.props.language}
          is_selected={this.state.selected_bay_item && assigned_bay.id == this.state.selected_bay_item.id}
          bay_item={assigned_bay}
          onSelectionChanged={this.handleBayItemSelection}
          colorLayouts={this.props.color_layouts}
          onDrop={assigned_bay.status == 'planning' ? this.handleDrop : null}
          onDragOver={assigned_bay.status == 'planning' ? this.handleAllowDrop : null} />
      }
      if (!assigned_bay && this.props.is_editable) {
        // add planning button
        div = <div className='rack-slot-add-planning' onDrop={this.handleDrop} onDragOver={this.handleAllowDrop}>
          <button type='button' className='btn' value={i} onClick={this.handleAddPlannedItem}>+</button>
        </div>
      }

      let button = null
      if (this.props.is_editable) {
        button = assigned_bay ? <button type='button' id={'delete_' + assigned_bay.id} onClick={this.handleBayItemRemove}>X</button> : null
      }
      const count_first_row = !assigned_bay || !assigned_bay.project_identifier
      assigned_rows.push(<tr key={'tr_' + i} id={'row_' + i}>
        <td className={classPlanning} rowSpan={count_first_row ? 1 : heightHe}>{heName}</td>
        <td className={classSelected} rowSpan={heightHe}>{div}</td>
        <td rowSpan={heightHe}>{mark} {button}</td>
      </tr>)
      if (heightHe > 1) {
        for (let j = 1; j < heightHe; j++) {
          i++
          assigned_hes.push(i)
          const first_td = count_first_row ? <td className={classPlanning}>{i}</td> : null
          assigned_rows.push(<tr key={'tr_' + i} id={'row_' + i}>
            {first_td}
          </tr>)
        }
      }
    }

    let thead = null
    if (this.props.is_editable) {
      thead = <thead>
        <tr>
          <td>{this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__HE}</td>
          <td colSpan={2}><input type='number' className='input input-small' value={this.state.rack_he} onChange={this.handleRackHeChange} /></td>
        </tr>
      </thead>
    }
    const rack = <div id='bay_editor_rack' key='bay_editor_rack'>
      <table>
        {thead}
        <tbody>
          {assigned_rows}
        </tbody>
      </table>
    </div>

    let properties = null
    if (this.state.selected_bay_item) {
      // set default values if needed
      const item = this.state.selected_bay_item
      item.short_name = item.short_name ? item.short_name : ''
      item.comment = item.comment ? item.comment : ''
      item.project_identifier = item.project_identifier ? item.project_identifier : ''
      item.color = item.color ? item.color : -1
      item.wire_bundles = item.wire_bundles ? item.wire_bundles : []

      properties = <BayItemProperties language={this.props.language}
        isEditable={this.props.is_editable}
        bayItem={item}
        onChangedBayItem={this.handleChangedBayItem}
        colorLayouts={this.props.color_layouts}
        bayHe={this.state.rack_he}
        assignedHes={assigned_hes} />
    }

    return <div id='bay_editor' key='bay_editor'>
      {bay_selector}
      {items}
      {rack}
      {properties}
    </div>
  }
}
