import * as React from 'react'
import { BayItem, ColorLayout } from './BayEditorCompose'

export interface BayItemCellProps {
  language: {[key: string]: string},
  is_selected: boolean,
  bay_item: BayItem,
  colorLayouts: ColorLayout[],
  onSelectionChanged: Function,
  onDrop?: Function,
  onDragOver?: Function,
}

export class BayItemCell extends React.Component<BayItemCellProps, {}> {
  constructor (props: BayItemCellProps) {
    super(props)
    this.handleSelection = this.handleSelection.bind(this)
    this.handleAllowDrop = this.handleAllowDrop.bind(this)
    this.handleDrop = this.handleDrop.bind(this)
  }

  handleStartDrag (e: React.DragEvent<HTMLDivElement>) {
    e.dataTransfer.setData('text', e.currentTarget.id)
  }

  handleAllowDrop (e: React.DragEvent<HTMLDivElement>) {
    if (this.props.onDragOver) {
      this.props.onDragOver(e)
    }
  }

  handleDrop (e: React.DragEvent<HTMLDivElement>) {
    if (this.props.onDrop) {
      this.props.onDrop(e)
    }
  }

  handleSelection () {
    this.props.onSelectionChanged(this.props.bay_item.id)
  }

  render () {
    const colors = this.props.colorLayouts.filter(c => c.id == this.props.bay_item.color)
    const defaultColor: ColorLayout = {
      id: -1,
      title: '',
      layout: 1,
      color_left_top: '#F8F8F8'
    }

    const CELL_HEIGHT = 30
    const BORDER_WIDTH = 1
    let fullCellHeight = CELL_HEIGHT * this.props.bay_item.he
    let halfCellHeight = fullCellHeight / 2
    fullCellHeight -= this.props.is_selected ? (2 * BORDER_WIDTH) : 0
    halfCellHeight -= this.props.is_selected ? BORDER_WIDTH : 0

    // left div with information
    const color = colors.length > 0 ? colors[0] : defaultColor
    let prevCableName: string = null
    const cableNames: JSX.Element[] = []
    let prevObjectName: string = null
    const objectNames: JSX.Element[] = []
    let wire: string = null
    let prevWire: number = null
    let wires: JSX.Element[] = []
    if (this.props.bay_item.wire_bundles && this.props.bay_item.wire_bundles.length > 0) {
      this.props.bay_item.wire_bundles.forEach(bundle => {
        if (!wire) {
          // setup first bundle
          cableNames.push(<p key={bundle.cable_name}>{bundle.cable_name}</p>)
          prevCableName = bundle.cable_name
          objectNames.push(<p key={bundle.sibling_object_title}>{bundle.sibling_object_title}</p>)
          prevObjectName = bundle.sibling_object_title
          wire = this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__FIBRE_WIRE + ': ' + bundle.wire_start
        }

        if (prevCableName != bundle.cable_name || prevObjectName != bundle.sibling_object_title) {
          // save bundle
          cableNames.push(<p key={bundle.cable_name}>{bundle.cable_name}</p>)
          prevCableName = bundle.cable_name

          objectNames.push(<p key={bundle.sibling_object_title}>{bundle.sibling_object_title}</p>)
          prevObjectName = bundle.sibling_object_title

          wire += '-' + prevWire
          wires.push(<p key={wire}>{wire}</p>)
          wire = this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__FIBRE_WIRE + ': ' + bundle.wire_start
        }
        prevWire = bundle.wire_end
      })
      // save last bundle
      wire += '-' + prevWire
      wires.push(<p key={wire}>{wire}</p>)
    }

    let leftBottom = null
    const styleLeftBottom = { backgroundColor: color.color_left_bottom, height: halfCellHeight }
    if (color.layout == 4) {
      leftBottom = <div className="rack-slot-left-100" style={styleLeftBottom}>&nbsp;</div>
    }

    const styleLeftTop = { height: halfCellHeight }
    let leftDiv = <div className='rack-slot-left' style={styleLeftTop}>
      <div className='rack-slot-left-33' style={styleLeftTop}>{cableNames}</div>
      <div className='rack-slot-left-33'>{objectNames}</div>
      <div className='rack-slot-left-33'>{this.props.bay_item.short_name}</div>
      {leftBottom}
    </div>

    // right div(s)
    let className = 'rack-slot assigned ' + (this.props.is_selected ? 'selected' : '')
    let styleLeft = { backgroundColor: color.color_left_top, height: fullCellHeight }
    let subDivs = null

    if (this.props.bay_item.status == 'planning') {
      wires = []
      className = 'rack-slot planning ' + (this.props.is_selected ? 'selected' : '')
      const c = color.color_left_top != '' && color.color_left_top != '#F8F8F8' ? color.color_left_top : '#BBB'
      styleLeft = { backgroundColor: c, height: fullCellHeight }
      leftDiv = <div>{this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_BAY_PLANNED_ITEM}</div>
    }

    if (color.layout == 1) {
      subDivs = <div className='rack-slot-right'>{wires}</div>
    }
    if (color.layout == 2) {
      const styleRightTop = { backgroundColor: color.color_right_top, height: fullCellHeight }
      subDivs = <div className='rack-slot-right' style={styleRightTop}>{wires}</div>
    }

    if (color.layout > 2) {
      const styleRight = { backgroundColor: color.color_right_top, height: fullCellHeight }
      const styleRightTop = { backgroundColor: color.color_right_top, height: halfCellHeight }
      const styleRightBottom = { backgroundColor: color.color_right_bottom, height: halfCellHeight }
      subDivs = <div className='rack-slot-right' style={styleRight}>
        <div className='rack-slot-right-top' style={styleRightTop}>{wires}</div>
        <div className='rack-slot-right-bottom' style={styleRightBottom} />
      </div>
    }

    return <div id={this.props.bay_item.id}
      draggable={true}
      onDragStart={this.handleStartDrag}
      onDrop={this.handleDrop}
      onDragOver={this.handleAllowDrop}
      className={className}
      title={this.props.bay_item.comment}
      style={styleLeft} onClick={this.handleSelection}>
      {subDivs}
      {leftDiv}
    </div>
  }
}
