import * as React from 'react'
import { BayItem, ColorLayout } from './BayEditorCompose'

export interface BayItemPropertiesProps {
  language: {[key: string]: string},
  isEditable: boolean,
  bayItem: BayItem,
  onChangedBayItem: Function,
  colorLayouts: ColorLayout[],
  bayHe: number,
  assignedHes: number[],
}

interface BayItemPropertiesState {
  error?: {[key: string]: string}
}

export class BayItemProperties extends React.Component<BayItemPropertiesProps, BayItemPropertiesState> {
  constructor (props: BayItemPropertiesProps) {
    super(props)

    this.state = {}
    this.handleChange = this.handleChange.bind(this)
    this.handleColorChange = this.handleColorChange.bind(this)
    this.renderWireBundles = this.renderWireBundles.bind(this)
  }

  handleChange (e: React.FormEvent<HTMLInputElement>) {
    // validate
    let errorMessage = ''
    switch (e.currentTarget.name) {
      case 'position':
        if (e.currentTarget.value == '') {
          errorMessage = this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_ERROR_NOT_EMPTY
        }

        const new_pos = parseInt(e.currentTarget.value)
        if (new_pos <= 0 || new_pos > this.props.bayHe) {
          errorMessage = this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_ERROR_NOT_BETWEEN
          errorMessage += ' 0 - ' + this.props.bayHe
        }

        this.props.assignedHes.forEach(he => {
          if (new_pos == he) {
            errorMessage = this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_ERROR_ALREADY_ASSIGNED
          }
        })
        break
      case 'he':
        if (e.currentTarget.value == '') {
          errorMessage = this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_ERROR_NOT_EMPTY
        }

        // cast values to int explicitly
        const pos = parseInt(this.props.bayItem.position.toString())
        const he = parseInt(e.currentTarget.value)
        if (he <= this.props.bayItem.he) {
          // new HE is lower
          break
        }

        if (he <= 0 || (pos + he - 1) > this.props.bayHe) {
          // HE outside of rack
          errorMessage = this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_ERROR_NOT_BETWEEN
          errorMessage += ' 0 - ' + this.props.bayHe
          break
        }

        // cast values to int explicitly
        const p: number = pos + parseInt(this.props.bayItem.he.toString())
        const heDiff = he - parseInt(this.props.bayItem.he.toString())

        // check HEs within new range
        for (let i = p; i < (p + heDiff); i++) {
          this.props.assignedHes.forEach(he => {
            if (i == he) {
              errorMessage = this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_ERROR_HE_IN_USE
            }
          })
        }
        break
    }

    // handle errors
    const error = errorMessage.length > 0 ? { [e.currentTarget.name]: errorMessage } : {}
    this.setState((prevState) => ({
      ...prevState,
      error: error
    }))
    if (errorMessage.length > 0) {
      // display error and exit
      return
    }

    // update bay item on success
    const bayItem = this.props.bayItem
    switch (e.currentTarget.name) {
      case 'position':
        const pos = parseInt(e.currentTarget.value)
        bayItem.position = pos
        break
      case 'he':
        bayItem.he = parseInt(e.currentTarget.value)
        break
      case 'short_name':
        bayItem.short_name = e.currentTarget.value
        break
      case 'project_identifier':
        bayItem.project_identifier = e.currentTarget.value
        break
      case 'comment':
        bayItem.comment = e.currentTarget.value
        break
    }
    this.props.onChangedBayItem(bayItem)
  }

  handleColorChange (e: React.ChangeEvent<HTMLSelectElement|HTMLInputElement>) {
    const value = parseInt(e.currentTarget.value)
    const bayItem = this.props.bayItem
    bayItem.color = value > 0 ? value : null
    this.props.onChangedBayItem(bayItem)
  }

  renderWireBundles () {
    if (!this.props.bayItem.wire_bundles) {
      return []
    }

    const wireBundles: JSX.Element[] = []
    this.props.bayItem.wire_bundles.forEach(b => {
      const wireState = b.wire_state ? '(' + b.wire_state + ')' : ''
      const suggested_colors = this.props.isEditable ? this.getSuggestedColors(b.device_colors) : null
      wireBundles.push(
        <div key={'cable_' + b.cable_name + '_' + b.wire_start} className='wire-bundle'>
          <span>
            {this.props.language.LC__SHD__TELECOMMUNICATION_CABLE}:
          </span>
          <a href={b.cable_url} target='_blank'>{b.cable_name}</a>
              &nbsp;- {this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__FIBRE_WIRE} {b.wire_start}-{b.wire_end} {wireState}
          <br />

          <span>
            {this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__CONNECTED_WITH}:
          </span>
          <a href={b.sibling_object_url} target='_blank'>
            {b.sibling_object_title}
          </a>&nbsp;
          ({b.sibling_pin_scope})
          {suggested_colors}
        </div>
      )
    })
    return wireBundles
  }

  getSuggestedColors (device_colors: {}) {
    let showHint = this.props.bayItem.color && parseInt(this.props.bayItem.color.toString()) > 0
    const deviceColorItems: JSX.Element[] = []
    Object.entries(device_colors).forEach(([c, value]) => {
      if (this.props.bayItem.color && parseInt(c) == this.props.bayItem.color) {
        // current color is in suggested colors, no hint needed
        showHint = false
      }

      let input = null
      if (this.props.isEditable) {
        input = <input type="radio"
          name={'device_color_' + c}
          value={c}
          defaultChecked={this.props.bayItem.color && parseInt(c) == this.props.bayItem.color}
          onChange={this.handleColorChange} />
      }
      deviceColorItems.push(
        <li key={c}>
          {input}
          {value}
        </li>
      )
    })
    const suggested_color = []
    if (deviceColorItems.length > 0) {
      suggested_color.push(<ul key="device_colors">
        {deviceColorItems}
      </ul>)
      if (showHint) {
        suggested_color.push(<span key="no_match" className="hint">{this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__SUGGESTED_COLOR_NOT_MATCH}</span>)
      }
    } else {
      suggested_color.push(<span key="-">{this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__NO_COLOR_SUGGESTED}</span>)
    }
    return <div className="suggested-color">
      <span>{this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__SUGGESTED_COLORS}:</span>
      {suggested_color}
    </div>
  }

  render () {
    const colorOptions: JSX.Element[] = []
    colorOptions.push(<option key={'color_0'} value={0}>-</option>)

    // sort options by title
    const sortable = this.props.colorLayouts.sort((a: ColorLayout, b: ColorLayout) => b.title.toLowerCase() < a.title.toLowerCase() ? 1 : -1)
    sortable.forEach((c) => {
      colorOptions.push(<option key={'color_' + c.id} value={c.id}>{c.title}</option>)
    })
    const wireBundles = this.renderWireBundles()

    let divShortName = null
    let divWireBundles = null
    if (!this.props.bayItem.status || this.props.bayItem.status != 'planning') {
      divShortName = <div>
        <span>
          {this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__SHORT_NAME}
        </span>
        <div>
          <input type='text'
                 name='short_name'
                 className='input input-small'
                 value={this.props.bayItem.short_name}
                 onChange={this.handleChange}
                 disabled={!this.props.isEditable} />
        </div>
      </div>

      divWireBundles = <div>
        <span>
          {this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__WIRE_BUNDLES}
        </span>
        <div>
          {wireBundles}
        </div>
      </div>
    }

    const divColor = <div>
      <span>
        {this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_DEVICE_COLOR}
      </span>
      <div>
        <select onChange={this.handleColorChange}
                className='input input-small'
                value={this.props.bayItem.color}
                disabled={!this.props.isEditable}>
          {colorOptions}
        </select>
      </div>
    </div>

    return <div id='bay_editor_properties' key='bay_editor_properties'>
      <span><h3>{this.props.bayItem.title}</h3></span>
      <div>
        <span>
          {this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__HE}
        </span>
        <div>
          <input type='text'
                 name='he'
                 className='input input-small'
                 value={this.props.bayItem.he}
                 onChange={this.handleChange}
                 disabled={!this.props.isEditable} />
          {this.state.error && this.state.error.he ? <span className='error'>{this.state.error.he}</span> : null}
        </div>
      </div>
      <div>
        <span>
          {this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__POSITION}
        </span>
        <div>
          <input type='text'
                 name='position'
                 className='input input-small'
                 value={this.props.bayItem.position}
                 onChange={this.handleChange}
                 disabled={!this.props.isEditable} />
          {this.state.error && this.state.error.position ? <span className='error'>{this.state.error.position}</span> : null}
        </div>
      </div>
      {divShortName}
      <div>
        <span>
          {this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__PROJECT_IDENTIFIER}
        </span>
        <div>
          <input type='text'
                 name='project_identifier'
                 className='input input-small'
                 value={this.props.bayItem.project_identifier}
                 onChange={this.handleChange}
                 disabled={!this.props.isEditable} />
        </div>
      </div>
      <div>
        <span>
          {this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__COMMENT}
        </span>
        <div>
          <input type='text'
                 name='comment'
                 className='input input-small'
                 value={this.props.bayItem.comment}
                 onChange={this.handleChange}
                 disabled={!this.props.isEditable} />
        </div>
      </div>
      {divColor}
      {divWireBundles}
    </div>
  }
}
