import * as React from 'react'
import { loadAjax } from '../ajax'
import { addObjects, CommonObject, dropObject, PositionalObject, removeObject } from './Common'

export interface ExportEditorComposeProps {
  ajax_url: string,
  session: string,
  language: {[key: string]: string},
  spinner_url: string,
  is_editable: boolean,
  objects: PositionalObject[],
  subway_lines: CommonObject[],
}

interface ExportEditorComposeState {
  available_stations: CommonObject[],
  available_objects: PositionalObject[],
  assigned_objects: PositionalObject[],
  selected_station: CommonObject,
  selected_object: PositionalObject,
  stations_loading: boolean,
  objects_loading: boolean,
}

export class ExportEditorCompose extends React.Component<ExportEditorComposeProps, ExportEditorComposeState> {
  constructor (props: ExportEditorComposeProps) {
    super(props)

    this.state = {
      available_stations: [],
      available_objects: [],
      assigned_objects: this.props.objects ? this.props.objects : [],
      selected_station: null,
      selected_object: null,
      stations_loading: false,
      objects_loading: false
    }

    this.handleLocationChange = this.handleLocationChange.bind(this)
    this.handleStationSelection = this.handleStationSelection.bind(this)
    this.handleAllowDrop = this.handleAllowDrop.bind(this)
    this.handleDrop = this.handleDrop.bind(this)
    this.handleObjectDrop = this.handleObjectDrop.bind(this)
    this.handleAddObject = this.handleAddObject.bind(this)
    this.handleAddAllObjects = this.handleAddAllObjects.bind(this)
    this.handleRemoveObject = this.handleRemoveObject.bind(this)
    this._updateState = this._updateState.bind(this)
    this.saveToDom = this.saveToDom.bind(this)
  }

  handleLocationChange (e: React.ChangeEvent<HTMLSelectElement>) {
    const locationId = e.currentTarget.value
    this.setState((prevState) => ({
      ...prevState,
      stations_loading: true,
      available_objects: [],
      selected_station: null,
      selected_object: null
    }), () =>
      loadAjax(this.props.ajax_url + '&func=get-stations&location_id=' + locationId, this.props.session)
        .then(res => {
          this.setState((prevState) => ({
            ...prevState,
            available_stations: res.data,
            stations_loading: false
          }))
        })
    )
  }

  handleStationSelection (e: React.MouseEvent<HTMLDivElement>) {
    if (this.state.objects_loading) {
      return
    }

    const stationId = parseInt(e.currentTarget.id)
    const stations = this.state.available_stations.filter(c => c.id == stationId)
    const station = stations.length > 0 ? stations[0] : null
    this.setState((prevState) => ({
      ...prevState,
      selected_station: station,
      objects_loading: true,
      selected_object: null
    }), () => {
      loadAjax(this.props.ajax_url + '&func=get-objects&station_id=' + stationId, this.props.session)
        .then(res => {
          this.setState((prevState) => ({
            ...prevState,
            available_objects: res.data,
            objects_loading: false
          }))
        })
    })
  }

  handleStartStationDrag (e: React.DragEvent<HTMLDivElement>) {
    const data = {
      id: e.currentTarget.id,
      type: 'station'
    }
    e.dataTransfer.setData('text', JSON.stringify(data))
  }

  handleStartObjectDrag (e: React.DragEvent<HTMLDivElement>) {
    const data = {
      id: e.currentTarget.id,
      type: 'object'
    }
    e.dataTransfer.setData('text', JSON.stringify(data))
  }

  handleAllowDrop (e: React.DragEvent<HTMLDivElement>) {
    if (this.props.is_editable) {
      e.currentTarget.className = 'target-field active'
      e.preventDefault()
    }
  }

  handleDragLeave (e: React.DragEvent<HTMLDivElement>) {
    e.currentTarget.className = 'target-field'
  }

  handleDrop (e: React.DragEvent<HTMLDivElement>) {
    if (!this.props.is_editable) {
      return
    }
    e.currentTarget.className = 'target-field'
    e.preventDefault()

    const json = e.dataTransfer.getData('text')
    const data = JSON.parse(json)
    if (data.type == 'object') {
      this.handleObjectDrop(data.id, e.currentTarget.id)
    } else if (data.type == 'station') {
      // todo: handle station drop
    }
  }

  handleObjectDrop (objectId: number, posString: string) {
    const newAssignedItems = dropObject(objectId, posString, this.state.available_objects, this.state.assigned_objects)
    this._updateState(newAssignedItems)
  }

  handleAddObject (e: React.MouseEvent<HTMLDivElement>) {
    const objectId = parseInt(e.currentTarget.parentElement.id)
    const objects = this.state.available_objects.filter(r => r.id == objectId)
    const object = objects.length > 0 ? objects[0] : null
    if (object) {
      const assignedRacks = addObjects([object], this.state.assigned_objects)
      this._updateState(assignedRacks)
    }
  }

  handleAddAllObjects () {
    if (this.state.available_objects.length > 0) {
      const assignedRacks = addObjects(this.state.available_objects, this.state.assigned_objects)
      this._updateState(assignedRacks)
    }
  }

  handleRemoveObject (e: React.MouseEvent<HTMLButtonElement>) {
    const object_id = parseInt(e.currentTarget.parentElement.id)
    const assignedObjects = removeObject(object_id, this.state.assigned_objects)
    this._updateState(assignedObjects)
  }

  _updateState (assignedObjects: PositionalObject[]) {
    this.setState((prevState) => ({
      ...prevState,
      assigned_objects: assignedObjects
    }), () => {
      this.saveToDom()
    })
  }

  saveToDom () {
    // save json to DOM
    const json = JSON.stringify({
      objects: this.state.assigned_objects
    })
    document.getElementById('C__CATG__SMVIEW_CABLE_ETSI_EXPORT').setAttribute('value', json)
  }

  render () {
    const subwayOptions: JSX.Element[] = []
    this.props.subway_lines.forEach(subway => {
      subwayOptions.push(<option
        key={'subway_' + subway.id}
        value={subway.id}>{subway.name}</option>)
    })

    const stations: JSX.Element[] = []
    if (this.state.stations_loading) {
      stations.push(<img key={'station_loading'} src={this.props.spinner_url} alt="Loading..." />)
    } else {
      const av_stations = this.state.available_stations ? this.state.available_stations : []
      av_stations.forEach(station => {
        const clsName = this.state.selected_station && station.id == this.state.selected_station.id ? 'selected' : ''
        stations.push(<div
          key={'station_' + station.id}
          draggable={false}
          onDragStart={this.handleStartStationDrag}
          id={station.id.toString()}
          onClick={this.handleStationSelection}
          className={clsName}>{station.name}</div>)
      })
    }

    let objectButton = null
    const availableObjects: JSX.Element[] = []
    if (this.state.objects_loading) {
      availableObjects.push(<img key={'object_loading'} src={this.props.spinner_url} alt="Loading..." />)
    } else {
      this.state.available_objects.forEach(obj => {
        const assignedItems = this.state.assigned_objects.filter(a => a.id == obj.id)
        if (assignedItems.length == 0) {
          availableObjects.push(
            <div
              key={'object_' + obj.id}
              draggable={this.props.is_editable && !this.state.objects_loading}
              onDragStart={this.handleStartObjectDrag}
              id={obj.id.toString()}>
              {obj.name}
              <span
                className="add-button"
                onClick={this.handleAddObject}
                title={this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_ADD_AS_LAST}>+</span>
            </div>)
        }
      })
      objectButton = availableObjects.length > 0 ? <button type="button"
                                                           className="btn"
                                                           onClick={this.handleAddAllObjects}>{this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_AUTO_ASSIGN_OBJECTS}</button> : null
    }

    const objects: JSX.Element[] = []
    let position = 1
    const sortedObjects = this.state.assigned_objects
    sortedObjects.sort((a: PositionalObject, b: PositionalObject) => {
      return a.position > b.position ? 1 : -1
    })

    if (this.props.is_editable) {
      objects.push(<div
        key={'pos' + position}
        className="target-field"
        id={'pos' + position}
        title={this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_DROP_ELEMENT_HERE}
        onDrop={this.handleDrop}
        onDragOver={this.handleAllowDrop}
        onDragLeave={this.handleDragLeave}>+</div>)
    }

    const removeButton = !this.props.is_editable ? null : <span className="remove-button"
      onClick={this.handleRemoveObject}
      title={this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_REMOVE_OBJECT}>X</span>
    sortedObjects.forEach(obj => {
      position++
      objects.push(
        <div
          key={'object_' + obj.id}
          className="assigned-object"
          draggable={this.props.is_editable}
          onDragStart={this.handleStartObjectDrag}
          id={obj.id.toString()}>
          {obj.name}
          {removeButton}
          <br /><br />
            Bays: {obj.item_count}
        </div>)
      if (this.props.is_editable) {
        objects.push(<div
          key={'pos' + position}
          className="target-field"
          id={'pos' + position}
          title={this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_DROP_ELEMENT_HERE}
          onDrop={this.handleDrop}
          onDragOver={this.handleAllowDrop}
          onDragLeave={this.handleDragLeave}>+</div>)
      }
    })

    const availableItems = !this.props.is_editable ? null : <div className="available-items">
      <div className="available-stations">
        <span>{this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_STATIONS}</span>
        <select onChange={this.handleLocationChange} disabled={this.state.stations_loading} className="input">
          {subwayOptions}
        </select><br />
        {stations}
      </div>
      <div className="available-objects">
        <span>{this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_OBJECTS_OF} {this.state.selected_station ? this.state.selected_station.name : ''}</span>
        {availableObjects}<br />
        {objectButton}
      </div>
    </div>

    return <div id='export_editor' key='export_editor'>
      {availableItems}
      <div className="assigned-objects">
        {objects}
      </div>
    </div>
  }
}
