import * as React from 'react'
import { loadAjax } from '../ajax'
import {
  addObjects,
  CommonObject,
  dropObject,
  PositionalObject,
  removeObject
} from '../etsi_export_editor/Common'

export interface OrderingEditorComposeProps {
  ajax_url: string,
  session: string,
  language: {[key: string]: string},
  spinner_url: string,
  is_editable: boolean,
  racks: PositionalObject[],
  object: CommonObject,
}

interface OrderingEditorComposeState {
  available_racks: PositionalObject[],
  assigned_racks: PositionalObject[],
  selected_object: CommonObject,
  selected_rack: PositionalObject,
  is_loading: boolean,
}

export class OrderingEditorCompose extends React.Component<OrderingEditorComposeProps, OrderingEditorComposeState> {
  constructor (props: OrderingEditorComposeProps) {
    super(props)

    this.state = {
      available_racks: [],
      assigned_racks: this.props.racks ? this.props.racks : [],
      selected_object: null,
      selected_rack: null,
      is_loading: false,
    }

    this.handleAllowDrop = this.handleAllowDrop.bind(this)
    this.handleDrop = this.handleDrop.bind(this)
    this.handleAddRack = this.handleAddRack.bind(this)
    this.handleAddAllRacks = this.handleAddAllRacks.bind(this)
    this.handleRemoveRack = this.handleRemoveRack.bind(this)
    this._updateState = this._updateState.bind(this)
    this._setLoadState = this._setLoadState.bind(this)
    this.saveToDom = this.saveToDom.bind(this)
  }

  componentDidMount () {
    this._setLoadState(true)
    loadAjax(this.props.ajax_url + '&func=get-racks&obj_id=' + this.props.object.id, this.props.session)
      .then(res => {
        this.setState((prevState) => ({
          ...prevState,
          selected_object: this.props.object,
          available_racks: res.data,
          selected_rack: null,
          is_loading: false,
        }))
      })
      .catch(() => {
        this._setLoadState(false)
      })
  }

  handleStartDrag (e: React.DragEvent<HTMLDivElement>) {
    const data = {
      id: e.currentTarget.id
    }
    e.dataTransfer.setData('text', JSON.stringify(data))
  }

  handleAllowDrop (e: React.DragEvent<HTMLDivElement>) {
    if (this.props.is_editable) {
      e.currentTarget.className = 'target-field active'
      e.preventDefault()
    }
  }

  handleDragLeave (e: React.DragEvent<HTMLDivElement>) {
    e.currentTarget.className = 'target-field'
  }

  handleDrop (e: React.DragEvent<HTMLDivElement>) {
    if (!this.props.is_editable) {
      return
    }
    e.currentTarget.className = 'target-field'
    e.preventDefault()

    const json = e.dataTransfer.getData('text')
    const data = JSON.parse(json)
    const rackId = data.id
    const posString = e.currentTarget.id
    const newAssignedItems = dropObject(rackId, posString, this.state.available_racks, this.state.assigned_racks)
    this._updateState(newAssignedItems)
  }

  handleAddRack (e: React.MouseEvent<HTMLDivElement>) {
    const rackId = parseInt(e.currentTarget.parentElement.id)
    const racks = this.state.available_racks.filter(r => r.id == rackId)
    const rack = racks.length > 0 ? racks[0] : null
    if (rack) {
      const assignedRacks = addObjects([rack], this.state.assigned_racks)
      this._updateState(assignedRacks)
    }
  }

  handleAddAllRacks () {
    if (this.state.available_racks.length > 0) {
      const assignedRacks = addObjects(this.state.available_racks, this.state.assigned_racks)
      this._updateState(assignedRacks)
    }
  }

  handleRemoveRack (e: React.MouseEvent<HTMLButtonElement>) {
    const rackId = parseInt(e.currentTarget.parentElement.id)
    const assignedRacks = removeObject(rackId, this.state.assigned_racks)
    this._updateState(assignedRacks)
  }

  _updateState (assignedRacks: PositionalObject[]) {
    this.setState((prevState) => ({
      ...prevState,
      assigned_racks: assignedRacks
    }), () => {
      this.saveToDom()
    })
  }

  _setLoadState(is_loading: boolean) {
    this.setState((prevState) => ({
      ...prevState,
      is_loading: is_loading,
    }));
  }

  saveToDom () {
    // save json to DOM
    const json = JSON.stringify({
      racks: this.state.assigned_racks
    })
    document.getElementById('C__CATG__SMVIEW_CABLE_ETSI_EXPORT').setAttribute('value', json)
  }

  render () {
    const availableRacks: JSX.Element[] = []
    this.state.available_racks.forEach(r => {
      const assignedItems = this.state.assigned_racks.filter(a => a.id == r.id)
      if (assignedItems.length == 0) {
        availableRacks.push(
          <div draggable={this.props.is_editable}
            onDragStart={this.handleStartDrag}
            id={r.id.toString()}
            key={r.id}>
            {r.name}
            <span className="add-button"
              onClick={this.handleAddRack}
              title={this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_ADD_AS_LAST}>+</span>
          </div>)
      }
    })

    const racks: JSX.Element[] = []
    let position = 1
    const sortedRacks = this.state.assigned_racks
    sortedRacks.sort((a: PositionalObject, b: PositionalObject) => {
      return a.position > b.position ? 1 : -1
    })

    if (this.props.is_editable) {
      racks.push(<div className="target-field"
        id={'pos' + position}
        key={'pos' + position}
        title={this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_DROP_ELEMENT_HERE}
        onDrop={this.handleDrop}
        onDragOver={this.handleAllowDrop}
        onDragLeave={this.handleDragLeave}>+</div>)
    }

    const removeButton = !this.props.is_editable ? null : <span className="remove-button"
      onClick={this.handleRemoveRack}
      title={this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_REMOVE_RACK}>X</span>
    sortedRacks.forEach(r => {
      position++
      racks.push(
        <div className="assigned-object"
          draggable={this.props.is_editable}
          onDragStart={this.handleStartDrag}
          id={r.id.toString()}
          key={r.id}>
          {r.name}
          {removeButton}
          <br /><br />
          {r.description}
          <br />
            ({r.item_count} {this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__HE})
        </div>)
      if (this.props.is_editable) {
        racks.push(<div className="target-field"
          id={'pos' + position}
          key={'pos' + position}
          title={this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_DROP_ELEMENT_HERE}
          onDrop={this.handleDrop}
          onDragOver={this.handleAllowDrop}
          onDragLeave={this.handleDragLeave}>+</div>)
      }
    })

    let spinner = null
    if (this.state.is_loading) {
      spinner = <img src={this.props.spinner_url} alt="..." />
    }
    const rackButton = availableRacks.length > 0 ? <button type="button" className="button" onClick={this.handleAddAllRacks}>{this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_AUTO_ASSIGN_RACKS}</button> : null
    const availableItems = !this.props.is_editable ? null : <div className="available-items">
      <div className="available-objects">
        <span>{this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ETSI_BAYS_OF} {this.state.selected_object ? this.state.selected_object.name : ''}</span>
        {spinner}
        {availableRacks}<br />
        {rackButton}
      </div>
    </div>

    return <div id='export_editor' key='export_editor'>
      {availableItems}
      <div className="assigned-objects">
        {racks}
      </div>
    </div>
  }
}
