import * as React from 'react'
import { loadAjax } from '../ajax'
import {IndexedRow} from "./Row/IndexedRow";
import {LastRow} from "./Row/LastRow";
import {TypeSelectionRow} from "./Row/TypeSelectionRow";
import {SectionRouteTracing} from "./SectionRouteTracing";

export interface SimpleElement {
    id: number,
    title: string,
}

export interface BaseElement {
  cable_id: number,
  cable_title: string,
  length: number,
  wire_id: number,
  wire_title: string,
  wire_scope: string,
  wire_pin_in_title: string,
  wire_pin_out_title: string,
  section_id: number,
  section_title: string,
  cable_free_wires: number,
  cable_patched_wires: number,
  cable_max_capacity: number,
  obj_in_id: number,
  obj_in_title: string,
  pin_in_id: number,
  pin_in_title: string,
  pin_in_scope: string,
  obj_out_id: number,
  obj_out_title: string,
  pin_out_id: number,
  pin_out_title: string,
  pin_out_scope: string,
}

export interface IndexedElement extends BaseElement {
  id?: number,
  index: number,
  associated_element?: IndexedElement,
  connections: BaseElement[],
}

export interface SectionRouteEditorComposeProps {
  ajax_url: string,
  session: string,
  language: {[key: string]: string},
  spinner_url: string,
  is_editable: boolean,
  elements: IndexedElement[],
  object_types: {[key: number]: string},
  routes: {[key: number]: string},
  is_route: boolean,
  is_virtual_route?: boolean,
  associated_route?: number,
}

interface SectionRouteEditorComposeState {
  is_loading: boolean,
  show_trace_setup: boolean,
  show_associated_setup: boolean,
  elements: IndexedElement[],
  last_element: SimpleElement,
  selectable_objects: BaseElement[],
  selectable_objects_by_type: SimpleElement[],
  selected_associated_route?: number,
  consider_patched_connections: boolean,
}

export class SectionRouteEditorCompose extends React.Component<SectionRouteEditorComposeProps, SectionRouteEditorComposeState> {
  constructor (props: SectionRouteEditorComposeProps) {
    super(props)

    this.state = {
      is_loading: false,
      show_trace_setup: false,
      show_associated_setup: this.props.is_editable && !!this.props.associated_route && !!this.props.is_virtual_route,
      elements: this.props.elements,
      last_element: {
        id: null,
        title: ''
      },
      selectable_objects: [],
      selectable_objects_by_type: [],
      selected_associated_route: this.props.associated_route,
      consider_patched_connections: this.props.is_virtual_route
    }

    this.getNewElement = this.getNewElement.bind(this)
    this.showTraceSetup = this.showTraceSetup.bind(this)
    this.showAssociatedSetup = this.showAssociatedSetup.bind(this)
    this.handleTraceStart = this.handleTraceStart.bind(this)
    this.handleDeleteClick = this.handleDeleteClick.bind(this)
    this.loadLastRow = this.loadLastRow.bind(this)
    this.handleObjectTypeSelection = this.handleObjectTypeSelection.bind(this)
    this.handleAssociatedRouteSelection = this.handleAssociatedRouteSelection.bind(this)
    this.setRouteElementConnection = this.setRouteElementConnection.bind(this)
    this._updateState = this._updateState.bind(this)
    this.saveToDom = this.saveToDom.bind(this)
  }

  componentDidMount () {
    if (this.props.is_editable) {
      this.loadLastRow()
    }
  }

  handleObjectTypeSelection(e: React.ChangeEvent<HTMLSelectElement>) {
    const type_id = parseInt(e.currentTarget.value)
    if (type_id <= 0) {
        return
    }

    this.setState((prevState) => ({
      ...prevState,
      is_loading: true,
      elements: [],
      last_element: {
        id: null,
        title: ''
      },
      selectable_objects: [],
      selectable_objects_by_type: []
    }), () => {
      loadAjax(this.props.ajax_url + '&func=get-objects-by-type&type_id=' + type_id, this.props.session)
          .then(res => {
              this.setState((prevState) => ({
                  ...prevState,
                  selectable_objects_by_type: res.data
              }))
          })
          .finally(() => {
              this.setState((prevState) => ({
                  ...prevState,
                  is_loading: false
              }))
          })
    })
  }

    handleAssociatedRouteSelection(e: React.ChangeEvent<HTMLSelectElement>) {
        let selected_associated_route = parseInt(e.currentTarget.value)
        if (selected_associated_route <= 0) {
            selected_associated_route = null
        }

        this.setState((prevState) => ({
            ...prevState,
            selected_associated_route
        }), () => this.saveToDom())
    }

  loadLastRow(selected_obj_id?: number) {
      this.setState((prevState) => ({
          ...prevState,
          is_loading: true
      }), () => {
          let params = '&func=get-connected-objects&obj_id='
          if (selected_obj_id) {
              params += selected_obj_id
          } else {
              let lastElement: IndexedElement = null
              this.state.elements.forEach(e => {
                  lastElement = !lastElement || e.index > lastElement.index ? e : lastElement
              })

              if (! lastElement) {
                  this.setState((prevState) => ({
                      ...prevState,
                      is_loading: false,
                  }))
                  return
              }
              params += lastElement.obj_in_id + '&wire_id=' + lastElement.wire_id + '&section_id=' + lastElement.section_id
          }
          if (this.state.selected_associated_route) {
              params += '&associated_route=' + this.state.selected_associated_route
          }
          if (this.props.is_virtual_route && this.state.consider_patched_connections) {
              params += '&consider_patched_connections=1'
          }

          loadAjax(this.props.ajax_url + params, this.props.session)
              .then(res => {
                  const data = {...res.data}
                  const obj_types = selected_obj_id && data.objects.length == 0 ? this.state.selectable_objects_by_type : []

                  // assign associated element to last element
                  let elements = this.state.elements
                  if (data.associated_element) {
                      const associatedElement = data.associated_element
                      let lastElement: IndexedElement = null
                      this.state.elements.forEach(e => {
                          lastElement = !lastElement || e.index > lastElement.index ? e : lastElement
                      })
                      associatedElement.index = lastElement.index
                      lastElement.associated_element = associatedElement
                      elements = this.state.elements.filter(e => e.wire_id != lastElement.wire_id)
                      elements.push(lastElement)
                  }

                  this.setState((prevState) => ({
                      ...prevState,
                      last_element: data.element,
                      selectable_objects: data.objects,
                      selectable_objects_by_type: obj_types,
                  }), () => {
                      this._updateState(elements ? elements : [])
                  })
              })
              .finally(() => {
                  this.setState((prevState) => ({
                      ...prevState,
                      is_loading: false,
                  }))
              })
      })
  }

  showTraceSetup() {
    this.setState((prevState) => ({
        ...prevState,
        show_trace_setup: ! this.state.show_trace_setup,
    }))
  }

  showAssociatedSetup() {
    this.setState((prevState) => ({
        ...prevState,
        show_associated_setup: ! this.state.show_associated_setup,
    }))
  }

  handleTraceStart(start_obj_id: number, end_obj_id: number, search_method: string) {
      this.setState((prevState) => ({
          ...prevState,
          is_loading: true,
          elements: [],
          selectable_objects: [],
      }), () => {
          let params = '&func=trace&start_obj_id=' + start_obj_id + '&end_obj_id=' + end_obj_id + '&search_method=' + search_method
          loadAjax(this.props.ajax_url + params, this.props.session)
              .then(res => {
                  const elements = res.data
                  this._updateState(elements ? elements : [])
              })
              .finally(() => {
                  this.setState((prevState) => ({
                      ...prevState,
                      is_loading: false,
                  }))
              })
      })
  }

  setRouteElementConnection(index: number, wire_id?: number, section_id?: number) {
    // get element
    let filteredElements = this.state.elements.filter(e => e.index == index)
    let element = filteredElements.length > 0 ? filteredElements[0] : null

    if (! element) {
      element = this.getNewElement(index);
    }

    // add connection data
    let current_selectable_objects = element.connections && element.connections.length > 0 ? element.connections : this.state.selectable_objects
    let connection = null;
    if (wire_id) {
      element.wire_id = wire_id;
      element.section_id = null;
      let connections = current_selectable_objects.filter(c => c.wire_id == wire_id);
      connection = connections.length > 0 ? connections[0] : null;

      element.cable_id = connection['cable_id'];
      element.cable_title = connection['cable_title'];
      element.wire_title = connection['wire_title'];
      element.wire_scope = connection['wire_scope'];
    }
    if (section_id) {
      element.section_id = section_id;
      element.wire_id = null;
      let connections = current_selectable_objects.filter(c => c.section_id == section_id);
      connection = connections.length > 0 ? connections[0] : null;
      element.section_title = connection['section_title'];
    }
    element.connections = current_selectable_objects;
    element = {...element, ...connection};

    // toggle in and out if needed
    const prevIndex = index - 1;
    filteredElements = this.state.elements.filter(e => e.index == prevIndex)
    if (filteredElements.length > 0) {
      let prev = filteredElements[0]
      if (prev.obj_in_id == element.obj_in_id) {
        element = this._toggleElementInOut(element);
      }
    } else {
      const nextIndex = index + 1;
      filteredElements = this.state.elements.filter(e => e.index == nextIndex)
      if (filteredElements.length > 0) {
        let next = filteredElements[0]
        if (next.obj_out_id == element.obj_out_id) {
          element = this._toggleElementInOut(element);
        }
      }
    }

    // add element to current list
    let elements = this.state.elements.filter(e => e.index != index)
    elements.push(element)
    this._updateState(elements, true)
  }

  _toggleElementInOut(element: IndexedElement) {
    let toggled_element = {
      'obj_in_id': element['obj_out_id'],
      'obj_in_title': element['obj_out_title'],
      'obj_out_id': element['obj_in_id'],
      'obj_out_title': element['obj_in_title'],
      'pin_in_id': element['pin_out_id'],
      'pin_in_scope': element['pin_out_scope'],
      'pin_in_title': element['pin_out_title'],
      'pin_out_id': element['pin_in_id'],
      'pin_out_scope': element['pin_in_scope'],
      'pin_out_title': element['pin_in_title'],
      'wire_pin_in_title': element['wire_pin_out_title'],
      'wire_pin_out_title': element['wire_pin_in_title'],
    };
    return {...element, ...toggled_element};
  }

  getNewElement(index: number): IndexedElement {
    return {
      id: null,
      index: index,
      cable_id: null,
      cable_title: null,
      cable_free_wires: null,
      cable_patched_wires: null,
      cable_max_capacity: null,
      length: null,
      wire_title: null,
      wire_scope: null,
      wire_pin_in_title: null,
      wire_pin_out_title: null,
      connections: [],
      obj_in_id: null,
      obj_in_title: null,
      pin_in_id: null,
      pin_in_scope: null,
      pin_in_title: null,
      obj_out_id: null,
      obj_out_title: null,
      pin_out_id: null,
      pin_out_scope: null,
      pin_out_title: null,
      wire_id: null,
      section_id: null,
      section_title: null,
    }
  }

  handleDeleteClick(index: number) {
    let elements = this.state.elements.filter(e => e.index != index)
    this._updateState(elements, true)
  }

  _updateState (elements: IndexedElement[], loadLastElement: boolean = false) {
    this.setState((prevState) => ({
      ...prevState,
      elements: elements
    }), () => {
      this.saveToDom()
      if (loadLastElement) {
        this.loadLastRow()
      }
    })
  }

  saveToDom () {
    // save json to DOM
    const json = JSON.stringify(this.state.elements)
    document.getElementById('C__CATG__SMVIEW_CABLE_ELEMENTS').setAttribute('value', json)
    if (! document.getElementById('C__CATG__SMVIEW_CABLE_ASSOCIATED_ROUTE__HIDDEN')) {
      return
    }

    if (! this.state.selected_associated_route) {
        return;
    }

    document.getElementById('C__CATG__SMVIEW_CABLE_ASSOCIATED_ROUTE__HIDDEN').setAttribute('value', this.state.selected_associated_route.toString())
    for (const key in this.props.routes) {
      if (key == this.state.selected_associated_route.toString()) {
        const title = this.props.routes[key]
        document.getElementById('C__CATG__SMVIEW_CABLE_ASSOCIATED_ROUTE__VIEW').setAttribute('value', title)
      }
    }
  }

  render () {
    let rows: JSX.Element[] = []
    let complete_length = 0
    let counter = 1
    let elements = this.state.elements
    elements.sort((a, b) => {
        return a.index > b.index ? 1 : -1
    })
    this.state.elements.forEach(element => {
      if (element.length) {
        complete_length += parseFloat(element.length.toString())
        complete_length.toFixed(2)
      }
      rows.push(<IndexedRow key={element.id}
                            is_editable={this.props.is_editable}
                            is_route={this.props.is_route}
                            element={element}
                            completeLength={complete_length}
                            isLastRow={counter == this.state.elements.length}
                            is_associated_element={false}
                            onConfiguredCableSelectionChange={this.setRouteElementConnection}
                            onDeleteClick={this.handleDeleteClick} />)
        if (element.associated_element) {
            rows.push(<IndexedRow key={'assoc_' + element.id}
                                  is_editable={false}
                                  is_route={this.props.is_route}
                                  element={element.associated_element}
                                  completeLength={complete_length}
                                  isLastRow={counter == this.state.elements.length}
                                  is_associated_element={true}
                                  onConfiguredCableSelectionChange={null}
                                  onDeleteClick={null} />)
        }
      counter++
    })

    if (this.props.is_editable) {
        if (this.state.selectable_objects_by_type.length > 0) {
            rows.push(<TypeSelectionRow key="object_types"
                                        is_loading={this.state.is_loading}
                                        spinner_url={this.props.spinner_url}
                                        elements={this.state.selectable_objects_by_type}
                                        onObjectSelectionChange={this.loadLastRow} />)
        } else if (this.state.selectable_objects.length > 0) {
            let index: number = 0
            this.state.elements.forEach(e => {
                const i = parseInt(e.index.toString())
                index = i > index ? i : index
            })
            rows.push(<LastRow key="last_row"
                               is_loading={this.state.is_loading}
                               is_route={this.props.is_route}
                               spinner_url={this.props.spinner_url}
                               last_index={index}
                               selectable_objects={this.state.selectable_objects}
                               last_element={this.state.last_element}
                               onCableSelectionChange={this.setRouteElementConnection} />)
        }
    }

    let startObjectType
    const spinner = this.state.is_loading ? <img key={'station_loading'} src={this.props.spinner_url} alt="Loading..." /> : null
    if (this.props.is_editable) {
        const objectTypeOptions: JSX.Element[] = SectionRouteTracing.getSortedOptions(this.props.object_types)
        const associatedRoute = this.props.is_route && ! this.props.is_virtual_route ? <a className="btn" onClick={this.showAssociatedSetup}>
            {this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__CONFIGURE_ASSOCIATED_ROUTE}
        </a> : null
        const considerPatchedConnections = this.props.is_virtual_route ?
            <div style={{display: 'inline-block'}}>
                <input type="checkbox"
                       className="input"
                       checked={this.state.consider_patched_connections}
                       onChange={() => this.setState({consider_patched_connections: ! this.state.consider_patched_connections})} />
                {this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__CONSIDER_PATCHED_CONNECTIONS}
            </div> : null
        startObjectType = <table className="contentTable">
            <tbody>
                <tr>
                    <td className="key">{this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__START_OBJECT_TYPES}</td>
                    <td className="value">
                        <div className="ml20">
                            <select className="select input input-small" onChange={this.handleObjectTypeSelection} disabled={this.state.is_loading}>
                                {objectTypeOptions}
                            </select>
                            &nbsp;
                            <a className="btn" onClick={this.showTraceSetup}>
                                {this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__AUTOMATIC_TRACING}
                            </a>
                            &nbsp;
                            {associatedRoute}
                            {considerPatchedConnections}
                            {spinner}
                        </div>
                    </td>
                </tr>
            </tbody>
        </table>
    }

    let traceSetup
    if (this.state.show_trace_setup) {
        traceSetup = <SectionRouteTracing is_loading={this.state.is_loading}
                                          ajax_url={this.props.ajax_url}
                                          session={this.props.session}
                                          language={this.props.language}
                                          spinner_url={this.props.spinner_url}
                                          object_types={this.props.object_types}
                                          onTraceStart={this.handleTraceStart} />
    }

    let associatedSetup
    if (this.state.show_associated_setup) {
      const associatedRouteOptions: JSX.Element[] = SectionRouteTracing.getSortedOptions(this.props.routes)
      associatedSetup = <table className="contentTable">
          <tbody>
              <tr>
                  <td className="key">{this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__REDUNDANT_ROUTE}</td>
                  <td className="value">
                      <div className="ml20">
                          <select className="select input input-small"
                                  onChange={this.handleAssociatedRouteSelection}
                                  defaultValue={this.state.selected_associated_route}
                                  disabled={this.state.is_loading || this.props.associated_route > 0}>
                            {associatedRouteOptions}
                          </select>
                          &nbsp;
                          {spinner}
                      </div>
                  </td>
              </tr>
          </tbody>
      </table>
    }

    return <div>
      {startObjectType}
      {traceSetup}
      {associatedSetup}

      <table id="route_elements" className="route_elements">
        <thead>
          <tr>
            <td>{this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ROUTE_INDEX}</td>
            <td>{this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ROUTE_OBJECT_OUTGOING}</td>
            <td>{this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ROUTE_OBJECT_INCOMING}</td>
            <td>{this.props.language.LC__SHD__TELECOMMUNICATION_CABLE}</td>
            <td>{this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ROUTE_ELEMENT_CAPACITY}</td>
            <td>{this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ROUTE_ELEMENT_LENGTH}</td>
            <td>{this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__ROUTE_OPTIONS}</td>
          </tr>
        </thead>
        <tbody>
          {rows}
        </tbody>
      </table>
    </div>
  }
}
