import * as React from 'react'
import {loadAjax} from "../ajax";
import {SimpleElement} from "./SectionRouteEditor";

interface SectionRouteTracingProps {
    is_loading: boolean,
    ajax_url: string,
    session: string,
    spinner_url: string,
    language: {[key: string]: string},
    object_types: {[key: number]: string},
    onTraceStart: Function,
}

interface SectionRouteTracingState {
    start_objects: SimpleElement[],
    end_objects: SimpleElement[],
    selected_start_object_id: number,
    selected_end_object_id: number,
    search_method: string,
}

export class SectionRouteTracing extends React.Component<SectionRouteTracingProps, SectionRouteTracingState> {
    constructor (props: SectionRouteTracingProps) {
        super(props)
        this.state = {
            start_objects: [],
            end_objects: [],
            selected_start_object_id: null,
            selected_end_object_id: null,
            search_method: 'shortest_path',
        }
        this.handleObjectTypeSelectionChange = this.handleObjectTypeSelectionChange.bind(this)
        this.handleObjectSelectionChange = this.handleObjectSelectionChange.bind(this)
        this.handleSearchMethodChange = this.handleSearchMethodChange.bind(this)
        this.handleTraceStart = this.handleTraceStart.bind(this)
    }

    handleObjectTypeSelectionChange(e: React.ChangeEvent<HTMLSelectElement>) {
        let obj_type_id = parseInt(e.currentTarget.value)
        let objSide = e.currentTarget.name == 'start_object_type' ? 'start_objects' : 'end_objects'
        loadAjax(this.props.ajax_url + '&func=get-objects-by-type&type_id=' + obj_type_id, this.props.session)
            .then(res => {
                this.setState((prevState) => ({
                    ...prevState,
                    [objSide]: res.data
                }))
            })
    }

    handleObjectSelectionChange(e: React.ChangeEvent<HTMLSelectElement>) {
        let obj_id = parseInt(e.currentTarget.value)
        let side = e.currentTarget.name == 'start_object' ? 'selected_start_object_id' : 'selected_end_object_id'
        this.setState((prevState) => ({
            ...prevState,
            [side]: obj_id > 0 ? obj_id : null
        }))
    }

    handleSearchMethodChange(e: React.MouseEvent<HTMLInputElement>) {
        let search_method = e.currentTarget.value
        this.setState((prevState) => ({
            ...prevState,
            search_method: search_method
        }))
    }

    public static getSortedOptions(unsorted_options: {[key: number]: string}) {
        // sort object types by name
        let sorted_option_tuples = [];
        if (! unsorted_options || Object.keys(unsorted_options).length === 0) {
            return []
        }

        for (const key in unsorted_options) sorted_option_tuples.push([key, unsorted_options[key]]);
        sorted_option_tuples.sort(function(a, b) {
            if (! a || ! b || ! a[1] || ! b[1]) {
                return 0
            }
            const a1 = typeof a[1] == 'string' ? a[1].toLowerCase() : ''
            const b1 = typeof b[1] == 'string' ? b[1].toLowerCase() : ''
            return a1 < b1 ? -1 : (a1 > b1 ? 1 : 0);
        });

        let sorted_options: JSX.Element[] = []
        sorted_options.push(<option key={-1} value={-1}>-</option>)
        for (let i = 0; i < sorted_option_tuples.length; i++) {
            const key = sorted_option_tuples[i][0];
            const value = sorted_option_tuples[i][1];
            sorted_options.push(<option key={key} value={key}>{value}</option>)
        }
        return sorted_options
    }

    getObjectOptions(objs: SimpleElement[]) {
        let objectOptions: JSX.Element[] = []
        objectOptions.push(<option key={-1} value={-1}>-</option>)
        objs.forEach(obj => {
            objectOptions.push(
                <option key={obj.id} value={obj.id.toString()}>
                    {obj.title}
                </option>)
        })
        return objectOptions
    }

    handleTraceStart() {
        this.props.onTraceStart(this.state.selected_start_object_id, this.state.selected_end_object_id, this.state.search_method)
    }

    render () {
        const objectTypeOptions: JSX.Element[] = SectionRouteTracing.getSortedOptions(this.props.object_types)
        const startObjectOptions: JSX.Element[] = this.getObjectOptions(this.state.start_objects)
        const endObjectOptions: JSX.Element[] = this.getObjectOptions(this.state.end_objects)
        const traceDisabled = this.state.selected_start_object_id <= 0 || this.state.selected_end_object_id <= 0 || this.props.is_loading
        const spinner = this.props.is_loading ? <img key={'station_loading'} src={this.props.spinner_url} alt="Loading..." /> : null

        return <table className="contentTable route_tracing">
            <tbody>
                <tr>
                    <td className="key">
                        {this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__AUTOMATIC_TRACING}
                    </td>
                    <td colSpan={2}>
                        <div className="ml20">
                            <input type="radio" name="search_method" value="shortest_path" defaultChecked={true} onClick={this.handleSearchMethodChange} />
                            {this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__SHORTEST_PATH}
                            &nbsp;
                            <input type="radio" name="search_method" value="least_hops" onClick={this.handleSearchMethodChange} />
                            {this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__LEAST_HOPS}
                        </div>
                    </td>
                    <td>&nbsp;</td>
                </tr>
                <tr>
                    <td className="key">&nbsp;</td>
                    <td className="value" colSpan={3}>
                        <div className="ml20">
                            <div style={{width: '40px'}}>
                                {this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__FROM}
                            </div>
                            <select className="input input-small" name="start_object_type" onChange={this.handleObjectTypeSelectionChange}>
                                {objectTypeOptions}
                            </select>
                            <select className="input input-small" name="start_object" onChange={this.handleObjectSelectionChange}>
                                {startObjectOptions}
                            </select>
                        </div>
                    </td>
                </tr>
                <tr>
                    <td className="key">&nbsp;</td>
                    <td className="value" colSpan={3}>
                        <div className="ml20">
                            <div style={{width: '40px'}}>
                                {this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__TO}
                            </div>
                            <select className="input input-small" name="end_object_type" onChange={this.handleObjectTypeSelectionChange}>
                                {objectTypeOptions}
                            </select>
                            <select className="input input-small" name="end_object" onChange={this.handleObjectSelectionChange}>
                                {endObjectOptions}
                            </select>
                            <button type="button" className="btn" onClick={this.handleTraceStart} disabled={traceDisabled}>
                                {this.props.language.LC__CMDB__CATG__SMVIEW_CABLE__START_AUTOMATIC_TRACING}
                            </button>
                            {spinner}
                        </div>
                    </td>
                </tr>
            </tbody>
        </table>
    }
}
