<?php

use idoit\Component\Property\Type\TextProperty;

/**
 * i-doit
 *
 * DAO: global category for SM-VIEW Cable assigned routes.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  2021 SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license MIT
 */
class isys_cmdb_dao_category_g_smview_cable_assigned_routes extends isys_cmdb_dao_category_global
{

    /**
     * Category's name. Will be used for the identifier, constant, main table, and many more.
     *
     * @var string
     */
    protected $m_category = 'smview_cable_assigned_routes';

    /**
     * Is category multi-valued or single-valued?
     *
     * @var boolean
     */
    protected $m_multivalued = true;

    /**
     * Category table
     *
     * @var string
     */
    protected $m_table = 'isys_catg_smview_cable_route_list';

    /**
     * Simple cache to store trace results.
     *
     * @var array|null
     */
    private $_traceResult = null;

    /**
     * Method for returning the properties.
     *
     * @return array
     */
    protected function properties()
    {
        return [
            'title' => (new TextProperty(
                'C__CATG__SMVIEW_PATCH_PARKING_DECK_INFO_OBJECT_TITLE',
                'LC__CMDB__CATG__SMVIEW_PATCH_PARKING_DECK_INFO__OBJECT_TITLE',
                'isys_obj__title',
                'isys_obj'
            ))->mergePropertyUiParams(
                [
                    'p_strClass'  => 'input',
                    'p_bDisabled' => true,
                ]
            ),
        ];
    }

    /**
     * Get data method, uses logical unit DAO.
     *
     * @param  int $p_catg_list_id
     * @param  int $p_obj_id
     * @param  string $p_condition
     * @param  array $p_filter
     * @param  int $p_status
     * @return isys_component_dao_result
     * @throws isys_exception_database
     */
    public function get_data($p_catg_list_id = null, $p_obj_id = null, $p_condition = '', $p_filter = null, $p_status = null)
    {
        $routes = $this->_trace_connection_by_obj_id($p_obj_id);
        $ids = [];
        foreach ($routes as $route) {
            $ids[] = $route['isys_obj__id'];
        }
        $idList = implode(',', $ids);
        $condition = ! empty($idList) ? " AND isys_obj__id IN ($idList) " : ' AND 1 = 0';
        return parent::get_data(null, null, $condition);
    }

    /**
     * Retrieves the number of saved category-entries to the given object.
     *
     * @param  int $p_obj_id
     * @return int
     * @throws isys_exception_database
     */
    public function get_count($p_obj_id = null)
    {
        $routes = $this->_trace_connection_by_obj_id($p_obj_id);
        return count($routes);
    }

    /**
     * Trace a connection by object ID until cable pin or last device is reached.
     *
     * @param  int $obj_id
     * @return array
     * @throws isys_exception_database
     */
    private function _trace_connection_by_obj_id($obj_id)
    {
        if (isset($this->_traceResult) && ! empty($this->_traceResult[$obj_id])) {
            return $this->_traceResult[$obj_id];
        }

        // add only incoming connectors
        $dao = new isys_cmdb_dao_category_g_connector($this->m_db);
        $connectors = $dao->get_data_as_array(null, $obj_id, ' AND isys_catg_connector_list.isys_catg_connector_list__type = 1');
        $routes = [];
        foreach ($connectors as $c) {
            $p_routes = $this->_trace_connection($c);
            $routes = array_merge($routes, $p_routes);
        }
        $this->_traceResult[$obj_id] = $routes;
        return $routes;
    }

    /**
     * Trace a connection until cable pin or last device is reached.
     *
     * @param  array $port
     * @param  array $tracedPortIds
     * @return array
     * @throws isys_exception_database
     */
    private function _trace_connection($port, $tracedPortIds = [])
    {
        $dao = new isys_cmdb_dao_category_g_connector($this->m_db);
        $tracedPortIds[] = $port['isys_catg_connector_list__id'];

        // check cable pin
        $pin_dao = new isys_cmdb_dao_smview_cable_pin($this->m_db);
        $wire_pin_dao = new isys_cmdb_dao_smview_cable_wire_pin($this->m_db);
        $route_element_dao = new isys_cmdb_dao_smview_cable_route_element($this->m_db);
        $route_dao = new isys_cmdb_dao_smview_cable_route($this->m_db);
        $pins = $pin_dao->filter_objects(['connector_id' => $port['isys_catg_connector_list__id']]);
        $routes = [];
        if (! empty($pins)) {
            foreach ($pins as $pin) {
                $wp = $wire_pin_dao->get_object('connected_pin_id', $pin['id']);
                if (empty($wp)) {
                    continue;
                }

                $re = $route_element_dao->get_object('wire_id', $wp['wire_id']);
                if (empty($re)) {
                    continue;
                }

                $r = $route_dao->get_object('id', $re['route_id']);
                if (empty($r)) {
                    continue;
                }
                $route = $dao->get_object($r['obj_id'])->get_row();
                $routes[] = $route;
            }
            return $routes;
        }

        $connector = $port;

        // try to check connected connector if not traced anymore
        $connected_connector = $connector['con_connector'];
        if (empty($connected_connector) || in_array($connected_connector, $tracedPortIds)) {
            // try to use sibling connector
            $connected_connector = null;
            if (! empty($port['isys_catg_connector_list__isys_catg_connector_list__id'])) {
                // connector has a sibling
                $sibling_connector = $port['isys_catg_connector_list__isys_catg_connector_list__id'];
                $connector = $dao->get_data($sibling_connector)->get_row();
                $connected_connector = $connector['con_connector'];
                $tracedPortIds[] = $connector['isys_catg_connector_list__id'];
            }
        }

        if (! empty($connected_connector) && ! in_array($connected_connector, $tracedPortIds)) {
            // trace connected connector
            $connected_port = $dao->get_data($connected_connector)->get_row();
            $routes = $this->_trace_connection($connected_port, $tracedPortIds);
        }
        return $routes;
    }

    /**
     * Creates new entity.
     *
     * @param  array $p_data Properties in a associative array with tags as keys and their corresponding values as values.
     * @return null  Returns created entity's identifier (int) or false (bool).
     */
    public function create_data($p_data)
    {
        // no op
        return null;
    }

    /**
     * Updates existing entity.
     *
     * @param  int $p_list_id Entity's identifier
     * @param  array $p_data Properties in a associative array with tags as keys and their corresponding values as values.
     * @return null
     */
    public function save_data($p_list_id, $p_data)
    {
        // no op
        return null;
    }
}
