<?php

use idoit\Module\Cable\SectionTracing;

/**
 * i-doit
 *
 * DAO: global category for SM-VIEW Cable virtual route elements.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  2022 SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class isys_cmdb_dao_category_g_smview_cable_virtual_route_element extends isys_cmdb_dao_category_g_smview_cable_route_element
{

    /**
     * Category's name. Will be used for the identifier, constant, main table, and many more.
     *
     * @var string
     */
    protected $m_category = 'smview_cable_virtual_route_element';

    /**
     * Table name.
     *
     * @var string
     */
    protected $m_table = 'isys_catg_smview_cable_vir_route_element_list';

    /**
     * Is category multi-valued or single-valued?
     *
     * @var boolean
     */
    protected $m_multivalued = true;

    /**
     * Get all virtual route elements.
     *
     * @return array
     * @throws isys_exception_database
     */
    public function get_all_elements()
    {
        $dao = new isys_cmdb_dao_smview_cable_virtual_route(isys_application::instance()->database);
        $routes = $dao->get_objects();

        $return = [];
        foreach ($routes as $route) {
            $return[$route["obj_id"]] = $this->get_elements_by_route_id($route["id"]);
        }
        return $return;
    }

    /**
     * Get virtual route elements by isys_obj_id.
     *
     * @param  int     $p_obj_id
     * @param  boolean $ignore_start_object
     * @param  boolean $add_connection_information
     * @return array
     * @throws isys_exception_database
     */
    public function get_elements_by_obj_id($p_obj_id, $ignore_start_object = false, $add_connection_information = true)
    {
        $dao = new isys_cmdb_dao_smview_cable_virtual_route(isys_application::instance()->database);
        $route = $dao->get_object("obj_id", $p_obj_id);
        return $this->get_elements_by_route_id($route["id"], $ignore_start_object, $add_connection_information);
    }

    /**
     * Get virtual route elements by route_id.
     *
     * @param  int     $route_id
     * @param  boolean $ignore_start_object
     * @param  boolean $add_connection_information
     * @param  boolean $consider_associated_route
     * @return array
     * @throws isys_exception_database
     */
    public function get_elements_by_route_id($route_id, $ignore_start_object = false, $add_connection_information = true, $consider_associated_route = false)
    {
        $dao = new isys_cmdb_dao_smview_cable_virtual_route(isys_application::instance()->database);
        $element_dao = new isys_cmdb_dao_smview_cable_virtual_route_element(isys_application::instance()->database);
        $route = $dao->get_object('id', $route_id);
        $routeElements = $element_dao->filter_objects(['route_id' => $route_id]);

        // get all connections
        $elements = [];
        foreach ($routeElements as $element) {
            if (! empty($element['section_id'])) {
                $section_id = $element['section_id'];
                $section = SectionTracing::get_section_endpoints($section_id);
                $element = array_merge($element, $section);
            } else {
                $element = isys_cmdb_dao_category_g_smview_cable_section_element::get_wire_connection($element);
            }
            $elements[$element['index']] = $element;
        }

        if (empty($elements)) {
            return $elements;
        }
        // sort elements by index
        ksort($elements);

        $start_obj_id = !empty($route['start_obj_id']) ? $route['start_obj_id'] : null;
        if ($ignore_start_object) {
            $start_obj_id = null;
        }
        $ordered_elements = isys_cmdb_dao_category_g_smview_cable_section_element::order_objects_within_connections($elements, $start_obj_id);

        // add connection information to each element
        $new_elements = [];
        foreach ($ordered_elements as $index => $element) {
            $new_elements[$index] = $add_connection_information ? self::add_connection_information($element) : $element;
        }

        // finally sort elements by index
        ksort($new_elements);
        return $new_elements;
    }

    /**
     * Re-order indices in case if the first one was deleted.
     *
     * @param  int $route_id
     * @throws isys_exception_database
     */
    public function re_order_indices($route_id)
    {
        $element_dao = new isys_cmdb_dao_smview_cable_virtual_route_element(isys_application::instance()->database);
        $elements = $this->get_elements_by_route_id($route_id, true);
        $currentIndex = 1;
        foreach ($elements as $index => $element) {
            if ($index != $currentIndex) {
                $element['index'] = $currentIndex;
                $element_dao->update_object($element['id'], $element);
            }
            $currentIndex++;
        }
    }

    /**
     * Patch route elements.
     *
     * @param int $route_id
     * @param array $current_elements
     */
    public function patch_route_elements($route_id, $current_elements = [])
    {
        // nothing will be patched on virtual route elements
    }
}
