<?php

/**
 * i-doit
 *
 * Abstract DAO for SM-VIEW Cable.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
abstract class isys_cmdb_dao_smview_cable_abstract extends isys_module_dao
{
    
    /**
     * Mapping of database and API keys
     *
     * @var array
     */
    protected $db_to_api_map = array();
    
    /**
     * Name of the table.
     *
     * @var string
     */
    protected $table_name = null;
    
    /**
     * Name of the primary key field.
     *
     * @var string
     */
    protected $primary_key_field = null;

    /**
     * Get data method.
     *
     * @return array
     * @throws isys_exception_database
     */
    public function get_data()
    {
        return $this->get_objects();
    }

    /**
     * Get all objects.
     *
     * @return array
     * @throws isys_exception_database
     */
    public function get_objects()
    {
        $tableName = $this->table_name;
        $query = "SELECT * FROM $tableName;";
        $res = $this->retrieve($query);
        return $this->_map_result_to_array($res);
    }

    /**
     * Get an object.
     *
     * @param  string $field
     * @param  string $value
     * @return array
     * @throws isys_exception_database
     */
    public function get_object($field, $value)
    {
        $tableName = $this->table_name;
        $db_field = array_search($field, $this->db_to_api_map);
        $query = "SELECT * FROM $tableName WHERE $db_field = '$value';";
        $res = $this->retrieve($query);
        $objects = $this->_map_result_to_array($res);
        return ! empty($objects) ? $objects[0] : array();
    }

    /**
     * Filter objects.
     *
     * @param  array $filter
     * @return array
     * @throws isys_exception_database
     */
    public function filter_objects($filter)
    {
        $tableName = $this->table_name;
        $query = "SELECT * FROM $tableName WHERE 1 = 1 ";
        
        foreach ($filter as $field => $value) {
            $db_field = array_search($field, $this->db_to_api_map);
            if (! $db_field) {
                continue;
            }
            if (strpos($value, '%') !== false) {
                $query .= " AND $db_field like '$value'";
            } else {
                $query .= " AND $db_field = '$value'";
            }
        }
        $query .= ';';
        
        $res = $this->retrieve($query);
        return $this->_map_result_to_array($res);
    }
    
    /**
     * Map database results to array.
     *
     * @param  isys_component_dao_result $res
     * @return array
     */
    protected function _map_result_to_array($res)
    {
        $objects = array();
        while ($row = $res->get_row()) {
            $obj = array();
            foreach ($this->db_to_api_map as $db_key => $api_key) {
                $obj[$api_key] = isset($row[$db_key]) ? $row[$db_key] : null;
            }
            
            $additional = $this->_get_additional_informations($obj);
            if ($additional) {
                $objects[] = array_merge($obj, $additional);
            } else {
                $objects[] = $obj;
            }
        }
        return $objects;
    }
    
    /**
     * Optional method to add additional informations to an object.
     *
     * @param  array $obj
     * @return array|null
     */
    protected function _get_additional_informations($obj)
    {
        return null;
    }

    /**
     * Create a new object.
     *
     * @param  array $params
     * @param  array $optionalParams
     * @return int|false
     * @throws isys_exception_dao
     */
    public function create_object($params, $optionalParams = [])
    {
        $tableName = $this->table_name;
        $sqlRows = $this->_getSqlRows($params, $optionalParams);
        
        if (empty($sqlRows)) {
            return null;
        }
        
        $sql = "INSERT INTO $tableName SET ";
        $sql .= implode(', ', $sqlRows) . ';';
        
        $id = false;
        if ($this->update($sql) && $this->apply_update()) {
            $id = $this->get_last_insert_id();
        }
        return $id;
    }

    /**
     * Update an existing object.
     *
     * @param  int $id
     * @param  array $params
     * @param  array $optionalParams
     * @return boolean
     * @throws isys_exception_dao
     */
    public function update_object($id, $params, $optionalParams = [])
    {
        $tableName = $this->table_name;
        $idField = $this->primary_key_field;
        $sqlRows = $this->_getSqlRows($params, $optionalParams);
        
        if (empty($sqlRows)) {
            return true;
        }

        $rows = implode(', ', $sqlRows);
        $sql = "UPDATE $tableName SET " . $rows . " WHERE $idField = " . $this->convert_sql_id($id);
        return $this->update($sql) && $this->apply_update();
    }
    
    /**
     * Get rows prepared for SQL statement.
     *
     * @param  array $params
     * @param  array $optionalParams
     * @return array
     */
    protected function _getSqlRows($params, $optionalParams)
    {
        $sqlRows = [];
        $parameters = $params;
        foreach ($optionalParams as $key => $value) {
            // do not ignore optional values, but set them to NULL
            //            if (empty($value)) {
            //                //continue;
            //            }
            $parameters[$key] = $value;
        }
        
        foreach ($parameters as $key => $value) {
            $db_field = array_search($key, $this->db_to_api_map);
            if ($db_field === false) {
                continue;
            }
            if (! isset($value)) {
                $sqlRows[] = "$db_field = NULL";
            } else {
                $sqlRows[] = "$db_field = " . $this->convert_sql_text($value);
            }
        }
        return $sqlRows;
    }

    /**
     * Delete an existing object.
     *
     * @param  int $id
     * @return boolean
     * @throws isys_exception_dao
     */
    public function delete_object($id)
    {
        $this->begin_update();

        $tableName = $this->table_name;
        $idField = $this->primary_key_field;
        $l_q = "DELETE FROM $tableName WHERE $idField = " . $this->convert_sql_id($id);
        $l_res = $this->update($l_q);

        return $l_res ? $this->apply_update() : false;
    }
}
