<?php

/**
 * i-doit
 *
 * Wire Pin DAO for SM-VIEW Cable.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class isys_cmdb_dao_smview_cable_wire_pin extends isys_cmdb_dao_smview_cable_abstract
{
    
    /**
     * Mapping of database and API keys
     *
     * @var array
     */
    protected $db_to_api_map = array(
        'isys_catg_smview_cable_wire_pin__id'               => 'id',
        'isys_catg_smview_cable_wire_pin__wire_id'          => 'wire_id',
        'isys_catg_smview_cable_wire_pin__description'      => 'description',
        'isys_catg_smview_cable_wire_pin__sibling_id'       => 'sibling_id',
        'isys_catg_smview_cable_wire_pin__connected_pin_id' => 'connected_pin_id',
        'isys_catg_smview_cable_wire_pin__status_id'        => 'status_id',
    );
    
    /**
     * Name of the table.
     *
     * @var string
     */
    protected $table_name = 'isys_catg_smview_cable_wire_pin';
    
    /**
     * Name of the primary key field.
     *
     * @var string
     */
    protected $primary_key_field = 'isys_catg_smview_cable_wire_pin__id';
    
    /**
     * Get possible states of a wire.
     *
     * @return array
     */
    public static function get_wire_states()
    {
        $l_status_dao = new isys_cmdb_dao_smview_cable_wire_pin_status(isys_application::instance()->database);
        $objects = $l_status_dao->get_objects();
        $status = [];
        foreach ($objects as $o) {
            $status[$o['id']] = $o['const'];
        }
        return $status;
    }

    /**
     * Create a new wire.
     *
     * @param  int    $wire_id
     * @param  string $description
     * @param  int    $status_id
     * @param  int    $sibling_id
     * @param  int    $connected_pin_id
     * @return int|false
     */
    public function create_wire_pin($wire_id, $description, $status_id = null, $sibling_id = null, $connected_pin_id = null)
    {
        $states = self::get_wire_states();
        if (empty($states[$status_id])) {
            $state = null;
        }
        
        $params = array(
            'wire_id'     => $wire_id,
            'description' => $description,
        );
        $optionalParams = array(
            'status_id'        => $status_id,
            'sibling_id'       => $sibling_id,
            'connected_pin_id' => $connected_pin_id,
        );
        return $this->create_object($params, $optionalParams);
    }

    /**
     * Update an existing wire.
     *
     * @param  int    $wire_pin_id
     * @param  string $description
     * @param  int    $status_id
     * @param  int    $sibling_id
     * @param  int    $connected_pin_id
     * @return boolean
     */
    public function update_wire_pin($wire_pin_id, $description, $status_id = null, $sibling_id = null, $connected_pin_id = null, $displayMessages = false)
    {
        $states = self::get_wire_states();
        if (empty($states[$status_id])) {
            $state = null;
        }
        
        // get sibling ID
        $obj = $this->get_object('id', $wire_pin_id);
        if (! empty($obj['sibling_id'])) {
            // keep sibling ID
            $sibling_id = $obj['sibling_id'];
        }
        
        $params = array(
            'description' => $description,
            'sibling_id'  => $sibling_id,
        );
        $optionalParams = array(
            'status_id'        => $status_id,
            'connected_pin_id' => $connected_pin_id,
        );

        // check, if connected pin ID is already assigned
        if ($connected_pin_id) {
            $pins = $this->filter_objects(array('connected_pin_id' => $connected_pin_id));
            foreach ($pins as $pin) {
                // remove current assignments
                $this->update_object($pin['id'], [], ['connected_pin_id' => null]);
            }
        }

        $return = $this->update_object($wire_pin_id, $params, $optionalParams);

        try {
            $this->_set_extended_location_on_cable($obj['wire_id'], $displayMessages);
        } catch (isys_exception_database | Exception $e) {
            isys_notify::error('Could not set extended location on cable: ' . $e->getMessage());
        }
        return $return;
    }

    /**
     * Reconnect a bunch of wires to new PINs after validation.
     *
     * @param  array $mapping
     * @throws \idoit\Exception\Exception
     * @throws isys_exception_dao
     * @throws isys_exception_database
     */
    public function reconnect_wires($mapping)
    {
        // check if PINs are available for reconnect
        $pin_dao = new isys_cmdb_dao_smview_cable_pin($this->m_db);
        foreach ($mapping as $pin_id => $wire_pin_id) {
            $pin = $pin_dao->get_object('id', $pin_id);
            if (empty($pin)) {
                throw new \idoit\Exception\Exception('pin does not exist');
            }

            $wire_pin = $this->get_object('id', $wire_pin_id);
            if (empty($wire_pin)) {
                throw new \idoit\Exception\Exception('wire pin does not exist');
            }

            $connected_wire_pins = $this->filter_objects(array('connected_pin_id' => $pin_id));
            if ($connected_wire_pins) {
                $connected_wire_pin = $connected_wire_pins[0];
                // PIN is already connected
                $isConnectedToAnotherWire = true;
                foreach ($mapping as $wp_id) {
                    if ($wp_id == $connected_wire_pin['id']) {
                        // PIN is connected to a wire within new scope -> ok
                        $isConnectedToAnotherWire = false;
                    }
                }

                if ($isConnectedToAnotherWire) {
                    // error, PIN is connected to a wire outside of new scope
                    throw new \idoit\Exception\Exception('PIN already connected to another wire');
                }
            }
        }
        $this->_reconnect_wires($mapping);
    }

    /**
     * Reconnect a bunch of wires to new PINs.
     *
     * @param  array $mapping
     * @throws isys_exception_dao
     * @throws \idoit\Exception\Exception
     * @throws isys_exception_database
     */
    private function _reconnect_wires($mapping)
    {
        $pin_dao = new isys_cmdb_dao_smview_cable_pin($this->m_db);
        $preCount = count($mapping);
        $unconnectedPins = [];
        $connectedPins = [];
        foreach ($mapping as $pin_id => $wire_pin_id) {
            $pin_id = (int) $pin_id;
            $wire_pin_id = (int) $wire_pin_id;

            $wire_pins = $this->filter_objects(array('connected_pin_id' => $pin_id));
            if ($wire_pins) {
                $connectedPins[$pin_id] = $wire_pin_id;
            } else {
                $unconnectedPins[$pin_id] = $wire_pin_id;
            }
        }

        $tableName = $this->table_name;
        $idField = $this->primary_key_field;
        foreach ($unconnectedPins as $pin_id => $wire_pin_id) {
            // get patch PIN of current connected PIN
            $wire_pin = $this->get_object('id', $wire_pin_id);
            $old_pin_id = $wire_pin['connected_pin_id'];
            $old_pin = $pin_dao->get_object('id', $old_pin_id);
            $old_sib_pin = $pin_dao->get_object('id', $old_pin['internal_pin_id']);
            $patch_pin_id = $old_sib_pin['patch_pin_id'];

            // reconnect PIN
            $sql = "UPDATE $tableName SET isys_catg_smview_cable_wire_pin__connected_pin_id = " . $pin_id
                . " WHERE $idField = " . $this->convert_sql_id($wire_pin_id);
            $this->update($sql) && $this->apply_update();

            if (! empty($patch_pin_id)) {
                // patch both connections
                $pin = $pin_dao->get_object('id', $pin_id);
                $sql = "UPDATE isys_catg_smview_cable_pin_list SET isys_catg_smview_cable_pin_list__patch_id = " . $patch_pin_id
                    . " WHERE isys_catg_smview_cable_pin_list__id = " . $this->convert_sql_id($pin['internal_pin_id']);
                $this->update($sql) && $this->apply_update();

                $sql = "UPDATE isys_catg_smview_cable_pin_list SET isys_catg_smview_cable_pin_list__patch_id = " . $pin['internal_pin_id']
                    . " WHERE isys_catg_smview_cable_pin_list__id = " . $this->convert_sql_id($patch_pin_id);
                $this->update($sql) && $this->apply_update();

                // remove old patch
                $sql = "UPDATE isys_catg_smview_cable_pin_list SET isys_catg_smview_cable_pin_list__patch_id = NULL"
                    . " WHERE isys_catg_smview_cable_pin_list__id = " . $this->convert_sql_id($old_sib_pin['id']);
                $this->update($sql) && $this->apply_update();
            }
        }

        if (count($connectedPins) == $preCount) {
            // this should not happen normally
            throw new \idoit\Exception\Exception('could not move wire to pin');
        }

        if (count($connectedPins) > 0) {
            // try to connect remaining PINs
            $this->_reconnect_wires($connectedPins);
        }
    }

    /**
     * Set wire pin status.
     *
     * @param  int $wire_pin_id
     * @param  int $status_id
     * @throws isys_exception_dao
     */
    public function set_wire_pin_status($wire_pin_id, $status_id)
    {
        $tableName = $this->table_name;
        $idField = $this->primary_key_field;
        $sql = "UPDATE $tableName SET isys_catg_smview_cable_wire_pin__status_id = " . $status_id
            . " WHERE $idField = " . $this->convert_sql_id($wire_pin_id);
        $this->update($sql) && $this->apply_update();
    }

    /**
     * Set extended location on a cable, based on extended location of its connected nodes.
     *
     * @param  int $wire_id
     * @param  boolean $displayMessages
     * @throws isys_exception_database
     * @throws Exception
     */
    private function _set_extended_location_on_cable($wire_id, $displayMessages = false)
    {
        // load both wire pins
        $wirePins = $this->filter_objects(array('wire_id' => $wire_id));
        if (empty($wirePins) || count($wirePins) != 2) {
            return;
        }

        // get extended location from both connected sides
        $lang = isys_application::instance()->container->get('language');
        $locationKey = 'isys_catg_smview_cable_extended_location_list__location_id';
        $pinDao = new isys_cmdb_dao_smview_cable_pin($this->m_db);
        $extendedLocationDao = new isys_cmdb_dao_category_g_smview_cable_extended_location($this->m_db);
        $extendedLocationId = null;
        foreach ($wirePins as $wirePin) {
            if (empty($wirePin['connected_pin_id'])) {
                // return if wire pin is not connected
                return;
            }

            // get connected pin
            $pin = $pinDao->get_object('id', $wirePin['connected_pin_id']);
            if (empty($pin['obj_id'])) {
                // return if pin is not assigned to an object
                return;
            }

            // get extended location of connected pin
            $location = $extendedLocationDao->get_data(null, $pin['obj_id'])->get_row();
            if (empty($location)) {
                // return if one connected pin has no extended location
                return;
            }

            if (empty($extendedLocationId)) {
                $extendedLocationId = $location[$locationKey];
            }

            if ($extendedLocationId != $location[$locationKey]) {
                // locations are not equal -> display warning message
                if ($displayMessages) {
                    isys_notify::warning($lang->get('LC__CMDB__CATG__SMVIEW_CABLE__WARNING_EXTENDED_LOCATIONS_NOT_EQUAL'), ['sticky' => true]);
                }
                return;
            }
        }

        // update location on cable
        $wireDao = new isys_cmdb_dao_smview_cable_wire($this->m_db);
        $wire = $wireDao->get_object('id', $wire_id);
        $cableObjId = $wire['obj_id'];
        $location = $extendedLocationDao->get_data(null, $cableObjId)->get_row();
        if (empty($location)) {
            // create extended location on cable
            $extendedLocationDao->create_data([
                'isys_obj__id' => $cableObjId,
                'location'     => $extendedLocationId,
            ]);
            if ($displayMessages) {
                isys_notify::success($lang->get('LC__CMDB__CATG__SMVIEW_CABLE__EXTENDED_LOCATION_SUCCESSFULLY_CREATED'), ['life' => 15]);
            }
        } elseif (empty($location[$locationKey])) {
            // update extended location on cable
            $extendedLocationDao->save_data($location['isys_catg_smview_cable_extended_location_list__id'], [
                'location'     => $extendedLocationId,
            ]);
            if ($displayMessages) {
                isys_notify::success($lang->get('LC__CMDB__CATG__SMVIEW_CABLE__EXTENDED_LOCATION_SUCCESSFULLY_UPDATED'), ['life' => 15]);
            }
        } else {
            // check if extended location matches currently
            if ($location[$locationKey] != $extendedLocationId && $displayMessages) {
                // locations are not equal anymore -> display warning message
                isys_notify::warning($lang->get('LC__CMDB__CATG__SMVIEW_CABLE__ERROR_EXTENDED_LOCATIONS_NOT_EQUAL'), ['life' => 15]);
            }
        }
    }
}
