<?php

/**
 * Cable ETSI AJAX handler
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class isys_ajax_handler_cable_etsi extends isys_ajax_handler
{

    /**
     * Init method, which gets called from the framework.
     *
     * @throws Exception
     */
    public function init()
    {
        // We set the header information because we don't accept anything than JSON.
        header('Content-Type: application/json');

        $l_return = [
            'success' => true,
            'message' => null,
            'data'    => null
        ];

        try {
            $canView = isys_auth_cmdb::instance()->is_allowed_to(isys_auth::VIEW, "CATEGORY/C__CATG__SMVIEW_CABLE_ETSI_BAY");
            switch ($this->m_get['func']) {
                case 'get-stations':
                    if ($canView) {
                        $location_id = $this->m_get['location_id'];
                        $l_return['data'] = $this->_get_stations($location_id);
                    }
                    break;
                case 'get-objects':
                    if ($canView) {
                        $station_id = $this->m_get['station_id'];
                        $l_return['data'] = $this->_get_objects($station_id);
                    }
                    break;
                case 'get-racks':
                    if ($canView) {
                        $obj_id = $this->m_get['obj_id'];
                        $l_return['data'] = $this->_get_racks($obj_id);
                    }
                    break;
            }
        }
        catch (Exception $e) {
            $l_return['success'] = false;
            $l_return['message'] = $e->getMessage();
        }

        echo isys_format_json::encode($l_return);

        $this->_die();
    }

    /**
     * This method defines, if the hypergate needs to be included for this request.
     *
     * @static
     * @return boolean
     */
    public static function needs_hypergate()
    {
        return true;
    }

    /**
     * Get stations of a location (subway line).
     *
     * @param  int $location_id
     * @return array
     * @throws isys_exception_database
     */
    private function _get_stations($location_id)
    {
        $station = defined_or_default('C__SHD__STATION');
        $dao = new isys_cmdb_dao_category_g_location($this->m_database_component);
        $res = $dao->get_data(null, null, "AND isys_obj_type__id=$station AND isys_catg_location_list__parentid=$location_id");

        // get parent locations of all stations
        $stations = [];
        while ($row = $res->get_row()) {
            $racks = $this->_get_objects($row['isys_obj__id']);
            if (count($racks) == 0) {
                continue;
            }

            $stations[$row['isys_obj__title']] = [
                'id'   => $row['isys_obj__id'],
                'name' => $row['isys_obj__title'],
            ];
        }
        ksort($stations);
        return array_values($stations);
    }

    /**
     * Get objects of a station.
     *
     * @param  int $station_id
     * @return array
     * @throws isys_exception_database
     */
    private function _get_objects($station_id)
    {
        $dao = new isys_cmdb_dao_category_g_smview_cable_etsi_bay($this->m_database_component);
        $export_dao = new isys_cmdb_dao_category_g_smview_cable_etsi_bay_export($this->m_database_component);
        $export_item_dao = new isys_cmdb_dao_smview_cable_etsi_bay_item_export($this->m_database_component);
        $location_dao = new isys_cmdb_dao_category_g_location($this->m_database_component);
        $res = $dao->get_data();
        $objects = [];
        while ($row = $res->get_row()) {
            $export_row = $export_dao->get_data(null, $row['isys_obj__id'])->get_row();
            if (empty($export_row)) {
                continue;
            }

            $path = $location_dao->get_location_path($row['isys_obj__id']);
            if (! in_array($station_id, $path)) {
                continue;
            }

            if (empty($objects[$row['isys_obj__id']])) {
                $items = $export_item_dao->filter_objects(['list_id' => $export_row['isys_catg_smview_cable_etsi_bay_export_list__id']]);
                $objects[$row['isys_obj__id']] = [
                    'id'         => $export_row['isys_catg_smview_cable_etsi_bay_export_list__id'],
                    'name'       => $row['isys_obj__title'],
                    'position'   => null,
                    'item_count' => count($items),
                ];
            }
        }
        return array_values($objects);
    }

    /**
     * Get racks of an object.
     *
     * @param  int $obj_id
     * @return array
     */
    private function _get_racks($obj_id)
    {
        $dao = new isys_cmdb_dao_category_g_smview_cable_etsi_bay($this->m_database_component);
        $res = $dao->get_data(null, $obj_id);
        $racks = [];
        while ($row = $res->get_row()) {
            $racks[] = [
                'id'          => $row['isys_catg_smview_cable_etsi_bay_list__id'],
                'name'        => $row['isys_catg_smview_cable_etsi_bay_list__name'],
                'description' => $row['isys_catg_smview_cable_etsi_bay_list__description'],
                'item_count'  => $row['isys_catg_smview_cable_etsi_bay_list__he'],
                'obj_title'   => $row['isys_obj__title'],
                'position'    => null,
            ];
        }
        return array_values($racks);
    }
}
