<?php

use idoit\Module\Cable\CableTracing;
use idoit\Module\Cable\ExcelExporter\DampingExporter;
use idoit\Module\Cable\Tracing;

/**
 * Cable section AJAX handler
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class isys_ajax_handler_cable_section extends isys_ajax_handler
{
    
    /**
     * Init method, which gets called from the framework.
     */
    public function init()
    {
        // We set the header information because we don't accept anything than JSON.
        header('Content-Type: application/json');

        $l_return = [
            'success' => true,
            'message' => null,
            'data'    => null
        ];

        try {
            $canView = isys_auth_cmdb::instance()->is_allowed_to(isys_auth::VIEW, "CATEGORY/C__CATG__SMVIEW_CABLE_SECTION");
            switch ($this->m_get['func']) {
                case 'get-objects-by-type':
                    if ($canView) {
                        $type_id = !empty($this->m_post['type_id']) ? $this->m_post['type_id'] : $this->m_get['type_id'];
                        $l_return['data'] = $this->_get_objects_by_type($type_id);
                    }
                    break;
                case 'get-connected-objects':
                    if ($canView) {
                        $obj_id = !empty($this->m_post['obj_id']) ? $this->m_post['obj_id'] : $this->m_get['obj_id'];
                        $tracing = new CableTracing($obj_id);
                        $ignored_wire_id = !empty($this->m_post['wire_id']) ? $this->m_post['wire_id'] : $this->m_get['wire_id'];
                        $l_return['data'] = $tracing->get_unpatched_connections($ignored_wire_id, $ignored_wire_id);
                    }
                    break;
                case 'trace':
                    if ($canView) {
                        $start_obj_id = !empty($this->m_post['start_obj_id']) ? $this->m_post['start_obj_id'] : $this->m_get['start_obj_id'];
                        $end_obj_id = !empty($this->m_post['end_obj_id']) ? $this->m_post['end_obj_id'] : $this->m_get['end_obj_id'];
                        $search_method = isset($this->m_post['search_method']) ? $this->m_post['search_method'] : 'shortest_path';
                        $l_return['data'] = $this->_trace($start_obj_id, $end_obj_id, $search_method);
                    }
                    break;
                case 'download':
                    if ($canView) {
                        $obj_id = !empty($this->m_post['obj_id']) ? $this->m_post['obj_id'] : $this->m_get['obj_id'];
                        self::downloadCalculatedDampingAsExcelFile($obj_id);
                    }
                    break;
                }
        }
        catch (isys_exception $e) {
            $l_return['success'] = false;
            $l_return['message'] = $e->getMessage();
        }

        echo isys_format_json::encode($l_return);

        $this->_die();
    }

    /**
     * Get objects by type.
     *
     * @param  int $type_id
     * @return array
     * @throws isys_exception_database
     */
    private function _get_objects_by_type($type_id)
    {
        $query = "SELECT 
                o.isys_obj__id,
                o.isys_obj__title
            FROM isys_catg_smview_cable_pin_list l
            JOIN isys_obj o ON l.isys_catg_smview_cable_pin_list__isys_obj__id = o.isys_obj__id
            JOIN isys_obj_type t ON o.isys_obj__isys_obj_type__id = t.isys_obj_type__id
            WHERE t.isys_obj_type__id=$type_id;";
        $dao = new isys_cmdb_dao_smview_cable_section(isys_application::instance()->database);
        $res = $dao->retrieve($query);
        
        $objects = array();
        while ($row = $res->get_row()) {
            $obj_id = $row['isys_obj__id'];
            $objects[$obj_id] = array(
                'id'    => $obj_id,
                'title' => $row['isys_obj__title'],
            );
        }
        
        uasort($objects, array($this, '_sortByTitle'));
        return array_values($objects);
    }

    /**
     * Sort array by title of an object.
     *
     * @param  array $a
     * @param  array $b
     * @return int
     */
    private function _sortByTitle($a, $b)
    {
        $left = $a['title'];
        $right = $b['title'];
        return strnatcmp($left, $right);
    }

    /**
     * Trace between two objects.
     *
     * @param  int $start_obj_id
     * @param  int $end_obj_id
     * @param  string $search_method
     * @return array
     */
    private function _trace($start_obj_id, $end_obj_id, $search_method)
    {
        $tracing = new Tracing($start_obj_id, false);
        if ($search_method == 'shortest_path') {
            $route = $tracing->get_shortest_path($end_obj_id);
        } else {
            $route = $tracing->get_least_number_of_hops($end_obj_id);
        }

        if (empty($route)) {
            return [];
        }

        $elements = [];
        $index = 1;
        foreach ($route as $part) {
            $wires = $part['wires'];
            $wire = null;
            // get first wire of shortest cable
            foreach ($wires as $w) {
                if (! $wire || $w['length'] < $wire['length']) {
                    $wire = $w;
                }
            }

            $element = array(
                'index'   => $index,
                'wire_id' => $wire['wire_id'],
            );
            $elements[$index] = isys_cmdb_dao_category_g_smview_cable_section_element::get_wire_connection($element);
            $index++;
        }

        // sort elements by index
        ksort($elements);
        $ordered_elements = isys_cmdb_dao_category_g_smview_cable_section_element::order_objects_within_connections($elements, $start_obj_id);

        // add connection information to each element
        $new_elements = [];
        foreach ($ordered_elements as $index => $element) {
            $new_elements[$index] =isys_cmdb_dao_category_g_smview_cable_section_element::add_connection_information($element);
        }

        // finally sort elements by index
        ksort($new_elements);
        return $new_elements;
    }

    /**
     * Download calculated damping as excel file.
     *
     * @param  int $obj_id
     * @throws \PhpOffice\PhpSpreadsheet\Exception
     * @throws \PhpOffice\PhpSpreadsheet\Writer\Exception
     * @throws isys_exception_database
     */
    public static function downloadCalculatedDampingAsExcelFile($obj_id)
    {
        $lang = isys_application::instance()->container->get('language');
        $dao = new isys_cmdb_dao_category_g_global(isys_application::instance()->database);
        $obj = $dao->get_data(null, $obj_id)->get_row();
        $exporter = new DampingExporter(isys_application::instance()->database);
        $path = $exporter->export_to_excel($obj);
        $title = $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__CALCULATED_DAMPING') . '-' . $obj['isys_obj__title'];

        header("Content-Description: File Transfer");
        header("Content-Disposition: attachment; filename=$title.xlsx");
        header("Content-Type: application/vnd.openxmlformats-officedocument.spreadsheetml.sheet");
        header("Content-Transfer-Encoding: binary");
        header("Content-length: " . filesize($path));
        header("Pragma: no-cache");
        header("Expires: 0");

        // read the file from disk
        readfile($path);
        die;
    }

    /**
     * This method defines, if the hypergate needs to be included for this request.
     *
     * @static
     * @return boolean
     */
    public static function needs_hypergate()
    {
        return true;
    }
}
