<?php

use idoit\Module\Cable\CableListImporter;
use idoit\Module\Cable\DependentObjectManager\AbstractManager;
use idoit\Module\Cable\DependentObjectManager\AdminUnitManager;
use idoit\Module\Cable\DependentObjectManager\CableTypeManager;
use idoit\Module\Cable\DependentObjectManager\ConnectionTypeManager;
use idoit\Module\Cable\DependentObjectManager\GeneratorManager;
use idoit\Module\Cable\DependentObjectManager\GroupManager;
use idoit\Module\Cable\DependentObjectManager\EtsiDeviceColorManager;
use idoit\Module\Cable\DependentObjectManager\EtsiExportManager;
use idoit\Module\Cable\DependentObjectManager\WirePinStatusManager;
use idoit\Module\Cable\DependentObjectManager\TechnologyManager;
use idoit\Module\Cable\DependentObjectManager\UserStateManager;
use idoit\Module\Cable\DependentObjectManager\RouteUsageManager;
use idoit\Module\Cable\ExcelExporter\EtsiBayExporter;
use idoit\Module\Cable\OrientDB;
use idoit\Module\Cable\ExcelExporter\RouteExporter;
use idoit\Module\Cable\SectionRouteValidator;
use PhpOrient\Exceptions\SocketException;

/**
 * i-doit
 *
 * SM-VIEW Cable module.
 *
 * @package i-doit
 * @subpackage smview_cable
 * @author Steve Groeber <steve.groeber@shd-online.de>
 * @copyright SHD System-Haus-Dresden GmbH
 * @license http://www.i-doit.com/license
 */
class isys_module_smview_cable extends isys_module {

    const DISPLAY_IN_MAIN_MENU = true;
    const DISPLAY_IN_SYSTEM_MENU = false;
    const DAMPING_1310 = 'smview_cable.kilometric_damping.1310nm';
    const DAMPING_1550 = 'smview_cable.kilometric_damping.1550nm';
    const OBJECT_TITLE_CHECK_ENABLED = 'smview_cable.object_title_check.enabled';
    const OBJECT_TITLE_CHECK_OBJECT_TYPES = 'smview_cable.object_title_check.object_types';
    const OBJECT_TITLE_CHECK_PROPERTIES = 'smview_cable.object_title_check.properties';
    const OBJECT_TITLE_CHECK_EXCLUDED_OBJECT_TYPES = 'smview_cable.object_title_check.excluded_object_types';
    const CAPACITY_COLOR_CURRENT_MODE = 'smview_cable.edge_color.current_mode';
    const EDGE_COLOR_ABSOLUTE_MODE = 'absolute';
    const CAPACITY_COLOR_RELATIVE_MODE = 'relative';
    const CAPACITY_COLOR_OK_COLOR = 'smview_cable.capacity_color.ok_color';
    const CAPACITY_COLOR_WARNING_COLOR = 'smview_cable.capacity_color.warning_color';
    const CAPACITY_COLOR_CRITICAL_COLOR = 'smview_cable.capacity_color.critical_color';
    const CAPACITY_COLOR_ABSOLUTE_WARNING = 'smview_cable.capacity_color.absolute.warning';
    const CAPACITY_COLOR_ABSOLUTE_CRITICAL = 'smview_cable.capacity_color.absolute.critical';
    const CAPACITY_COLOR_RELATIVE_WARNING = 'smview_cable.capacity_color.relative.warning';
    const CAPACITY_COLOR_RELATIVE_CRITICAL = 'smview_cable.capacity_color.relative.critical';
    const LINE_COLOR_SETTINGS = 'smview_cable.line_color.settings';

    /**
     * @var bool
     */
    protected static bool $m_licenced = true;

    /**
     * Variable which holds the database component.
     *
     * @var isys_component_database|null
     */
    protected ?isys_component_database $m_db = null;

    /**
     * Variable which holds the template component.
     *
     * @var isys_component_template|null
     */
    protected ?isys_component_template $m_tpl = null;
    
    /**
     * Variable which holds the template component.
     *
     * @var isys_component_tree|null
     */
    protected ?isys_component_tree $m_tree = null;

    /**
     * Property mapping.
     *
     * @var array
     */
    protected array $_config_mapping = array(
        'address'  => 'C__SMVIEW_CABLE__ADDRESS',
        'port'     => 'C__SMVIEW_CABLE__PORT',
        'database' => 'C__SMVIEW_CABLE__DATABASE',
        'username' => 'C__SMVIEW_CABLE__USERNAME',
        'password' => 'C__SMVIEW_CABLE__PASSWORD',
    );

    /**
     * Static method for retrieving the path, to the modules templates.
     *
     * @static
     * @return  string
     */
    public static function get_tpl_dir(): string {
        return __DIR__ . '/templates/';
    }

    /**
     * Represents the name, of the corresponding auth class - "false" means, there is none.
     * @var  string
     */
    const AUTH_CLASS_NAME = 'isys_auth_module_smview_cable';

    /**
     * Method for retrieving the path to the module directory (needed for includes).
     *
     * @static
     * @return  string
     */
    public static function get_module_dir(): string {
        return __DIR__ . DS;
    }
    
    /**
     * Method for retrieving the www-path to the assets directory (needed for the frontend).
     *
     * @static
     * @return  string
     */
    public static function get_assets_www_dir(): string {
        return isys_application::instance()->www_path . 'src/classes/modules/smview_cable/assets/';
    }

    /**
     * Initializes the module.
     *
     * @param isys_module_request & $p_req
     *
     * @return isys_module_smview_cable
     * @throws \PhpOffice\PhpSpreadsheet\Exception
     * @throws \PhpOffice\PhpSpreadsheet\Writer\Exception
     * @throws isys_exception_dao
     * @throws isys_exception_database
     */
    public function init(isys_module_request &$p_req): isys_module_smview_cable {
        $this->m_db = $p_req->get_database();
        $this->m_tpl = $p_req->get_template();
        $this->m_tree = $p_req->get_menutree();

        // add export button on list view mode of a route
        if ($_GET['objTypeID'] == defined_or_default('C__SHD__ROUTE') &&
            $_GET['viewMode'] == defined_or_default('C__CMDB__VIEW__LIST_OBJECT')) {

            // get row padding
            $userSettings = isys_application::instance()->container->get('settingsUser');
            switch ($userSettings->get('gui.category.padding', 'l')) {
                case 's':
                    $height = 340;
                    break;
                case 'm':
                    $height = 370;
                    break;
                case 'l':
                default:
                    $height = 400;
            }

            $lang = isys_application::instance()->container->get('language');
            $p_req->get_navbar()->append_button(
                $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__PATCH_EXPORT_TO_EXCEL'),
                'C__NAVBAR__BUTTON__EXPORT', [
                    'active'        => true,
                    'tooltip'       => $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__PATCH_EXPORT_TO_EXCEL'),
                    'icon'          => 'icons/silk/page_white_excel.png',
                    'icon_inactive' => 'icons/silk/page_white_excel.png',
                    'navmode'       => defined_or_default('C__NAVMODE__DUPLICATE'), // important to prevent JS syntax error
                    'js_onclick'    => " onclick=\"get_popup('smview_cable_route_export', null, '550', '$height', {'func':'export'});\""
                ]);
        }

        // handle route export
        if ($_POST['navMode'] && $_POST['navMode'] == 'export_route') {
            $route_ids = ! empty($_POST['route_ids']) ? $_POST['route_ids'] : null;
            $title = ! empty($_POST['title']) ? $_POST['title'] : null;
            $mergeRoutes = !empty($_POST['merge_routes']) && $_POST['merge_routes'] == 'on';
            $addDamping = !empty($_POST['add_damping']) && $_POST['add_damping'] == 'on';
            $addRedundantRoutes = !empty($_POST['add_redundant_routes']) && $_POST['add_redundant_routes'] == 'on';
            if ($route_ids) {
                $exporter = new RouteExporter($this->m_db);
                $routes = explode(',', $route_ids);
                $path = $exporter->export_to_excel($routes, $title, $mergeRoutes, $addDamping, $addRedundantRoutes);

                header("Content-Description: File Transfer");
                header("Content-Disposition: attachment; filename=$title.xlsx");
                header("Content-Type: application/vnd.openxmlformats-officedocument.spreadsheetml.sheet");
                header("Content-Transfer-Encoding: binary");
                header("Content-length: " . filesize($path));
                header("Pragma: no-cache");
                header("Expires: 0");

                // read the file from disk
                readfile($path);
                die;
            }
        }

        // handle ETSI export
        if ($_POST['navMode'] && $_POST['navMode'] == 'export_etsi_rack') {
            $export_config = ! empty($_POST['export_configuration']) ? $_POST['export_configuration'] : null;
            $bay_export_config = ! empty($_POST['bay_export_configuration']) ? $_POST['bay_export_configuration'] : null;
            if ($export_config || $bay_export_config) {
                $exporter = new EtsiBayExporter($this->m_db);
                if ($bay_export_config) {
                    $export_dao = new isys_cmdb_dao_category_g_smview_cable_etsi_bay_export($this->m_db);
                    $export = $export_dao->get_data($bay_export_config)->get_row();
                    $title = $export['isys_obj__title'];
                    $path = $exporter->export_bays_of_object($bay_export_config, $title);
                } else {
                    $export_dao = new isys_cmdb_dao_smview_cable_etsi_export($this->m_db);
                    $export = $export_dao->get_object('id', $export_config);
                    $title = $export['name'];
                    $path = $exporter->export_config($export_config);
                }

                header("Content-Description: File Transfer");
                header("Content-Disposition: attachment; filename=$title.xlsx");
                header("Content-Type: application/vnd.openxmlformats-officedocument.spreadsheetml.sheet");
                header("Content-Transfer-Encoding: binary");
                header("Content-length: " . filesize($path));
                header("Pragma: no-cache");
                header("Expires: 0");

                // read the file from disk
                readfile($path);
                die;
            }
        }

        return $this;
    }

    /**
     * Start method.
     *
     * @desc   Starts module process
     * @throws isys_exception_dao
     * @throws isys_exception_database
     * @throws Exception
     */
    public function start() {
        global $g_dirs;
        
        $l_template = $this->m_tpl;
        $l_tree = $this->m_tree;
        
        # index
        $language = isys_application::instance()->container->get('language');
        $moduleId = defined_or_default('C__MODULE__SMVIEW_CABLE');
        $l_tree->add_node(0, -1, $language->get("LC__CMDB__CATG__CABLE__TREE_TITLE"),
            isys_helper_link::create_url([
                C__GET__MODULE_ID => $moduleId,
                C__GET__TREE_NODE => 0,
            ])
        );
        
        # cable type management
        $l_tree->add_node(1, 0, $language->get("LC__CMDB__CATG__CABLE__TREE_CABLE_TYPE_MANAGEMENT"),
            isys_helper_link::create_url([
                C__GET__MODULE_ID => $moduleId,
                C__GET__TREE_NODE => 1,
                'request' => 'cable_type',
            ]),
            '',
            $g_dirs['images'] . 'icons/silk/application_view_icons.png'
        );
        
        # admin unit management
        $l_tree->add_node(2, 0, $language->get("LC__CMDB__CATG__CABLE__TREE_ADMIN_UNIT_MANAGEMENT"),
            isys_helper_link::create_url([
                C__GET__MODULE_ID => $moduleId,
                C__GET__TREE_NODE => 2,
                'request' => 'admin_unit',
            ]),
            '',
            $g_dirs['images'] . 'icons/silk/report.png'
        );
        
        # group management
        $l_tree->add_node(3, 0, $language->get("LC__CMDB__CATG__CABLE__TREE_GROUP_MANAGEMENT"),
            isys_helper_link::create_url([
                C__GET__MODULE_ID => $moduleId,
                C__GET__TREE_NODE => 3,
                'request' => 'group',
            ]),
            '',
            $g_dirs['images'] . 'icons/silk/table_multiple.png'
        );
        
        # connection type management
        $l_tree->add_node(4, 0, $language->get("LC__CMDB__CATG__CABLE__TREE_CABLE_CONNECTION_TYPE_MANAGEMENT"),
            isys_helper_link::create_url([
                C__GET__MODULE_ID => $moduleId,
                C__GET__TREE_NODE => 4,
                'request' => 'connection_type',
            ]),
            '',
            $g_dirs['images'] . 'icons/silk/connect-rotated.png'
        );
        
        # generator management
        $l_tree->add_node(5, 0, $language->get("LC__CMDB__CATG__CABLE__TREE_CABLE_GENERATOR_MANAGEMENT"),
            isys_helper_link::create_url([
                C__GET__MODULE_ID => $moduleId,
                C__GET__TREE_NODE => 5,
                'request' => 'generator',
            ]),
            '',
            $g_dirs['images'] . 'icons/silk/application_cascade.png'
        );
        
        # user status management
        $l_tree->add_node(6, 0, $language->get("LC__CMDB__CATG__CABLE__TREE_CABLE_USER_STATE_MANAGEMENT"),
            isys_helper_link::create_url([
                C__GET__MODULE_ID => $moduleId,
                C__GET__TREE_NODE => 6,
                'request' => 'user_state',
            ]),
            '',
            $g_dirs['images'] . 'icons/silk/group.png'
        );
        
        # technology management
        $l_tree->add_node(7, 0, $language->get("LC__CMDB__CATG__CABLE__TREE_TECHNOLOGY_MANAGEMENT"),
            isys_helper_link::create_url([
                C__GET__MODULE_ID => $moduleId,
                C__GET__TREE_NODE => 7,
                'request' => 'technology',
            ]),
            '',
            $g_dirs['images'] . 'icons/silk/telephone.png'
        );

        # route usage management
        $l_tree->add_node(8, 0, $language->get("LC__CMDB__CATG__CABLE__TREE_ROUTE_USAGE_MANAGEMENT"),
            isys_helper_link::create_url([
                C__GET__MODULE_ID => $moduleId,
                C__GET__TREE_NODE => 8,
                'request' => 'route_usage',
            ]),
            '',
            $g_dirs['images'] . 'icons/silk/chart_graph_blue.png'
        );

        # pin status management
        $l_tree->add_node(9, 0, $language->get("LC__CMDB__CATG__CABLE__TREE_WIRE_PIN_STATUS_MANAGEMENT"),
            isys_helper_link::create_url([
                C__GET__MODULE_ID => $moduleId,
                C__GET__TREE_NODE => 9,
                'request' => 'wire_pin_status',
            ]),
            '',
            $g_dirs['images'] . 'icons/silk/lightbulb.png'
        );

        # color management
        $l_tree->add_node(10, 0, $language->get("LC__CMDB__CATG__CABLE__TREE_COLOR_MANAGEMENT"),
            null,
            '',
            $g_dirs['images'] . 'icons/silk/color_swatch.png'
        );

        # ETSI device color management
        $l_tree->add_node(11, 10, $language->get("LC__CMDB__CATG__CABLE__TREE_ETSI_DEVICE_COLOR_MANAGEMENT"),
            isys_helper_link::create_url([
                C__GET__MODULE_ID => $moduleId,
                C__GET__TREE_NODE => 11,
                'request' => 'etsi_device_color',
            ]),
            '',
            $g_dirs['images'] . 'icons/silk/color_swatch.png'
        );

        # ETSI export management
        $l_tree->add_node(12, 0, $language->get("LC__CMDB__CATG__CABLE__TREE_ETSI_EXPORT_MANAGEMENT"),
            isys_helper_link::create_url([
                C__GET__MODULE_ID => $moduleId,
                C__GET__TREE_NODE => 12,
                'request' => 'etsi_export',
            ]),
            '',
            $g_dirs['images'] . 'icons/silk/page_white_get.png'
        );

        if (isys_auth_auth::instance()->is_allowed_to(isys_auth::EDIT, "MODULE/C__MODULE__SMVIEW_CABLE")) {
            $l_tree->add_node(
                13, 0, $language->get("LC__CMDB__CATG__SMVIEW_CABLE__CABLE_LIST_IMPORT_TITLE"),
                isys_helper_link::create_url(
                    [
                        C__GET__MODULE_ID => $moduleId,
                        C__GET__TREE_NODE => 13,
                        'request'         => 'importer',
                    ]
                ),
                '',
                $g_dirs['images'] . 'icons/silk/database_copy.png'
            );
        }

        # line color management
        $l_tree->add_node(14, 10, $language->get("LC__CMDB__CATG__CABLE__TREE_LINE_COLOR_MANAGEMENT"),
            isys_helper_link::create_url([
                C__GET__MODULE_ID => $moduleId,
                C__GET__TREE_NODE => 14,
                'request'         => 'line_colors',
            ]),
            '',
            $g_dirs['images'] . 'icons/silk/color_swatch.png'
        );

        # capacity color management
        $l_tree->add_node(15, 10, $language->get("LC__CMDB__CATG__CABLE__TREE_CAPACITY_COLOR_MANAGEMENT"),
            isys_helper_link::create_url([
                C__GET__MODULE_ID => $moduleId,
                C__GET__TREE_NODE => 15,
                'request'         => 'capacity_colors',
            ]),
            '',
            $g_dirs['images'] . 'icons/silk/color_swatch.png'
        );
        
        $l_template->assign("menu_tree", $l_tree->process($_GET[C__GET__TREE_NODE]));
        $l_posts = isys_module_request::get_instance()->get_posts();
        switch ($l_posts[C__GET__NAVMODE]) {
            case defined_or_default('C__NAVMODE__NEW'):
            case defined_or_default('C__NAVMODE__EDIT'):
                $canEdit = isys_auth_auth::instance()->is_allowed_to(isys_auth::EDIT, "MODULE/C__MODULE__SMVIEW_CABLE");
                isys_component_template_navbar::getInstance()
                    ->set_active($canEdit, defined_or_default('C__NAVBAR_BUTTON__SAVE'))
                    ->set_active($canEdit, defined_or_default('C__NAVBAR_BUTTON__CANCEL'));
                break;
        }
        
        if (isset($_GET['request'])) {
            $instance_id = $_GET['instance_id'] ?? null;
            $instance_id = $_POST['instance_id'] ?? $instance_id;
            $this->route($_GET['request'], $instance_id);
        } else {
            $this->route('/');
        }
    }

    /**
     * Routing.
     *
     * @param string $p_request
     * @param int|null $p_instance_id
     *
     * @throws isys_exception_dao
     * @throws isys_exception_database
     */
    public function route(string $p_request, $p_instance_id = null) {
        $lang = isys_application::instance()->container->get('language');
        $canEdit = isys_auth_auth::instance()->is_allowed_to(isys_auth::EDIT, "MODULE/C__MODULE__SMVIEW_CABLE");
        $l_template = $this->m_tpl;
        $l_template->assign('assets_dir', $this->get_assets_www_dir());
        $l_posts = isys_module_request::get_instance()->get_posts();
        
        if (empty($p_instance_id)) {
            $items = $l_posts['id'] ?? [];
        } else {
            $items = array($p_instance_id);
        }

        switch ($p_request) {
            case 'admin_unit':
                $manager = new AdminUnitManager();
                $l_template->include_template('contentbottomcontent', __DIR__ . '/templates/admin_unit_management.tpl');
                $this->_manage($manager, $l_posts, $p_instance_id, $items);
                break;
            case 'cable_type':
                $manager = new CableTypeManager();
                $l_template->include_template('contentbottomcontent', __DIR__ . '/templates/cable_type_management.tpl');
                $this->_manage($manager, $l_posts, $p_instance_id, $items);
                break;
            case 'group':
                $manager = new GroupManager();
                $l_template->include_template('contentbottomcontent', __DIR__ . '/templates/group_management.tpl');
                $this->_manage($manager, $l_posts, $p_instance_id, $items);
                break;
            case 'connection_type':
                $manager = new ConnectionTypeManager();
                $l_template->include_template('contentbottomcontent', __DIR__ . '/templates/connection_type_management.tpl');
                $this->_manage($manager, $l_posts, $p_instance_id, $items);
                break;
            case 'generator':
                $manager = new GeneratorManager();
                $l_template->include_template('contentbottomcontent', __DIR__ . '/templates/generator_management.tpl');
                $this->_manage($manager, $l_posts, $p_instance_id, $items);
                break;
            case 'user_state':
                $manager = new UserStateManager();
                $l_template->include_template('contentbottomcontent', __DIR__ . '/templates/user_state_management.tpl');
                $this->_manage($manager, $l_posts, $p_instance_id, $items);
                break;
            case 'technology':
                $manager = new TechnologyManager();
                $l_template->include_template('contentbottomcontent', __DIR__ . '/templates/technology_management.tpl');
                $this->_manage($manager, $l_posts, $p_instance_id, $items);
                break;
            case 'route_usage':
                $manager = new RouteUsageManager();
                $l_template->include_template('contentbottomcontent', __DIR__ . '/templates/route_usage_management.tpl');
                $this->_manage($manager, $l_posts, $p_instance_id, $items);
                break;
            case 'wire_pin_status':
                $manager = new WirePinStatusManager();
                $l_template->include_template('contentbottomcontent', __DIR__ . '/templates/wire_pin_status_management.tpl');
                $this->_manage($manager, $l_posts, $p_instance_id, $items);
                break;
            case 'etsi_device_color':
                $manager = new EtsiDeviceColorManager();
                $l_template->include_template('contentbottomcontent', __DIR__ . '/templates/etsi_device_color_management.tpl');
                $data = $_POST['C__CATG__SMVIEW_CABLE_ETSI_DEVICE_COLOR'];
                $data = json_decode($data, true);
                $this->_manage($manager, $l_posts, $p_instance_id, $items, $data);
                break;
            case 'etsi_export':
                $l_navbar = isys_component_template_navbar::getInstance();
                $l_navbar->append_button(
                    $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__ETSI_EXPORT_TO_EXCEL'),
                    'C__NAVBAR__BUTTON__EXPORT', [
                        'tooltip'       => $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__ETSI_EXPORT_TO_EXCEL'),
                        "icon"          => 'icons/silk/page_white_excel.png',
                        'icon_inactive' => 'icons/silk/page_white_excel.png',
                        "navmode"       => defined_or_default('C__NAVMODE__DUPLICATE'), // important to prevent JS syntax error
                        "js_onclick"    => " onclick=\"get_popup('smview_cable_etsi_export', null, '480', '200', {'func':'export'});\""
                    ]);

                $manager = new EtsiExportManager();
                $l_template->include_template('contentbottomcontent', __DIR__ . '/templates/etsi_export_management.tpl');
                $data = $_POST['C__CATG__SMVIEW_CABLE_ETSI_EXPORT'];
                $data = json_decode($data, true);
                $this->_manage($manager, $l_posts, $p_instance_id, $items, $data);
                break;
            case 'importer':
                if ($canEdit) {
                    $this->_handle_cable_list_import();
                }
                break;
            case 'line_colors':
                $this->m_tpl->include_template('contentbottomcontent', __DIR__ . '/templates/line_colors.tpl');

                $colors = self::get_line_colors();
                if (isys_glob_get_param(C__GET__NAVMODE) == defined_or_default('C__NAVMODE__SAVE') && $canEdit) {
                    foreach ($colors as $color) {
                        $id = $color['id'];
                        $c = $_POST['line-' . $id];
                        $c = strlen($c) == 6 ? $c : 'CCCCCC';
                        try {
                            self::set_line_color($color['title'], $c);
                        } catch (isys_exception_validation $e) {
                            isys_notify::error($lang->get('LC__UNIVERSAL__ERROR') . ' ' . $e->getMessage());
                        }
                    }
                    isys_notify::success($lang->get('LC__UNIVERSAL__SUCCESSFULLY_SAVED'), ['life' => 15]);
                }

                $colors = self::get_line_colors();
                $l_template->assign('lines', $colors);

                $l_navbar = isys_component_template_navbar::getInstance();
                $l_navbar->set_active($canEdit, defined_or_default('C__NAVBAR_BUTTON__SAVE'));
                $l_navbar->set_active($canEdit, defined_or_default('C__NAVBAR_BUTTON__CANCEL'));
                break;
            case 'capacity_colors':
                $this->m_tpl->include_template('contentbottomcontent', __DIR__ . '/templates/capacity_colors.tpl');

                $settings = self::get_capacity_color_settings();
                if (isys_glob_get_param(C__GET__NAVMODE) == defined_or_default('C__NAVMODE__SAVE') && $canEdit) {
                    $params = [];
                    foreach (array_keys($settings) as $key) {
                        $postKey = str_replace('.', '_', $key);
                        $value = $_POST[$postKey] ?? null;
                        switch ($key) {
                            case 'absolute.warning':
                            case 'absolute.critical':
                            case 'relative.warning':
                            case 'relative.critical':
                                $value = floatval($value);
                                break;
                        }
                        $params[$key] = $value;
                    }
                    try {
                        self::set_capacity_color_settings($params);
                        isys_notify::success($lang->get('LC__UNIVERSAL__SUCCESSFULLY_SAVED'), ['life' => 15]);
                    } catch (isys_exception_validation $e) {
                        isys_notify::error($lang->get('LC__UNIVERSAL__ERROR') . ' ' . $e->getMessage());
                    }
                }

                $settings = self::get_capacity_color_settings();
                foreach ($settings as $key => $value) {
                    $key = str_replace('.', '_', $key);
                    $l_template->assign($key, $value ?? '');
                }

                $l_navbar = isys_component_template_navbar::getInstance();
                $l_navbar->set_active($canEdit, defined_or_default('C__NAVBAR_BUTTON__SAVE'));
                $l_navbar->set_active($canEdit, defined_or_default('C__NAVBAR_BUTTON__CANCEL'));
                break;
            default:
                if (! empty($l_posts['start_validation']) && isset($l_posts['validate_configuration']) && $canEdit) {
                    $validator = new SectionRouteValidator($this->m_db);
                    $fixErrors = isset($l_posts['fix_errors']) && $l_posts['fix_errors'] == 'on';
                    switch ($l_posts['validate_configuration']) {
                        case 'section': $errors = $validator->validateSections($l_posts['limit'], $fixErrors); break;
                        case 'route': $errors = $validator->validateRoutes($l_posts['limit'], $fixErrors); break;
                        case 'patch': $errors = $validator->validatePatches($l_posts['limit'], $fixErrors); break;
                    }
                    $l_template->assign('validation_errors', $errors ?? null);
                    $l_template->assign('validation_type', $l_posts['validate_configuration']);
                }
                if ($l_posts[C__GET__NAVMODE] == defined_or_default('C__NAVMODE__SAVE')) {
                    $this->_save_config();
                    isys_notify::success($lang->get('LC__UNIVERSAL__SUCCESSFULLY_SAVED'), ['life' => 15]);
                }
                $this->_process_config_page();
                $l_template->include_template('contentbottomcontent', __DIR__ . '/templates/index.tpl');
                break;
        }
    }

    /**
     * Manage common CRUD operations.
     *
     * @param AbstractManager $manager
     * @param array           $l_posts
     * @param int|null        $p_instance_id
     * @param array|null      $items
     * @param array|null      $data
     */
    private function _manage(AbstractManager $manager, array $l_posts, $p_instance_id, $items, $data = null) {
        $data = $data ?: $_POST;
        switch ($l_posts[C__GET__NAVMODE]) {
            case defined_or_default('C__NAVMODE__NEW'):
                $manager->handleNewMode();
                $manager->processNewPage();
                break;
            case defined_or_default('C__NAVMODE__SAVE'):
                $manager->saveObject($p_instance_id, [], $data);
                $lang = isys_application::instance()->container->get('language');
                isys_notify::success($lang->get('LC__UNIVERSAL__SUCCESSFULLY_SAVED'), ['life' => 15]);
                $manager->processListing();
                break;
            case defined_or_default('C__NAVMODE__DELETE'):
                foreach ($items as $id) {
                    $manager->deleteObject($id);
                }
                $manager->processListing();
                break;
            default:
                if (! empty($items)) {
                    $id = $items[0];
                    $manager->processPage($id);
                } else {
                    $manager->processListing();
                }
        }
    }

    /**
     * Handle import of uploaded excel file.
     */
    private function _handle_cable_list_import()
    {
        global $g_dirs;

        $l_template = $this->m_tpl;
        $l_template->include_template('contentbottomcontent', __DIR__ . '/templates/cable_list_importer.tpl');
        $l_template->assign('encType', 'multipart/form-data');
        if ($_SERVER['REQUEST_METHOD'] === 'POST') {
            try {
                if ($_FILES['C__CATS__FILE_UPLOAD_CABLES']['size'] > 0) {
                    $key = 'C__CATS__FILE_UPLOAD_CABLES';
                    $new_file = $this->_save_uploaded_file($key);
                    $importer = new CableListImporter($this->m_db);
                    $log = $importer->import_cables_by_path($g_dirs["temp"] . $new_file);
                    $l_template->assign('log', $log);
                }
                $l_template->assign('success', true);
            } catch (Exception $e) {
                $l_template->assign('error', $e->getMessage());
            }
        } else {
            // this is necessary for the file-upload
            isys_component_template_navbar::getInstance()
                ->set_save_mode('formsubmit')
                ->append_button('LC__CMDB__CATG__SMVIEW_CABLE__NAVBAR_BUTTON_UPLOAD','upload', [
                    "icon"       => "icons/silk/page_save.png",
                    "js_onclick" => "document.isys_form.submit();",
                ]);
        }
    }

    /**
     * Save uploaded file.
     *
     * @param string $key
     *
     * @return string
     */
    private function _save_uploaded_file(string $key): string
    {
        global $g_dirs;

        $newFileName = isys_helper_upload::prepare_filename('cable_list__' . time() . '__' . $_FILES[$key]['name']);
        try {
            isys_helper_upload::save($_FILES[$key], $newFileName, $g_dirs["temp"], 0644);
        } catch (isys_exception_filesystem $e) {
            isys_notify::error($e->getMessage() . '<br />' . isys_helper_upload::get_error($key) . '<br />');
        }
        return $newFileName;
    }

    /**
     * Process cache configuration page.
     */
    private function _process_config_page()
    {
        $language = isys_application::instance()->container->get('language');
        $dao = new isys_cmdb_dao_smview_cable_orientdb($this->m_db);
        try {
            $type = $dao->get_object('id', 1);
        } catch (isys_exception_database $e) {
            $type = [];
        }
        $l_rules = [];
        foreach ($this->_config_mapping as $db => $field) {
            $l_rules[$field]['p_strValue'] = $type[$db] ?: '';
        }
        $l_rules['C__SMVIEW_CABLE__ADDRESS']['p_strPlaceholder'] = '1.2.3.4';
        $l_rules['C__SMVIEW_CABLE__PORT']['p_strPlaceholder'] = '2424';

        $l_navbar = isys_component_template_navbar::getInstance();
        $canEdit = isys_auth_auth::instance()->is_allowed_to(isys_auth::EDIT, "MODULE/C__MODULE__SMVIEW_CABLE");
        if ($_POST[C__GET__NAVMODE] == defined_or_default('C__NAVMODE__EDIT')) {
            $l_navbar->set_selected($canEdit, defined_or_default('C__NAVBAR_BUTTON__EDIT'))
                ->set_active($canEdit, defined_or_default('C__NAVBAR_BUTTON__SAVE'))
                ->set_active($canEdit, defined_or_default('C__NAVBAR_BUTTON__CANCEL'));
        } else {
            $l_navbar->set_active($canEdit, defined_or_default('C__NAVBAR_BUTTON__EDIT'))
                ->set_visible($canEdit, defined_or_default('C__NAVBAR_BUTTON__EDIT'));
        }

        $info = '';
        $error = '';
        try {
            $orient_db = new OrientDB();
            $data = $orient_db->get_last_update_info();
            if (! empty($data)) {
                $info = $data['last_update'] . ' (' . $data['duration'] . 's)';
            }
            $error = $language->get("LC__CMDB__CATG__CABLE__CONNECTION_SUCCEEDED");
        } catch (\PhpOrient\Exceptions\PhpOrientException $e) {
            if (strpos($e->getMessage(), 'Wrong user/password')) {
                $error = $language->get("LC__CMDB__CATG__CABLE__INVALID_USER_OR_PASSWORD");
            }
            if (strpos($e->getMessage(), 'Connection refused')) {
                $error = $language->get("LC__CMDB__CATG__CABLE__CONNECTION_REFUSED");
            }
            if (strpos($e->getMessage(), 'Operation now in progress')) {
                $error = $language->get("LC__CMDB__CATG__CABLE__HOST_UNREACHABLE");
            }
        } catch (SocketException $e) {
            $error = $language->get("LC__CMDB__CATG__CABLE__HOST_UNREACHABLE");
        }

        $damping_1310 = isys_settings::get(self::DAMPING_1310, null);
        $damping_1550 = isys_settings::get(self::DAMPING_1550, null);
        $l_rules['C__SMVIEW_CABLE__KILOMETRIC_DAMPING_1310']['p_strValue'] = $damping_1310 ?: '';
        $l_rules['C__SMVIEW_CABLE__KILOMETRIC_DAMPING_1550']['p_strValue'] = $damping_1550 ?: '';

        $check_enabled = isys_settings::get(self::OBJECT_TITLE_CHECK_ENABLED, null);
        $l_rules['C__SMVIEW_CABLE__OBJECT_TITLE_CHECK_ENABLED'] = [
            'p_arData'        => get_smarty_arr_YES_NO(),
            'p_strClass'      => 'input-small',
            'p_strSelectedID' => $check_enabled ?: null,
        ];

        $check_object_types = isys_settings::get(self::OBJECT_TITLE_CHECK_OBJECT_TYPES, null);
        $l_rules['C__SMVIEW_CABLE__OBJECT_TITLE_CHECK_OBJECT_TYPES'] = [
            'p_strTable'      => 'isys_obj_type',
            'p_strSelectedID' => $check_object_types,
            'multiselect'     => true,
        ];

        $check_properties = isys_settings::get(self::OBJECT_TITLE_CHECK_PROPERTIES, null);
        $properties = [];
        $selectableProperties = ['purpose', 'category', 'cmdb_status', 'status'];
        $global_dao = isys_cmdb_dao_category_g_global::instance($this->m_db);
        $global_properties = $global_dao->get_properties();
        foreach ($global_properties as $id => $p) {
            if (in_array($id, $selectableProperties)) {
                $properties[$id] = $language->get($p[C__PROPERTY__INFO][C__PROPERTY__INFO__TITLE]);
            }
        }
        $location_dao = isys_cmdb_dao_category_g_location::instance($this->m_db);
        $location_properties = $location_dao->get_properties();
        if ($location_properties['parent']) {
            $p = $location_properties['parent'];
            $properties['location'] = $language->get($p[C__PROPERTY__INFO][C__PROPERTY__INFO__TITLE]);
        }

        $l_rules['C__SMVIEW_CABLE__OBJECT_TITLE_CHECK_PROPERTIES'] = [
            'p_arData'        => $properties,
            'p_strSelectedID' => $check_properties,
            'multiselect'     => true,
        ];

        $excluded_object_types = isys_settings::get(self::OBJECT_TITLE_CHECK_EXCLUDED_OBJECT_TYPES, null);
        $l_rules['C__SMVIEW_CABLE__OBJECT_TITLE_CHECK_EXCLUDED_OBJECT_TYPES'] = [
            'p_strTable'      => 'isys_obj_type',
            'p_strSelectedID' => $excluded_object_types,
            'multiselect'     => true,
        ];

        isys_application::instance()->template
            ->assign('last_update', $info)
            ->assign('connection_info', $error)
            ->assign('image_url', isys_helper_link::get_base() . 'images/')
            ->assign('ajax_url', isys_helper_link::create_url([C__GET__AJAX => 1, C__GET__AJAX_CALL => 'cable_tracing', 'func' => 'sync-cache']))
            ->smarty_tom_add_rules("tom.content.bottom.content", $l_rules);
    }

    /**
     * Save cache configuration.
     *
     * @throws isys_exception_dao
     * @throws isys_exception_database
     */
    private function _save_config()
    {
        $dao = new isys_cmdb_dao_smview_cable_orientdb(isys_application::instance()->database);
        isys_component_template_navbar::getInstance()->set_active(isys_auth_auth::instance()->is_allowed_to(isys_auth::EDIT, "MODULE/C__MODULE__SMVIEW_CABLE"),
            defined_or_default('C__NAVBAR_BUTTON__EDIT'));

        $params = [];
        foreach ($this->_config_mapping as $db => $field) {
            $value = $_POST[$field] ?? null;
            $params[$db] = $value;
        }

        try {
            $type = $dao->get_object('id', 1);
            if (empty($type)) {
                $dao->create_object($params);
            } else {
                $dao->update_object(1, $params);
            }
        } catch (UnexpectedValueException $e) {
            $_POST["navMode"] = defined_or_default('C__NAVMODE__EDIT');
        }

        $damping_1310 = $this->_get_posted_damping('C__SMVIEW_CABLE__KILOMETRIC_DAMPING_1310');
        $damping_1550 = $this->_get_posted_damping('C__SMVIEW_CABLE__KILOMETRIC_DAMPING_1550');
        isys_settings::set(self::DAMPING_1310, $damping_1310);
        isys_settings::set(self::DAMPING_1550, $damping_1550);

        $enabled = $_POST['C__SMVIEW_CABLE__OBJECT_TITLE_CHECK_ENABLED'] ?? null;
        $object_types = $_POST['C__SMVIEW_CABLE__OBJECT_TITLE_CHECK_OBJECT_TYPES'] ?? [];
        $object_types = implode(',', $object_types);
        $properties = $_POST['C__SMVIEW_CABLE__OBJECT_TITLE_CHECK_PROPERTIES'] ?? [];
        $properties = implode(',', $properties);
        $excluded_object_types = $_POST['C__SMVIEW_CABLE__OBJECT_TITLE_CHECK_EXCLUDED_OBJECT_TYPES'] ?? [];
        $excluded_object_types = implode(',', $excluded_object_types);
        isys_settings::set(self::OBJECT_TITLE_CHECK_ENABLED, $enabled);
        isys_settings::set(self::OBJECT_TITLE_CHECK_OBJECT_TYPES, $object_types);
        isys_settings::set(self::OBJECT_TITLE_CHECK_PROPERTIES, $properties);
        isys_settings::set(self::OBJECT_TITLE_CHECK_EXCLUDED_OBJECT_TYPES, $excluded_object_types);

        $l_rules = [];
        foreach ($this->_config_mapping as $db => $field) {
            $l_rules[$field]['p_strValue'] = $params[$db];
        }
        isys_application::instance()->template->smarty_tom_add_rules("tom.content.bottom.content", $l_rules);
    }

    /**
     * Get posted damping value and convert it to float.
     *
     * @param string $fieldName
     *
     * @return float|null
     */
    private function _get_posted_damping(string $fieldName): ?float
    {
        $damping = $_POST[$fieldName] ?? null;
        $damping = str_replace(',', '.', $damping);
        return ! empty($damping) ? (float) $damping : null;
    }

    /**
     * Get capacity color settings of railway network.
     *
     * @return array
     */
    public static function get_capacity_color_settings(): array
    {
        return [
            'current_mode'      => isys_settings::get(self::CAPACITY_COLOR_CURRENT_MODE, self::CAPACITY_COLOR_RELATIVE_MODE),
            'ok_color'          => isys_settings::get(self::CAPACITY_COLOR_OK_COLOR, 'FF0000'),
            'warning_color'     => isys_settings::get(self::CAPACITY_COLOR_WARNING_COLOR, 'FFD800'),
            'critical_color'    => isys_settings::get(self::CAPACITY_COLOR_CRITICAL_COLOR, 'FF0000'),
            'absolute.warning'  => isys_settings::get(self::CAPACITY_COLOR_ABSOLUTE_WARNING, 10),
            'absolute.critical' => isys_settings::get(self::CAPACITY_COLOR_ABSOLUTE_CRITICAL, 5),
            'relative.warning'  => isys_settings::get(self::CAPACITY_COLOR_RELATIVE_WARNING, 50),
            'relative.critical' => isys_settings::get(self::CAPACITY_COLOR_RELATIVE_CRITICAL, 80),
        ];
    }

    /**
     * Set capacity color settings of railway network.
     *
     * @param array $params
     * @return void
     * @throws isys_exception_validation
     */
    public static function set_capacity_color_settings(array $params)
    {
        $mapping = [
            'current_mode'      => self::CAPACITY_COLOR_CURRENT_MODE,
            'ok_color'          => self::CAPACITY_COLOR_OK_COLOR,
            'warning_color'     => self::CAPACITY_COLOR_WARNING_COLOR,
            'critical_color'    => self::CAPACITY_COLOR_CRITICAL_COLOR,
            'absolute.warning'  => self::CAPACITY_COLOR_ABSOLUTE_WARNING,
            'absolute.critical' => self::CAPACITY_COLOR_ABSOLUTE_CRITICAL,
            'relative.warning'  => self::CAPACITY_COLOR_RELATIVE_WARNING,
            'relative.critical' => self::CAPACITY_COLOR_RELATIVE_CRITICAL,
        ];

        // validate
        $errors = [];
        foreach ($params as $key => $value) {
            switch ($key) {
                case 'current_mode':
                    if ($value != self::EDGE_COLOR_ABSOLUTE_MODE && $value != self::CAPACITY_COLOR_RELATIVE_MODE) {
                        $errors[$key] = 'unknown mode';
                    }
                    break;
                case 'ok_color':
                case 'warning_color':
                case 'critical_color':
                    if (! ctype_xdigit($value)) {
                        $errors[$key] = 'value is not an hexadecimal value';
                    }
                    // hex
                    break;
                case 'absolute.warning':
                case 'absolute.critical':
                case 'relative.warning':
                case 'relative.critical':
                    if (!is_int($value) && ! is_float($value)) {
                        $errors[$key] = 'value is not a float value';
                    }
                    break;
            }
        }

        if (! empty($errors)) {
            $message = 'could not save edge color settings: ' . implode(', ', $errors);
            throw new isys_exception_validation($message, $errors);
        }

        foreach ($params as $key => $value) {
            if (isset($mapping[$key])) {
                isys_settings::set($mapping[$key], $value);
            }
        }
    }

    /**
     * Get all railway line colors.
     *
     * @return array
     */
    public static function get_line_colors(): array
    {
        // get all lines by object type
        $raw_settings = isys_settings::get(self::LINE_COLOR_SETTINGS, null);
        $settings = [];
        if (! empty($raw_settings)) {
            $settings = is_string($raw_settings) ? json_decode($raw_settings, true) : $raw_settings;
        }

        $dao = isys_cmdb_dao_category_g_global::instance(isys_application::instance()->database);
        $type = defined_or_default('C__SHD__LINE');
        $lines = $dao->get_objects_by_type($type);

        // add color from settings
        $return = [];
        while ($row = $lines->get_row()) {
            $title = $row['isys_obj__title'];
            $return[$title] = [
                'id'    => $row['isys_obj__id'],
                'title' => $title,
                'color' => $settings[$title] ?? null,
            ];
        }
        ksort($return);
        return array_values($return);
    }

    /**
     * Set color of a railway line.
     *
     * @param string $line
     * @param string $color
     * @return void
     * @throws isys_exception_validation
     */
    public static function set_line_color(string $line, string $color)
    {
        // get all lines by object type
        $raw_settings = isys_settings::get(self::LINE_COLOR_SETTINGS, null);
        $settings = [];
        if (! empty($raw_settings)) {
            $settings = is_string($raw_settings) ? json_decode($raw_settings, true) : $raw_settings;
        }

        $dao = isys_cmdb_dao_category_g_global::instance(isys_application::instance()->database);
        $type = defined_or_default('C__SHD__LINE');
        $lines = $dao->get_objects_by_type($type);

        // try to find line by its title
        $found_line = false;
        while ($row = $lines->get_row()) {
            if ($row['isys_obj__title'] == $line) {
                $found_line = true;
            }
        }

        $errors = [];
        if (! $found_line) {
            $errors['line'] = 'could not find line by title';
        }

        if (! ctype_xdigit($color)) {
            $errors['color'] = 'color is not an hexadecimal value';
        }

        if (! empty($errors)) {
            $message = 'could not save line color: ' . implode(', ', $errors);
            throw new isys_exception_validation($message, $errors);
        }

        $settings[$line] = $color;
        $raw_settings = json_encode($settings);
        isys_settings::set(self::LINE_COLOR_SETTINGS, $raw_settings);
    }
}
