<?php

namespace idoit\Module\Cable\Algorithms;

use SplDoublyLinkedList;
use SplQueue;

require_once('IAlgorithm.php'); // needed for PHPUnit

/**
 * Least-Number-Of-Hops algorithm class.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class BreadthFirstSearch implements IAlgorithm
{

    /**
     * Graph.
     *
     * @var array
     */
    protected $graph;

    /**
     * Visited nodes.
     *
     * @var array
     */
    protected $visited = array();

    /**
     * Constructor.
     *
     * @param array $graph

        Structure of graph:

        $graph = array(
          'A' => array('B', 'F'),
          'B' => array('A', 'D', 'E'),
          'C' => array('F'),
          'D' => array('B', 'E'),
          'E' => array('B', 'D', 'F'),
          'F' => array('A', 'E', 'C'),
        );

     */
    public function __construct($graph)
    {
        $this->graph = $graph;
    }

    /**
     * Get least number of hops (using breadthFirstSearch).
     *
     * @param int $origin
     * @param int $destination
     * @return array
     */
    public function calculate($origin, $destination)
    {
        // mark all nodes as unvisited
        foreach ($this->graph as $vertex => $adj) {
            $this->visited[$vertex] = false;
        }

        // create an empty queue
        $q = new SplQueue();

        // enqueue the origin vertex and mark as visited
        $q->enqueue($origin);
        $this->visited[$origin] = true;

        // this is used to track the path back from each node
        $path = array();
        $path[$origin] = new SplDoublyLinkedList();
        $path[$origin]->setIteratorMode(
            SplDoublyLinkedList::IT_MODE_FIFO|SplDoublyLinkedList::IT_MODE_KEEP
        );

        $path[$origin]->push($origin);

        // while queue is not empty and destination not found
        while (!$q->isEmpty() && $q->bottom() != $destination) {
            $t = $q->dequeue();
            if (empty($this->graph[$t])) {
                continue;
            }
            
            // for each adjacent neighbor
            foreach ($this->graph[$t] as $vertex) {
                if (!$this->visited[$vertex]) {
                  // if not yet visited, enqueue vertex and mark as visited
                    $q->enqueue($vertex);
                    $this->visited[$vertex] = true;
                  
                  // add vertex to current path
                    $path[$vertex] = clone $path[$t];
                    $path[$vertex]->push($vertex);
                }
            }
        }

        $paths = [];
        if (isset($path[$destination])) {
            foreach ($path[$destination] as $vertex) {
                $paths[] = $vertex;
            }
        }
        
        return $paths;
    }
}
