<?php

namespace idoit\Module\Cable;

use isys_application;
use isys_cmdb_dao_smview_cable_user_state;
use isys_cmdb_dao_smview_cable_wire;
use isys_exception_database;

/**
 * Cable Capacity class.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  2024 SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class CableCapacity
{

    /**
     * User state names, which mark a wire as patched.
     */
    const PATCHED_USER_STATES = [
        'LWL defekt',
        'LWL gesperrt Umschaltung',
        'LWL reserviert',
    ];

    /**
     * Section to route mapping.
     *
     * @var array
     */
    private $sectionRouteMapping = [];

    /**
     * List of patched user state IDs.
     *
     * @var array
     */
    private $patchedUserStateIds = [];

    /**
     * Instance of this class.
     *
     * @var CableCapacity
     */
    private static $instance = null;

    /**
     * Return singleton instance of current class.
     *
     * @return CableCapacity
     */
    public static function instance()
    {
        if (! isset(self::$instance)) {
            $class = static::class;
            self::$instance = new $class();
        }
        return self::$instance;
    }

    /**
     * Constructor.
     */
    private function __construct()
    {
        $userStateDao = isys_cmdb_dao_smview_cable_user_state::instance(isys_application::instance()->database);
        $this->patchedUserStateIds = [];
        foreach (self::PATCHED_USER_STATES as $patchedUserState) {
            $userState = $userStateDao->get_object('name', $patchedUserState);
            if (! empty($userState)) {
                $this->patchedUserStateIds[] = $userState['id'];
            }
        }

        // todo: consider archived and deleted elements
        $res = $userStateDao->retrieve("SELECT * FROM isys_catg_smview_cable_route_element");
        $this->sectionRouteMapping = [];
        while ($row = $res->get_row()) {
            $this->sectionRouteMapping[$row['isys_catg_smview_cable_route_element__section_id']] = $row['isys_catg_smview_cable_route_element__route_id'];
        }
    }

    /**
     * Get capacity of a cable by its ID.
     *
     * @param int $cableObjId
     * @return array
     * @throws isys_exception_database
     */
    public function get_capacity_by_cable_id($cableObjId)
    {
        // load cable wires
        $wireDao = new isys_cmdb_dao_smview_cable_wire(isys_application::instance()->database);
        $sql = "SELECT w.*,
                       se.isys_catg_smview_cable_section_element__id,
                       se.isys_catg_smview_cable_section_element__section_id,
                       re.isys_catg_smview_cable_route_element__id
                FROM isys_catg_smview_cable_wire_list w
                LEFT JOIN isys_catg_smview_cable_section_element se ON se.isys_catg_smview_cable_section_element__wire_id = w.isys_catg_smview_cable_wire_list__id
                LEFT JOIN isys_catg_smview_cable_route_element re ON re.isys_catg_smview_cable_route_element__wire_id = w.isys_catg_smview_cable_wire_list__id
                WHERE w.isys_catg_smview_cable_wire_list__isys_obj__id = " . $wireDao->convert_sql_id($cableObjId);
        $res = $wireDao->retrieve($sql);
        $wireCount = $res->count();

        $patchedWireCount = 0;
        while ($row = $res->get_row()) {
            $routeElementId = $row['isys_catg_smview_cable_route_element__id'];

            if (! empty($row['isys_catg_smview_cable_section_element__section_id'])) {
                // check if this section is used within a route
                $sectionId = $row['isys_catg_smview_cable_section_element__section_id'];
                $routeElementId = $this->sectionRouteMapping[$sectionId];
            }

            if (! empty($routeElementId)) {
                // this wire is wired
                $patchedWireCount++;
            } elseif (in_array($row['isys_catg_smview_cable_wire_list__user_state_id'], $this->patchedUserStateIds)) {
                $patchedWireCount++;
            }
        }

        return array(
            'free_wires'    => $wireCount - $patchedWireCount,
            'patched_wires' => $patchedWireCount,
        );
    }
}
