<?php

namespace idoit\Module\Cable;

use Exception;
use isys_api_model_cmdb_dialog;
use isys_application;
use isys_cache;
use isys_cmdb_dao_category_g_global;
use isys_cmdb_dao_category_g_location;
use isys_cmdb_dao_smview_cable;
use isys_cmdb_dao_smview_cable_generator;
use isys_cmdb_dao_smview_cable_group;
use isys_cmdb_dao_smview_cable_pin;
use isys_cmdb_dao_smview_cable_type;
use isys_cmdb_dao_smview_cable_wire;
use isys_cmdb_dao_smview_cable_wire_pin;
use isys_component_database;
use isys_exception_api;
use isys_exception_auth;
use isys_exception_dao;
use isys_exception_dao_cmdb;
use isys_exception_database;
use isys_tenantsettings;
use PhpOffice\PhpSpreadsheet\Reader\Xlsx;

/**
 * Cable List Importer class.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  2023 SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class CableListImporter
{
    const CABLE_NUMBER = 'Kabelnummer';
    const CABLE_DESCRIPTION = 'Kabelbeschreibung';
    const CABLE_GROUP = 'Kabelgruppe';
    const CABLE_TYPE = 'Kabeltyp';
    const GENERATOR_TYPE = 'Generierungstyp';
    const CABLE_SCOPE = 'Leiterbereich';
    const WIRE_COUNT = 'Anzahl';

    const OBJECT_OUT = 'Objekt abg.';
    const OBJECT_TYPE_OUT = 'Objekttyp abg.';
    const OBJECT_CATEGORY_OUT = 'Kategorie abg.';
    const OBJECT_LOCATION_OUT = 'Standort abg.';
    const PIN_GENERATOR_OUT = 'Generierungstyp abg.';
    const PIN_SCOPE_OUT     = 'Bereich abg.';
    const START_NUMBER_OUT  = 'Startnummer abg.';

    const OBJECT_IN = 'Objekt ank.';
    const OBJECT_TYPE_IN = 'Objekttyp ank.';
    const OBJECT_CATEGORY_IN = 'Kategorie ank.';
    const OBJECT_LOCATION_IN = 'Standort ank.';
    const PIN_GENERATOR_IN = 'Generierungstyp ank.';
    const PIN_SCOPE_IN     = 'Bereich ank.';
    const START_NUMBER_IN  = 'Startnummer ank.';

    const CSV_DELIMITER = ';';

    /**
     * List of required parameters.
     *
     * @var array
     */
    private $_requiredParameters = [
        self::CABLE_NUMBER,
        self::CABLE_GROUP,
        self::CABLE_SCOPE,
        self::CABLE_DESCRIPTION,
        self::OBJECT_OUT,
        self::OBJECT_LOCATION_OUT,
        self::OBJECT_IN,
        self::OBJECT_LOCATION_IN,
    ];

    /**
     * List of CMDB locations.
     *
     * @var array
     */
    private $_cmdb_locations = [];

    /**
     * Logbook.
     *
     * @var array
     */
    private $_log = [];

    /**
     * Variable which holds the database component.
     *
     * @var isys_component_database
     */
    protected $_database = null;

    /**
     * Constructor.
     *
     * @param  isys_component_database $database
     * @throws Exception
     */
    public function __construct($database)
    {
        $this->_database = $database;
        $this->_cmdb_locations = $this->_get_cmdb_locations();
    }

    /**
     * Get a list of all locations below the root location.
     *
     * @return array
     * @throws Exception
     */
    private function _get_cmdb_locations()
    {
        ini_set('memory_limit', '1024M');

        // try to get locations from cache
        $m_cache_object = isys_cache::keyvalue();
        $m_cache_object->ns('bvg_cable_list');
        $cacheKey = "locations";
        $jsonLocations = $m_cache_object->get($cacheKey);
        if (! empty($jsonLocations)) {
            return json_decode($jsonLocations, true);
        }

        $dao = new isys_cmdb_dao_category_g_location(isys_application::instance()->database);
        $query = isys_cmdb_dao_category_g_location::build_location_path_query(8, 255, 40, '', true);
        $query = substr_replace($query, 'SELECT objMain.isys_obj__id, objMain.isys_obj__title, ', strpos($query, 'SELECT '), strlen('SELECT '));
        try {
            $res = $dao->retrieve($query);
        } catch (Exception $e) {
            // compatibility hack for version 1.19 (method parameters has changed)
            $query = isys_cmdb_dao_category_g_location::build_location_path_query(8, 40, '', true);
            $query = substr_replace($query, 'SELECT objMain.isys_obj__id, objMain.isys_obj__title, ', strpos($query, 'SELECT '), strlen('SELECT '));
            try {
                $res = $dao->retrieve($query);
            } catch (isys_exception_database $e) {
                return [];
            }
        }
        $locations = [];
        $separator = isys_tenantsettings::get('gui.separator.location', ' > ');
        $lang = isys_application::instance()->container->get('language');
        $translatedRootLocation = $lang->get('LC__OBJ__ROOT_LOCATION');
        while ($row = $res->get_row()) {
            if (empty($row['title']) || $row['isys_obj__title'] == $translatedRootLocation) {
                continue;
            }

            $parts = [];
            if ($row['title'] != $translatedRootLocation) {
                $parts = explode($separator, $row['title']);

                // remove 'Root location'
                array_shift($parts);
            }

            // add current object name
            $parts[] = $row['isys_obj__title'];

            $cleanParts = [];
            foreach ($parts as $part) {
                $cleanParts[] = trim($part);
            }

            $path = implode($separator, $cleanParts);
            $locations[$path] = (int) $row['isys_obj__id'];
        }

        // save locations to cache
        $m_cache_object->set($cacheKey, json_encode($locations), 1200);

        return $locations;
    }

    /**
     * Import cables.
     *
     * @param  string $path
     * @return array
     * @throws \PhpOffice\PhpSpreadsheet\Exception
     * @throws \PhpOffice\PhpSpreadsheet\Reader\Exception
     * @throws isys_exception_dao
     * @throws isys_exception_database
     */
    public function import_cables_by_path($path)
    {
        $this->_log = [];
        $cables = $this->_getElementsByPath($path);
        foreach ($cables as $rowIndex => $cable) {
            $this->_importCable($cable, $rowIndex);
        }
        return $this->_log;
    }

    /**
     * Import cables.
     *
     * @param  string $csv
     * @return array
     * @throws isys_exception_dao
     * @throws isys_exception_database
     */
    public function import_cables_by_csv($csv)
    {
        $this->_log = [];
        $cables = $this->_getElementsByCsv($csv);
        foreach ($cables as $rowIndex => $cable) {
            $this->_importCable($cable, $rowIndex);
        }
        return $this->_log;
    }

    /**
     * Get elements from excel file.
     *
     * @param  string $path
     * @param  string $startColumn
     * @param  int $currentRow
     * @return array
     * @throws \PhpOffice\PhpSpreadsheet\Exception
     * @throws \PhpOffice\PhpSpreadsheet\Reader\Exception
     */
    private function _getElementsByPath($path, $startColumn = 'A', $currentRow = 1)
    {
        $reader = new Xlsx();
        $reader->setReadDataOnly(true);
        $spreadsheet = $reader->load($path);
        $sheet = $spreadsheet->getActiveSheet();

        $columns = [];
        for ($i = 1; $i < 99; $i++) {
            $cell = $sheet->getCell($startColumn . $currentRow, false);
            if (!$cell) {
                break;
            }
            $columns[$startColumn] = $cell->getValue();
            $startColumn++;
        }

        $elements = [];
        $currentRow++;
        while (true) {
            $element = [];
            foreach ($columns as $index => $name) {
                $cell = $sheet->getCell($index . $currentRow, false);
                $value = $cell ? $cell->getValue() : null;
                $element[$name] = $value;

                if ($index == 'A' && $value == null) {
                    // break on first empty row
                    break 2;
                }
            }
            $elements[] = $element;
            $currentRow++;
        }
        return $elements;
    }

    /**
     * Get elements from CSV file.
     *
     * @param  string $csv
     * @return array
     */
    private function _getElementsByCsv($csv)
    {
        $rows = explode("\n", $csv);

        // get and cleanup first row
        $firstRow = array_shift($rows);
        $firstRow = explode(self::CSV_DELIMITER, $firstRow);
        $header = [];
        foreach ($firstRow as $i => $name) {
            if (! empty($name)) {
                $header[] = $name;
            }
        }

        $elements = [];
        foreach ($rows as $rowString) {
            $row = explode(self::CSV_DELIMITER, $rowString);
            $element = [];
            foreach ($header as $index => $columnName) {
                $value = $row[$index] ?? null;
                $element[$columnName] = $value;
            }
            $elements[] = $element;
        }
        return $elements;
    }

    /**
     * Import a cable.
     *
     * @param array $cable
     * @param int   $rowIndex
     *
     * @throws isys_exception_api
     * @throws isys_exception_auth
     * @throws isys_exception_dao
     * @throws isys_exception_dao_cmdb
     * @throws isys_exception_database
     * @throws Exception
     */
    private function _importCable($cable, $rowIndex)
    {
        $global_dao = new isys_cmdb_dao_category_g_global($this->_database);
        $location_dao = new isys_cmdb_dao_category_g_location($this->_database);
        $wire_dao = new isys_cmdb_dao_smview_cable_wire($this->_database);
        foreach ($this->_requiredParameters as $p) {
            if (!array_key_exists($p, $cable)) {
                // do not import incomplete cable
                $this->_log[$rowIndex][] = "ERROR: required parameter '$p' is missing";
                return;
            }
        }

        $cableName = $cable[self::CABLE_NUMBER];
        $rows = $global_dao->get_data_as_array(
            null,
            null,
            "AND isys_obj.isys_obj__title= '$cableName'",
            null,
            defined_or_default('C__RECORD_STATUS__NORMAL')
        );

        if (count($rows) > 1) {
            $this->_log[$rowIndex][] = "ERROR: multiple cables with name '$cableName' exists";
            return;
        }

        if (count($rows) == 1) {
            // get existing cable
            $row = $rows[0];
            $cableObjId = $row['isys_obj__id'];
        } else {
            // create new cable
            $cableObjId = $location_dao->create_object($cableName, defined_or_default('C__SHD__TELECOMMUNICATION_CABLE'));
            $this->_log[$rowIndex][] = "INFO: created new cable";
        }

        $wireCount = $cable[self::WIRE_COUNT];
        $generatorName = $cable[self::GENERATOR_TYPE];
        $generatorNameOut = $cable[self::PIN_GENERATOR_OUT];
        $generatorStartOut = $cable[self::START_NUMBER_OUT];
        $generatorNameIn = $cable[self::PIN_GENERATOR_IN];
        $generatorStartIn = $cable[self::START_NUMBER_IN];

        $wireNames = $this->_getGeneratedNames($generatorName, $wireCount, $rowIndex);
        $pinNamesOut = $this->_getGeneratedNames($generatorNameOut, $wireCount, $rowIndex, $generatorStartOut);
        $pinNamesIn = $this->_getGeneratedNames($generatorNameIn, $wireCount, $rowIndex, $generatorStartIn);

        $cableObject = $this->_addOrUpdateCableProperties($cable, $cableObjId, $rowIndex);
        $pinsOut = $this->_addOrUpdateObject($cable, $cableObject, $pinNamesOut, 'out', $rowIndex);
        $pinsIn = $this->_addOrUpdateObject($cable, $cableObject, $pinNamesIn, 'in', $rowIndex);
        $wires = $wire_dao->filter_objects(array('obj_id' => $cableObjId));
        foreach ($wireNames as $wireName) {
            $pinOutId = array_shift($pinsOut);
            $pinInId = array_shift($pinsIn);
            $scope = $cable[self::CABLE_SCOPE];
            $this->_importWire($wires, $wireName, $cableObjId, $scope, $rowIndex, $pinOutId, $pinInId);
        }
    }

    /**
     * Add or update cable properties.
     *
     * @param array $cable
     * @param int   $cableObjId
     * @param int   $rowIndex
     *
     * @return array
     * @throws isys_exception_dao
     * @throws isys_exception_database
     */
    private function _addOrUpdateCableProperties($cable, $cableObjId, $rowIndex)
    {
        $cable_dao = new isys_cmdb_dao_smview_cable($this->_database);
        $cableObject = $cable_dao->get_object('obj_id', $cableObjId);
        if (empty($cableObject)) {
            $params = [
                'obj_id' => $cableObjId,
            ];
            $cable_dao->create_object($params, []);
            $cableObject = $cable_dao->get_object('obj_id', $cableObjId);
        }

        // check description
        $global_dao = new isys_cmdb_dao_category_g_global($this->_database);
        $obj = $global_dao->get_object($cableObjId)->get_row();
        $description = $cable[self::CABLE_DESCRIPTION];
        if ($obj['isys_obj__description'] != $description) {
            // update object description
            $global_dao->update_object($cableObjId, null, null, $description);
            $this->_log[$rowIndex][] = "INFO: updated cable description";
        }

        $this->_addOrUpdateCableGroup($cableObject, $cable[self::CABLE_GROUP], $rowIndex);
        $this->_addOrUpdateCableType($cableObject, $cable[self::CABLE_TYPE], $cable[self::WIRE_COUNT], $rowIndex);
        return $cableObject;
    }

    /**
     * Add or update cable properties.
     *
     * @param array  $cable
     * @param array  $cableObject
     * @param array  $pinNames
     * @param string $type
     * @param int    $rowIndex
     *
     * @throws isys_exception_api
     * @throws isys_exception_auth
     * @throws isys_exception_dao
     * @throws isys_exception_dao_cmdb
     * @throws isys_exception_database
     */
    private function _addOrUpdateObject($cable, $cableObject, $pinNames, $type, $rowIndex)
    {
        $objectNameOut = $type == 'out' ? $cable[self::OBJECT_OUT] : $cable[self::OBJECT_IN];
        $objectTypeOut = $type == 'out' ? $cable[self::OBJECT_TYPE_OUT] : $cable[self::OBJECT_TYPE_IN];
        $locationPathOut = $type == 'out' ? $cable[self::OBJECT_LOCATION_OUT] : $cable[self::OBJECT_LOCATION_IN];
        $locationSide = $type == 'out' ? 'location_out_id' : 'location_in_id';
        $objOutId = $this->_addOrUpdateCableLocation($objectNameOut, $objectTypeOut, $locationPathOut, $cableObject, $locationSide, $rowIndex);

        $categoryOut = $type == 'out' ? $cable[self::OBJECT_CATEGORY_OUT] : $cable[self::OBJECT_CATEGORY_IN];
        $this->_addOrUpdateObjectCategory($objOutId, $categoryOut);

        $pinType = $type == 'out' ? isys_cmdb_dao_smview_cable_pin::PIN_TYPE_OUTPUT : isys_cmdb_dao_smview_cable_pin::PIN_TYPE_INPUT;
        $pinScope = $type == 'out' ? $cable[self::PIN_SCOPE_OUT] : $cable[self::PIN_SCOPE_IN];
        return $this->_addOrUpdateObjectPins($objOutId, $pinNames, $pinType, $pinScope, $rowIndex);
    }

    /**
     * Add or update cable location.
     *
     * @param string $name
     * @param string $typeName
     * @param string $path
     * @param array  $cableObject
     * @param string $param
     * @param int    $rowIndex
     *
     * @return int
     * @throws isys_exception_dao_cmdb
     */
    private function _addOrUpdateCableLocation($name, $typeName, $path, $cableObject, $param, $rowIndex)
    {
        // cleanup path and replace separators
        $parts = explode('/', $path);
        $parts[] = $name;
        $cleanParts = [];
        foreach ($parts as $part) {
            $part = trim($part);
            if (! empty($part)) {
                $cleanParts[] = $part;
            }
        }

        $separator = isys_tenantsettings::get('gui.separator.location', ' > ');
        $path = implode($separator, $cleanParts);
        $locationsObjId = $this->_cmdb_locations[$path];
        if (empty($locationsObjId)) {
            $global_dao = new isys_cmdb_dao_category_g_global($this->_database);
            $objType = null;
            switch ($typeName) {
                case 'LAN':
                    $objType = defined_or_default('C__BVG__DVT');
                    break;
                case 'KAG':
                    $objType = defined_or_default('C__BVG__KAG');
                    break;
                case 'LWL':
                    $objType = defined_or_default('C__BVG__LWL');
                    break;
                case 'Endgeräte':
                    $objType = defined_or_default('C__BVG__LINEUNIT');
                    break;
            }

            // create a new object
            $locationsObjId = $global_dao->create_object($name, $objType);
            $this->_log[$rowIndex][] = "INFO: added new object '$name'";

            // check if parent location is available
            array_pop($cleanParts);
            $parentPath = implode($separator, $cleanParts);
            $parentLocationsObjId = $this->_cmdb_locations[$parentPath];
            if (! empty($parentLocationsObjId)) {
                // assign new object to parent location
                $location_dao = new isys_cmdb_dao_category_g_location($this->_database);
                $location_dao->create($locationsObjId, $parentLocationsObjId);
            }
        }

        if ($locationsObjId != $cableObject[$param]) {
            $cable_dao = new isys_cmdb_dao_smview_cable($this->_database);
            $cable_dao->update_object($cableObject['id'], [$param => $locationsObjId]);
            $this->_log[$rowIndex][] = "INFO: updated cable location";
        }
        return $locationsObjId;
    }

    /**
     * Add or update existing cable group.
     *
     * @param array  $cableObject
     * @param string $name
     * @param int    $rowIndex
     *
     * @throws isys_exception_dao
     * @throws isys_exception_database
     */
    private function _addOrUpdateCableGroup($cableObject, $name, $rowIndex)
    {
        $cable_dao = new isys_cmdb_dao_smview_cable($this->_database);
        $cable_group_dao = new isys_cmdb_dao_smview_cable_group($this->_database);
        $elements = $cable_group_dao->filter_objects(array('description' => $name));
        if (count($elements) == 0) {
            // create cable group first
            $groupId = $cable_group_dao->create_object([
                'description'    => $name,
                'classification' => $name,
            ]);
            $this->_log[$rowIndex][] = "INFO: added new cable group '$name'";
        } else {
            // use first cable type
            $element = $elements[0];
            $groupId = $element['id'];
        }
        if ($groupId != $cableObject['group_id']) {
            $cable_dao->update_object($cableObject['id'], ['group_id' => $groupId]);
            $this->_log[$rowIndex][] = "INFO: updated cable group";
        }
    }

    /**
     * Add or update existing cable type.
     *
     * @param array  $cableObject
     * @param string $name
     * @param int    $wireCount
     * @param int    $rowIndex
     *
     * @throws isys_exception_dao
     * @throws isys_exception_database
     */
    private function _addOrUpdateCableType($cableObject, $name, $wireCount, $rowIndex)
    {
        $cable_dao = new isys_cmdb_dao_smview_cable($this->_database);
        $cable_type_dao = new isys_cmdb_dao_smview_cable_type($this->_database);
        $elements = $cable_type_dao->filter_objects(array('description' => $name));
        if (count($elements) == 0) {
            // create cable type first
            $typeId = $cable_type_dao->create_object([
                'description' => $name,
                'wire_count'  => $wireCount,
            ]);
            $this->_log[$rowIndex][] = "INFO: added new cable type '$name'";
        } else {
            // use first cable type
            $element = $elements[0];
            $typeId = $element['id'];
        }

        if ($typeId != $cableObject['type_id']) {
            $cable_dao->update_object($cableObject['id'], ['type_id' => $typeId]);
            $this->_log[$rowIndex][] = "INFO: updated cable type";
        }
    }

    /**
     * Add or update existing objects global category assignment.
     *
     * @param int $objId
     * @param string $categoryName
     *
     * @throws isys_exception_api
     * @throws isys_exception_auth
     * @throws isys_exception_database
     */
    private function _addOrUpdateObjectCategory($objId, $categoryName)
    {
        if (empty($categoryName)) {
            return;
        }

        $globalDao = new isys_cmdb_dao_category_g_global($this->_database);
        $obj = $globalDao->get_data(null, $objId)->get_row();
        if ($obj['isys_catg_global_category__title'] == $categoryName) {
            return;
        }

        $sql =  "SELECT * FROM isys_catg_global_category WHERE isys_catg_global_category__title = '$categoryName' AND isys_catg_global_category__status = " . C__RECORD_STATUS__NORMAL;
        $row = $globalDao->retrieve($sql)->get_row();
        if (empty($row)) {
            $cmdbDialog = new isys_api_model_cmdb_dialog($globalDao);
            $res = $cmdbDialog->create([
                'catgID'   => defined_or_default('C__CATG__GLOBAL'),
                'property' => 'category',
                'value'    => $categoryName,
            ]);
            $dialogItemId = $res['entry_id'];
        } else {
            $dialogItemId = $row['isys_catg_global_category__id'];
        }
        $globalDao->save_category($objId, $dialogItemId);
    }

    /**
     * Add or update cable pins on an object.
     *
     * @param int    $objId
     * @param array  $pinNames
     * @param string $pinType
     * @param string $pinScope
     * @param int    $rowIndex
     *
     * @return array
     * @throws isys_exception_database
     */
    private function _addOrUpdateObjectPins($objId, $pinNames, $pinType, $pinScope, $rowIndex)
    {
        $pin_dao = new isys_cmdb_dao_smview_cable_pin($this->_database);
        $currentPins = $pin_dao->filter_objects(['obj_id' => $objId]);
        $pins = [];
        foreach ($pinNames as $pinName) {
            $pinId = null;
            foreach ($currentPins as $pin) {
                if ($pin['description'] == $pinName) {
                    $pinId = $pin['id'];
                }
            }

            if (! $pinId) {
                $pinId = $pin_dao->create_pin($objId, $pinName, (int) $pinName, $pinType, null, null, null, $pinScope);
                $this->_log[$rowIndex][] = "INFO: created new wire '$pinName'";
            }
            $pins[$pinName] = $pinId;
        }
        return $pins;
    }

    /**
     * Get generated names by generator name.
     *
     * @param string      $generatorName
     * @param int         $wireCount
     * @param int         $rowIndex
     * @param string|null $startNumber
     *
     * @return array
     * @throws Exception
     */
    private function _getGeneratedNames($generatorName, $wireCount, $rowIndex, $startNumber = null)
    {
        $generator_dao = new isys_cmdb_dao_smview_cable_generator($this->_database);
        $generatorOut = $generator_dao->get_object('title', $generatorName);
        if (empty($generatorOut)) {
            $this->_log[$rowIndex][] = "WARN: could not find generator " . $generatorName;
            $startNumber = isset($startNumber) ? (int) $startNumber : 1;
            return range($startNumber, $wireCount);
        }

        $start = $startNumber ?? $generatorOut['counter_min'];
        $name_generator = new LabelGenerator($generatorOut['id']);
        return $name_generator->generate_names_by_generator($start, null, $wireCount);
    }

    /**
     * Import a wire.
     *
     * @param array    $existingWires
     * @param string   $wireName
     * @param int      $cableObjId
     * @param string   $scope
     * @param int      $rowIndex
     * @param int|null $pinIdOut
     * @param int|null $pinIdIn
     *
     * @throws isys_exception_dao
     * @throws isys_exception_database
     */
    private function _importWire($existingWires, $wireName, $cableObjId, $scope, $rowIndex, $pinIdOut = null, $pinIdIn = null)
    {
        $wire_dao = new isys_cmdb_dao_smview_cable_wire($this->_database);
        $wire_pin_dao = new isys_cmdb_dao_smview_cable_wire_pin($this->_database);
        $wireId = null;
        foreach ($existingWires as $wire) {
            if ($wire['description'] == $wireName) {
                $wireId = $wire['id'];

                // check scope
                if ($wire['scope'] != $scope) {
                    // update scope
                    $wire_dao->update_object($wire['id'], ['scope' => $scope]);
                    $this->_log[$rowIndex][] = "INFO: updated wire scope";
                }
                break;
            }
        }

        if (! $wireId) {
            // create new wire
            $wireId = $wire_dao->create_wire($cableObjId, $wireName, null, $scope);
            $this->_log[$rowIndex][] = "INFO: created new wire '$wireName'";
        }

        // update wire pins
        $pins = $wire_pin_dao->filter_objects(array('wire_id' => $wireId));
        $pinIn = $pins[0];
        $wire_pin_dao->update_object($pinIn['id'], [
            'description'      => $wireName,
            'connected_pin_id' => $pinIdIn,
        ]);

        $pinOut = $pins[1];
        $wire_pin_dao->update_object($pinOut['id'], [
            'description'      => $wireName,
            'connected_pin_id' => $pinIdOut,
        ]);
    }
}
