<?php

namespace idoit\Module\Cable;

use isys_application;
use isys_cmdb_dao_category_g_global;
use isys_cmdb_dao_smview_cable;
use isys_cmdb_dao_smview_cable_connection_view;
use isys_cmdb_dao_smview_cable_patch_view;
use isys_cmdb_dao_smview_cable_user_state;
use isys_exception_database;

/**
 * Cable Tracing class.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class CableTracing
{

    /**
     * User state names, which mark a wire as patched.
     */
    const PATCHED_USER_STATES = [
        'LWL defekt',
        'LWL gesperrt Umschaltung',
        'LWL reserviert',
    ];

    /**
     * Object ID.
     *
     * @var int
     */
    private $_objId = null;
    
    /**
     * List of connections.
     *
     * @var array
     */
    private $_connections = [];
    
    /**
     * List of patched wires.
     *
     * @var array
     */
    private $_patchedWires = [];

    /**
     * List of connected cables with its capacity.
     *
     * @var array
     */
    private $_cables = [];
    
    /**
     * Constructor.
     *
     * @param int $obj_id
     */
    public function __construct($obj_id)
    {
        $this->_objId = $obj_id;
        
        $dao_view = new isys_cmdb_dao_smview_cable_connection_view(isys_application::instance()->database);

        // filter open connections
        $all_connections = $dao_view->filter_objects(array('obj_id_in' => $obj_id));
        $connections = [];
        foreach ($all_connections as $c) {
            if (empty($c['pin_id_in']) || empty($c['pin_id_out'])) {
                continue;
            }
            $connections[] = $c;
        }
        $all_connections = $dao_view->filter_objects(array('obj_id_out' => $obj_id));
        foreach ($all_connections as $c) {
            if (empty($c['pin_id_in']) || empty($c['pin_id_out'])) {
                continue;
            }
            $connections[] = $c;
        }
        $this->_connections = $connections;

        $this->_cables = $this->_get_connected_cables();
    }

    /**
     * Get connections to this object.
     *
     * @return array
     */
    public function get_connections()
    {
        return $this->_connections;
    }

    /**
     * Get connected cables by object ID.
     * $selected_wire_id will be used to filter cable technology.
     *
     * @param  int $selected_wire_id
     * @return array
     */
    public function get_connected_cables($selected_wire_id = null)
    {
        $cables = [];

        // find cable type by selected wire id
        $cable_technology = $this->_get_technology_by_wire($selected_wire_id);
        foreach ($this->_cables as $cable_id => $cable) {
            $tc = $cable['technology_id'];
            if (isset($cable_technology) && $tc != $cable_technology) {
                // this cable uses another technology
                continue;
            }
            $cables[$cable_id] = $cable;
        }

        uasort($cables, array($this, '_sortByCableTitle'));
        return $cables;
    }

    /**
     * Get connected cables of this object.
     *
     * @return array
     */
    private function _get_connected_cables()
    {
        $connected_cables = [];
        foreach ($this->_connections as $c) {
            $cable_id = $c['cable_id'];
            $connected_cables[$cable_id] = array(
                'cable_title'   => $c['cable_title'],
                'technology_id' => $c['technology_id'],
            );
        }

        $cable_dao = isys_cmdb_dao_smview_cable::instance(isys_application::instance()->database);
        $cables = [];
        foreach ($connected_cables as $cable_id => $c) {
            // load cable object
            $cable_cat = $cable_dao->get_object('obj_id', $cable_id);
            $max_capacity = (int) $cable_cat['capacity'];

            $cableCapacity = CableCapacity::instance();
            $capacity = $cableCapacity->get_capacity_by_cable_id($cable_id);
            $capacity['cable_id'] = $cable_id;
            $capacity['cable_title'] = $c['cable_title'];
            $capacity['technology_id'] = $c['technology_id'];
            $capacity['max_capacity'] = $max_capacity;
            $cables[$cable_id] = $capacity;
        }

        uasort($cables, array($this, '_sortByCableTitle'));
        return $cables;
    }
    
    /**
     * Get cable technology by wire.
     *
     * @param  int|null $selected_wire_id
     * @return int|null
     */
    private function _get_technology_by_wire($selected_wire_id)
    {
        if (empty($selected_wire_id)) {
            return null;
        }
        
        // find cable type by selected wire id
        foreach ($this->_connections as $c) {
            $wire_id = $c['wire_id'];
            if ($wire_id == $selected_wire_id) {
                return $c['technology_id'];
            }
        }
        return null;
    }

    /**
     * Get unpatched connections.
     *
     * @param  int|null $ignored_wire_id            Will be ignored in connection listing
     * @param  int|null $tech_wire_id               Will be used to filter cable technology
     * @param  boolean $handleAssociatedRoute       Handles creation of an associated route
     * @param  boolean $considerPatchedConnections  Allow usage of patched connections
     * @return array
     * @throws isys_exception_database
     */
    public function get_unpatched_connections($ignored_wire_id = null, $tech_wire_id = null, $handleAssociatedRoute = false, $considerPatchedConnections = false)
    {
        $obj_dao = isys_cmdb_dao_category_g_global::instance(isys_application::instance()->database);
        $obj = $obj_dao->get_object_by_id($this->_objId)->get_row();

        $associated_element = null;
        if ($handleAssociatedRoute && ! empty($ignored_wire_id)) {
            $associated_element = $this->_get_next_wire($ignored_wire_id);
        }

        // get all unpatched connections of the incoming object
        $objects = $this->_get_unpatched_connections_by_wire_id($ignored_wire_id, $tech_wire_id, $considerPatchedConnections);
        return array(
            'element' => array(
                'id'    => $this->_objId,
                'title' => $obj['isys_obj__title'],
            ),
            'associated_element' => $associated_element,
            'objects' => array_values($objects),
        );
    }

    /**
     * Try to get next wire.
     *
     * @param  int $wire_id
     * @return array|null
     */
    private function _get_next_wire($wire_id)
    {
        $connection_dao = isys_cmdb_dao_smview_cable_connection_view::instance(isys_application::instance()->database);
        $wire = $connection_dao->get_object('wire_id', $wire_id);

        // try to get next wire number
        $wireNr = (int) $wire['wire_title'];
        $wireNr++;
        $nextWireTitle = str_pad($wireNr, strlen($wire['wire_title']), '0', STR_PAD_LEFT);
        $next_elements = $connection_dao->filter_objects(['cable_id' => $wire['cable_id'], 'wire_title' => $nextWireTitle]);
        $associated_element = count($next_elements) > 0 ? $next_elements[0] : null;
        if (! $associated_element) {
            return null;
        }

        $type_left = 'in';
        $type_right = 'out';
        if ($associated_element['obj_id_in'] == $this->_objId) {
            $type_left = 'out';
            $type_right = 'in';
        }
        return $this->_get_wire_information($associated_element, $type_left, $type_right);
    }
    
    /**
     * Sort array by title of an object.
     *
     * @param  array $a
     * @param  array $b
     * @return int
     */
    private function _sortByCableTitle($a, $b)
    {
        $left = $a['cable_title'];
        $right = $b['cable_title'];
        return strnatcmp($left, $right);
    }
    
    /**
     * Get unpatched connections by wire ID.
     *
     * @param  int $ignored_wire_id
     * @param  int $tech_wire_id
     * @param  boolean $considerPatchedConnections
     * @return array
     */
    private function _get_unpatched_connections_by_wire_id($ignored_wire_id = null, $tech_wire_id = null, $considerPatchedConnections = false)
    {
        $connections = [];
        $blockedCmdbStatus = [
            defined_or_default('C__CMDB_STATUS__BLOCKED', 0),
            defined_or_default('C__CMDB_STATUS__INOPERATIVE', 0),
        ];

        $userStateDao = new isys_cmdb_dao_smview_cable_user_state(isys_application::instance()->database);
        $defectState = $userStateDao->get_object('name', 'D');
        $defectStateId = $defectState ? $defectState['id'] : -1;
        
        // get connected object
        $connected_object = null;
        if (isset($ignored_wire_id)) {
            $connection_dao = isys_cmdb_dao_smview_cable_connection_view::instance(isys_application::instance()->database);
            $connection = $connection_dao->get_object('wire_id', $ignored_wire_id);
            $connected_object = $connection['obj_id_out'] == $this->_objId ? $connection['obj_id_in'] : $connection['obj_id_out'];
        }
        
        // find cable type by ignored/selected wire id
        $cable_technology = $this->_get_technology_by_wire($tech_wire_id);

        $this->_get_patched_wires($considerPatchedConnections);
        
        foreach ($this->_connections as $c) {
            $wire_id = $c['wire_id'];
            if (isset($ignored_wire_id) && $wire_id == $ignored_wire_id) {
                // ignore current wire
                continue;
            }
            
            if (isset($this->_patchedWires[$wire_id])) {
                // this connection is already patched
                continue;
            }

            if (! empty($c['user_state_id'])) {
                // do not list connections with user state
                continue;
            }

            if ($c['cable_user_state_id'] == $defectStateId) {
                continue;
            }

            if (in_array($c['cable_cmdb_status_id'], $blockedCmdbStatus)) {
                continue;
            }
            
            $tc = $c['technology_id'];
            if (isset($cable_technology) && $tc != $cable_technology) {
                // this cable uses another technology
                continue;
            }
                
            $type_left = 'in';
            $type_right = 'out';
            if ($c['obj_id_in'] == $this->_objId) {
                $type_left = 'out';
                $type_right = 'in';
            }
            
            if (empty($c['obj_id_' . $type_right])) {
                // ignore open connections
                continue;
            }
            
            if ($connected_object && $c['obj_id_' . $type_right] == $connected_object) {
                // ignore connections back to connected object
                continue;
            }
            $connections[$c['wire_id']] = $this->_get_wire_information($c, $type_left, $type_right);
        }
        return array_values($connections);
    }

    /**
     * Get a list of all patched wires.
     *
     * @param boolean $considerPatchedConnections
     */
    private function _get_patched_wires($considerPatchedConnections = false)
    {
        if ($considerPatchedConnections) {
            // set patched wires empty if all connections should be considered
            $this->_patchedWires = [];
            return;
        }

        if (! empty($this->_patchedWires)) {
            return;
        }

        // filter patched connections
        // (its not necessary to filter obj_id_out, because view contains redundant information for "in" and "out")
        $patch_view = new isys_cmdb_dao_smview_cable_patch_view(isys_application::instance()->database);
        $patches = $patch_view->filter_objects(array('obj_id_in' => $this->_objId));

        foreach ($patches as $p) {
            $this->_patchedWires[$p['wire_id']] = $p['wire_id'];
        }
    }

    /**
     * Get unpatched connections by object ID.
     *
     * @param  int $wire_id
     * @return array
     */
    public function get_connection_by_wire_id($wire_id)
    {
        $dao_view = new isys_cmdb_dao_smview_cable_connection_view(isys_application::instance()->database);
        $connections = $dao_view->filter_objects(array('wire_id' => $wire_id));
        if (empty($connections)) {
            return [];
        }

        $connection = $connections[0];
        $type_left = 'in';
        $type_right = 'out';
        if ($connection['obj_id_in'] == $this->_objId) {
            $type_left = 'out';
            $type_right = 'in';
        }
        return $this->_get_wire_information($connection, $type_left, $type_right);
    }

    /**
     * Get wire information with additional cable information.
     *
     * @param  array  $connection
     * @param  string $type_left
     * @param  string $type_right
     * @return array
     */
    private function _get_wire_information($connection, $type_left, $type_right)
    {
        $cable = $this->_cables[$connection['cable_id']];

        // always return requested pin on "obj_out" and matching connection on "obj_in"
        return array(
            'cable_id'            => $connection['cable_id'],
            'cable_title'         => $connection['cable_title'],
            'cable_free_wires'    => $cable['free_wires'],
            'cable_patched_wires' => $cable['patched_wires'],
            'cable_max_capacity'  => $cable['max_capacity'],
            'technology_id'       => $connection['technology_id'],
            'length'              => $connection['length'],
            'wire_id'             => $connection['wire_id'],
            'wire_title'          => $connection['wire_title'],
            'wire_scope'          => $connection['wire_scope'],
            'user_state_id'       => $connection['user_state_id'],
            'user_state'          => $connection['user_state'],
            'obj_in_id'           => $connection['obj_id_' . $type_left],
            'obj_in_title'        => $connection['obj_title_' . $type_left],
            'pin_in_title'        => $connection['pin_title_' . $type_left],
            'pin_in_scope'        => $connection['pin_scope_' . $type_left],
            'obj_out_id'          => $connection['obj_id_' . $type_right],
            'obj_out_title'       => $connection['obj_title_' . $type_right],
            'pin_out_title'       => $connection['pin_title_' . $type_right],
            'pin_out_scope'       => $connection['pin_scope_' . $type_right],
        );
    }
}
