<?php

namespace idoit\Module\Smview_Cable\Console\Command;

use Exception;
use idoit\Console\Command\AbstractCommand;
use isys_application;
use isys_cmdb_dao_category_g_global;
use isys_cmdb_dao_category_g_location;
use isys_cmdb_dao_dialog_admin;
use isys_cmdb_dao_object;
use isys_cmdb_dao_smview_cable;
use isys_cmdb_dao_smview_cable_etsi_device_color;
use isys_cmdb_dao_smview_cable_generator;
use isys_cmdb_dao_smview_cable_generator_algorithm;
use isys_cmdb_dao_smview_cable_pin;
use isys_cmdb_dao_smview_cable_technology;
use isys_cmdb_dao_smview_cable_type;
use isys_cmdb_dao_smview_cable_wire;
use isys_cmdb_dao_smview_cable_wire_pin;
use isys_cmdb_dao_smview_cable_wire_pin_status;
use isys_module_smview_cable;
use isys_settings;
use Symfony\Component\Console\Input\InputDefinition;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;

/**
 * Command to generate some test data.
 *
 * @package    i-doit
 * @subpackage smview
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  2023 SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class TestDataGeneratorCommand extends AbstractCommand
{

    /**
     * Get name for command
     *
     * @return string
     */
    public function getCommandName()
    {
        return 'smview_cable_test_data_generator';
    }

    /**
     * Get description for command
     *
     * @return string
     */
    public function getCommandDescription()
    {
        return 'Command to generate some SM-VIEW cable test data.';
    }

    /**
     * Retrieve Command InputDefinition
     *
     * @return InputDefinition
     */
    public function getCommandDefinition()
    {
        return new InputDefinition();
    }

    /**
     * Checks if a command can have a config file via --config
     *
     * @return bool
     */
    public function isConfigurable()
    {
        return false;
    }

    /**
     * Execute command.
     *
     * @param  InputInterface  $input
     * @param  OutputInterface $output
     * @return int
     * @throws Exception
     */
    protected function execute(InputInterface $input, OutputInterface $output)
    {
        try {
            $output->writeln("start\n");

            // add generator
            $generator = [
                'title'       => 'DA',
                'description' => 'DA',
                'mask'        => '#',
                'counter_min' => '01',
                'counter_max' => '02',
            ];
            $generator_dao = new isys_cmdb_dao_smview_cable_generator(isys_application::instance()->database);
            $obj = $generator_dao->get_object('title', 'DA');
            if (empty($obj)) {
                $generator_id = $generator_dao->create_object($generator);
                $algorithm_dao = isys_cmdb_dao_smview_cable_generator_algorithm::instance(isys_application::instance()->database);
                $params = array(
                    'generator_id'  => $generator_id,
                    'index'         => 1,
                    'mask_position' => 1,
                );
                $optional_params = array(
                    'type'        => 'n',
                    'length'      => 2,
                    'increment'   => 1,
                );
                $algorithm_dao->create_object($params, $optional_params);
                $output->writeln("added new Generator: 'DA'\n");
            }

            // add cable type
            $tech_dao = new isys_cmdb_dao_smview_cable_technology(isys_application::instance()->database);
            $tech_id = $tech_dao->create_object(array(
                'description' => 'LWL',
                'technology'  => 'LWL',
            ));

            $type_dao = new isys_cmdb_dao_smview_cable_type(isys_application::instance()->database);
            $type_id = $type_dao->create_object(array(
                'description'   => 'LWL',
                'wire_count'    => 14,
                'technology_id' => $tech_id,
            ));

            // add wire pin status
            $wire_pin_status_dao = new isys_cmdb_dao_smview_cable_wire_pin_status(isys_application::instance()->database);
            $din = $wire_pin_status_dao->create_object(array(
                'const'                    => 'DIN',
                'description'              => 'DIN',
                'upper_transition_damping' => 0.7,
            ));
            $wire_pin_status_id = $wire_pin_status_dao->create_object(array(
                'const'                    => 'E2000',
                'description'              => 'E2000',
                'upper_transition_damping' => 0.5,
            ));

            // add default cable
            $cable_dao = isys_cmdb_dao_object::instance(isys_application::instance()->database);
            $cable_id = $cable_dao->get_obj_id_by_title('Default-Cable', C__SHD__TELECOMMUNICATION_CABLE);
            $wire1_id = null;
            $wire2_id = null;
            if ($cable_id === false) {
                $cable_id = $cable_dao->create_object('Default-Cable', C__SHD__TELECOMMUNICATION_CABLE);

                $cable_catg_dao = new isys_cmdb_dao_smview_cable(isys_application::instance()->database);
                $cable_catg_dao->create_object(array(
                    'obj_id' => $cable_id,
                    'type_id' => $type_id,
                    'length'  => 400,
                ));

                $dao = new isys_cmdb_dao_smview_cable_wire(isys_application::instance()->database);
                $wire1_id = $dao->create_wire($cable_id, 'wire01');
                $wire2_id = $dao->create_wire($cable_id, 'wire02');
                $dao->create_wire($cable_id, 'wire03');
                $dao->create_wire($cable_id, 'wire04');

                $output->writeln("added new Cable: 'Default-Cable'\n");
            }

            // add second cable
            $cable_id = $cable_dao->get_obj_id_by_title('Second-Cable', C__SHD__TELECOMMUNICATION_CABLE);
            if ($cable_id === false) {
                $cable_id = $cable_dao->create_object('Second-Cable', C__SHD__TELECOMMUNICATION_CABLE);

                $cable_catg_dao = new isys_cmdb_dao_smview_cable(isys_application::instance()->database);
                $cable_catg_dao->create_object(array(
                    'obj_id' => $cable_id,
                    'type_id' => $type_id,
                ));

                $dao = new isys_cmdb_dao_smview_cable_wire(isys_application::instance()->database);
                $dao->create_wire($cable_id, 'wire01');
                $dao->create_wire($cable_id, 'wire02');
                $dao->create_wire($cable_id, 'wire03');
                $dao->create_wire($cable_id, 'wire04');

                $output->writeln("added new Cable: 'Second-Cable'\n");
            }

            // add LWLs
            $lwl2_2pin = null;
            for ($i = 1; $i <= 4; $i++) {
                $name = "LWL0$i";
                $lwl_dao = isys_cmdb_dao_object::instance(isys_application::instance()->database);
                $lwl01Id = $lwl_dao->get_obj_id_by_title($name, C__BVG__LWL);
                if ($lwl01Id === false) {
                    $lwl01Id = $lwl_dao->create_object($name, C__BVG__LWL);
                    $output->writeln("added new LWL: '$name'\n");

                    $dao = new isys_cmdb_dao_smview_cable_pin(isys_application::instance()->database);
                    $type_in = isys_cmdb_dao_smview_cable_pin::PIN_TYPE_INPUT;
                    $pin_in1_id = $dao->create_pin($lwl01Id, 'PIN01', 1, $type_in, null, null, null, 'Bu01Box01');
                    $pin_in2_id = $dao->create_pin($lwl01Id, 'PIN03', 3, $type_in, null, null, null, 'Bu01Box01');
                    $pin_in3_id = $dao->create_pin($lwl01Id, 'PIN05', 5, $type_in, null, null, null, 'Bu01Box02');
                    $pin_in4_id = $dao->create_pin($lwl01Id, 'PIN07', 7, $type_in, null, null, null, 'Bu01Box02');
                    $pin_in5_id = $dao->create_pin($lwl01Id, 'PIN09', 9, $type_in, null, null, null, 'Bu02Box01');
                    $pin_in6_id = $dao->create_pin($lwl01Id, 'PIN11', 11, $type_in, null, null, null, 'Bu02Box01');

                    $type_out = isys_cmdb_dao_smview_cable_pin::PIN_TYPE_OUTPUT;
                    $dao->create_pin($lwl01Id, 'PIN02', 2, $type_out, $pin_in1_id, null, null, 'Bu01Box01');
                    $dao->create_pin($lwl01Id, 'PIN04', 4, $type_out, $pin_in2_id, null, null, 'Bu01Box01');
                    $dao->create_pin($lwl01Id, 'PIN06', 6, $type_out, $pin_in3_id, null, null, 'Bu01Box02');
                    $dao->create_pin($lwl01Id, 'PIN08', 8, $type_out, $pin_in4_id, null, null, 'Bu01Box02');
                    $dao->create_pin($lwl01Id, 'PIN10', 10, $type_out, $pin_in5_id, null, null, 'Bu02Box01');
                    $dao->create_pin($lwl01Id, 'PIN12', 12, $type_out, $pin_in6_id, null, null, 'Bu02Box01');

                    if ($i == 2) {
                        $lwl2_2pin = $pin_in2_id;
                    }
                }
            }

            // connect PINs
            $dao = new isys_cmdb_dao_smview_cable_wire_pin(isys_application::instance()->database);
            $wire_pins = $dao->filter_objects(array('wire_id' => $wire1_id));
            $wire_pin_in = $wire_pins[0]['id'] < $wire_pins[1]['id'] ? $wire_pins[0] : $wire_pins[1];
            $wire_pin_out = $wire_pins[0]['id'] < $wire_pins[1]['id'] ? $wire_pins[1] : $wire_pins[0];
            $dao->set_wire_pin_status($wire_pin_in['id'], $din);
            $dao->set_wire_pin_status($wire_pin_out['id'], $din);

            $wire_pins = $dao->filter_objects(array('wire_id' => $wire2_id));
            $wire_pin_in = $wire_pins[0]['id'] < $wire_pins[1]['id'] ? $wire_pins[0] : $wire_pins[1];
            $wire_pin_out = $wire_pins[0]['id'] < $wire_pins[1]['id'] ? $wire_pins[1] : $wire_pins[0];
            $dao->update_wire_pin($wire_pin_in['id'], 'already connected', $wire_pin_status_id, null, $lwl2_2pin);
            $dao->set_wire_pin_status($wire_pin_out['id'], $wire_pin_status_id);

            // add ETSI setup
            $dao = new isys_cmdb_dao_category_g_location(isys_application::instance()->database);

            // add structure
            $subwayId = $dao->create_object('U-Bahn', defined_or_default('C__SHD__STRUCTURE'));
            $dao->create($subwayId, 1);

            // add subway line
            $u1Id = $dao->create_object('U1', defined_or_default('C__SHD__LINE'));
            $u2Id = $dao->create_object('U2', defined_or_default('C__SHD__LINE'));
            $dao->create($u1Id, $subwayId);
            $dao->create($u2Id, $subwayId);

            // add some purposes
            $purposeDao = isys_cmdb_dao_dialog_admin::instance(isys_application::instance()->database);
            $purpose1Id = $purposeDao->create('isys_purpose', 'ISDN', null, null, C__RECORD_STATUS__NORMAL);

            // assign purpose to subway lines
            $globalDao = new isys_cmdb_dao_category_g_global(isys_application::instance()->database);
            $catRow = $globalDao->get_data(null, $u1Id)->get_row();
            $globalDao->save_data($catRow['isys_catg_global_list__id'], [
                'purpose' => $purpose1Id,
            ]);
            $catRow = $globalDao->get_data(null, $u2Id)->get_row();
            $globalDao->save_data($catRow['isys_catg_global_list__id'], [
                'purpose' => $purpose1Id,
            ]);

            // add station
            $stationHoId = $dao->create_object('1136 Ho', defined_or_default('C__SHD__STATION'));
            $stationNuId = $dao->create_object('1141 Nu', defined_or_default('C__SHD__STATION'));
            $dao->create($stationHoId, $u1Id);
            $dao->create($stationNuId, $u1Id);

            // add LWL locations
            $name = "LWL01";
            $lwl01Id = $dao->get_obj_id_by_title($name, C__BVG__LWL);
            $dao->create($lwl01Id, $stationHoId);

            // add ETSI device colors
            $dao = new isys_cmdb_dao_smview_cable_etsi_device_color(isys_application::instance()->database);
            $dao->create_object([
                'description'        => 'Überlängenbox',
                'layout'             => '1',
                'title_left_top'     => 'Box',
                'color_left_top'     => '#D28212',
            ]);
            $dao->create_object([
                'description'        => 'Einzelspleiß + Pigtail',
                'layout'             => '3',
                'title_left_top'     => 'Spleiß',
                'color_left_top'     => '#FF2C9A',
                'title_right_top'    => '',
                'color_right_top'    => '#A88812',
                'title_right_bottom' => 'Pigtail',
                'color_right_bottom' => '#D28212',
            ]);

            $object_types = implode(',', [C__SHD__LINE, C__SHD__STATION]);
            $properties = implode(',', ['purpose']);
            isys_settings::set(isys_module_smview_cable::OBJECT_TITLE_CHECK_ENABLED, true);
            isys_settings::set(isys_module_smview_cable::OBJECT_TITLE_CHECK_OBJECT_TYPES, $object_types);
            isys_settings::set(isys_module_smview_cable::OBJECT_TITLE_CHECK_PROPERTIES, $properties);
            $output->writeln("added object title check settings'\n");

            $output->writeln("finished\n");
        } catch (Exception $e) {
            $output->writeln("");
            $output->writeln($e);
            $output->writeln("");
        }
        return 0;
    }

    /**
     * Returns an array of command usages
     *
     * @return array
     */
    public function getCommandUsages(): array
    {
        return [
            'smview_cable_test_data_generator -u admin -p admin -i 1',
        ];
    }
}
