<?php

namespace idoit\Module\Cable\DependentObjectManager;

use Exception;
use isys_application;
use isys_cmdb_dao_smview_cable_etsi_device_color;
use isys_cmdb_dao_smview_cable_etsi_device_color_wire_status;
use isys_cmdb_dao_smview_cable_wire_pin_status;
use isys_exception_dao;
use isys_exception_database;
use isys_module_smview_cable;

/**
 * ETSI Device Color Manager class.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class EtsiDeviceColorManager extends AbstractManager
{

    /**
     * Property mapping.
     *
     * @var array
     */
    protected $_mapping = array(
        'description'        => 'C__CATG__SMVIEW_ETSI_DEVICE_COLOR_DESCRIPTION',
        'layout'             => 'layout',
        'title_left_top'     => 'title_left_top',
        'color_left_top'     => 'color_left_top',
        'title_left_bottom'  => 'title_left_bottom',
        'color_left_bottom'  => 'color_left_bottom',
        'title_right_top'    => 'title_right_top',
        'color_right_top'    => 'color_right_top',
        'title_right_bottom' => 'title_right_bottom',
        'color_right_bottom' => 'color_right_bottom',
        'wire_count'         => 'C__CATG__SMVIEW_ETSI_DEVICE_COLOR_WIRE_COUNT',
    );

    /**
     * Constructor.
     */
    public function __construct()
    {
        $this->_dao = new isys_cmdb_dao_smview_cable_etsi_device_color(isys_application::instance()->database);
    }

    /**
     * Process generator listing.
     * @throws Exception
     */
    public function processListing()
    {
        $lang = isys_application::instance()->container->get('language');
        $title = $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__ETSI_DEVICE_COLORS');
        $headers = [
            "id"          => "ID",
            "description" => $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__ETSI_DEVICE_COLOR_DESCRIPTION'),
            "layout"      => $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__ETSI_DEVICE_COLOR_LAYOUT'),
            "wire_count"  => $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__ETSI_DEVICE_COLOR_WIRE_COUNT'),
            "wire_states" => $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__ETSI_DEVICE_COLOR_WIRE_STATUS'),
        ];
        $this->_processListing($title, $headers);
    }

    /**
     * Process new page.
     *
     * @param  array $additional_fields
     * @throws Exception
     */
    public function processNewPage($additional_fields = [])
    {
        $data = [
            'layout'             => '1',
            'title_left_top'     => null,
            'color_left_top'     => '#6E94FF',
            'title_left_bottom'  => null,
            'color_left_bottom'  => '#4E74DF',
            'title_right_top'    => null,
            'color_right_top'    => '#9EB7FF',
            'title_right_bottom' => null,
            'color_right_bottom' => '#CCD9FF',
            'wire_count'         => null,
            'wire_status_id'     => null,
        ];
        $js_config = [
            'language'        => $this->_get_js_translations(),
            'data'            => $data,
            'wire_states'     => [],
            'wire_pin_status' => $this->_getWirePinStatus(),
        ];

        $additional_fields['js_config'] = json_encode($js_config);
        $additional_fields['C__CATG__SMVIEW_CABLE_ETSI_DEVICE_COLOR'] = json_encode($data);
        $additional_fields['assets_dir'] = isys_module_smview_cable::get_assets_www_dir();

        parent::processNewPage($additional_fields);
    }

    /**
     * Process page.
     *
     * @param  int $element_id
     * @param  array $additional_fields
     * @throws Exception
     */
    public function processPage($element_id, $additional_fields = [])
    {
        $row = $this->_dao->get_object('id', $element_id);
        unset($row['description']);

        $dao = new isys_cmdb_dao_smview_cable_etsi_device_color_wire_status(isys_application::instance()->database);
        $wire_status = $dao->filter_objects(['device_id' => $element_id]);

        $states = [];
        foreach ($wire_status as $index => $ws) {
            $states[] = [
                'id'                 => $index + 1,
                'wire_pin_status_id' => $ws['status_id'],
                'operator'           => $ws['operator'],
                'not'                => $ws['not'],
            ];
        }

        $row['wire_states'] = $states;
        $js_config = [
            'language'        => $this->_get_js_translations(),
            'data'            => $row,
            'wire_pin_status' => $this->_getWirePinStatus(),
        ];

        $additional_fields['js_config'] = json_encode($js_config);
        $additional_fields['C__CATG__SMVIEW_CABLE_ETSI_DEVICE_COLOR'] = json_encode($row);
        $additional_fields['assets_dir'] = isys_module_smview_cable::get_assets_www_dir();

        parent::processPage($element_id, $additional_fields);
    }

    /**
     * Get translated strings to use with javascript.
     *
     * @return array
     * @throws Exception
     */
    private function _get_js_translations()
    {
        // language templates
        $language_templates = [
            'LC__CMDB__CATG__SMVIEW_CABLE__ETSI_DEVICE_COLOR_DESCRIPTION',
            'LC__CMDB__CATG__SMVIEW_CABLE__ETSI_DEVICE_COLOR_PLACEHOLDER',
            'LC__CMDB__CATG__SMVIEW_CABLE__ETSI_DEVICE_COLOR_NOT',
            'LC__CMDB__CATG__SMVIEW_CABLE__ETSI_DEVICE_COLOR_WIRE_STATUS',
            'LC__CMDB__CATG__SMVIEW_CABLE__ETSI_DEVICE_COLOR_OPERAND',
            'LC__CMDB__CATG__SMVIEW_CABLE__ETSI_DEVICE_COLOR_ADD_RULE',
            'LC__CMDB__CATG__SMVIEW_CABLE__ETSI_DEVICE_COLOR_REMOVE_RULE',
            'LC__CMDB__CATG__SMVIEW_CABLE__ETSI_DEVICE_COLOR_AND',
            'LC__CMDB__CATG__SMVIEW_CABLE__ETSI_DEVICE_COLOR_OR',
        ];

        // translate
        $language = [];
        $lang = isys_application::instance()->container->get('language');
        foreach ($language_templates as $l) {
            $language[$l] = $lang->get($l);
        }
        return $language;
    }

    /**
     * Get all wire pin status.
     *
     * @return array
     * @throws isys_exception_database
     */
    private function _getWirePinStatus()
    {
        $dao = new isys_cmdb_dao_smview_cable_wire_pin_status(isys_application::instance()->database);
        $all_status = $dao->get_objects();
        $status = [];
        foreach ($all_status as $s) {
            $status[] = [
                'id'    => $s['id'],
                'const' => $s['const'],
            ];
        }
        return $status;
    }

    /**
     * Save object.
     *
     * @param  int|null $element_id
     * @param  array $optionalParams
     * @param  array|null $data
     * @throws isys_exception_dao
     * @throws isys_exception_database
     */
    public function saveObject($element_id = null, $optionalParams = [], $data = null)
    {
        parent::saveObject($element_id, $optionalParams, $data);
        $id = isys_application::instance()->template->get_template_vars("instance_id");

        // remove current wire status
        $dao = new isys_cmdb_dao_smview_cable_etsi_device_color_wire_status(isys_application::instance()->database);
        $ws = $dao->filter_objects(['device_id' => $id]);
        foreach ($ws as $s) {
            $dao->delete_object($s['id']);
        }

        // save wire status
        foreach ($data['wire_states'] as $ws) {
            if (! empty($ws['wire_pin_status_id'])) {
                $dao->create_object([
                    'device_id' => $id,
                    'status_id' => $ws['wire_pin_status_id'],
                    'operator'  => $ws['operator'] ?? null,
                    'not'       => ! empty($ws['not']) ? 1 : null,
                ]);
            }
        }
    }
}
