<?php

namespace idoit\Module\Cable\DependentObjectManager;

use Exception;
use isys_application;
use isys_cmdb_dao_category_g_location;
use isys_cmdb_dao_category_g_smview_cable_etsi_bay;
use isys_cmdb_dao_category_g_smview_cable_etsi_bay_export;
use isys_cmdb_dao_smview_cable_etsi_bay_item_export;
use isys_cmdb_dao_smview_cable_etsi_export;
use isys_cmdb_dao_smview_cable_etsi_export_object;
use isys_exception_database;
use isys_helper_link;
use isys_module_smview_cable;

/**
 * ETSI Export Manager class.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  2020 SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class EtsiExportManager extends AbstractManager
{

    /**
     * Property mapping.
     *
     * @var array
     */
    protected $_mapping = array(
        'name'        => 'C__CATG__SMVIEW_ETSI_EXPORT_NAME',
        'description' => 'C__CATG__SMVIEW_ETSI_EXPORT_DESCRIPTION',
    );

    /**
     * Constructor.
     */
    public function __construct()
    {
        $this->_dao = new isys_cmdb_dao_smview_cable_etsi_export(isys_application::instance()->database);
    }

    /**
     * Process generator listing.
     * @throws Exception
     */
    public function processListing()
    {
        $lang = isys_application::instance()->container->get('language');
        $title = $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__ETSI_EXPORT');
        $headers = [
            "id"          => "ID",
            "name"        => $lang->get('LC_UNIVERSAL__NAME'),
            "description" => $lang->get('LC__UNIVERSAL__DESCRIPTION')
        ];
        $this->_processListing($title, $headers);
    }

    /**
     * Process new page.
     *
     * @param  array $additional_fields
     * @throws Exception
     */
    public function processNewPage($additional_fields = [])
    {
        $additional_fields['js_config'] = json_encode($this->_get_js_config());
        $additional_fields['C__CATG__SMVIEW_CABLE_ETSI_EXPORT'] = json_encode([]);
        $additional_fields['assets_dir'] = isys_module_smview_cable::get_assets_www_dir();
        parent::processNewPage($additional_fields);
    }

    /**
     * Process page.
     *
     * @param  int $element_id
     * @param  array $additional_fields
     * @throws Exception
     */
    public function processPage($element_id, $additional_fields = [])
    {
        $dao = new isys_cmdb_dao_smview_cable_etsi_export_object(isys_application::instance()->database);
        $bay_dao = new isys_cmdb_dao_category_g_smview_cable_etsi_bay(isys_application::instance()->database);
        $bay_export_dao = new isys_cmdb_dao_category_g_smview_cable_etsi_bay_export(isys_application::instance()->database);
        $export_item_dao = new isys_cmdb_dao_smview_cable_etsi_bay_item_export(isys_application::instance()->database);
        $rack_objects = $dao->filter_objects(['export_id' => $element_id]);
        $objects = [];
        foreach ($rack_objects as $r) {
            $bay_export = $bay_export_dao->get_data($r['bay_export_list_id'])->get_row();
            $bay = $bay_dao->get_data(null, $bay_export['isys_catg_smview_cable_etsi_bay_export_list__isys_obj__id'])->get_row();
            $items = $export_item_dao->filter_objects(['list_id' => $bay_export['isys_catg_smview_cable_etsi_bay_export_list__id']]);
            $objects[] = [
                'id'          => $r['bay_export_list_id'],
                'name'        => $bay['isys_obj__title'],
                'description' => $bay['isys_catg_smview_cable_etsi_bay_list__description'],
                'item_count'  => count($items),
                'obj_title'   => $bay['isys_obj__title'],
                'position'    => $r['position'],
            ];
        }

        $additional_fields['js_config'] = json_encode($this->_get_js_config($objects));
        $additional_fields['C__CATG__SMVIEW_CABLE_ETSI_EXPORT'] = json_encode(['objects' => $objects]);
        $additional_fields['assets_dir'] = isys_module_smview_cable::get_assets_www_dir();
        parent::processPage($element_id, $additional_fields);
    }

    /**
     * Get subway lines.
     *
     * @return array
     * @throws isys_exception_database
     */
    private function _get_subway_lines()
    {
        $station = defined_or_default('C__SHD__STATION');
        $dao = new isys_cmdb_dao_category_g_location(isys_application::instance()->database);
        $res = $dao->get_data(null, null, "AND isys_obj_type__id=$station");

        // get parent locations of all stations
        $locations = [];
        while ($row = $res->get_row()) {
            $location_obj_id = $row['isys_catg_location_list__parentid'];
            $obj = $dao->get_object($location_obj_id)->get_row();
            $name = $obj['isys_obj__title'];
            $locations[$name] = [
                'id'   => $obj['isys_obj__id'],
                'name' => $name,
            ];
        }
        ksort($locations);
        return array_values($locations);
    }

    /**
     * Get translated strings to use with javascript.
     *
     * @param  array $objects
     * @return array
     * @throws Exception
     */
    private function _get_js_config($objects = [])
    {
        // language templates
        $language_templates = [
            'LC__CMDB__CATG__SMVIEW_CABLE__ETSI_ADD_AS_LAST',
            'LC__CMDB__CATG__SMVIEW_CABLE__ETSI_STATIONS',
            'LC__CMDB__CATG__SMVIEW_CABLE__ETSI_OBJECTS_OF',
            'LC__CMDB__CATG__SMVIEW_CABLE__ETSI_BAYS_OF',
            'LC__CMDB__CATG__SMVIEW_CABLE__ETSI_AUTO_ASSIGN_STATIONS',
            'LC__CMDB__CATG__SMVIEW_CABLE__ETSI_AUTO_ASSIGN_OBJECTS',
            'LC__CMDB__CATG__SMVIEW_CABLE__ETSI_AUTO_ASSIGN_RACKS',
            'LC__CMDB__CATG__SMVIEW_CABLE__ETSI_DROP_ELEMENT_HERE',
            'LC__CMDB__CATG__SMVIEW_CABLE__ETSI_REMOVE_OBJECT',
        ];

        // translate
        $language = [];
        $lang = isys_application::instance()->container->get('language');
        foreach ($language_templates as $l) {
            $language[$l] = $lang->get($l);
        }

        return [
            'ajax_url'     => isys_helper_link::create_url([C__GET__AJAX => 1, C__GET__AJAX_CALL => 'cable_etsi']),
            'session'      => $_COOKIE['PHPSESSID'],
            'language'     => $language,
            'spinner_url'  => isys_helper_link::get_base() . '/images/ajax-loading.gif',
            'objects'      => $objects,
            'subway_lines' => $this->_get_subway_lines(),
        ];
    }

    /**
     * Save object.
     *
     * @param int $element_id
     * @param array $optionalParams
     * @param array|null $data
     */
    public function saveObject($element_id = null, $optionalParams = [], $data = null)
    {
        parent::saveObject($element_id, $optionalParams, $data);
        $instance_id = isys_application::instance()->template->get_template_vars("instance_id");

        // save rack items
        $dao = new isys_cmdb_dao_smview_cable_etsi_export_object(isys_application::instance()->database);
        $rack_objects = $dao->filter_objects(['export_id' => $element_id]);
        foreach ($rack_objects as $r) {
            $dao->delete_object($r['id']);
        }

        $newItems = $data['objects'];
        foreach ($newItems as $r) {
            $dao->create_object([
                'export_id'          => $instance_id,
                'bay_export_list_id' => $r['id'],
                'position'           => $r['position'],
            ]);
        }
    }
}
