<?php

namespace idoit\Module\Cable\DependentObjectManager;

use Exception;
use isys_application;
use isys_auth;
use isys_auth_auth;
use isys_cmdb_dao_smview_cable_wire_pin_status;
use isys_component_template_navbar;
use isys_exception_database;
use UnexpectedValueException;

/**
 * PIN Status Manager class.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class WirePinStatusManager extends AbstractManager
{

    /**
     * Property mapping.
     *
     * @var array
     */
    protected $_mapping = array(
        'const'                    => 'C__CATG__SMVIEW_CABLE_CONST',
        'description'              => 'C__CATG__SMVIEW_CABLE_DESCRIPTION',
        'obj_id'                   => 'C__CATG__FILE_OBJ_FILE',
        'color'                    => 'C__CATG__SMVIEW_CABLE_COLOR',
        'upper_transition_damping' => 'C__CATG__SMVIEW_CABLE_UPPER_TRANSITION_DAMPING',
        'wire_not_applied'         => 'C__CATG__SMVIEW_CABLE_WIRE_NOT_APPLIED',
    );

    /**
     * Constructor.
     */
    public function __construct()
    {
        $this->_dao = new isys_cmdb_dao_smview_cable_wire_pin_status(isys_application::instance()->database);
    }

    /**
     * Process PIN status listing.
     *
     * @throws Exception
     */
    public function processListing()
    {
        $lang = isys_application::instance()->container->get('language');
        $title = $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__WIRE_PIN_STATUS');
        $headers = [
            "id"                       => "ID",
            "const"                    => $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__CONST'),
            "description"              => $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__DESCRIPTION'),
            "filename_original"        => $lang->get('LC__CMDB__CATG__SMVIEW_PATCH__CONNECTION_TYPE_FILE'),
            "color"                    => $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__COLOR'),
            "upper_transition_damping" => $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__UPPER_TRANSITION_DAMPING'),
            "wire_not_applied"         => $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__WIRE_NOT_APPLIED'),
        ];
        $this->_processListing($title, $headers);
    }

    /**
     * Modify row method will be called for each row to alter its content.
     *
     * @param  array &$p_row
     * @throws isys_exception_database
     */
    public function modify_row(&$p_row)
    {
        $row = $this->_dao->get_object('id', $p_row['id']);
        $url = isys_cmdb_dao_smview_cable_wire_pin_status::get_file_url($row['obj_id']);

        if (! empty($url)) {
            $alt = $p_row['filename_original'];
            $title = $row['title'];
            $img = "<img src='$url' alt='$alt' title='$title' style='max-width: 128px; max-height: 32px;' />";
            $p_row['filename_original'] = $img;
        }
        if ($p_row['wire_not_applied'] != '') {
            $lang = isys_application::instance()->container->get('language');
            $p_row['wire_not_applied'] = $p_row['wire_not_applied'] == 1 ? $lang->get('LC__UNIVERSAL__YES') : $lang->get('LC__UNIVERSAL__NO');
        }
    }

    /**
     * Process page.
     *
     * @param  int $element_id
     * @param  array $additional_fields
     * @throws isys_exception_database
     */
    public function processPage($element_id, $additional_fields = [])
    {
        global $g_config;

        parent::processPage($element_id, $additional_fields);

        isys_component_template_navbar::getInstance()
            ->set_visible(false, defined_or_default('C__NAVBAR_BUTTON__DELETE'));

        $row = $this->_dao->get_object('id', $element_id);
        $url = isys_cmdb_dao_smview_cable_wire_pin_status::get_file_url($row['obj_id']);
        $l_rules['C__CATG__SMVIEW_CABLE_WIRE_NOT_APPLIED']['p_strSelectedID'] = $row['wire_not_applied'];
        $l_rules['C__CATG__SMVIEW_CABLE_WIRE_NOT_APPLIED']['p_arData'] = get_smarty_arr_YES_NO();
        isys_application::instance()->template
            ->assign("dir_tools", $g_config["www_dir"] . "src/tools/")
            ->assign("C__CATG__SMVIEW_CABLE_COLOR", $row['color'])
            ->assign('url', $url)
            ->smarty_tom_add_rules("tom.content.bottom.content", $l_rules);
    }

    /**
     * Save object.
     *
     * @param  int $instance_id
     * @param  array $optionalParams
     * @param array|null $data
     */
    public function saveObject($instance_id = null, $optionalParams = [], $data = null)
    {
        isys_component_template_navbar::getInstance()
            ->set_active(isys_auth_auth::instance()->is_allowed_to(isys_auth::EDIT, "MODULE/C__MODULE__SMVIEW_CABLE"),
                defined_or_default('C__NAVBAR_BUTTON__EDIT'));

        foreach ($this->_mapping as $db => $field) {
            $value = isset($_POST[$field]) ? $_POST[$field] : null;
            $params[$db] = $value;
        }
        if (! empty($params['upper_transition_damping'])) {
            $damping = $params['upper_transition_damping'];
            $damping = str_replace(',', '.', $damping);
            $params['upper_transition_damping'] = (float) $damping;
        }
        if (isset($params['wire_not_applied'])) {
            $params['wire_not_applied'] = $params['wire_not_applied'] == 1 ? 1 : 0;
        }
        $params['obj_id'] = ! empty($_POST['C__CATG__FILE_OBJ_FILE__HIDDEN']) ? $_POST['C__CATG__FILE_OBJ_FILE__HIDDEN'] : null;

        try {
            if (!isset($instance_id) || $instance_id == "") {
                $instance_id = $this->_dao->create_object($params);
            } else {
                $this->_dao->update_object($instance_id, $params);
            }
        } catch (UnexpectedValueException $e) {
            $_POST["navMode"] = defined_or_default('C__NAVMODE__EDIT');
        }

        $l_rules = [];
        foreach ($this->_mapping as $db => $field) {
            $value = $params[$db];
            $l_rules[$field]['p_strValue'] = $value;
        }

        isys_application::instance()->template->assign("instance_id", $instance_id)
            ->smarty_tom_add_rules("tom.content.bottom.content", $l_rules);
    }
}
