<?php

namespace idoit\Module\Cable\ExcelExporter;

use isys_component_database;
use PhpOffice\PhpSpreadsheet\Exception;
use PhpOffice\PhpSpreadsheet\Style\Alignment;
use PhpOffice\PhpSpreadsheet\Style\Border;
use PhpOffice\PhpSpreadsheet\Style\Fill;
use PhpOffice\PhpSpreadsheet\Worksheet\Worksheet;

/**
 * Damping Exporter class.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  2021 SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
abstract class AbstractExporter
{

    const DEFAULT_ALIGNMENT = 'center';
    const DEFAULT_FONT_SIZE = 9;

    /**
     * Variable which holds the database component.
     *
     * @var isys_component_database
     */
    protected $_database = null;

    /**
     * Current worksheet.
     *
     * @var Worksheet
     */
    protected $_sheet = null;

    /**
     * Constructor.
     *
     * @param isys_component_database $database
     */
    public function __construct($database)
    {
        $this->_database = $database;
    }

    /**
     * Add cell to sheet.
     *
     * @param  int $column
     * @param  int $row
     * @param  string $content
     * @param  int $width
     * @param  string $alignment
     * @param  boolean $isBold
     * @param  string $color
     * @param  string $bgColor
     * @param  int $fontSize
     * @throws Exception
     */
    protected function _add_cell($column, $row, $content, $width = 1, $withBorder = false, $alignment = self::DEFAULT_ALIGNMENT, $isBold = false, $color = '000000', $bgColor = 'FFFFFF', $fontSize = self::DEFAULT_FONT_SIZE)
    {
        $this->_sheet->setCellValueByColumnAndRow($column, $row, $content);
        $fill = [];
        if ($bgColor != 'FFFFFF') {
            $fill = [
                'fillType' => Fill::FILL_SOLID,
                'startColor' => [
                    'argb' => 'FF' . $bgColor,
                ],
            ];
        }
        switch ($alignment) {
            case 'left': $align = Alignment::HORIZONTAL_LEFT; break;
            case 'right': $align = Alignment::HORIZONTAL_RIGHT; break;
            default: $align = Alignment::HORIZONTAL_CENTER; break;
        }

        $color = ! empty($color) ? $color : '000000';
        $styleArray = [
            'font' => [
                'size'  => $fontSize,
                'color' => [
                    'argb' => 'FF' . $color,
                ],
                'bold'  => $isBold,
            ],
            'alignment' => [
                'horizontal' => $align,
                'vertical'   => Alignment::VERTICAL_CENTER,
            ],
            'fill' => $fill,
        ];
        $this->_sheet->getStyleByColumnAndRow($column, $row)->applyFromArray($styleArray);

        $columnX = $column + $width - 1;
        $rowY = $row;
        $this->_sheet->mergeCellsByColumnAndRow($column, $row, $columnX, $rowY);

        if ($withBorder) {
            $this->_add_border($column, $row, $columnX, $rowY);
        }
    }

    /**
     * Add border around coordinates.
     *
     * @param int $x1
     * @param int $y1
     * @param int $x2
     * @param int $y2
     * @param string $thickness
     */
    protected function _add_border($x1, $y1, $x2, $y2, $thickness = Border::BORDER_THIN)
    {
        $styleArray = [
            'borders' => [
                'top' => [
                    'borderStyle' => $thickness,
                ],
                'right' => [
                    'borderStyle' => $thickness,
                ],
                'bottom' => [
                    'borderStyle' => $thickness,
                ],
                'left' => [
                    'borderStyle' => $thickness,
                ],
            ],
        ];
        $this->_sheet->getStyleByColumnAndRow($x1, $y1, $x2, $y2)->applyFromArray($styleArray);
    }
}

