<?php

namespace idoit\Module\Cable\ExcelExporter;

use isys_application;
use isys_cmdb_dao_category_g_smview_cable_route;
use isys_cmdb_dao_category_g_smview_cable_route_element;
use isys_cmdb_dao_category_g_smview_cable_section;
use isys_cmdb_dao_category_g_smview_cable_section_element;
use isys_exception_database;
use isys_module_smview_cable;
use isys_settings;
use PhpOffice\PhpSpreadsheet\Exception;
use PhpOffice\PhpSpreadsheet\Spreadsheet;
use PhpOffice\PhpSpreadsheet\Worksheet\Worksheet;
use PhpOffice\PhpSpreadsheet\Writer\Xlsx;

/**
 * Damping Exporter class.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  2021 SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class DampingExporter extends AbstractExporter
{

    const DEFAULT_ALIGNMENT = 'left';

    /**
     * Get elements from excel file.
     *
     * @param  array $obj
     * @return string
     * @throws Exception
     * @throws \PhpOffice\PhpSpreadsheet\Writer\Exception
     * @throws isys_exception_database
     */
    public function export_to_excel($obj)
    {
        $spreadsheet = new Spreadsheet();
        $this->_sheet = $spreadsheet->getActiveSheet();
        $title = $obj['isys_obj__title'];
        $title = str_replace('\\', ' ', $title); // remove back slash
        $title = substr($title, 0, 31); // reduce length to 31 characters
        $this->_sheet->setTitle($title);

        // export damping data
        $this->export_to_sheet($this->_sheet, $obj, 1, 1, $title);

        // create file
        $path = "/var/www/html/i-doit/temp/$title.xlsx";
        $writer = new Xlsx($spreadsheet);
        $writer->save($path);
        return $path;
    }

    /**
     * Export to existing sheet.
     *
     * @param Worksheet $sheet
     * @param array $obj
     * @param int $column
     * @param int $row
     * @param string $title
     * @return int
     * @throws Exception
     * @throws isys_exception_database
     */
    public function export_to_sheet($sheet, $obj, $column = 1, $row = 1, $title = null)
    {
        $lang = isys_application::instance()->container->get('language');
        $this->_sheet = $sheet;

        $this->_sheet->getColumnDimensionByColumn($column)->setWidth(16);
        $this->_sheet->getColumnDimensionByColumn($column + 1)->setWidth(8);
        $this->_sheet->getColumnDimensionByColumn($column + 2)->setWidth(10);
        $this->_sheet->getColumnDimensionByColumn($column + 3)->setWidth(12);

        if ($title) {
            // add report title
            $this->_add_cell($column, $row, $title, 4, false, 'center', true, '000000', 'FFFFFF', 12);
            $row++;
        }

        $elements = [];
        $section_dao = new isys_cmdb_dao_category_g_smview_cable_section(isys_application::instance()->database);
        $damping_1310_factor = isys_settings::get(isys_module_smview_cable::DAMPING_1310, 0);
        $damping_1550_factor = isys_settings::get(isys_module_smview_cable::DAMPING_1550, 0);
        if ($obj['isys_obj_type__const'] == 'C__SHD__ROUTE') {
            $route_dao = new isys_cmdb_dao_category_g_smview_cable_route(isys_application::instance()->database);
            $route = $route_dao->get_data(null, $obj['isys_obj__id'])->get_row();
            $dao = new isys_cmdb_dao_category_g_smview_cable_route_element(isys_application::instance()->database);
            $elements = $dao->get_elements_by_route_id($route['isys_catg_smview_cable_route_list__id'], false, false);
        }

        if ($obj['isys_obj_type__const'] == 'C__SHD__SECTION') {
            $section = $section_dao->get_data(null, $obj['isys_obj__id'])->get_row();
            $dao = new isys_cmdb_dao_category_g_smview_cable_section_element(isys_application::instance()->database);
            $elements = $dao->get_elements_by_section_id($section['isys_catg_smview_cable_section_list__id'], false, false, false);
        }
        $damping = $section_dao->calculate_damping($elements, $damping_1310_factor, $damping_1550_factor);

        // add header
        $this->_add_cell($column, $row, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__DESCRIPTION'), 1, true, self::DEFAULT_ALIGNMENT, true, '000000', 'E4E4E4');
        $this->_add_cell($column + 1, $row, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__COUNT'), 1, true, self::DEFAULT_ALIGNMENT, true, '000000', 'E4E4E4');
        $this->_add_cell($column + 2, $row, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__TYPICAL_DAMPING'), 1, true, self::DEFAULT_ALIGNMENT, true, '000000', 'E4E4E4');
        $this->_add_cell($column + 3, $row, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__DAMPING_DB'), 1, true, self::DEFAULT_ALIGNMENT, true, '000000', 'E4E4E4');

        // add connectors
        $row++;
        foreach ($damping['connectors'] as $c) {
            $this->_add_cell($column, $row, $c['description'], 1, true, self::DEFAULT_ALIGNMENT, true, '000000', 'E4E4E4');
            $this->_add_cell($column + 1, $row, $c['count'], 1, true, self::DEFAULT_ALIGNMENT);
            $this->_add_cell($column + 2, $row, $c['factor'], 1, true, self::DEFAULT_ALIGNMENT);
            $this->_add_cell($column + 3, $row, ($c['count'] * $c['factor']), 1, true, self::DEFAULT_ALIGNMENT);
            $row++;
        }

        // length
        $this->_add_cell($column, $row, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__LENGTH_M'), 1, true, self::DEFAULT_ALIGNMENT, true, '000000', 'E4E4E4');
        $this->_add_cell($column + 1, $row, $damping['length'], 3, true, self::DEFAULT_ALIGNMENT);
        $row++;

        // wave length 1310
        $this->_add_cell($column, $row, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__WAVE_LENGTH_1310'), 1, true, self::DEFAULT_ALIGNMENT, true, '000000', 'E4E4E4');
        $this->_add_cell($column + 1, $row, '', 1, true, self::DEFAULT_ALIGNMENT);
        $this->_add_cell($column + 2, $row, $damping_1310_factor, 1, true, self::DEFAULT_ALIGNMENT);
        $this->_add_cell($column + 3, $row, $damping['length_1310'], 1, true, self::DEFAULT_ALIGNMENT);
        $row++;

        // wave length 1550
        $this->_add_cell($column, $row, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__WAVE_LENGTH_1550'), 1, true, self::DEFAULT_ALIGNMENT, true, '000000', 'E4E4E4');
        $this->_add_cell($column + 1, $row, '', 1, true, self::DEFAULT_ALIGNMENT);
        $this->_add_cell($column + 2, $row, $damping_1550_factor, 1, true, self::DEFAULT_ALIGNMENT);
        $this->_add_cell($column + 3, $row, $damping['length_1550'], 1, true, self::DEFAULT_ALIGNMENT);
        $row++;

        // sum 1310
        $this->_add_cell($column, $row, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__DAMPING_SUM_1310'), 3, true, self::DEFAULT_ALIGNMENT, true, '000000', 'E4E4E4');
        $this->_add_cell($column + 3, $row, $damping['sum_1310'], 1, true, self::DEFAULT_ALIGNMENT);
        $row++;

        // sum 1550
        $this->_add_cell($column, $row, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__DAMPING_SUM_1550'), 3, true, self::DEFAULT_ALIGNMENT, true, '000000', 'E4E4E4');
        $this->_add_cell($column + 3, $row, $damping['sum_1550'], 1, true, self::DEFAULT_ALIGNMENT);
        $row++;

        if (! empty($damping['empty_wire_damping'])) {
            // hint on empty wire damping
            $this->_add_cell($column, $row, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__DAMPING_EMPTY_WIRES'), 4, true, self::DEFAULT_ALIGNMENT, true);
            $row++;
        } elseif (! empty($damping['not_applied_wires'])) {
            // hint on not applied wires
            $this->_add_cell($column, $row, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__DAMPING_NOT_APPLIED_WIRES'), 4, true, self::DEFAULT_ALIGNMENT, true);
            $row++;
        }
        return $row;
    }
}

