<?php

namespace idoit\Module\Cable\ExcelExporter;

use isys_cmdb_dao_category_g_global;
use isys_cmdb_dao_category_g_location;
use isys_cmdb_dao_category_g_smview_cable_etsi_bay;
use isys_cmdb_dao_category_g_smview_cable_etsi_bay_export;
use isys_cmdb_dao_smview_cable_etsi_bay_item_export;
use isys_cmdb_dao_smview_cable_etsi_device_color;
use isys_cmdb_dao_smview_cable_etsi_export;
use isys_cmdb_dao_smview_cable_etsi_export_object;
use isys_component_database;
use isys_exception_dao;
use isys_exception_database;
use PhpOffice\PhpSpreadsheet\Exception;
use PhpOffice\PhpSpreadsheet\Spreadsheet;
use PhpOffice\PhpSpreadsheet\Style\Alignment;
use PhpOffice\PhpSpreadsheet\Style\Border;
use PhpOffice\PhpSpreadsheet\Style\Fill;
use PhpOffice\PhpSpreadsheet\Worksheet\PageSetup;
use PhpOffice\PhpSpreadsheet\Worksheet\Worksheet;
use PhpOffice\PhpSpreadsheet\Writer\Xlsx;

/**
 * ETSI Bay Exporter class.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class EtsiBayExporter
{
    /*
     * Default settings.
     */
    const WIDTH_HE = 2;
    const WIDTH_CABLE_NAME = 3;
    const WIDTH_OBJECT = 3;
    const WIDTH_SHORT_NAME = 3;
    const WIDTH_WIRE = 5;
    const HEIGHT_ITEM = 2;
    const DEFAULT_COLUMN_WIDTH = 2.15;
    const DEFAULT_ROW_HEIGHT = 8.25;
    const DEFAULT_FONT_SIZE = 8.5;
    const DEFAULT_PAGE_MARGIN = 0.4;

    /*
     * Column types.
     */
    const COLUMN_OBJECT = 'object';

    /**
     * Variable which holds the database component.
     *
     * @var isys_component_database
     */
    protected $_database = null;

    /**
     * Current worksheet.
     *
     * @var Worksheet
     */
    private $_sheet = null;

    /**
     * Constructor.
     *
     * @param isys_component_database $database
     */
    public function __construct($database)
    {
        $this->_database = $database;
    }

    /**
     * Export a complete export configuration.
     *
     * @param  int $export_id
     * @return string
     * @throws Exception
     * @throws \PhpOffice\PhpSpreadsheet\Writer\Exception
     * @throws isys_exception_dao
     * @throws isys_exception_database
     */
    public function export_config($export_id)
    {
        $export_dao = new isys_cmdb_dao_smview_cable_etsi_export($this->_database);
        $export = $export_dao->get_object('id', $export_id);

        $rack_dao = new isys_cmdb_dao_smview_cable_etsi_export_object($this->_database);
        $export_objects = $rack_dao->filter_objects(['export_id' => $export_id]);

        $colorLayouts = $this->_get_color_layouts();
        $spreadsheet = $this->_get_new_spreadsheet();
        $this->_sheet = $spreadsheet->getActiveSheet();
        $this->_setup_sheet($this->_sheet, $export['name']);

        // enable this option to fit complete height
//        $this->_sheet->getPageSetup()->setFitToHeight(1);

        $export_ids = [];
        foreach ($export_objects as $o) {
            $export_ids[] = $o['bay_export_list_id'];
        }
        $this->_export_bays_of_object($export_ids, $colorLayouts);

        $path = '/var/www/html/i-doit/temp/etsi-config-export.xlsx';
        $writer = new Xlsx($spreadsheet);
        $writer->save($path);
        return $path;
    }

    /**
     * Get elements from excel file.
     *
     * @param  int $export_id
     * @param  string $title
     * @return string
     * @throws Exception
     * @throws \PhpOffice\PhpSpreadsheet\Writer\Exception
     * @throws isys_exception_dao
     * @throws isys_exception_database
     */
    public function export_bays_of_object($export_id, $title)
    {
        $colorLayouts = $this->_get_color_layouts();
        $spreadsheet = $this->_get_new_spreadsheet();
        $this->_sheet = $spreadsheet->getActiveSheet();
        $this->_setup_sheet($this->_sheet, $title);

        $this->_export_bays_of_object([$export_id], $colorLayouts);

        $path = '/var/www/html/i-doit/temp/etsi-bay-export.xlsx';
        $writer = new Xlsx($spreadsheet);
        $writer->save($path);
        return $path;
    }

    /**
     * Get a new spreadsheet.
     *
     * @return Spreadsheet
     */
    private function _get_new_spreadsheet()
    {
        $spreadsheet = new Spreadsheet();
        $spreadsheet->getProperties()->setCreator('SM-VIEW Cable');
        $spreadsheet->getProperties()->setCompany('SHD System-Haus-Dresden GmbH');
        $spreadsheet->getProperties()->setCreated(time());
        $spreadsheet->getProperties()->setSubject('ETSI Export');
        return $spreadsheet;
    }

    /**
     * Setup sheet with default settings.
     *
     * @param Worksheet $sheet
     * @param string $title
     */
    private function _setup_sheet($sheet, $title)
    {
        $sheet->setTitle($title);
        $sheet->getDefaultColumnDimension()->setWidth(self::DEFAULT_COLUMN_WIDTH);
        $sheet->getDefaultRowDimension()->setRowHeight(self::DEFAULT_ROW_HEIGHT);
        $sheet->getPageMargins()->setLeft(self::DEFAULT_PAGE_MARGIN);
        $sheet->getPageMargins()->setRight(self::DEFAULT_PAGE_MARGIN);
        $sheet->getPageMargins()->setTop(self::DEFAULT_PAGE_MARGIN);
        $sheet->getPageMargins()->setBottom(self::DEFAULT_PAGE_MARGIN);
        $sheet->getPageSetup()->setPaperSize(PageSetup::PAPERSIZE_A4);
        $sheet->getPageSetup()->setOrientation(PageSetup::ORIENTATION_LANDSCAPE);
    }

    /**
     * Export all configured ETSI bays of an object.
     *
     * @param  array $export_bay_ids
     * @param  array $colorLayouts
     * @throws Exception
     * @throws isys_exception_dao
     * @throws isys_exception_database
     */
    private function _export_bays_of_object($export_bay_ids, $colorLayouts)
    {
        $bay_export_dao = new isys_cmdb_dao_category_g_smview_cable_etsi_bay_export($this->_database);
        $export_item_dao = new isys_cmdb_dao_smview_cable_etsi_bay_item_export($this->_database);

        $bayCount = 1;
        $wholeWith = self::WIDTH_HE + self::WIDTH_CABLE_NAME + self::WIDTH_OBJECT + self::WIDTH_SHORT_NAME + self::WIDTH_WIRE;
        foreach ($export_bay_ids as $id) {
            $bay_export = $bay_export_dao->get_data($id)->get_row();
            $items = $export_item_dao->filter_objects(['list_id' => $bay_export['isys_catg_smview_cable_etsi_bay_export_list__id']]);
            foreach ($items as $r) {
                $startColumn = ($wholeWith + 1) * ($bayCount - 1);
                $startColumn++;

                $this->_export_bay($r['bay_item_id'], $colorLayouts, $startColumn);
                $bayCount++;

                if ($bayCount > 1) {
                    $c = ($wholeWith + 1) * ($bayCount - 1);
                    $w = $bayCount % 4 == 1 ? 2 : 1;
                    $this->_sheet->getColumnDimensionByColumn($c)->setWidth($w);
                }
            }
        }
    }

    /**
     * Export a single bay.
     *
     * @param  int $bay_item_id
     * @param  array $colorLayouts
     * @param  int $startColumn
     * @throws Exception
     * @throws isys_exception_dao
     * @throws isys_exception_database
     */
    private function _export_bay($bay_item_id, $colorLayouts, $startColumn = 1)
    {
        $wholeWith = self::WIDTH_HE + self::WIDTH_CABLE_NAME + self::WIDTH_OBJECT + self::WIDTH_SHORT_NAME + self::WIDTH_WIRE;

        $bay_dao = new isys_cmdb_dao_category_g_smview_cable_etsi_bay($this->_database);
        $bay = $bay_dao->get_bay_by_id($bay_item_id);

        $row = $this->_add_header_information($bay_item_id, $startColumn, $wholeWith, $bay['title']);
        for ($i = 1; $i <= $bay['he']; $i++) {
            $column = $startColumn;
            $heightItem = self::HEIGHT_ITEM;
            $row2 = ($row + self::HEIGHT_ITEM - 1);
            $assignedBayItem = null;
            foreach ($bay['items'] as $item) {
                if ($item['position'] == $i) {
                    $assignedBayItem = $item;
                    $heightItem *= $assignedBayItem['he'];
                    $row2 = ($row + $heightItem - 1);
                    break;
                }
            }

            // he
            $this->_sheet->getColumnDimensionByColumn($column)->setWidth(1);
            $heColor = ! empty($assignedBayItem['project_identifier']) ? '00FFFF' : 'CCCCCC';
            if (! empty($assignedBayItem['project_identifier'])) {
                // use whole item height
                $this->_add_cell($column, $row, $assignedBayItem['project_identifier'], $heColor, self::WIDTH_HE, $heightItem);
            } else {
                $hes = ! empty($assignedBayItem['he']) ? $assignedBayItem['he'] : 1;
                $heRow = $row;
                $heIndex = $i;
                for ($j = 1; $j <= $hes; $j++) {
                    // separate HE counter on row-span elements
                    $this->_add_cell($column, $heRow, $heIndex, $heColor, self::WIDTH_HE, self::HEIGHT_ITEM);
                    if ($hes > 1) {
                        $heRow += self::HEIGHT_ITEM;
                        $heIndex++;
                    }
                }
            }

            if (! $assignedBayItem) {
                // add empty cell block
                $column += self::WIDTH_HE;
                $this->_add_cell($column, $row, '', 'F8F8F8', $wholeWith - self::WIDTH_HE, $heightItem);
                $this->_add_border($startColumn, $row, $startColumn + $wholeWith - 1, $row + ($heightItem - 1));

                $row += $heightItem;
                continue;
            }

            if ($assignedBayItem['status'] == 'planning') {
                $column += self::WIDTH_HE;
                $title = $assignedBayItem['title'];
                $title .= $assignedBayItem['comment'] ? ' (' . $assignedBayItem['comment'] . ')' : '';
                $this->_add_cell($column, $row, $title, 'AAAAAA', $wholeWith - self::WIDTH_HE, $heightItem);
                $this->_add_border($startColumn, $row, $startColumn + $wholeWith - 1, $row + ($heightItem - 1));

                $row += $heightItem;
                continue;
            }

            $color = $colorLayouts[$assignedBayItem['color']];
            $colorLeftTop = $color['color_left_top'];
            $colorLeftBottom = $color['color_left_bottom'];
            $colorRightTop = $color['color_left_top'];
            $colorRightBottom = $color['color_left_top'];
            $colorLayout = (int) $color['layout'];
            if ($colorLayout >= 2) {
                $colorRightTop = $color['color_right_top'];
                $colorRightBottom = $color['color_right_top'];
            }
            if ($colorLayout >= 3) {
                $colorRightBottom = $color['color_right_bottom'];
            }
            $colorLeftTop = substr($colorLeftTop, 1, strlen($colorLeftTop));
            $colorLeftBottom = substr($colorLeftBottom, 1, strlen($colorLeftBottom));
            $colorRightTop = substr($colorRightTop, 1, strlen($colorRightTop));
            $colorRightBottom = substr($colorRightBottom, 1, strlen($colorRightBottom));

            // handle wire bundles
            $prevCableName = null;
            $cableNames = [];
            $prevObjectName = null;
            $objectNames = [];
            $wire = null;
            $prevWire = null;
            $wires = [];
            $wireBundles = $assignedBayItem['wire_bundles'];
            foreach ($wireBundles as $bundle) {
                if (! $wire) {
                    // setup first bundle
                    $cableNames[] = $bundle['cable_name'];
                    $prevCableName = $bundle['cable_name'];
                    $objectNames[] = $bundle['sibling_object_title'];
                    $prevObjectName = $bundle['sibling_object_title'];
                    $wire = 'Faser ' . $bundle['wire_start'];
                }

                if ($prevCableName != $bundle['cable_name'] || $prevObjectName != $bundle['sibling_object_title']) {
                    // save bundle
                    $cableNames[] = $bundle['cable_name'];
                    $prevCableName = $bundle['cable_name'];

                    $objectNames[] = $bundle['sibling_object_title'];
                    $prevObjectName = $bundle['sibling_object_title'];

                    $wire .= '-' . $prevWire;
                    $wires[] = $wire;
                    $wire = 'Faser ' . $bundle['wire_start'];
                }
                $prevWire = $bundle['wire_end'];
            }
            if ($wire) {
                // save last bundle
                $wire .= '-' . $prevWire;
                $wires[] = $wire;
            }

            // cable name
            $column += self::WIDTH_HE;
            if ($colorLayout == 4) {
                $this->_add_cell($column, $row, implode("\n", $cableNames), $colorLeftTop, self::WIDTH_CABLE_NAME, $heightItem / 2);
                $this->_add_cell($column, $row + 1, '', $colorLeftBottom, self::WIDTH_CABLE_NAME, $heightItem / 2);
            } else {
                $this->_add_cell($column, $row, implode("\n", $cableNames), $colorLeftTop, self::WIDTH_CABLE_NAME, $heightItem);
            }

            // object
            $column += self::WIDTH_CABLE_NAME;
            if ($colorLayout == 4) {
                $this->_add_cell($column, $row, implode("\n", $objectNames), $colorLeftTop, self::WIDTH_OBJECT, $heightItem / 2);
                $this->_add_cell($column, $row + 1, '', $colorLeftBottom, self::WIDTH_OBJECT, $heightItem / 2);
            } else {
                $this->_add_cell($column, $row, implode("\n", $objectNames), $colorLeftTop, self::WIDTH_OBJECT, $heightItem, false, self::COLUMN_OBJECT);
            }

            // short name
            $column += self::WIDTH_OBJECT;
            if ($colorLayout == 4) {
                $this->_add_cell($column, $row, $assignedBayItem['short_name'], $colorLeftTop, self::WIDTH_SHORT_NAME, $heightItem / 2);
                $this->_add_cell($column, $row + 1, '', $colorLeftBottom, self::WIDTH_SHORT_NAME, $heightItem / 2);
            } else {
                $this->_add_cell($column, $row, $assignedBayItem['short_name'], $colorLeftTop, self::WIDTH_SHORT_NAME, $heightItem);
            }

            // wire
            $column += self::WIDTH_SHORT_NAME;
            $wire = count($wires) > 0 ? $wires[0] : '';
            if ($colorLayout >= 3) {
                $this->_add_cell($column, $row, $wire, $colorRightTop, self::WIDTH_WIRE, 1);
                $wire = count($wires) > 1 ? implode("\n", array_slice($wires, 1)) : '';
                $this->_add_cell($column, $row2, $wire, $colorRightBottom, self::WIDTH_WIRE, 1);
            } else {
                if (count($wires) <= 1) {
                    $this->_add_cell($column, $row, $wire, $colorRightTop, self::WIDTH_WIRE, $heightItem);
                } else {
                    // slit into two cells on more than one wire
                    $this->_add_cell($column, $row, $wire, $colorRightTop, self::WIDTH_WIRE, $heightItem / 2);
                    $wire = count($wires) > 1 ? implode("\n", array_slice($wires, 1)) : '';
                    $this->_add_cell($column, $row + 1, $wire, $colorRightTop, self::WIDTH_WIRE, $heightItem / 2);
                }
            }
            $this->_add_border($startColumn, $row, $startColumn + $wholeWith - 1, $row + ($heightItem - 1));

            $i += ($heightItem / self::HEIGHT_ITEM) - 1;
            $row += $heightItem;
        }

        // add a thick border around whole rack
        $this->_add_border($startColumn, 1, $startColumn + $wholeWith - 1, $row - 1, Border::BORDER_THICK);
    }

    /**
     * Add common header information of a bay.
     *
     * @param  int $bay_item_id
     * @param  int $startColumn
     * @param  int $wholeWith
     * @param  string $title
     * @return int
     * @throws Exception
     * @throws isys_exception_database
     */
    private function _add_header_information($bay_item_id, $startColumn, $wholeWith, $title)
    {
        $row = 1;
        $bay_dao = new isys_cmdb_dao_category_g_smview_cable_etsi_bay($this->_database);
        $bay_data = $bay_dao->get_data($bay_item_id)->get_row();
        $obj_id = $bay_data['isys_catg_smview_cable_etsi_bay_list__isys_obj__id'];

        // add location
        $location_dao = new isys_cmdb_dao_category_g_location($this->_database);
        $global_dao = new isys_cmdb_dao_category_g_global($this->_database);
        $path = $location_dao->get_location_path($obj_id);
        if (count($path) > 2) {
            // remove first 2 location items
            $path = array_slice($path, 0, count($path) - 2);
        }
        $path = array_reverse($path);
        $location = [];
        foreach ($path as $p) {
            $obj = $global_dao->get_object($p)->get_row();
            $location[] = $obj['isys_obj__title'];
        }
        $location = implode(' > ', $location);

        $this->_add_cell($startColumn, $row, $location, 'FFFFFF', $wholeWith, 1, true);
        $this->_add_border($startColumn, $row, $startColumn + ($wholeWith - 1), $row);
        $row++;

        // add title
        $this->_add_cell($startColumn, $row, $title, 'FFFFFF', $wholeWith, 1, true);
        $this->_add_border($startColumn, $row, $startColumn + ($wholeWith - 1), $row);
        $row++;

        // add header information
        $description = $bay_data['isys_catg_smview_cable_etsi_bay_list__description'] ?? '';
        $this->_add_cell($startColumn, $row, $description, 'FFFFFF', $wholeWith, 1);
        $this->_add_border($startColumn, $row, $startColumn + ($wholeWith - 1), $row);
        $row++;
        return $row;
    }

    /**
     * Get all color layouts.
     *
     * @return array
     * @throws isys_exception_database
     */
    private function _get_color_layouts()
    {
        $dao = new isys_cmdb_dao_smview_cable_etsi_device_color($this->_database);
        $colors = $dao->get_data();
        $colorLayouts = [];
        foreach ($colors as $c) {
            $colorLayouts[$c['id']] = $c;
        }
        return $colorLayouts;
    }

    /**
     * Add cell to sheet.
     *
     * @param  int $column
     * @param  int $row
     * @param  string $content
     * @param  string $color
     * @param  int $width
     * @param  int $heightItem
     * @param  boolean $isBold
     * @param  string $columnType
     * @throws Exception
     */
    private function _add_cell($column, $row, $content, $color, $width, $heightItem, $isBold = false, $columnType = null)
    {
        $content = trim($content);
        $color = ! empty($color) ? $color : 'FFFFFF';
        $this->_sheet->setCellValueByColumnAndRow($column, $row, $content);

        $styleArray = [
            'font' => [
                'size' => self::DEFAULT_FONT_SIZE,
                'bold' => $isBold,
            ],
            'alignment' => [
                'horizontal' => Alignment::HORIZONTAL_CENTER,
                'vertical'   => Alignment::VERTICAL_CENTER,
            ],
            'fill' => [
                'fillType' => Fill::FILL_SOLID,
                'startColor' => [
                    'argb' => 'FF' . $color,
                ],
            ],
        ];
        $this->_sheet->getStyleByColumnAndRow($column, $row)->applyFromArray($styleArray);

        // todo: add a dedicated handling of column types (cable, object, wires, ...)
        if (strlen($content) > 9 && $width <= 3) {
            // handle large content on small cells
            $separator = strpos($content, ' ') !== false ? ' ' : null;
            $separator = strpos($content, "\n") !== false ? "\n" : $separator;
            $style = $this->_sheet->getStyleByColumnAndRow($column, $row);
            if ($separator) {
                // content contains spaces or line breaks
                if ($heightItem == 2) {
                    // split cell into two rows
                    $parts = explode($separator, $content);
                    $cnt = 0;
                    $firstPart = '';
                    $lastPart = '';
                    foreach ($parts as $index => $p) {
                         $cnt += strlen($p);
                         if ($cnt > 9) {
                             $index = $index == 0 ? 1 : $index;
                             $firstPart = implode($separator, array_slice($parts, 0, $index));
                             $lastPart = implode($separator, array_slice($parts, $index, count($parts)));
                             break;
                         }
                        $cnt += strlen($separator); // add length of separator
                    }

                    // overwrite current cell content
                    $this->_sheet->setCellValueByColumnAndRow($column, $row, $firstPart);

                    if ($columnType == self::COLUMN_OBJECT) {
                        $this->_sheet->getStyleByColumnAndRow($column, $row)->getAlignment()->setHorizontal(Alignment::HORIZONTAL_LEFT);
                    }

                    // add second cell
                    $this->_add_cell($column, $row + 1, $lastPart, $color, $width, 1, $isBold);
                    $heightItem = 1;
                } else {
                    // wrap text if it contains at least one space
                    $style->getAlignment()->setWrapText(true);
                    if (strlen($content) > 14) {
                        $style->getFont()->setSize((self::DEFAULT_FONT_SIZE - 1));
                    }
                }
            } else {
                // shrink large content
                $style->getAlignment()->setShrinkToFit(true);
            }
        }

        $columnX = $column + $width - 1;
        $rowY = $row + $heightItem - 1;
        $this->_sheet->mergeCellsByColumnAndRow($column, $row, $columnX, $rowY);
    }

    /**
     * Add border around coordinates.
     *
     * @param int $x1
     * @param int $y1
     * @param int $x2
     * @param int $y2
     * @param string $border
     */
    private function _add_border($x1, $y1, $x2, $y2, $border = Border::BORDER_THIN)
    {
        $styleArray = [
            'borders' => [
                'top' => [
                    'borderStyle' => $border,
                ],
                'right' => [
                    'borderStyle' => $border,
                ],
                'bottom' => [
                    'borderStyle' => $border,
                ],
                'left' => [
                    'borderStyle' => $border,
                ],
            ],
        ];
        $this->_sheet->getStyleByColumnAndRow($x1, $y1, $x2, $y2)->applyFromArray($styleArray);
    }
}
