<?php

namespace idoit\Module\Cable\ExcelExporter;

use DateTime;
use isys_application;
use isys_cmdb_dao_category_g_global;
use isys_cmdb_dao_category_g_smview_cable_route;
use isys_cmdb_dao_category_g_smview_cable_route_element;
use isys_cmdb_dao_category_g_smview_cable_section;
use isys_cmdb_dao_category_g_smview_cable_section_element;
use isys_cmdb_dao_smview_cable_route;
use isys_cmdb_dao_smview_cable_section;
use isys_cmdb_dao_smview_cable_wire_pin;
use isys_cmdb_dao_smview_cable_wire_pin_status;
use isys_exception_database;
use PhpOffice\PhpSpreadsheet\Cell;
use PhpOffice\PhpSpreadsheet\Exception;
use PhpOffice\PhpSpreadsheet\Spreadsheet;
use PhpOffice\PhpSpreadsheet\Style\Border;
use PhpOffice\PhpSpreadsheet\Worksheet\PageSetup;
use PhpOffice\PhpSpreadsheet\Worksheet\Worksheet;
use PhpOffice\PhpSpreadsheet\Writer\Xlsx;

/**
 * Route Exporter class.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class RouteExporter extends AbstractExporter
{

    /**
     * Default settings.
     */
    const DEFAULT_COLUMN_WIDTH = 6;
    const DEFAULT_PAGE_MARGIN = 0.4;

    /**
     * List of used wire states.
     *
     * @var array
     */
    private $_usedWireStates = [];

    /**
     * Get elements from excel file.
     *
     * @param  array $routes
     * @param  string $title
     * @param  boolean $mergeRoutes
     * @param  boolean $addDamping
     * @param  boolean $addRedundantRoutes
     * @return string
     * @throws Exception
     * @throws \PhpOffice\PhpSpreadsheet\Writer\Exception
     * @throws isys_exception_database
     */
    public function export_to_excel($routes, $title, $mergeRoutes = false, $addDamping = false, $addRedundantRoutes = false)
    {
        $spreadsheet = $this->_get_new_spreadsheet();
        $this->_sheet = $spreadsheet->getActiveSheet();
        $this->_setup_sheet($this->_sheet, $title);

        if ($addRedundantRoutes) {
            $dao = new isys_cmdb_dao_category_g_smview_cable_route($this->_database);
            $updatedRoutes = [];
            foreach ($routes as $routeId) {
                $updatedRoutes[] = $routeId;
                $res = $dao->get_assigned_redundant_routes($routeId);
                while ($row = $res->get_row()) {
                    $updatedRoutes[] = $row['isys_obj__id'];
                }
            }
            $routes = $updatedRoutes;
        }

        $startRow = 2;
        $maxWidth = 0;
        $allRouteIds = [];
        if ($mergeRoutes) {
            $similarRouteBundles = $this->_get_similar_routes($routes);
            foreach ($similarRouteBundles as $routeBundle) {
                // save all routes with start height
                foreach (array_keys($routeBundle) as $routeId) {
                    $allRouteIds[$routeId] = $startRow;
                }

                $dimensions = $this->_export_routes($routeBundle, $startRow);
                $width = $dimensions[0] + 1;
                $startRow = $dimensions[1] + 7;
                $maxWidth = $width > $maxWidth ? $width : $maxWidth;
            }
        } else {
            foreach ($routes as $route) {
                $similarRouteBundles = $this->_get_similar_routes([$route]);
                $simRoutes = array_shift($similarRouteBundles);
                // save all routes with start height
                foreach (array_keys($simRoutes) as $routeId) {
                    $allRouteIds[$routeId] = $startRow;
                }

                $dimensions = $this->_export_routes($simRoutes, $startRow);
                $width = $dimensions[0] + 1;
                $startRow = $dimensions[1] + 7;
                $maxWidth = $width > $maxWidth ? $width : $maxWidth;
            }
        }

        // add report title
        $this->_add_cell(1, 1, $title, $maxWidth, false, 'center', true, '000000', 'FFFFFF', 14);
        $this->_sheet->getColumnDimensionByColumn(1)->setWidth(12);

        // add damping if needed
        if ($addDamping) {
            $exporter = new DampingExporter($this->_database);
            $dao = new isys_cmdb_dao_category_g_global(isys_application::instance()->database);
            $lastStartRow = null;
            foreach ($allRouteIds as $routeId => $sRow) {
                $obj = $dao->get_data(null, $routeId)->get_row();
                $title = null;
                if (count($routes) > 1) {
                    // add title on one than one routes
                    $title = $obj['isys_obj__title'];
                    $title = str_replace('\\', ' ', $title); // remove back slash
                    $title = substr($title, 0, 31); // reduce length to 31 characters
                }
                $column = $maxWidth + 3;
                if ($lastStartRow && $lastStartRow == $sRow) {
                    $column += 5;
                }
                $row = $title ? $sRow : $sRow + 1;
                $exporter->export_to_sheet($this->_sheet, $obj, $column, $row, $title);
                $lastStartRow = $sRow;
            }
        }

        // add legend
        $this->_add_legend($startRow + 1);

        $path = "/var/www/html/i-doit/temp/$title.xlsx";
        $writer = new Xlsx($spreadsheet);
        $writer->save($path);
        return $path;
    }

    /**
     * Get a new spreadsheet.
     *
     * @return Spreadsheet
     */
    private function _get_new_spreadsheet()
    {
        $spreadsheet = new Spreadsheet();
        $spreadsheet->getProperties()->setCreator('SM-VIEW Cable');
        $spreadsheet->getProperties()->setCompany('SHD System-Haus-Dresden GmbH');
        $spreadsheet->getProperties()->setCreated(time());
        return $spreadsheet;
    }

    /**
     * Setup sheet with default settings.
     *
     * @param Worksheet $sheet
     * @param string $title
     */
    private function _setup_sheet($sheet, $title)
    {
        $title = str_replace('\\', ' ', $title); // remove back slash
        $title = substr($title, 0, 31); // reduce length to 31 characters
        $sheet->setTitle($title);
        $sheet->getDefaultColumnDimension()->setWidth(self::DEFAULT_COLUMN_WIDTH);
        $sheet->getPageMargins()->setLeft(0.7)
            ->setRight(self::DEFAULT_PAGE_MARGIN)
            ->setTop(self::DEFAULT_PAGE_MARGIN)
            ->setBottom(self::DEFAULT_PAGE_MARGIN);
        $sheet->getPageSetup()->setPaperSize(PageSetup::PAPERSIZE_A4)
            ->setOrientation(PageSetup::ORIENTATION_LANDSCAPE)
            ->setFitToWidth(1);  // fit to complete width
    }

    /**
     * Try to get similar routes.
     *
     * @param  array $routes
     * @return array
     * @throws isys_exception_database
     */
    private function _get_similar_routes($routes)
    {
        $routeElementCatgDao = new isys_cmdb_dao_category_g_smview_cable_route_element($this->_database);
        $sectionDao = new isys_cmdb_dao_smview_cable_section($this->_database);
        $sectionElementCatgDao = new isys_cmdb_dao_category_g_smview_cable_section_element($this->_database);

        $refCables = [];
        $similarRoutes = [];
        $cableSetNr = 1;
        foreach ($routes as $routeId) {
            $cables = [];
            $rawRouteElements = [];
            $elements = $routeElementCatgDao->get_elements_by_obj_id($routeId, false, false);
            $prevElement = null;
            foreach ($elements as $e) {
                if ($e['section_id']) {
                    // divide section into its wires
                    $section = $sectionDao->get_object('id', $e['section_id']);
                    $sectionElements = $sectionElementCatgDao->get_elements_by_obj_id($section['obj_id'], false, false, false);

                    $sectionElements = isys_cmdb_dao_category_g_smview_cable_section::checkAndFixConnectedSectionSide($prevElement, $sectionElements);
                    foreach ($sectionElements as $s) {
                        $cables[] = $s['cable_id'];
                        $rawRouteElements[] = $s;
                    }
                } else {
                    $cables[] = $e['cable_id'];
                    $rawRouteElements[] = $e;
                }
                $prevElement = $e;
            }

            if (empty($refCables)) {
                $key = "set_$cableSetNr";
                $refCables[$key] = $cables;
                $similarRoutes[$key][$routeId] = $rawRouteElements;
                $cableSetNr++;
            } else {
                // check cables
                $key = null;
                foreach ($refCables as $k => $ref_cable) {
                    if ($ref_cable === $cables) {
                        $key = $k;
                        break;
                    }
                }
                if (! $key) {
                    $key = "set_$cableSetNr";
                    $refCables[$key] = $cables;
                    $cableSetNr++;
                }
                $similarRoutes[$key][$routeId] = $rawRouteElements;
            }
        }
        return $similarRoutes;
    }

    /**
     * Export a single bay.
     *
     * @param  array $routes
     * @param  int $startRow
     * @return array Width and height of this route
     * @throws Exception
     * @throws isys_exception_database
     */
    private function _export_routes($routes, $startRow = 2)
    {
        $lang = isys_application::instance()->container->get('language');
        $routeDao = new isys_cmdb_dao_smview_cable_route($this->_database);
        $routeObjects = [];
        foreach (array_keys($routes) as $rId) {
            $r = $routeDao->get_object('obj_id', $rId);
            $routeObjects[] = $r;
        }

        // first column
        $firstRouteId = $routeObjects[0]['obj_id'];
        $firstRoute = $routeDao->get_object('obj_id', $firstRouteId);
        $startColumn = 1;
        $column = $startColumn;
        $this->_add_cell($column, $startRow + 1, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__PATCH_EXPORT_COMPLETE_LENGTH'));
        $measuredLength = $firstRoute['measured_length'] ? $firstRoute['measured_length'] . ' m' : '';
        $this->_add_cell($column, $startRow + 2, $measuredLength);
        $this->_add_cell($column, $startRow + 3, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__PATCH_EXPORT_MEASURED'));
        $this->_add_border($column, $startRow + 1, $column, $startRow + 3, Border::BORDER_THICK);
        $offset = 4;
        $routeElementCatgDao = new isys_cmdb_dao_category_g_smview_cable_route_element($this->_database);
        foreach (array_keys($routes) as $routeId) {
            $obj = $routeElementCatgDao->get_object_by_id($routeId)->get_row();
            $this->_add_cell($column, $startRow + $offset, $obj['isys_obj__title'], 1, true, self::DEFAULT_ALIGNMENT, true);
            $offset++;
        }
        $this->_add_border($column, $startRow + 4, $column, $startRow + $offset, Border::BORDER_THICK);
        $this->_add_cell($column, $startRow + $offset, '----->');
        $this->_add_border($column, $startRow + $offset, $column, $startRow + $offset, Border::BORDER_THICK);
        $this->_add_cell($column, $startRow + $offset + 1, '<-----');
        $this->_add_border($column, $startRow + $offset + 1, $column, $startRow + $offset + 1, Border::BORDER_THICK);

        // second column
        $column += 1;
        $this->_add_cell($column, $startRow + 1, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__PATCH_EXPORT_DAMPING_MEASUREMENT'), 2, false, 'left');
        $this->_add_cell($column, $startRow + 2, 'E <----- S', 2);
        $this->_add_cell($column, $startRow + 3, '1310nm');
        $this->_add_cell($column + 1, $startRow + 3, '1550nm');
        $this->_add_border($column, $startRow + 1, $column + 1, $startRow + 3, Border::BORDER_THICK);
        $offset = 4;
        foreach ($routeObjects as $r) {
            $dampingIn1310 = $r['damping_left1'] ?: '';
            $dampingIn1550 = $r['damping_left2'] ?: '';
            $this->_add_cell($column, $startRow + $offset, $dampingIn1310, 1, true);
            $this->_add_cell($column + 1, $startRow + $offset, $dampingIn1550, 1, true);
            $offset++;
        }
        $this->_add_cell($column, $startRow + $offset, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__PATCH_EXPORT_DATE'), 2);
        $this->_add_border($column, $startRow + $offset, $column + 1, $startRow + $offset, Border::BORDER_THICK);
        $date = $firstRoute['measure_date'] ?: '';
        if (! empty($date)) {
            // format date
            $datetime = DateTime::createFromFormat ('Y-m-d' , $date);
            $date = strftime('%d.%m.%Y', $datetime->getTimestamp());
        }
        $this->_add_cell($column, $startRow + $offset + 1, $date, 2);
        $this->_add_border($column, $startRow + $offset + 1, $column + 1, $startRow + $offset + 1, Border::BORDER_THICK);

        // third column
        $column += 2;
        $this->_add_cell($column, $startRow + 1, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__PATCH_EXPORT_DAMPING_MEASUREMENT'), 2, false, 'left');
        $this->_add_cell($column, $startRow + 2, 'E -----> S', 2);
        $this->_add_cell($column, $startRow + 3, '1310nm');
        $this->_add_cell($column + 1, $startRow + 3, '1550nm');
        $this->_add_border($column, $startRow + 1, $column + 1, $startRow + 3, Border::BORDER_THICK);
        $offset = 4;
        foreach ($routeObjects as $r) {
            $dampingOut1310 = $r['damping_right1'] ?: '';
            $dampingOut1550 = $r['damping_right2'] ?: '';
            $this->_add_cell($column, $startRow + $offset, $dampingOut1310, 1, true);
            $this->_add_cell($column + 1, $startRow + $offset, $dampingOut1550, 1, true);
            $offset++;
        }
        $this->_add_cell($column, $startRow + $offset, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__PATCH_EXPORT_LENGTH_SUM'), 3, false, 'right');
        $this->_add_border($column, $startRow + $offset, $column + 2, $startRow + $offset, Border::BORDER_THICK);
        $this->_add_cell($column, $startRow + $offset + 1, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__PATCH_EXPORT_REFERENCE_WIRE'), 3, false, 'right');
        $this->_add_border($column, $startRow + $offset + 1, $column + 2, $startRow + $offset + 1, Border::BORDER_THICK);

        // fourth column
        $column += 2;
        $this->_add_cell($column, $startRow + 1, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__PATCH_EXPORT_LOCATION'), 1, true);
        $this->_add_cell($column, $startRow + 2, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__PATCH_EXPORT_CABLE'), 1, true);
        $this->_add_border($column, $startRow + 1, $column, $startRow + 2, Border::BORDER_THICK);
        $this->_add_cell($column, $startRow + 3, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__PATCH_EXPORT_LENGTH'));
        $this->_add_border($column, $startRow + 3, $column, $startRow + 3, Border::BORDER_THICK);
        $offset = 4;
        for ($i = 1; $i <= count($routeObjects); $i++) {
            $this->_add_cell($column, $startRow + $offset, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__PATCH_EXPORT_WIRES'), 1, true);
            $offset++;
        }
        $this->_add_border($column, $startRow + 4, $column, $startRow + $offset, Border::BORDER_THICK);

        // add elements
        $firstObject = null;
        $prevObject = null;
        $column = 7;
        $row = $startRow + 1;
        $prevColumn = 0;
        $prevRow = 0;
        $elementCount = 0;
        $lineBreaks = 0;
        $firstElements = $routes[$firstRouteId];
        foreach ($firstElements as $e) {
            if (! $prevObject) {
                // add first block
                $currentObject = [
                    'id'    => $e['obj_out_id'],
                    'title' => $e['obj_out_title'],
                ];
                $firstObject = $currentObject;
                $wires = array_fill(0, count($routes), null);
                $this->_add_block($column, $row, $currentObject['title'], $wires, $prevColumn, $prevRow);
                $this->_add_border($column, $row + 2, $column, $row + 2, Border::BORDER_THICK);
                $prevColumn = $column;
                $prevRow = $row;
            }

            $currentObject = [
                'id'    => $e['obj_in_id'],
                'title' => $e['obj_in_title'],
            ];
            $column += 2;

            // get all wires of same routes
            $wires = [];
            foreach ($routes as $els) {
                $wires[] = $els[$elementCount];
            }

            if ($elementCount == 7 || ($elementCount > 10 && ($elementCount - 7) % 10 == 0)) {
                // add line break
                $remainingItems = count($firstElements) - $elementCount;
                if ($remainingItems < 10) {
                    // add an offset to display last blocks right-aligned
                    $diff = (2 * 10) - ($remainingItems * 2);
                    $column = $startColumn + $diff;
                } else {
                    $column = $startColumn;
                }
                $row += 6 + count($routes);

                // add labels
                $offset = 3 + count($routeObjects);
                $this->_add_cell($column, $row + $offset, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__PATCH_EXPORT_LENGTH_SUM'), 2, false, 'right');
                $this->_add_border($column, $row + $offset, $column + 1, $row + $offset, Border::BORDER_THICK);
                $this->_add_cell($column, $row + $offset + 1, $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__PATCH_EXPORT_REFERENCE_WIRE'), 2, false, 'right');
                $this->_add_border($column, $row + $offset + 1, $column + 1, $row + $offset + 1, Border::BORDER_THICK);
                $column += 2;
                $lineBreaks++;
            }

            $isLastBlock = ($elementCount + 1) == count($firstElements);
            $this->_add_block($column, $row, $currentObject['title'], $wires, $prevColumn, $prevRow, $isLastBlock);
            $prevColumn = $column;
            $prevRow = $row;
            $prevObject = $currentObject;
            $elementCount++;
        }

        // add last cable length cell
        $this->_add_cell($column + 1, $row + 2, '');
        $this->_add_border($column + 1, $row + 2, $column + 1, $row + 2, Border::BORDER_THICK);

        // add route title
        $routeTitle = $firstObject['title'] . ' - ' . $prevObject['title'];
        $width = $elementCount > 7 ? 21 : $column + 1;
        $this->_add_cell($startColumn, $startRow, $routeTitle, $width, false, 'left', true, '000000', 'FFFFFF', 14);
        return [$width, $row + (count($routes) - 1)];
    }

    /**
     * Add a block for each object.
     *
     * @param  int $x
     * @param  int $y
     * @param  string $objectName
     * @param  array $wires
     * @throws Exception
     * @throws isys_exception_database
     */
    private function _add_block($x, $y, $objectName, $wires, $prevX, $prevY, $isLastBlock = false)
    {
        $wirePins = [
            'in' => [
                'title'          => '',
                'color'          => '000000',
                'bg_color'       => 'FFFFFF',
                'location_color' => 'FFFFFF',
            ],
            'out' => [
                'title'          => '',
                'color'          => '000000',
                'bg_color'       => 'FFFFFF',
                'location_color' => 'FFFFFF',
            ],
        ];

        $firstWire = count($wires) > 0 ? $wires[0] : null;
        if ($firstWire) {
            // cable title of last object right
            $cableName = $firstWire['cable_title'];
            if (strpos($cableName, 'F ') === 0) {
                $cableName = substr($cableName, 2);
            }
            if ($y == $prevY) {
                $this->_add_cell($x - 1, $y + 1, $cableName);
            } else {
                // handle line break
                $this->_add_cell($prevX + 1, $prevY + 1, $cableName);
            }

            // calculate alphabetical column name
            $columnIndex = $prevX + 1;
            $columnLetter = Cell\Coordinate::stringFromColumnIndex($columnIndex);

            // cable title of current object left
            $lengthFromLeftColumn = '=' . $columnLetter . ($prevY + 1);
            $this->_add_cell($x, $y + 1, $lengthFromLeftColumn);
            // empty title of current object right
            $this->_add_cell($x + 1, $y + 1, '');

            // cable length
            $length = $firstWire['length'] ?: '0';
            $specialBackgroundColor = empty($length) ? 'FF7060' : 'FFFFFF';
            if ($y == $prevY) {
                $this->_add_cell($x - 1, $y + 2, $length, 1, false, 'right', false, '000000', $specialBackgroundColor);
                $this->_add_cell($x, $y + 2, 'm', 1, false, 'left');
                $this->_add_border($x - 1, $y + 2, $x, $y + 2, Border::BORDER_THICK);
            } else {
                $this->_add_cell($prevX + 1, $prevY + 2, $length, 1, false, 'right', false, '000000', $specialBackgroundColor);
                $this->_add_cell($prevX + 2, $prevY + 2, 'm', 1, false, 'left');
                $this->_add_border($prevX + 1, $prevY + 2, $prevX + 2, $prevY + 2, Border::BORDER_THICK);
                $this->_add_cell($x, $y + 2, $length . ' m');
                $this->_add_border($x, $y + 2, $x, $y + 2, Border::BORDER_THICK);
            }

            $yOffset = 3;
            foreach ($wires as $wire) {
                // add wire titles and wire PIN status colors
                if ($wire['wire_id']) {
                    $wireTitle = ltrim($wire['wire_title'], '0');
                    $wirePins = $this->_update_wire_pin_colors_by_wire($wire['wire_id'], $wireTitle, $wirePins);
                }

                if ($y == $prevY) {
                    $this->_add_cell($x - 1, $y + $yOffset, $wirePins['in']['title'], 1, true, self::DEFAULT_ALIGNMENT, false, $wirePins['in']['color'], $wirePins['in']['bg_color']);
                    $this->_add_partial_border($x - 1, $y + $yOffset, false, true, false, false, Border::BORDER_THICK);
                } else {
                    // handle line break
                    $this->_add_cell($prevX + 1, $prevY + $yOffset, $wirePins['in']['title'], 1, false, self::DEFAULT_ALIGNMENT, false, $wirePins['in']['color'], $wirePins['in']['bg_color']);
                }
                $prevWireColumn = '=' . $columnLetter . ($prevY + $yOffset);
                $this->_add_cell($x, $y + $yOffset, $prevWireColumn, 1, true, self::DEFAULT_ALIGNMENT, false, $wirePins['out']['color'], $wirePins['out']['bg_color']);
                $this->_add_cell($x + 1, $y + $yOffset, 'LC', 1, true, self::DEFAULT_ALIGNMENT);
                $this->_add_partial_border($x, $y + $yOffset, false, false, false, true, Border::BORDER_THICK);
                $this->_add_partial_border($x + 1, $y + $yOffset, false, true, false, false, Border::BORDER_THICK);
                $yOffset++;
            }
        } else {
            // add empty cell on first object
            $this->_add_cell($x, $y + 1, '');
            $this->_add_cell($x, $y + 2, '');
            for ($i = 0; $i < count($wires); $i++) {
                $this->_add_cell($x, $y + 3 + $i, 'LC');
            }
        }

        // object title
        $this->_add_cell($x, $y, $objectName, 2, false, 'left', true, '000000', $wirePins['out']['location_color']);
        $cell = $this->_sheet->getCellByColumnAndRow($x - 2, $y);
        $prevTitle = $cell->getValue();
        if ($prevTitle && $wirePins['in']['location_color'] != 'FFFFFF') {
            // update last object title, if background color has changed
            $this->_add_cell($x - 2, $y, $prevTitle, 2, false, 'left', true, '000000', $wirePins['in']['location_color']);
        }
        $this->_add_border($x, $y, $x + 1, $y + 1, Border::BORDER_THICK);

        // add lengths
        $this->_add_summarized_length($x, $y, count($wires), $prevX, $prevY, $isLastBlock);
    }

    /**
     * Add summarized length on each block.
     *
     * @param  int $x
     * @param  int $y
     * @param  int $wireCount
     * @param  int $prevX
     * @param  int $prevY
     * @param  boolean $isLastBlock
     * @throws Exception
     */
    private function _add_summarized_length($x, $y, $wireCount, $prevX, $prevY, $isLastBlock = false)
    {
        $offset = $wireCount + 3;
        $lengthFromLeftColumn = '0';

        // current length of this block
        $columnLetter = Cell\Coordinate::stringFromColumnIndex($prevX +1);
        $currentLengthCell = $columnLetter . ($prevY + 2);
        if (! empty($prevX)) {
            // summarize current length and previous length
            $columnLetter = Cell\Coordinate::stringFromColumnIndex($prevX);
            $lengthFromLeftColumn = '=' . $currentLengthCell . '+' . $columnLetter . ($prevY + $offset);
        }
        $this->_add_cell($x, $y + $offset, $lengthFromLeftColumn, 1, false, 'right');
        $this->_add_cell($x + 1, $y + $offset, 'm ->', 1, false, 'left');
        $this->_add_border($x, $y + $offset, $x + 1, $y + $offset, Border::BORDER_THICK);

        if (! empty($prevX)) {
            // set previous length from right
            $lengthFromRightColumn = '=' . $currentLengthCell;
            if (! $isLastBlock) {
                // summarized with current cell
                $columnLetter = Cell\Coordinate::stringFromColumnIndex($x);
                $lengthFromRightColumn .= '+' . $columnLetter . ($y + 1 + $offset);
            }
            $this->_add_cell($prevX, $prevY + 1 + $offset, $lengthFromRightColumn, 1, false, 'right');
            $this->_add_cell($prevX + 1, $prevY + 1 + $offset, 'm <-', 1, false, 'left');
        }

        // set current length from right empty
        $this->_add_cell($x, $y + 1 + $offset, '0', 1, false, 'right');
        $this->_add_cell($x + 1, $y + 1 + $offset, 'm <-', 1, false, 'left');
        $this->_add_border($x, $y + 1 + $offset, $x + 1, $y + 1 + $offset, Border::BORDER_THICK);
    }

    /**
     * Get colors of the wire PINs.
     *
     * @param  int $wireId
     * @param  string $wireTitle
     * @param  array $wirePins
     * @return array
     * @throws isys_exception_database
     */
    private function _update_wire_pin_colors_by_wire($wireId, $wireTitle, $wirePins)
    {
        if (! empty($wireId)) {
            $wirePinDao = new isys_cmdb_dao_smview_cable_wire_pin($this->_database);
            $wirePinStatusDao = new isys_cmdb_dao_smview_cable_wire_pin_status($this->_database);
            $wPins = $wirePinDao->filter_objects(['wire_id' => $wireId]);
            foreach ($wPins as $w) {
                $s = $wirePinStatusDao->get_object('id', $w['status_id']);
                // todo: validate side
                $colors = $this->_get_wire_pin_colors($s);
                $side = $w['id'] < $w['sibling_id'] ? 'in' : 'out';
                $wirePins[$side] = $colors;
            }
        }
        $wirePins['in']['title'] = $wireTitle;
        $wirePins['out']['title'] = $wireTitle;
        return $wirePins;
    }

    /**
     * Set wire pin colors.
     *
     * @param  array $s
     * @return array
     */
    private function _get_wire_pin_colors($s)
    {
        $colors = [
            'color'          => '000000',
            'bg_color'       => 'FFFFFF',
            'location_color' => 'FFFFFF',
        ];
        $colors['color'] = $s['color'];
        if ($s['const'] == 'H' || $s['const'] == 'VH') {
            // use color to set background on special wire pin status
            $colors['color'] = '000000';
            $colors['bg_color'] = $s['color'];
            $colors['location_color'] = $s['color'];
        } else {
            $this->_usedWireStates[$s['id']] = $s;
        }
        return $colors;
    }

    /**
     * Add border around coordinates.
     *
     * @param int $x1
     * @param int $y1
     * @param boolean $top
     * @param boolean $right
     * @param boolean $bottom
     * @param boolean $left
     * @param string $thickness
     */
    private function _add_partial_border($x1, $y1, $top = false, $right = false, $bottom = false, $left = false, $thickness = Border::BORDER_THIN)
    {
        $style = $this->_sheet->getStyleByColumnAndRow($x1, $y1);
        if ($top) {
            $style->getBorders()->getTop()->setBorderStyle($thickness);
        }
        if ($right) {
            $style->getBorders()->getRight()->setBorderStyle($thickness);
        }
        if ($bottom) {
            $style->getBorders()->getBottom()->setBorderStyle($thickness);
        }
        if ($left) {
            $style->getBorders()->getLeft()->setBorderStyle($thickness);
        }
    }

    /**
     * Add legend.
     *
     * @param  int $cellY
     * @throws Exception
     */
    private function _add_legend($cellY)
    {
        $offset = 0;
        foreach ($this->_usedWireStates as $s) {
            if (! in_array($s['const'], ['H', 'VH'])) {
                $this->_add_cell(2, $cellY + $offset, $s['const'], 1, false, 'left', false, $s['color']);
                $offset++;
            }
        }
    }
}
