<?php

namespace idoit\Module\Cable;

use Exception;
use InvalidArgumentException;
use isys_application;
use isys_cmdb_dao_smview_cable_generator;

/**
 * Label Generator class.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class LabelGenerator
{
    
    /**
     * Generator.
     *
     * @var array
     */
    private $_generator;
    
    /**
     * Algorithms.
     *
     * @var array
     */
    private $_algorithms;
    
    /**
     * Output of name generator.
     *
     * @var array
     */
    private $_output = [];
    
    /**
     * Current separators.
     *
     * @var array
     */
    private $_separators = [];
    
    /**
     * End-pattern.
     *
     * @var string
     */
    private $_end;
    
    /**
     * Maximum count of generated elements.
     *
     * @var int
     */
    private $_max_count;
    
    /**
     * Constructor.
     *
     * @param int $generator_id
     */
    public function __construct($generator_id)
    {
        $dao = new isys_cmdb_dao_smview_cable_generator(isys_application::instance()->database);
        $this->_generator = $dao->get_object('id', $generator_id);
        $this->_algorithms = $dao->get_assigned_algorithms($generator_id);
        $this->_separators = $this->_getSeparatorsByMask();
    }

    /**
     * Generate names by generator.
     *
     * @param  string $start
     * @param  string $end
     * @param  int $maxCount
     * @return array
     * @throws Exception
     */
    public function generate_names_by_generator($start, $end = null, $maxCount = null)
    {
        $this->_output = [];
        $this->_end = $end;
        $this->_max_count = $maxCount;
        
        $this->_validatePatterns($start, $end);
        
        $counterMin = $this->_generator['counter_min'];
        $counterMax = $this->_generator['counter_max'];
        
        $initParts = $this->_splitBySeparator($start);
        $finishParts = $this->_splitBySeparator($end);
        $startParts = $this->_splitBySeparator($counterMin);
        $endParts = $this->_splitBySeparator($counterMax);
        
        // initialize counters
        $counters = $this->_getCounters($startParts, $endParts);
        
        if (! empty($end)) {
            $this->_validateValueRange($initParts, $finishParts, $counters);
        }
        
        $current_values = [];
        $current_ends = [];
        foreach (array_keys($counters) as $index) {
            // sort by occurency within mask
            $current_values[$index] = $initParts[$index - 1];
            if (! empty($end)) {
                $current_ends[$index] = $finishParts[$index - 1];
            } else {
                // use value of last counter if no end pattern was provided
                $current_ends[$index] = $endParts[$index - 1];
            }
        }
        
        // starting with last algorithm
        $start_algorithm = count($this->_separators) + 1;
        $this->_runAlgorithm($start_algorithm, $counters, $current_values, $current_values, $current_ends);
        
        return $this->_output;
    }
    
    /**
     * Get separators by mask.
     *
     * @return array
     */
    private function _getSeparatorsByMask()
    {
        $mask = $this->_generator['mask'];
        $elementCount = substr_count($mask, '#');
        $separators = [];
        for ($i = 0; $i < $elementCount; $i++) {
            $pos = strpos($mask, '#');
            if ($i > 0) {
                $separators[] = substr($mask, 0, $pos);
            }
            $mask = substr($mask, $pos + 1, strlen($mask));
        }
        return $separators;
    }
    
    /**
     * Split string by separators.
     *
     * @param  string $string
     * @return array
     */
    private function _splitBySeparator($string)
    {
        $parts = [];
        $algos = $this->_getAlgorithms();
        foreach ($this->_separators as $index => $separator) {
            if ($separator == '') {
                // use algorithm to get character length
                $algo = $algos[$index + 1];
                $pos = (int) $algo['length'];
            } else {
                $pos = strpos($string, $separator);
            }
            $parts[] = substr($string, 0, $pos);
            $string = substr($string, ($pos + strlen($separator)), strlen($string));
        }
        $parts[] = $string;
        return $parts;
    }
    
    /**
     * Get algorithms ordered by position within mask.
     *
     * @return array
     */
    private function _getAlgorithms()
    {
        $maskElementCount = count($this->_separators) + 1;
        $algorithms = [];
        for ($i = 1; $i <= $maskElementCount; $i++) {
            // get matching pos from algo
            foreach ($this->_algorithms as $a) {
                if ($a['pos'] == $i) {
                    // sorted by occurency within mask
                    $algorithms[$i] = $a;
                    break;
                }
            }
        }
        return $algorithms;
    }

    /**
     * Validate if separators and patterns matches.
     *
     * @param  string $start
     * @param  string $end
     * @throws InvalidArgumentException
     * @throws Exception
     */
    private function _validatePatterns($start, $end = null)
    {
        $language = isys_application::instance()->container->get('language');
        if (empty($this->_separators)) {
            return;
        }
        
        $maskElementCount = count($this->_separators) + 1;
        $initParts = $this->_splitBySeparator($start);
        $finishParts = $this->_splitBySeparator($end);
        
        // validate pattern count
        if ($maskElementCount != count($initParts)) {
            throw new InvalidArgumentException(
                $language->get('LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_VALIDATION_START_PARAM_MISMATCH')
            );
        }
        
        if (! empty($end) && $maskElementCount != count($finishParts)) {
            throw new InvalidArgumentException(
                $language->get('LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_VALIDATION_END_PARAM_MISMATCH')
            );
        }
        
        $algos = $this->_getAlgorithms();
        foreach ($this->_separators as $index => $separator) {
            if ($separator == '') {
                // use algorithm to get character length
                $algo = $algos[$index + 1];
                $pos = (int) $algo['length'];
                $start = substr($start, 0, $pos - 1);
                $end = substr($end, 0, $pos - 1);
            } else {
                $sPos = strpos($start, $separator);
                $sSep = substr($start, $sPos, strlen($separator));
                $start = substr($start, ($sPos + strlen($separator)), strlen($start));
                
                $ePos = strpos($end, $separator);
                $eSep = substr($end, $ePos, strlen($separator));
                $end = substr($end, ($ePos + strlen($separator)), strlen($end));

                if ($sSep != $separator) {
                    throw new InvalidArgumentException(
                        $language->get('LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_VALIDATION_START_SEP_MISMATCH')
                    );
                }
                if (! empty($end) && $eSep != $separator) {
                    throw new InvalidArgumentException(
                        $language->get('LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_VALIDATION_END_SEP_MISMATCH')
                    );
                }
            }
        }
    }

    /**
     * Validate if values are within counter range.
     *
     * @param  array $startParts
     * @param  array $endParts
     * @param  array $counters
     * @throws InvalidArgumentException
     * @throws Exception
     */
    private function _validateValueRange($startParts, $endParts, $counters)
    {
        $language = isys_application::instance()->container->get('language');
        foreach ($counters as $i => $counter) {
            $startPart = $startParts[$i - 1];
            $endPart = $endParts[$i - 1];
            if ($counter->getType() == CounterItem::TYPE_NUMERIC) {
                $rangeStart = (int) $counter->getFirst();
                $rangeEnd = (int) $counter->getLast();
                
                $startPart = (int) $startPart;
                $endPart = (int) $endPart;
                if ($startPart < $rangeStart || $startPart > $rangeEnd) {
                    throw new InvalidArgumentException(
                        $language->get('LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_VALIDATION_VALUE') .
                        " '$startPart' " .
                        $language->get('LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_VALIDATION_NOT_IN_RANGE') .
                        " $rangeStart / $rangeEnd"
                    );
                }
                if ($endPart < $rangeStart || $endPart > $rangeEnd) {
                    throw new InvalidArgumentException(
                        $language->get('LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_VALIDATION_VALUE') .
                        " '$endPart' " .
                        $language->get('LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_VALIDATION_NOT_IN_RANGE') .
                        " $rangeStart / $rangeEnd"
                    );
                }
            } else {
                $enum = $counter->getEnum();
                if (! in_array($startPart, $enum)) {
                    throw new InvalidArgumentException(
                        $language->get('LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_VALIDATION_VALUE') .
                        " '$startPart' " .
                        $language->get('LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_VALIDATION_NOT_IN_ENUM')
                    );
                }
                if (! in_array($endPart, $enum)) {
                    throw new InvalidArgumentException(
                        $language->get('LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_VALIDATION_VALUE') .
                        " '$endPart' " .
                        $language->get('LC__CMDB__CATG__SMVIEW_CABLE__GENERATOR_VALIDATION_NOT_IN_ENUM')
                    );
                }
            }
        }
    }
    
    /**
     * Initialize counters.
     *
     * @param  array $startParts
     * @param  array $endParts
     * @return array
     */
    private function _getCounters($startParts, $endParts)
    {
        $counters = [];
        $algos = $this->_getAlgorithms();
        foreach ($algos as $i => $algorithm) {
            $type = $algorithm['type'];
            $start = $startParts[$i - 1];
            $end = $endParts[$i - 1];
            if ($type == CounterItem::TYPE_NUMERIC) {
                $start = (int) $start;
                $end = (int) $end;
            }
            $length = (int) $algorithm['length'];
            $enum = null;
            if (! empty($algorithm['elements'])) {
                $enum = explode(';', $algorithm['elements']);
            }
            $counter = new CounterItem($start, $type, $length, $end, $enum);
            
            // sorted by occurency within mask
            $counters[$i] = $counter;
        }
        return $counters;
    }
    
    /**
     * Run algorithm.
     *
     * @param  int   $algo_index
     * @param  array $counters
     * @param  array $currents
     * @param  array $startParts
     * @param  array $finishParts
     * @return array
     */
    private function _runAlgorithm($algo_index, $counters, $currents, $startParts, $finishParts)
    {
        $algo_setup = $this->_algorithms[$algo_index];
        $pattern_pos = $algo_setup['pos'];
        $counter = $counters[$pattern_pos];
        
        $start_part = $startParts[$pattern_pos];
        $end_part = $finishParts[$pattern_pos];
        
        $next = $currents[$pattern_pos];
        while (true) {
            // save current item
            $currents[$pattern_pos] = $next;
            
            if (($algo_index - 1) >= 1) {
                // if there is an algorithm before, execute it first
                $currents = $this->_runAlgorithm($algo_index - 1, $counters, $currents, $startParts, $finishParts);
            } else {
                $this->_saveOutput($currents, $counters);
            }
            
            if ($start_part == $end_part) {
                // exit if start-part matches end-part
                break;
            }
            
            // finish if output matches end-pattern
            $lastOutput = end($this->_output);
            if (isset($this->_end) && $lastOutput == $this->_end) {
                return $currents;
            }
            
            // finish on maximum count
            if (isset($this->_max_count) && count($this->_output) >= $this->_max_count) {
                return $currents;
            }
            
            // finish on more than 49.999 labels
            if (count($this->_output) > 49999) {
                return $currents;
            }
            
            if ($counter->isLast($next)) {
                // exit on last element
                break;
            }
            
            // get next element
            $next = $counter->getNext($next);
        }
        
        if ($start_part == $end_part) {
            // keep value
            $currents[$pattern_pos] = $start_part;
        } else {
            // reset to first element
            $currents[$pattern_pos] = $counter->getFirst();
        }
        return $currents;
    }

    /**
     * Save output to internal array.
     *
     * @param  array $currents
     * @param  array $counters
     * @return string
     */
    private function _saveOutput($currents, $counters)
    {
        $values = [];
        foreach ($currents as $index => $value) {
            $counter = $counters[$index];
            $val = $value;
            if ($counter && $counter->getType() == CounterItem::TYPE_NUMERIC) {
                // add leading zeros
                $length = $counter->getLength();
                $val = str_pad($value, $length, '0', STR_PAD_LEFT);
            }
            $values[] = $val;
        }
        
        // add separators
        $output = '';
        for ($i = 0; $i < count($values); $i++) {
            $output .= $values[$i];
            if (isset($this->_separators[$i])) {
                $output .= $this->_separators[$i];
            }
        }
        $this->_output[] = $output;
        return $output;
    }
}

class CounterItem
{
    
    const TYPE_NUMERIC = 'n';
    const TYPE_ENUMERATION = 'e';
    
    private $start = null;
    private $end = null;
    private $length = null;
    private $type = null;
    private $enum = [];
    
    /**
     * Constructor.
     *
     * @param int|string $start
     * @param string     $type
     * @param int        $length
     * @param int|string $end
     * @param array      $enum
     */
    public function __construct($start, $type, $length, $end = null, $enum = [])
    {
        $this->start = $start;
        $this->end = $end;
        $this->length = $length;
        $this->type = $type;
        $this->enum = $enum;
    }
    
    /**
     * Get first item.
     *
     * @return int|string
     */
    public function getFirst()
    {
        return $this->start;
    }
    
    /**
     * Get last item.
     *
     * @return int|string
     */
    public function getLast()
    {
        return $this->end;
    }
    
    /**
     * Get enum.
     *
     * @return array
     */
    public function getEnum()
    {
        return $this->enum;
    }
    
    /**
     * Get length.
     *
     * @return int
     */
    public function getLength()
    {
        return $this->length;
    }
    
    /**
     * Get type.
     *
     * @return string
     */
    public function getType()
    {
        return $this->type;
    }
    
    /**
     * Returns if this item is the last one or not.
     *
     * @param  int|string $item
     * @return boolean
     */
    public function isLast($item)
    {
        $last = $this->_getLast($this->start);
        return $item == $last;
    }
    
    /**
     * Get next item.
     *
     * @param  int|string $item
     * @return int|string
     */
    public function getNext($item)
    {
        if ($this->isLast($item)) {
            return $this->start;
        }
        if ($this->type == self::TYPE_NUMERIC) {
            $item = (int) $item;
            return $item + 1;
        } else {
            // get current position
            $pos = array_search($item, $this->enum);
            
            // get next
            $pos++;
            return $this->enum[$pos];
        }
    }
    
    /**
     * Get end-pattern, last enum or highest possible number.
     *
     * @param  int $item
     * @return int|string
     */
    private function _getLast($item)
    {
        if ($this->type != self::TYPE_NUMERIC) {
            // get last enum
            return end($this->enum);
        }
        
        $last = $item;
        $length = (int) $this->length;
        if (empty($this->end)) {
            // get highest number
            $next = $item + 1;
            if (strlen($next) <= $length) {
                $last = $this->_getLast($next);
            }
        } else {
            // return end-pattern
            $last = $this->end;
        }
        return $last;
    }
}
