<?php

namespace idoit\Module\Cable;

use idoit\Module\Cable\Algorithms\AllPaths;
use isys_application;
use isys_cmdb_dao_category_g_global;
use isys_cmdb_dao_category_g_location;
use isys_cmdb_dao_category_g_virtual_object;
use isys_cmdb_dao_smview_cable;
use isys_cmdb_dao_smview_cable_connection_view;
use isys_exception_database;

/**
 * Location Tracing class.
 *
 * @package i-doit
 * @subpackage smview_cable
 * @author Steve Groeber <steve.groeber@shd-online.de>
 * @copyright SHD System-Haus-Dresden GmbH
 * @license http://www.i-doit.com/license
 */
class LocationTracing
{
    
    /**
     * Connection list.
     *
     * @var array
     */
    private $_connections = [];

    /**
     * List of handled objects.
     *
     * @var array
     */
    private $_handled_objects = [];

    /**
     * Get nodes and edges between locations.
     *
     * @param  int $startLocation
     * @param  int $endLocation
     * @return array
     * @throws isys_exception_database
     */
    public function get_nodes_between_locations($startLocation, $endLocation)
    {
        $obj_dao = isys_cmdb_dao_category_g_virtual_object::instance(isys_application::instance()->database);
        $startObjects = $obj_dao->get_data(null, $startLocation)->__as_array();
        $endObjects = $obj_dao->get_data(null, $endLocation)->__as_array();
        
        $nodes = $edges = [];
        $flat_connections = [];
        foreach ($endObjects as $endObject) {
            $endObjId = $endObject['isys_obj__id'];
            foreach ($startObjects as $startObject) {
                $objId = $startObject['isys_obj__id'];
                if (! isset($flat_connections[$objId])) {
                    $flat_connections[$objId] = $this->get_connections($objId);
                }
                $flat_connection = $flat_connections[$objId];
                
                $nodeAndEdgeList = $this->get_nodes_between_objects($objId, $endObjId, $flat_connection);
                foreach ($nodeAndEdgeList['nodes'] as $k => $v) {
                    $nodes[$k] = $v;
                }
                foreach ($nodeAndEdgeList['edges'] as $k => $v) {
                    $edges[$k] = $v;
                }
            }
        }
        
        return array(
            'nodes' => $nodes,
            'edges' => $edges,
        );
    }

    /**
     * Get nodes and edges between objects.
     *
     * @param  int $objId
     * @param  int $endObjId
     * @param  array $flat_connections
     * @return array
     * @throws isys_exception_database
     */
    public function get_nodes_between_objects($objId, $endObjId, $flat_connections)
    {
        $obj_dao = isys_cmdb_dao_category_g_global::instance(isys_application::instance()->database);
        $allPaths = new AllPaths($flat_connections);
        $routes = $allPaths->calculate($objId, $endObjId);

        $nodes = $edges = [];
        foreach ($routes as $route) {
            $prevObjId = null;
            foreach ($route as $objId) {
                $obj = $obj_dao->get_object_by_id($objId)->get_row();
                $nodes[$objId] = array(
                    'id'    => $objId,
                    'title' => $obj['isys_obj__title'],
                );
                $key = $prevObjId . '-' . $objId;
                $wires = $this->_get_wires_between_objects($prevObjId, $objId);
                $edges[$key] = array(
                    'id'    => $key,
                    'from'  => $prevObjId,
                    'to'    => $objId,
                    'title' => implode("\n", $wires),
                );
                $prevObjId = $objId;
            }
        }
        
        return array(
            'nodes' => $nodes,
            'edges' => $edges,
        );
    }

    /**
     * Get connections by object ID.
     *
     * @param  int $objId
     * @return array
     */
    public function get_connections($objId)
    {
        $this->_connections = [];
        $this->_handled_objects = array(
            $objId => true,
        );
        
        $this->_add_object_connections($objId);

        $flat_connections = [];
        foreach ($this->_connections as $k => $v) {
            $flat_connections[$k] = array_keys($v);
        }
        
        return $flat_connections;
    }
    
    /**
     * Add connections from object.
     *
     * @param int $objId
     */
    private function _add_object_connections($objId)
    {
        $dao_view = new isys_cmdb_dao_smview_cable_connection_view(isys_application::instance()->database);
        $this->_handled_objects[$objId] = true;

        $connections_in = $dao_view->filter_objects(array('obj_id_in' => $objId));
        $connections_out = $dao_view->filter_objects(array('obj_id_out' => $objId));
        
        $this->_add_to_connection_list($objId, $connections_in, 'obj_id_out');
        $this->_add_to_connection_list($objId, $connections_out, 'obj_id_in');
    }
    
    /**
     * Add connections to connection list.
     *
     * @param int $objId
     * @param array $connection_list
     * @param string $identifier
     */
    private function _add_to_connection_list($objId, $connection_list, $identifier)
    {
        foreach ($connection_list as $c) {
            $objId2 = $c[$identifier];
            if (isset($this->_handled_objects[$objId2])) {
                continue;
            }

            $lowerId = $objId;
            $higherId = $objId2;
            if ($objId2 < $objId) {
                $lowerId = $objId2;
                $higherId = $objId;
            }
            $this->_connections[$lowerId][$higherId] = true;
            // save also reverse connection
            $this->_connections[$higherId][$lowerId] = true;
            
            $this->_add_object_connections($objId2);
        }
    }
    
    /**
     * Get wires between objects.
     *
     * @param  int $prevObjId
     * @param  int $objId
     * @return array
     */
    private function _get_wires_between_objects($prevObjId, $objId)
    {
        $dao_view = new isys_cmdb_dao_smview_cable_connection_view(isys_application::instance()->database);
        $connections_in = $dao_view->filter_objects(array('obj_id_in' => $prevObjId, 'obj_id_out' => $objId));
        $connections_out = $dao_view->filter_objects(array('obj_id_in' => $objId, 'obj_id_out' => $prevObjId));
        $wires = [];
        foreach ($connections_in as $c) {
            $wires[$c['wire_id']] = $c;
        }
        foreach ($connections_out as $c) {
            $wires[$c['wire_id']] = $c;
        }
        
        $titles = [];
        foreach ($wires as $c) {
            $titles[] = $c['cable_title'] . ' - ' . $c['wire_title'];
        }
        return $titles;
    }

    /**
     * Returns the routing tree between two locations.
     *
     * @param  int $sourceLocationId
     * @param  int $targetLocationId
     * @return array
     * @throws isys_exception_database
     */
    public function get_routing_tree($sourceLocationId, $targetLocationId)
    {
        $dao = isys_cmdb_dao_smview_cable::instance(isys_application::instance()->database);
        $obj_dao = isys_cmdb_dao_category_g_global::instance(isys_application::instance()->database);
        $matchingLocationRoutes = $this->get_routes_based_on_locations($sourceLocationId, $targetLocationId);
        
        $nodes = $edges = [];
        foreach ($matchingLocationRoutes as $route) {
            $locations = explode('-', $route);
            $prevLocation = null;
            foreach ($locations as $location_id) {
                $obj_location = $obj_dao->get_object_by_id($location_id)->get_row();
                $nodes[$location_id] = array(
                    'id'    => $location_id,
                    'title' => $obj_location['isys_obj__title'],
                );
                
                if (! $prevLocation) {
                    $prevLocation = $location_id;
                    continue;
                }
                
                $key = $prevLocation . '-' . $location_id;
                if (! isset($edges[$key])) {
                    $count = $dao->get_cable_count($prevLocation, $location_id);
                    $edges[$key] = array(
                        'id'    => $key,
                        'from'  => $prevLocation,
                        'to'    => $location_id,
                        'title' => "Cable Count: $count",
                    );
                }
                $prevLocation = $location_id;
            }
        }
        return array(
            'nodes' => $nodes,
            'edges' => $edges,
        );
    }
    
    /**
     * Try to get all routes between two locations.
     *
     * @param  int $sourceLocationId
     * @param  int $targetLocationId
     * @return array
     */
    public function get_routes_based_on_locations($sourceLocationId, $targetLocationId)
    {
        $dao = isys_cmdb_dao_smview_cable::instance(isys_application::instance()->database);
        $startingCables = $dao->filter_objects(array('location_in_id' => $sourceLocationId));
        $endingCables = $dao->filter_objects(array('location_out_id' => $targetLocationId));
        if (empty($startingCables) || empty($endingCables)) {
            // no cables starting or ending at this locations
            return [];
        }
        
        // prepare mappings
        $location_targets = $this->_get_location_targets();
        
        // get routes, starting at source location
        $source = "$sourceLocationId";
        $targets = $location_targets[$sourceLocationId];
        $routes = [];
        $this->_get_routes($source, $targets, $location_targets, $routes);
        
        // get routes, ending at specified target location
        $matchingLocationRoutes = [];
        foreach ($routes as $route) {
            $parts = explode('-', $route);
            $lastId = array_pop($parts);
            if ($lastId == $targetLocationId) {
                $matchingLocationRoutes[] = $route;
            }
        }
        return $matchingLocationRoutes;
    }
    
    /**
     * Get a list with all targets of the locations.
     *
     * @return array
     */
    private function _get_location_targets()
    {
        $location_targets = [];
        $dao = isys_cmdb_dao_smview_cable::instance(isys_application::instance()->database);
        $locations = $dao->get_objects();
        foreach ($locations as $l) {
            $location_in = $l['location_in_id'];
            $location_out = $l['location_out_id'];
            if (empty($location_in) || empty($location_out)) {
                continue;
            }
            $location_targets[$location_in][$location_out] = $location_out;
        }
        
        return $location_targets;
    }
    
    /**
     * Get all routes, starting at base source.
     *
     * @param string $baseSource
     * @param array $targets
     * @param array $location_targets
     * @param array $routes
     */
    private function _get_routes($baseSource, $targets, $location_targets, &$routes)
    {
        foreach ($targets as $target_id) {
            $source = "$baseSource-$target_id";
            $routes[$source] = $source;
            if (! isset($location_targets[$target_id])) {
                continue;
            }
            $sub_targets = $location_targets[$target_id];
            $this->_get_routes($source, $sub_targets, $location_targets, $routes);
        }
    }
    
    /**
     * Get objects by location.
     *
     * @param  int $location_id
     * @return array
     */
    public function get_objects_by_location($location_id)
    {
        $dao_view = new isys_cmdb_dao_smview_cable_connection_view(isys_application::instance()->database);
        $res = $dao_view->get_data();
        $objects = [];
        foreach ($res as $c) {
            $obj_id = $c['obj_id_in'];
            try {
                $title = $this->_get_location_title($obj_id, $location_id);
            } catch (isys_exception_database $e) {
            }
            if (empty($title)) {
                continue;
            }

            $obj_id = $c['obj_id_out'];
            try {
                $title = $this->_get_location_title($obj_id, $location_id);
            } catch (isys_exception_database $e) {
            }
            if (empty($title)) {
                continue;
            }
            
            $objects[$obj_id] = array(
                'id'    => $obj_id,
                'title' => $title,
            );
        }
        
        uasort($objects, function ($a, $b) {
            return strnatcmp($a['title'], $b['title']);
        });
        return array_values($objects);
    }

    /**
     * Get location title by object ID.
     *
     * @param  int $obj_id
     * @param  int $location_id
     * @return string|null
     * @throws isys_exception_database
     */
    private function _get_location_title($obj_id, $location_id)
    {
        if (empty($obj_id)) {
            return null;
        }
        
        $location_dao = isys_cmdb_dao_category_g_location::instance(isys_application::instance()->database);
        $obj_dao = isys_cmdb_dao_category_g_global::instance(isys_application::instance()->database);
        
        $locations = $location_dao->get_data(null, $obj_id)->__as_array();
        if (empty($locations)) {
            return null;
        }

        $location = $locations[0];
        if ($location['isys_catg_location_list__parentid'] != $location_id) {
            return null;
        }

        $obj = $obj_dao->get_object_by_id($obj_id)->get_row();
        if ($obj['isys_obj__status'] != defined_or_default('C__RECORD_STATUS__NORMAL')) {
            return null;
        }
        return $obj['isys_obj__title'];
    }
}
