<?php

namespace idoit\Module\Cable;

use Exception;
use isys_application;
use isys_cmdb_dao_category;
use isys_cmdb_dao_category_g_global;
use isys_cmdb_dao_category_g_location;
use isys_component_dao_result;
use isys_exception_database;
use isys_module_smview_cable;
use isys_notify;
use isys_settings;

/**
 * Object Title Validator class.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  2023 SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class ObjectTitleValidator
{

    /**
     * Validate an object after a category was saved.
     *
     * @param  isys_cmdb_dao_category $dao
     * @param  int $catId
     * @param  array $result
     * @param  int $objId
     * @param  array $post
     * @param  array $log
     * @throws isys_exception_database
     */
    public static function validateAfterCategorySave($dao, $catId, $result, $objId, $post, $log)
    {
        if (! ($dao instanceof isys_cmdb_dao_category_g_global || $dao instanceof isys_cmdb_dao_category_g_location)) {
            return;
        }

        $updatedObject = $dao->get_data(null, $objId)->get_row();
        $value = $updatedObject['isys_obj__title'];
        $propertiesToCheck = [
            'title' => $value ?? null,
        ];

        $result = self::validate($objId, null, $propertiesToCheck);
        if (! $result['success']) {
            // rename object to avoid duplicate property-value
            $newValue = self::_renameObjectWithCounter($objId, 1, $propertiesToCheck);

            $lang = isys_application::instance()->container->get('language');
            $message = $result['message'];
            $message .= "<br />" . $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__OBJECT_TITLE_CHECK_RENAMED') . ": $value -> $newValue.";

            // display error message
            isys_notify::error($message);
        }
    }

    /**
     * Validate if a list of property-values is unique.
     *
     * @param int|null $objId
     * @param int|null $objTypeId
     * @param array    $objectProperties
     *
     * @return array
     * @throws Exception
     */
    public static function validate($objId = null, $objTypeId = null, $objectProperties = [])
    {
        $lang = isys_application::instance()->container->get('language');
        $check_enabled = isys_settings::get(isys_module_smview_cable::OBJECT_TITLE_CHECK_ENABLED, null);
        if (! $check_enabled) {
            return [
                'success' => true,
                'message' => $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__OBJECT_TITLE_CHECK_DISABLED'),
            ];
        }

        // try to get CMDB status by ID or constant
        if (! empty($objectProperties['cmdb_status'])) {
            $statusDao = new \isys_cmdb_dao_status(isys_application::instance()->database);
            $statusRow = $statusDao->get_cmdb_status($objectProperties['cmdb_status'])->get_row();
            $objectProperties['cmdb_status'] = $statusRow ? $statusRow['isys_cmdb_status__const'] : 'C__CMDB_STATUS__IN_OPERATION';
        }

        // get object
        $dao = new isys_cmdb_dao_category_g_global(isys_application::instance()->database);
        if ($objId) {
            // merge with current object properties
            $object = self::_get_data($objId)->get_row();
            $objectProperties = [
                'title'       => $objectProperties['title'] ?? $object['isys_obj__title'],
                'purpose'     => $objectProperties['purpose'] ?? $object['isys_purpose__title'],
                'category'    => $objectProperties['category'] ?? $object['isys_catg_global_category__title'],
                'cmdb_status' => $objectProperties['cmdb_status'] ?? $object['isys_cmdb_status__const'],
                'status'      => $objectProperties['status'] ?? $object['isys_obj__status'],
                'location'    => $objectProperties['location'] ?? $object['isys_catg_location_list__parentid'],
            ];

            if (empty($objTypeId)) {
                $objTypeId = $object['isys_obj_type__id'];
            }
        }

        $check_properties_string = isys_settings::get(isys_module_smview_cable::OBJECT_TITLE_CHECK_PROPERTIES, null);
        $check_properties = explode(',', $check_properties_string);
        if (! in_array('title', $check_properties)) {
            // add title by default
            $check_properties[] = 'title';
        }

        // keep only object properties to be checked
        $relevantObjectProperties = [];
        foreach ($objectProperties as $property => $value) {
            if (in_array($property, $check_properties)) {
                $relevantObjectProperties[$property] = $value;
            }
        }

        if (! $objTypeId) {
            // object type could not be determined
            return [
                'success' => false,
                'message' => $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__ERROR_OBJECT_TYPE_NOT_FOUND'),
            ];
        }

        if (empty($objectProperties['title'])) {
            // object title is empty
            return [
                'success' => false,
                'message' => $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__ERROR_OBJECT_TITLE_IS_EMPTY'),
            ];
        }

        $excluded_object_types_string = isys_settings::get(isys_module_smview_cable::OBJECT_TITLE_CHECK_EXCLUDED_OBJECT_TYPES, null);
        $excluded_object_types = explode(',', $excluded_object_types_string);
        if (in_array($objTypeId, $excluded_object_types)) {
            // object type is excluded
            return [
                'success' => true,
                'message' => $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__OBJECT_TYPE_CHECK_DISABLED'),
            ];
        }

        $check_object_types_string = isys_settings::get(isys_module_smview_cable::OBJECT_TITLE_CHECK_OBJECT_TYPES, null);
        $check_object_types = explode(',', $check_object_types_string);
        if (! in_array($objTypeId, $check_object_types)) {
            // object type should not be checked -> check if title is unique for this object type
            $condition = " AND isys_obj_type__id = $objTypeId AND isys_obj__title = " . $dao->convert_sql_text($objectProperties['title']);

            // use only title property
            $check_properties = ['title'];
            $relevantObjectProperties = ['title' => $relevantObjectProperties['title']];
        } else {
            // get all objects of all selected object types containing given title
            $check_object_types_string = !empty($check_object_types_string) ? $check_object_types_string : '0';
            $condition = " AND isys_obj_type__id IN ($check_object_types_string) AND isys_obj__title = " . $dao->convert_sql_text($objectProperties['title']);
        }

        try {
            $res = self::_get_data(null, $condition);
        } catch (isys_exception_database $e) {
            return [
                'success' => false,
                'message' => $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__ERROR_PROPERTY_NOT_FOUND') . ": " . $objectProperties['title'],
            ];
        }

        if ($res->count() > 0) {
            $checkResult = self::_checkMultipleObjects($res, $check_properties, $relevantObjectProperties, $objId);
            if (! empty($checkResult['properties'])) {
                // get and translate all used properties
                $properties = [];
                $global_properties = $dao->get_properties();
                foreach ($global_properties as $id => $p) {
                    $properties[$id] = $lang->get($p[C__PROPERTY__INFO][C__PROPERTY__INFO__TITLE]);
                }
                $locationDao = new isys_cmdb_dao_category_g_location(isys_application::instance()->database);
                $location_properties = $locationDao->get_properties();
                if ($location_properties['parent']) {
                    $p = $location_properties['parent'];
                    $properties['location'] = $lang->get($p[C__PROPERTY__INFO][C__PROPERTY__INFO__TITLE]);
                }

                // list all duplicate properties
                $propMsg = [];
                foreach ($checkResult['properties'] as $k => $v) {
                    $k = $properties[$k] ?? $k;
                    $propMsg[] = "$k = '$v'";
                }

                // create message
                $message = $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__ERROR_MULTIPLE_OBJECTS') . ' ';
                $message .= implode(', ', $propMsg) . ": " . implode(', ', $checkResult['object_ids']);
                return [
                    'success' => false,
                    'message' => $message,
                ];
            }
        }

        return [
            'success' => true,
            'message' => $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__OBJECT_TITLE_CHECK_SUCCESS'),
        ];
    }

    /**
     * Check multiple objects with same base property-value.
     *
     * @param  isys_component_dao_result $res
     * @param  array                     $check_properties
     * @param  array                     $objectProperties
     * @param  int                       $objId
     *
     * @return array
     */
    private static function _checkMultipleObjects($res, $check_properties, $objectProperties, $objId = null)
    {
        // first group all objects with same property-value by check-property
        $objectListing = [];
        while ($row = $res->get_row()) {
            $currentObjId = $row['isys_obj__id'];
            foreach ($check_properties as $p) {
                $valueKey = null;
                switch ($p) {
                    case 'purpose':
                        $valueKey = $row['isys_purpose__title'];
                        break;
                    case 'cmdb_status':
                        $valueKey = $row['isys_cmdb_status__const'];
                        break;
                    case 'category':
                        $valueKey = $row['isys_catg_global_category__title'];
                        break;
                    case 'status':
                        $valueKey = $row['isys_obj__status'];
                        break;
                    case 'location':
                        $valueKey = $row['isys_catg_location_list__parentid'];
                        break;
                    case 'title':
                        $valueKey = $row['isys_obj__title'];
                        break;
                }
                $objectListing[$currentObjId][$p] = $valueKey;
            }
        }

        // then check if any object ID appear more than once
        $duplicateProperties = [];
        $duplicateObjIds = [];
        foreach ($objectListing as $id => $object) {
            $diffs = array_diff_assoc($objectProperties, $object);
            if (count($diffs) == 0) {
                // objects are equal
                if (empty($objId) || $id != $objId) {
                    // there is already an object with these properties in database
                    $duplicateProperties = $objectProperties;
                    $duplicateObjIds[] = $id;
                }
            }
        }

        return [
            'properties' => $duplicateProperties,
            'object_ids' => $duplicateObjIds,
        ];
    }

    /**
     * Fetches global and location category data from database.
     *
     * @param  int|null   $p_obj_id
     * @param  string $p_condition
     * @return isys_component_dao_result
     * @throws isys_exception_database
     */
    private static function _get_data($p_obj_id = null, $p_condition = '')
    {
        $dao = new isys_cmdb_dao_category_g_global(isys_application::instance()->database);
        $l_sql = "SELECT * FROM isys_obj " . "LEFT JOIN isys_catg_global_list ON isys_catg_global_list__isys_obj__id = isys_obj__id " .
            "INNER JOIN isys_cmdb_status ON isys_obj__isys_cmdb_status__id = isys_cmdb_status__id " .
            "INNER JOIN isys_obj_type ON isys_obj_type__id = isys_obj__isys_obj_type__id " .
            "LEFT JOIN isys_catg_location_list ON isys_catg_location_list__isys_obj__id = isys_obj__id " .
            "LEFT OUTER JOIN isys_catg_global_category ON isys_catg_global_list__isys_catg_global_category__id = isys_catg_global_category__id " .
            "LEFT OUTER JOIN isys_purpose ON isys_catg_global_list__isys_purpose__id = isys_purpose__id " . "WHERE TRUE " . $p_condition . " ";

        if ($p_obj_id !== null) {
            $l_sql .= $dao->get_object_condition($p_obj_id);
        }
        return $dao->retrieve($l_sql);
    }

    /**
     * Rename object with a prefixed counter.
     *
     * @param  int $objId
     * @param  int $counter
     * @param  array $propertiesToCheck
     * @return string
     * @throws isys_exception_database
     */
    private static function _renameObjectWithCounter($objId, $counter, $propertiesToCheck)
    {
        $dao = new isys_cmdb_dao_category_g_global(isys_application::instance()->database);
        $obj = $dao->get_object($objId)->get_row();
        $newTitle = $obj['isys_obj__title'] . "($counter)";
        $propertiesToCheck['title'] = $newTitle;
        $result = self::validate($objId, null, $propertiesToCheck);
        if (! $result['success']) {
            $newTitle = self::_renameObjectWithCounter($objId, $counter + 1, $propertiesToCheck);
        } else {
            $dao->save_title($objId, $newTitle);
        }
        return $newTitle;
    }
}
