<?php

namespace idoit\Module\Cable;

use idoit\Exception\Exception;
use isys_application;
use isys_cmdb_dao_smview_cable_orientdb;
use isys_exception_database;
use PhpOrient\Exceptions\PhpOrientException;
use PhpOrient\Exceptions\SocketException;
use PhpOrient\PhpOrient;

/**
 * OrientDB class.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class OrientDB
{
    /**
     * OrientDB client.
     *
     * @var PhpOrient
     */
    private $_client;

    /**
     * OrientDB constructor.
     *
     * @throws Exception
     */
    public function __construct()
    {
        $dao = new isys_cmdb_dao_smview_cable_orientdb(isys_application::instance()->database);
        try {
            $config = $dao->get_object('id', 1);
        } catch (isys_exception_database $e) {
            $config = [];
        }

        if (empty($config['address'])) {
            return;
        }

        $this->_client = new PhpOrient($config['address'], $config['port']);
        $this->_client->username = $config['username'];
        $this->_client->password = $config['password'];

        // enable persistent connections
        $this->_client->setSessionToken(true);

        $this->_client->connect();
        if (! $this->_client->dbExists($config['database'], PhpOrient::DATABASE_TYPE_GRAPH)) {
            $this->_client->dbCreate(
                $config['database'],
                PhpOrient::STORAGE_TYPE_PLOCAL,
                PhpOrient::DATABASE_TYPE_GRAPH
            );

            $sync = new OrientDBSync($this->_client);
            $sync->create_schema_if_needed();
        }

        try {
            $this->_client->dbOpen($config['database'], $config['username'], $config['password']);
        } catch (SocketException $e) {
            // re-throw exception
            throw new Exception($e->getMessage());
        }
    }

    /**
     * Get last update information.
     *
     * @return array
     */
    public function get_last_update_info()
    {
        if (! $this->_client) {
            return [];
        }

        try {
            $info = $this->_client->query( "select @rid, last_update_end, last_update_end_string, duration from update_info" );
        } catch (PhpOrientException $e) {
            return [];
        }
        if (! empty($info)) {
            foreach ($info as $i) {
                $data = $i->getOData();
                return [
                    'last_update' => $data['last_update_end_string'],
                    'duration'    => $data['duration'],
                ];
            }
        }
        return [];
    }

    /**
     * Get least number of hops.
     *
     * @param  int $from
     * @param  int $to
     * @param  boolean $with_sections
     * @return array
     * @throws isys_exception_database
     */
    public function get_least_number_of_hops($from, $to, $with_sections = true)
    {
        if ($with_sections) {
            $sql = 'SELECT 
        @rid,
        obj_id,
        out_connection.cable_id AS out_cable_id,
        out_connection.wire_id AS out_wire_id,
        out_connection.section_id AS out_section_id,
        out_connection.length AS out_length,
        out_connection.in AS out_in_obj,
        out_connection.out AS out_out_obj,
        in_connection.cable_id AS in_cable_id,
        in_connection.wire_id AS in_wire_id,
        in_connection.section_id AS in_section_id,
        in_connection.length AS in_length,
        in_connection.in AS in_in_obj,
        in_connection.out AS in_out_obj
    FROM (SELECT expand(path) FROM (
      SELECT shortestPath($from, $to, "both") AS path 
      LET 
        $from = (SELECT FROM object WHERE obj_id='.$from.'), 
        $to = (SELECT FROM object WHERE obj_id='.$to.') 
      UNWIND path
    ))';
        } else {
            $sql = 'SELECT 
        @rid,
        obj_id,
        out_wire.cable_id AS out_cable_id,
        out_wire.wire_id AS out_wire_id,
        out_wire.section_id AS out_section_id,
        out_wire.length AS out_length,
        out_wire.in AS out_in_obj,
        out_wire.out AS out_out_obj,
        in_wire.cable_id AS in_cable_id,
        in_wire.wire_id AS in_wire_id,
        in_wire.section_id AS in_section_id,
        in_wire.length AS in_length,
        in_wire.in AS in_in_obj,
        in_wire.out AS in_out_obj
    FROM (SELECT expand(path) FROM (
      SELECT shortestPath($from, $to, "both") AS path 
      LET 
        $from = (SELECT FROM wire_object WHERE obj_id='.$from.'), 
        $to = (SELECT FROM wire_object WHERE obj_id='.$to.') 
      UNWIND path
    ))';
        }

        $raw_route = $this->_execute_query($sql);
        return $this->_add_cable_info($raw_route);
    }

    /**
     * Get shortest path.
     *
     * @param  int $from
     * @param  int $to
     * @param  boolean $with_sections
     * @return array
     * @throws isys_exception_database
     */
    public function get_shortest_path($from, $to, $with_sections = true)
    {
//        $sql = 'SELECT cable_id, cable_title, wire_id, wire_title, distance, in.name, out.name FROM (SELECT expand(in_wire) FROM (SELECT expand(path) FROM (
//  SELECT dijkstra($from, $to, "length", "both") AS path
//  LET
//    $from = (SELECT FROM bvg WHERE id=30705),
//    $to = (SELECT FROM bvg WHERE id=31453)
//  UNWIND path
//)))';

        if ($with_sections) {
            $sql = 'SELECT
    @rid,
    obj_id,
    out_connection.cable_id AS out_cable_id,
    out_connection.wire_id AS out_wire_id,
    out_connection.section_id AS out_section_id,
    out_connection.length AS out_length,
    out_connection.in AS out_in_obj,
    out_connection.out AS out_out_obj,
    in_connection.cable_id AS in_cable_id,
    in_connection.wire_id AS in_wire_id,
    in_connection.section_id AS in_section_id,
    in_connection.length AS in_length,
    in_connection.in AS in_in_obj,
    in_connection.out AS in_out_obj
FROM (SELECT expand(path) FROM (
  SELECT dijkstra($from, $to, "length", "both") AS path 
  LET 
    $from = (SELECT FROM object WHERE obj_id=' . $from . '), 
    $to = (SELECT FROM object WHERE obj_id=' . $to . ') 
  UNWIND path
))';
        } else {
            $sql = 'SELECT
    @rid,
    obj_id,
    out_wire.cable_id AS out_cable_id,
    out_wire.wire_id AS out_wire_id,
    out_wire.section_id AS out_section_id,
    out_wire.length AS out_length,
    out_wire.in AS out_in_obj,
    out_wire.out AS out_out_obj,
    in_wire.cable_id AS in_cable_id,
    in_wire.wire_id AS in_wire_id,
    in_wire.section_id AS in_section_id,
    in_wire.length AS in_length,
    in_wire.in AS in_in_obj,
    in_wire.out AS in_out_obj
FROM (SELECT expand(path) FROM (
  SELECT dijkstra($from, $to, "length", "both") AS path 
  LET 
    $from = (SELECT FROM wire_object WHERE obj_id=' . $from . '), 
    $to = (SELECT FROM wire_object WHERE obj_id=' . $to . ') 
  UNWIND path
))';
        }

        $raw_route = $this->_execute_query($sql);
        return $this->_add_cable_info($raw_route);
    }

    /**
     * Execute query and prepare result.
     *
     * @param  string $sql
     * @return array
     */
    private function _execute_query($sql)
    {
        try {
            $elements = $this->_client->query($sql);
        } catch (PhpOrientException $e) {
            // could not find a path
            return [];
        }

        $path = [];
        $current = null;
        $prev = null;
        foreach ($elements as $l) {
            $d = $l->getOData();
            $rid = $d['rid']->__toString();

            $current = [
                'rid'       => $rid,
                'obj_id'    => $d['obj_id'],
//                'obj_title' => $d['obj_title'],
            ];

            if (! $prev) {
                $prev = $current;
                continue;
            }

            $wires = $this->_get_wires($d, 'in', $rid, $prev['rid']);
            $wires = $this->_get_wires($d, 'out', $rid, $prev['rid'], $wires);
            $path[] = array(
                'from'  => $prev['obj_id'],
                'to'    => $current['obj_id'],
                'wires' => $wires,
            );
            $prev = $current;
        }
        return $path;
    }

    /**
     * Get wires by direction.
     *
     * @param  array $d
     * @param  string $direction
     * @param  string $rid
     * @param  string $prev_rid
     * @param  array $wires
     * @return array
     */
    private function _get_wires($d, $direction, $rid, $prev_rid, $wires = [])
    {
        foreach ($d[$direction . '_wire_id'] as $index => $wire_id) {
            $obj_in = $d[$direction . '_in_obj'][$index]->__toString();
            $obj_out = $d[$direction . '_out_obj'][$index]->__toString();

            if (($obj_in == $prev_rid && $obj_out == $rid) ||
                ($obj_out == $prev_rid && $obj_in == $rid)) {
                // this wire connects both objects
                $wires[] = [
                    'cable_id' => $d[$direction . '_cable_id'][$index],
                    'wire_id'  => $wire_id,
                    'section_id' => $d[$direction . '_section_id'][$index],
                    'length' => $d[$direction . '_length'][$index],
                ];
            }
        }
        return $wires;
    }

    /**
     * Add cable information.
     *
     * @param  array $raw_route
     * @return array
     * @throws isys_exception_database
     */
    private function _add_cable_info($raw_route)
    {
        $cable_info = [];
        $wire_info = [];
        $section_info = [];
        $route = [];

        $cable_dao = new \isys_cmdb_dao_category_g_global(isys_application::instance()->database);
        $wire_dao = new \isys_cmdb_dao_smview_cable_wire(isys_application::instance()->database);
        foreach ($raw_route as $r) {
            $wires = [];
            foreach ($r['wires'] as $w) {
                if (! $cable_info[$w['cable_id']]) {
                    $cable_info[$w['cable_id']] = $cable_dao->get_object_by_id($w['cable_id'])->get_row();
                }
                $cable = $cable_info[$w['cable_id']];

                if (! $wire_info[$w['wire_id']]) {
                    $wire_info[$w['wire_id']] = $wire_dao->get_object('id', $w['wire_id']);
                }
                $wire = $wire_info[$w['wire_id']];

                if (! $section_info[$w['section_id']]) {
                    $section_info[$w['section_id']] = $cable_dao->get_object_by_id($w['section_id'])->get_row();
                }
                $section = $section_info[$w['section_id']];

                $wires[$w['wire_id']] = array(
                    'cable_id'    => $w['cable_id'],
                    'cable_title' => $cable['isys_obj__title'],
                    'length'      => $w['length'],
                    'wire_id'     => $w['wire_id'],
                    'wire_title'  => $wire['description'],
                    'wire_scope'  => $wire['scope'],
                    'section_id'  => $w['section_id'],
                    'section_title' => $section['isys_obj__title'],
                );
            }
            $route[] = [
                'from'  => $r['from'],
                'to'    => $r['to'],
                'wires' => $wires,
            ];
        }
        return $route;
    }

    /**
     * Synchronize connections between DB and OrientDB.
     */
    public function sync_connections()
    {
        $sync = new OrientDBSync($this->_client);
        $sync->sync();
    }
}
