<?php

namespace idoit\Module\Cable;

use isys_application;
use isys_cmdb_dao_category_g_global;
use isys_cmdb_dao_smview_cable_free_connection_view;
use isys_cmdb_dao_smview_cable_free_section_view;
use isys_exception_database;
use PhpOrient\PhpOrient;
use PhpOrient\Protocols\Binary\Data\ID;

/**
 * OrientDB Synchronizer class.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class OrientDBSync
{
    const BATCH_SIZE = 300;

    /**
     * OrientDB client.
     *
     * @var PhpOrient
     */
    private $_client;

    /**
     * Current wire (connection) data.
     *
     * @var array
     */
    private $_wire_data;

    /**
     * Current section (connection) data.
     *
     * @var array
     */
    private $_section_data;

    /**
     * Global DAO.
     *
     * @var isys_cmdb_dao_category_g_global
     */
    private $_global_dao;

    private $_log = [];

    /**
     * OrientDB constructor.
     *
     * @param PhpOrient $client
     */
    public function __construct($client)
    {
        $this->_client = $client;
        $this->_global_dao = new isys_cmdb_dao_category_g_global(isys_application::instance()->database);

        $dao = new isys_cmdb_dao_smview_cable_free_connection_view(isys_application::instance()->database);
        $this->_wire_data = $dao->get_data();

        $dao = new isys_cmdb_dao_smview_cable_free_section_view(isys_application::instance()->database);
        $this->_section_data = $dao->get_data();
    }

    /**
     * Synchronize connections between DB and OrientDB.
     * @throws isys_exception_database
     */
    public function sync()
    {
        $this->_log = [];
        $this->_log[] = "start: " . strftime('%Y-%m-%d %H:%M:%S');

        // only start sync if no other one is running
        if ($this->_is_syncing()) {
            $this->_log[] = "sync already running - abort";
            return;
        }

        $this->create_schema_if_needed();
        $this->_update_info_class();

        // switch between bulk- or sync-mode
        if ($this->_use_bulk_mode()) {
            // todo: cluster wechseln

            // sync only wire data
            $this->_delete_vertex(false);
            $this->_bulk_sync_objects($this->_wire_data, 'obj_id_in', 'obj_id_out', false);
            $this->_bulk_sync_connections(false);

            // sync wire and section data
            $this->_delete_vertex(true);
            $used_object_ids = $this->_bulk_sync_objects($this->_wire_data, 'obj_id_in', 'obj_id_out', true);
            $this->_bulk_sync_objects($this->_section_data, 'start_obj_id', 'end_obj_id', true, $used_object_ids);
            $this->_bulk_sync_connections();
            $this->_bulk_sync_sections();
        } else {
            ini_set('memory_limit', '2048M');

            $this->_sync_connections_only();
            $this->_sync_connections_and_sections();
        }
        $this->_update_info_class(true);
        $this->_log[] = "finish: " . strftime('%Y-%m-%d %H:%M:%S');
    }

    /**
     * Check if OrientDB is already syncing.
     *
     * @return boolean
     */
    private function _is_syncing()
    {
        $info = $this->_client->query( "select @rid, is_locked from update_info" );
        if (! empty($info)) {
            $data1 = $info[0]->getOData();
            $is_locked = $data1['is_locked'];
            if ($is_locked) {
                return true;
            }
        }
        return false;
    }

    /**
     * Check configuration if it is better to use bulk mode.
     *
     * @return boolean
     */
    private function _use_bulk_mode()
    {
        $useBulkMode = false;
        $count = $this->_get_count_of_orient_items('object');
        if ($count > 100) {
            // max. 100 objects
            $useBulkMode = true;
        }

        $count = $this->_get_count_of_orient_items('connection');
        if ($count > 1000) {
            // max. 1000 connections
            $useBulkMode = true;
        }
        return $useBulkMode;
    }

    /**
     * Get count of an OrientDB item type.
     *
     * @param  string $type
     * @return int
     */
    private function _get_count_of_orient_items($type)
    {
        $countRes = $this->_client->query( "select count(*) as count from $type" );
        if (! empty($countRes)) {
            $data1 = $countRes[0]->getOData();
            return (int) $data1['count'];
        }
        return 0;
    }

    /**
     * Delete a vertex.
     *
     * @param boolean $with_sections
     */
    private function _delete_vertex($with_sections = true)
    {
        $vertexName = $with_sections ? 'object' : 'wire_object';
        $this->_client->command( "delete vertex $vertexName" );
        $this->_log[] = "deleted vertex $vertexName: " . strftime('%Y-%m-%d %H:%M:%S');
    }

    /**
     * Synchronize connections between DB and OrientDB.
     * @throws isys_exception_database
     */
    private function _bulk_sync_sections()
    {
        $this->_log[] = "start sync sections: " . strftime('%Y-%m-%d %H:%M:%S');

        // get current objects
        $db_object_ids = $this->_get_orientdb_objects(true);
        $this->_log[] = "read objects from DB: " . strftime('%Y-%m-%d %H:%M:%S');

        // sync sections
        $count = 0;
        $cmd = 'begin;';
        foreach ($this->_section_data as $d) {
            $in_id = (int) $d['start_obj_id'];
            $out_id = (int) $d['end_obj_id'];
            $length = (int) $d['length'];
            $length = $length == 0 ? 999999 : $length;
            $section_id = (int) $d['id'];

            // add edges
            $in = $db_object_ids[$in_id];
            $out = $db_object_ids[$out_id];
            $cmd .= "create edge connection from $in to $out set length = $length, section_id = $section_id;";
            $count++;

            if ($count > 0 && $count % self::BATCH_SIZE == 0) {
                $cmd .= 'commit retry 100;';
                $this->_client->sqlBatch( $cmd );
                $this->_log[] = "add section batch: " . strftime('%Y-%m-%d %H:%M:%S');
                $cmd = 'begin;';
            }
        }

        $cmd .= 'commit retry 100;';
        $this->_client->sqlBatch( $cmd );

        $this->_log[] = "finish sync sections: " . strftime('%Y-%m-%d %H:%M:%S');
    }

    /**
     * Synchronize connections.
     *
     * @param  boolean $with_sections
     * @throws isys_exception_database
     */
    private function _bulk_sync_connections($with_sections = true)
    {
        $this->_log[] = "start sync edges: " . strftime('%Y-%m-%d %H:%M:%S');

        // get current objects
        $db_object_ids = $this->_get_orientdb_objects($with_sections);
        $this->_log[] = "read objects from DB: " . strftime('%Y-%m-%d %H:%M:%S');

        $count = 0;
        $cmd = 'begin;';
        $connectionVertexName = $with_sections ? 'connection' : 'wire';
        foreach ($this->_wire_data as $d) {
            $in_id = (int) $d['obj_id_in'];
            $out_id = (int) $d['obj_id_out'];
            $length = (int) $d['length'];
            $length = $length == 0 ? 999999 : $length;
            $cable_id = (int) $d['cable_id'];
            $wire_id = (int) $d['wire_id'];

            // add connection
            $in = $db_object_ids[$in_id];
            $out = $db_object_ids[$out_id];
            $cmd .= "create edge $connectionVertexName from $in to $out set wire_id = $wire_id, cable_id = $cable_id, length = $length;";
            $count++;

            if ($count > 0 && $count % self::BATCH_SIZE == 0) {
                $cmd .= 'commit retry 100;';
                $this->_client->sqlBatch( $cmd );
                $this->_log[] = "add wire batch: " . strftime('%Y-%m-%d %H:%M:%S');
                $cmd = 'begin;';
            }
        }

        $cmd .= 'commit retry 100;';
        $this->_client->sqlBatch( $cmd );

        $this->_log[] = "finish sync edges: " . strftime('%Y-%m-%d %H:%M:%S');
    }

    /**
     * Synchronize objects.
     *
     * @param  $data
     * @param  $obj_in_key
     * @param  $obj_out_key
     * @param  boolean $with_sections
     * @param  array $used_object_ids
     * @return array
     * @throws isys_exception_database
     */
    private function _bulk_sync_objects($data, $obj_in_key, $obj_out_key, $with_sections = true, $used_object_ids = [])
    {
        $vertexName = $with_sections ? 'object' : 'wire_object';
        $count = 0;
        $cmd = 'begin;';
        foreach ($data as $d) {
            $in_id = (int) $d[$obj_in_key];
            $out_id = (int) $d[$obj_out_key];

            // add objects
            if (! isset($used_object_ids[$in_id])) {
                $in_obj = $this->_global_dao->get_object_by_id($in_id)->get_row();
                $title = $this->_escape_string($in_obj['isys_obj__title']);
                $cmd .= "insert into $vertexName set obj_id = '$in_id', obj_title = '$title';";

                $used_object_ids[$in_id] = $in_id;
                $count++;
            }

            if (! isset($used_object_ids[$out_id])) {
                $out_obj = $this->_global_dao->get_object_by_id($out_id)->get_row();
                $title = $this->_escape_string($out_obj['isys_obj__title']);
                $cmd .= "insert into $vertexName set obj_id = '$out_id', obj_title = '$title';";

                $used_object_ids[$out_id] = $out_id;
                $count++;
            }

            if ($count > 0 && $count % self::BATCH_SIZE == 0) {
                $cmd .= 'commit retry 100;';
                $this->_client->sqlBatch( $cmd );
                $this->_log[] = "add object batch: " . strftime('%Y-%m-%d %H:%M:%S');

                $count = 0;
                $cmd = 'begin;';
            }
        }

        $cmd .= 'commit retry 100;';
        $this->_client->sqlBatch( $cmd );

        $this->_log[] = "add last object batch: " . strftime('%Y-%m-%d %H:%M:%S');
        return $used_object_ids;
    }

    /**
     * Synchronize connections between DB and OrientDB.
     * @throws isys_exception_database
     */
    private function _sync_connections_only()
    {
        // get current data
        $db_object_ids = $this->_get_orientdb_objects(false);

        // get current wires
        $res = $this->_get_orientdb_edges(false);
        $db_wire_ids = $res['db_wire_ids'];

        // sync edges
        $res = $this->_sync_edges($db_object_ids, $db_wire_ids, false);
        $used_object_ids = $res['used_object_ids'];
        $used_wire_ids = $res['used_wire_ids'];
        $db_object_ids = $res['db_object_ids'];
        $db_wire_ids = $res['db_wire_ids'];

        $this->_delete_unused_elements($db_object_ids, $used_object_ids, $db_wire_ids, $used_wire_ids);
    }

    /**
     * Synchronize connections between DB and OrientDB.
     * @throws isys_exception_database
     */
    private function _sync_connections_and_sections()
    {
        // get current objects
        $db_object_ids = $this->_get_orientdb_objects();

        // get current connections
        $res = $this->_get_orientdb_edges();
        $db_wire_ids = $res['db_wire_ids'];
        $db_section_ids = $res['db_section_ids'];

        // sync edges
        $res = $this->_sync_edges($db_object_ids, $db_wire_ids);
        $used_object_ids = $res['used_object_ids'];
        $used_wire_ids = $res['used_wire_ids'];
        $db_object_ids = $res['db_object_ids'];
        $db_wire_ids = $res['db_wire_ids'];

        // sync sections
        $used_section_ids = [];
        foreach ($this->_section_data as $d) {
            $in_id = (int) $d['start_obj_id'];
            $out_id = (int) $d['end_obj_id'];
            $length = (int) $d['length'];
            $length = $length == 0 ? 999999 : $length;
            $section_id = (int) $d['id'];

            // add objects
            $db_object_ids = $this->_add_object_if_new($in_id, $db_object_ids);
            $used_object_ids[$in_id] = $in_id;
            $db_object_ids = $this->_add_object_if_new($out_id, $db_object_ids);
            $used_object_ids[$out_id] = $out_id;

            // add edges
            $in = $db_object_ids[$in_id];
            $out = $db_object_ids[$out_id];
            if (! isset($db_section_ids[$section_id])) {
                $ret3 = $this->_client->command( "create edge connection from $in to $out set length = $length, section_id = $section_id" );
                $w = $ret3->getOData();
                $w['rid'] = $ret3->getRid();
                $db_section_ids[$section_id] = $w;
            } else {
                $wire = $db_section_ids[$section_id];
                $rid = $wire['rid']->__toString();
                $obj_in = $wire['in']->__toString();
                $obj_out = $wire['out']->__toString();
                if (($in != $obj_in && $out != $obj_in) || ($in != $obj_out && $out != $obj_out) || $length != $wire['length'] || $section_id != $wire['section_id']) {
                    $this->_client->command( "update connection set in = '$in', out = '$out', section_id = $section_id, length = $length where @rid=$rid" );
                }
            }
            $used_section_ids[$section_id] = $section_id;
        }

        $this->_delete_unused_elements($db_object_ids, $used_object_ids, $db_wire_ids, $used_wire_ids, $db_section_ids, $used_section_ids);
    }

    /**
     * Synchronize edges.
     *
     * @param  array $db_object_ids
     * @param  array $db_wire_ids
     * @param  boolean $with_sections
     * @return array
     * @throws isys_exception_database
     */
    private function _sync_edges($db_object_ids, $db_wire_ids, $with_sections = true)
    {
        $used_object_ids = [];
        $used_wire_ids = [];
        foreach ($this->_wire_data as $d) {
            $in_id = (int) $d['obj_id_in'];
            $out_id = (int) $d['obj_id_out'];
            $length = (int) $d['length'];
            $length = $length == 0 ? 999999 : $length;
            $cable_id = (int) $d['cable_id'];
            $wire_id = (int) $d['wire_id'];

            // add objects
            $db_object_ids = $this->_add_object_if_new($in_id, $db_object_ids, $with_sections);
            $used_object_ids[$in_id] = $in_id;
            $db_object_ids = $this->_add_object_if_new($out_id, $db_object_ids, $with_sections);
            $used_object_ids[$out_id] = $out_id;

            // add connection
            $in = $db_object_ids[$in_id];
            $out = $db_object_ids[$out_id];
            $vertexName = $with_sections ? 'connection' : 'wire';
            if (! isset($db_wire_ids[$wire_id])) {
                $wire = $this->_client->command( "create edge $vertexName from $in to $out set wire_id = $wire_id, cable_id = $cable_id, length = $length" );
                $w = $wire->getOData();
                $w['rid'] = $wire->getRid();
                $db_wire_ids[$wire_id] = $w;
            } else {
                $wire = $db_wire_ids[$wire_id];
                $obj_in = $wire['in']->__toString();
                $obj_out = $wire['out']->__toString();
                if (($in != $obj_in && $out != $obj_in) || ($in != $obj_out && $out != $obj_out) || $length != $wire['length'] || $cable_id != $wire['cable_id']) {
                    $rid = $wire['rid']->__toString();
                    $this->_client->command( "update $vertexName set in = '$in', out = '$out', wire_id = $wire_id, cable_id = $cable_id, length = $length where @rid=$rid" );
                }
            }
            $used_wire_ids[$wire_id] = $wire_id;
        }
        return [
            'db_object_ids'   => $db_object_ids,
            'db_wire_ids'     => $db_wire_ids,
            'used_object_ids' => $used_object_ids,
            'used_wire_ids'   => $used_wire_ids,
        ];
    }

    /**
     * Add database object if it is new.
     *
     * @param  int $obj_id
     * @param  array $db_object_ids
     * @param  boolean $with_sections
     * @return array
     * @throws isys_exception_database
     */
    private function _add_object_if_new($obj_id, $db_object_ids, $with_sections = true)
    {
        if (! isset($db_object_ids[$obj_id])) {

            $in_obj = $this->_global_dao->get_object_by_id($obj_id)->get_row();
            $title = $this->_escape_string($in_obj['isys_obj__title']);
            $object = $with_sections ? 'object' : 'wire_object';

            $ret1 = $this->_client->command( "insert into $object set obj_id = '$obj_id', obj_title = '$title'" );
            $rid = $ret1->getRid()->__toString();
            $db_object_ids[$obj_id] = $rid;
        }
        return $db_object_ids;
    }

    /**
     * Get OrientDB objects.
     *
     * @param  boolean $with_sections
     * @return array
     */
    private function _get_orientdb_objects($with_sections = true)
    {
        $db_object_ids = [];
        $limit = 5000;
        $offset = 0;
        $object = $with_sections ? 'object' : 'wire_object';
        while (true) {
            $db_objects = $this->_client->query( "select @rid, obj_id from $object limit $limit offset $offset" );
            $offset += $limit;
            if (empty($db_objects)) {
                break;
            }

            foreach ($db_objects as $l) {
                $data1 = $l->getOData();
                $id = (int) $data1['obj_id'];
                $rid = $data1['rid']->__toString();
                $db_object_ids[$id] = $rid;
            }
        }
        return $db_object_ids;
    }

    /**
     * Get OrientDB edges.
     *
     * @param  boolean $with_sections
     * @return array
     */
    private function _get_orientdb_edges($with_sections = true)
    {
        $db_wire_ids = [];
        $db_section_ids = [];
        $limit = 25000;
        $offset = 0;
        while (true) {
            $sql = "select @rid, wire_id, in, out, length, cable_id from wire limit $limit offset $offset";
            if ($with_sections) {
                $sql = "select @rid, wire_id, in, out, length, cable_id, section_id from connection limit $limit offset $offset";
            }
            $db_wires = $this->_client->query( $sql );
            $offset += $limit;
            if (empty($db_wires)) {
                break;
            }

            foreach ($db_wires as $w) {
                $data1 = $w->getOData();
                $wire_id = (int) $data1['wire_id'];
                $section_id = (int) $data1['section_id'];
                if (empty($section_id)) {
                    $db_wire_ids[$wire_id] = $data1;
                } else {
                    $db_section_ids[$section_id] = $data1;
                }
            }
        }
        return [
            'db_wire_ids'    => $db_wire_ids,
            'db_section_ids' => $db_section_ids,
        ];
    }

    /**
     * Escape some special characters.
     *
     * @param  string $string
     * @return string
     */
    private function _escape_string($string)
    {
        $string = str_replace("'", "", $string);
        $string = str_replace("\r", "", $string);
        $string = str_replace("\n", "", $string);
        return str_replace("\\", "", $string);
    }

    /**
     * Delete unused elements.
     *
     * @param array $db_object_ids
     * @param array $used_object_ids
     * @param array $db_wire_ids
     * @param array $used_wire_ids
     * @param array $db_section_ids
     * @param array $used_section_ids
     */
    private function _delete_unused_elements($db_object_ids, $used_object_ids, $db_wire_ids, $used_wire_ids, $db_section_ids = [], $used_section_ids = [])
    {
        foreach ($db_object_ids as $obj_id => $rid) {
            if (! isset ($used_object_ids[$obj_id])) {
                $rid = str_replace('#', '', $rid);
                $p = explode(':', $rid);
                $this->_client->recordDelete(new ID( $p[0], $p[1] ));
            }
        }

        foreach ($db_wire_ids as $wire_id => $wire) {
            if (! isset ($used_wire_ids[$wire_id])) {
                $this->_client->recordDelete($wire['rid']);
            }
        }

        foreach ($db_section_ids as $section_id => $section) {
            if (! isset ($used_section_ids[$section_id])) {
                $this->_client->recordDelete($section['rid']);
            }
        }
    }

    /**
     * Create OrientDB schema if needed.
     */
    public function create_schema_if_needed()
    {
        if (empty($this->_client)) {
            return;
        }

        $res = $this->_client->command("select expand(properties) from (select expand(classes) from metadata:schema) where name = 'wire_object'");
        if (empty($res)) {
            $this->_client->command("CREATE CLASS wire_object EXTENDS V CLUSTERS 1");
            $this->_client->command("CREATE PROPERTY wire_object.obj_id INTEGER");
            $this->_client->command("CREATE PROPERTY wire_object.obj_title STRING");
        }

        $res = $this->_client->command("select expand(properties) from (select expand(classes) from metadata:schema) where name = 'wire'");
        if (empty($res)) {
            $this->_client->command("CREATE CLASS wire EXTENDS E CLUSTERS 1");
            $this->_client->command("CREATE PROPERTY wire.cable_id INTEGER");
            $this->_client->command("CREATE PROPERTY wire.wire_id INTEGER");
            $this->_client->command("CREATE PROPERTY wire.length INTEGER");
            $this->_client->command("CREATE PROPERTY wire.in LINK wire_object");
            $this->_client->command("CREATE PROPERTY wire.out LINK wire_object");
        }

        $res = $this->_client->command("select expand(properties) from (select expand(classes) from metadata:schema) where name = 'object'");
        if (empty($res)) {
            $this->_client->command("CREATE CLASS object EXTENDS V CLUSTERS 1");
            $this->_client->command("CREATE PROPERTY object.obj_id INTEGER");
            $this->_client->command("CREATE PROPERTY object.obj_title STRING");
        }

        $res = $this->_client->command("select expand(properties) from (select expand(classes) from metadata:schema) where name = 'connection'");
        if (empty($res)) {
            $this->_client->command("CREATE CLASS connection EXTENDS E CLUSTERS 1");
            $this->_client->command("CREATE PROPERTY connection.cable_id INTEGER");
            $this->_client->command("CREATE PROPERTY connection.wire_id INTEGER");
            $this->_client->command("CREATE PROPERTY connection.section_id INTEGER");
            $this->_client->command("CREATE PROPERTY connection.length INTEGER");
            $this->_client->command("CREATE PROPERTY connection.in LINK object");
            $this->_client->command("CREATE PROPERTY connection.out LINK object");
        }

        $res = $this->_client->command("select expand(properties) from (select expand(classes) from metadata:schema) where name = 'update_info'");
        if (empty($res)) {
            $this->_client->command("CREATE CLASS update_info CLUSTERS 1");
            $this->_client->command("CREATE PROPERTY update_info.last_update_start INTEGER");
            $this->_client->command("CREATE PROPERTY update_info.last_update_start_string STRING");
            $this->_client->command("CREATE PROPERTY update_info.last_update_end INTEGER");
            $this->_client->command("CREATE PROPERTY update_info.last_update_end_string STRING");
            $this->_client->command("CREATE PROPERTY update_info.duration INTEGER");
            $this->_client->command("CREATE PROPERTY update_info.is_locked BOOLEAN");
        }
    }

    /**
     * Update info class.
     *
     * @param boolean $update_has_finished
     */
    private function _update_info_class($update_has_finished = false)
    {
        $info = $this->_client->query( "select @rid, last_update_start from update_info" );
        $now = time();
        $now_string = date('Y-m-d H:i:s', $now);
        if (empty($info)) {
            $this->_client->command( "insert into update_info set last_update_start = '$now', last_update_start_string = '$now_string', is_locked = true" );
        } else {
            $data1 = $info[0]->getOData();
            $rid = $data1['rid']->__toString();
            $sql = "update update_info set last_update_start = '$now', last_update_start_string = '$now_string' ";
            if ($update_has_finished) {
                $start = isset($data1['last_update_start']) ? $data1['last_update_start'] : 0;
                $duration = $now - $start;
                $sql = "update update_info set last_update_end = '$now', last_update_end_string = '$now_string', duration = '$duration', is_locked = false ";
            }
            $this->_client->command( $sql . " where @rid=$rid" );
        }
    }
}
