<?php

namespace idoit\Module\Cable;

use isys_cmdb_dao_category_g_global;
use isys_cmdb_dao_category_g_smview_cable_pin;
use isys_cmdb_dao_category_g_smview_cable_route_element;
use isys_cmdb_dao_category_g_smview_cable_section_element;
use isys_cmdb_dao_smview_cable_pin;
use isys_cmdb_dao_smview_cable_route;
use isys_cmdb_dao_smview_cable_route_element;
use isys_cmdb_dao_smview_cable_section;
use isys_component_database;
use isys_exception_dao;
use isys_exception_database;
use isys_helper_link;

/**
 * Section Route Validator class.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class SectionRouteValidator
{

    /**
     * Variable which holds the database component.
     *
     * @var isys_component_database
     */
    private $_database;

    /**
     * Validation errors.
     *
     * @var array
     */
    private $_errors = [];

    /**
     * Constructor.
     *
     * @param isys_component_database $database
     */
    public function __construct($database)
    {
        $this->_database = $database;
    }

    /**
     * Validate sections.
     *
     * @param  int $limit
     * @param  boolean $fixElements
     * @return array
     * @throws isys_exception_dao
     * @throws isys_exception_database
     */
    public function validateSections($limit = 250, $fixElements = false)
    {
        $section_dao = new isys_cmdb_dao_smview_cable_section($this->_database);
        $masters = $section_dao->get_objects();
        $this->_errors = [];
        foreach ($masters as $master) {
            $this->_validateElement('section', $master['id'], $master, $fixElements);
            if (count($this->_errors) >= $limit) {
                break;
            }
        }
        return $this->_errors;
    }

    /**
     * Validate routes.
     *
     * @param  int $limit
     * @param  boolean $fixElements
     * @return array
     * @throws isys_exception_dao
     * @throws isys_exception_database
     */
    public function validateRoutes($limit = 250, $fixElements = false)
    {
        $route_dao = new isys_cmdb_dao_smview_cable_route($this->_database);
        $masters = $route_dao->get_objects();
        $this->_errors = [];
        foreach ($masters as $master) {
            $this->_validateElement('route', $master['id'], $master, $fixElements);
            if (count($this->_errors) >= $limit) {
                break;
            }
        }
        return $this->_errors;
    }

    /**
     * Validate and optionally fix element.
     *
     * @param  string $type
     * @param  int $masterId
     * @param  array $master
     * @param  boolean $fixElement
     * @throws isys_exception_database
     * @throws isys_exception_dao
     */
    private function _validateElement($type, $masterId, $master = [], $fixElement = false)
    {
        $route_element_dao = new isys_cmdb_dao_category_g_smview_cable_route_element($this->_database);
        $route_elements_dao = new isys_cmdb_dao_smview_cable_route_element($this->_database);
        $section_element_dao = new isys_cmdb_dao_category_g_smview_cable_section_element($this->_database);
        $pin_dao = new isys_cmdb_dao_smview_cable_pin($this->_database);

        $used_in_route = false;
        if ($type == 'section') {
            $route_elements = $route_elements_dao->filter_objects(['section_id' => $masterId]);
            $used_in_route = count($route_elements) > 0;
        }

        $elements = [];
        if ($type == 'section') {
            $elements = $section_element_dao->get_elements_by_section_id($masterId, false, false, false);
        } elseif ($type == 'route') {
            $elements = $route_element_dao->get_elements_by_route_id($masterId, false, false);
        }

        $last = null;
        $counter = 1;
        foreach ($elements as $element) {
            $pin = $pin_dao->get_object('id', $element['pin_out_id']);
            $sib_pin = $pin_dao->get_object('id', $pin['internal_pin_id']);
            if ($counter == 1) {
                // first element
                if (! empty($sib_pin['patch_pin_id']) && ! $used_in_route) {
                    // error, first pin is patched (only on routes)
                    if ($fixElement) {
                        $message = $this->_removePatch($sib_pin['id']);
                    } else {
                        $message = 'first PIN should not be patched: ' . $sib_pin['description'];
                    }
                    $this->_add_error($type, $master, $element, [], $sib_pin, [], $message);
                }
            } else {
                $last_pin = $pin_dao->get_object('id', $last['pin_in_id']);
                $sib_last_pin = $pin_dao->get_object('id', $last_pin['internal_pin_id']);
                if ($sib_last_pin['patch_pin_id'] != $sib_pin['id'] || $sib_pin['patch_pin_id'] != $sib_last_pin['id']) {
                    if ($fixElement) {
                        $message = $this->_fixPatches($sib_pin, $sib_last_pin);
                    } elseif (empty($sib_pin) || empty($sib_last_pin)) {
                        $message = 'at least one PIN could not be found - this is not fixable';
                    } else {
                        $message = 'PIN ' . $sib_last_pin['description'] . ' is not patched to PIN ' . $sib_pin['description'];
                    }
                    $this->_add_error($type, $master, $element, $last, $sib_pin, $sib_last_pin, $message);
                }
            }

            if ($counter == count($elements)) {
                // last element
                $pin = $pin_dao->get_object('id', $element['pin_in_id']);
                $sib_pin = $pin_dao->get_object('id', $pin['internal_pin_id']);
                if (! empty($sib_pin['patch_pin_id']) && ! $used_in_route) {
                    // error, last pin is patched (only on routes)
                    if ($fixElement) {
                        $message = $this->_removePatch($sib_pin['id']);
                    } else {
                        $message = 'last PIN should not be patched: ' . $sib_pin['description'];
                    }
                    $this->_add_error($type, $master, [], $element, [], $sib_pin, $message);
                }
            }

            $last = $element;
            $counter++;
        }
    }

    /**
     * Add error.
     *
     * @param  string $type
     * @param  array $master
     * @param  array $element
     * @param  array $last
     * @param  array $sib_pin
     * @param  array $sib_last_pin
     * @param  string $message
     * @throws isys_exception_database
     */
    private function _add_error($type, $master, $element, $last, $sib_pin, $sib_last_pin, $message)
    {
        $index = $last['index'] ? $last['index'] : $element['index'];
        $objectId = ! empty($element) ? $element['obj_out_id'] : $last['obj_in_id'];
        $objectTitle = ! empty($element) ? $element['obj_out_title'] : $last['obj_in_title'];
        $pinOutScope = ! empty($element) ? $element['pin_out_scope'] : $last['pin_in_scope'];
        $cableId = ! empty($element) ? $element['cable_id'] : $last['cable_id'];

        $dao = new isys_cmdb_dao_category_g_global($this->_database);
        $obj = $dao->get_object($master['obj_id'])->get_row();
        $catGSection = defined_or_default('C__CATG__SMVIEW_CABLE_SECTION');
        $catGRoute = defined_or_default('C__CATG__SMVIEW_CABLE_ROUTE');
        $masterUrl = isys_helper_link::get_base() . isys_helper_link::create_url([
                C__CMDB__GET__OBJECT => $master['obj_id'],
                C__CMDB__GET__CATG   => $type == 'section' ? $catGSection : $catGRoute,
            ]);
        $objectUrl = isys_helper_link::get_base() . isys_helper_link::create_url([
                C__CMDB__GET__OBJECT => $objectId,
                C__CMDB__GET__CATG   => defined_or_default('C__CATG__SMVIEW_CABLE_PIN'),
                'filter'             => $pinOutScope,
            ]);
        $cable1Url = isys_helper_link::get_base() . isys_helper_link::create_url([
                C__CMDB__GET__OBJECT => $cableId,
                C__CMDB__GET__CATG   => defined_or_default('C__CATG__SMVIEW_CABLE_WIRE'),
            ]);
        $pin1Url = '';
        if ($sib_last_pin['id']) {
            $pin1Url = isys_helper_link::get_base() . isys_helper_link::create_url([
                    C__CMDB__GET__OBJECT   => $objectId,
                    C__CMDB__GET__CATG     => defined_or_default('C__CATG__SMVIEW_CABLE_PIN'),
                    C__CMDB__GET__CATLEVEL => $sib_last_pin['id']
                ]);
        }
        $cable2Url = '';
        if ($last['cable_id']) {
            $cable2Url = isys_helper_link::get_base() . isys_helper_link::create_url([
                    C__CMDB__GET__OBJECT => $last['cable_id'],
                    C__CMDB__GET__CATG   => defined_or_default('C__CATG__SMVIEW_CABLE_WIRE'),
                ]);
        }
        $pin2Url = '';
        if ($sib_pin['id']) {
            $pin2Url = isys_helper_link::get_base() . isys_helper_link::create_url([
                    C__CMDB__GET__OBJECT   => $objectId,
                    C__CMDB__GET__CATG     => defined_or_default('C__CATG__SMVIEW_CABLE_PIN'),
                    C__CMDB__GET__CATLEVEL => $sib_pin['id']
                ]);
        }

        $this->_errors[] = [
            'type'          => $type,
            'master'        => $obj['isys_obj__title'],
            'master_id'     => $master['id'],
            'master_url'    => $masterUrl,
            'element_index' => $index,
            'object'        => $objectTitle ? $objectTitle : '?',
            'object_url'    => $objectUrl,
            'pin1_title'    => $sib_last_pin['description'],
            'pin1_url'      => $pin1Url,
            'wire1_title'   => $last['wire_title'] ? $last['wire_title'] : '?',
            'cable1_title'  => $last['cable_title'] ? $last['cable_title'] : '?',
            'cable1_url'    => $cable2Url,
            'pin2_title'    => $sib_pin['description'],
            'pin2_url'      => $pin2Url,
            'wire2_title'   => $element['wire_title'] ? $element['wire_title'] : '?',
            'cable2_title'  => $element['cable_title'] ? $element['cable_title'] : '?',
            'cable2_url'    => $cable1Url,
            'message'       => $message,
        ];
    }

    /**
     * Validate patch connections.
     *
     * @param  int $limit
     * @param  boolean $removeInvalidPatches
     * @return array
     * @throws isys_exception_database
     */
    public function validatePatches($limit, $removeInvalidPatches = false)
    {
        $pin_dao = new isys_cmdb_dao_category_g_smview_cable_pin($this->_database);
        $sql = "SELECT
pin1_out.isys_catg_smview_cable_pin_list__isys_obj__id as obj_id,
o.isys_obj__title AS obj_title,

pin1_out.isys_catg_smview_cable_pin_list__id AS pin1_out_id,
pin1_out.isys_catg_smview_cable_pin_list__description AS pin1_out_title,
pin1_out.isys_catg_smview_cable_pin_list__sibling_id AS pin1_in_id,
pin1_in.isys_catg_smview_cable_pin_list__description AS pin1_in_title,

pin2_out.isys_catg_smview_cable_pin_list__id AS pin2_out_id,
pin2_out.isys_catg_smview_cable_pin_list__description AS pin2_out_title,
pin2_out.isys_catg_smview_cable_pin_list__sibling_id AS pin2_in_id,
pin2_in.isys_catg_smview_cable_pin_list__description AS pin2_in_title,

wp1.isys_catg_smview_cable_wire_pin__wire_id AS wire1_id,
wp2.isys_catg_smview_cable_wire_pin__wire_id AS wire2_id,
w1.isys_catg_smview_cable_wire_list__description AS wire1_title,
w2.isys_catg_smview_cable_wire_list__description AS wire2_title,

w1.isys_catg_smview_cable_wire_list__isys_obj__id AS cable1_id,
w2.isys_catg_smview_cable_wire_list__isys_obj__id AS cable2_id,
c1.isys_obj__title AS cable1_title,
c2.isys_obj__title AS cable2_title,

se1.isys_catg_smview_cable_section_element__id AS section_element1_id,
re1.isys_catg_smview_cable_route_element__id AS route_element1_id,

se2.isys_catg_smview_cable_section_element__id AS section_element2_id,
re2.isys_catg_smview_cable_route_element__id AS route_element2_id

FROM isys_catg_smview_cable_pin_list pin1_out

JOIN isys_obj o ON o.isys_obj__id = pin1_out.isys_catg_smview_cable_pin_list__isys_obj__id
JOIN isys_catg_smview_cable_pin_list pin1_in ON pin1_in.isys_catg_smview_cable_pin_list__id = pin1_out.isys_catg_smview_cable_pin_list__sibling_id
JOIN isys_catg_smview_cable_pin_list pin2_out ON pin2_out.isys_catg_smview_cable_pin_list__id = pin1_out.isys_catg_smview_cable_pin_list__patch_id
JOIN isys_catg_smview_cable_pin_list pin2_in ON pin2_in.isys_catg_smview_cable_pin_list__id = pin2_out.isys_catg_smview_cable_pin_list__sibling_id

LEFT JOIN isys_catg_smview_cable_wire_pin wp1 ON wp1.isys_catg_smview_cable_wire_pin__connected_pin_id = pin1_out.isys_catg_smview_cable_pin_list__sibling_id
LEFT JOIN isys_catg_smview_cable_wire_pin wp2 ON wp2.isys_catg_smview_cable_wire_pin__connected_pin_id = pin2_out.isys_catg_smview_cable_pin_list__sibling_id

LEFT JOIN isys_catg_smview_cable_wire_list w1 ON w1.isys_catg_smview_cable_wire_list__id = wp1.isys_catg_smview_cable_wire_pin__wire_id
LEFT JOIN isys_catg_smview_cable_wire_list w2 ON w2.isys_catg_smview_cable_wire_list__id = wp2.isys_catg_smview_cable_wire_pin__wire_id

LEFT JOIN isys_obj c1 ON c1.isys_obj__id = w1.isys_catg_smview_cable_wire_list__isys_obj__id
LEFT JOIN isys_obj c2 ON c2.isys_obj__id = w2.isys_catg_smview_cable_wire_list__isys_obj__id

LEFT JOIN isys_catg_smview_cable_section_element se1 ON se1.isys_catg_smview_cable_section_element__wire_id = wp1.isys_catg_smview_cable_wire_pin__wire_id
LEFT JOIN isys_catg_smview_cable_route_element re1 ON re1.isys_catg_smview_cable_route_element__wire_id = wp1.isys_catg_smview_cable_wire_pin__wire_id

LEFT JOIN isys_catg_smview_cable_section_element se2 ON se2.isys_catg_smview_cable_section_element__wire_id = wp2.isys_catg_smview_cable_wire_pin__wire_id
LEFT JOIN isys_catg_smview_cable_route_element re2 ON re2.isys_catg_smview_cable_route_element__wire_id = wp2.isys_catg_smview_cable_wire_pin__wire_id

WHERE pin1_out.isys_catg_smview_cable_pin_list__patch_id IS NOT NULL
GROUP BY pin1_out.isys_catg_smview_cable_pin_list__id;";

        $res = $pin_dao->retrieve($sql);
        $invalidPatches = [];
        while ($row = $res->get_row()) {
            if (empty($row['section_element1_id']) && empty($row['route_element1_id'])) {
                $invalidPatches[] = $row;
            } elseif (empty($row['section_element2_id']) && empty($row['route_element2_id'])) {
                $invalidPatches[] = $row;
            }

            if (count($invalidPatches) >= $limit) {
                break;
            }
        }

        $this->_errors = [];
        foreach ($invalidPatches as $p) {
            $objectUrl = isys_helper_link::get_base() . isys_helper_link::create_url([
                    C__CMDB__GET__OBJECT => $p['obj_id'],
                    C__CMDB__GET__CATG   => $catGSection = defined_or_default('C__CATG__SMVIEW_CABLE_PIN'),
                ]);

            $cable1Url = isys_helper_link::get_base() . isys_helper_link::create_url([
                    C__CMDB__GET__OBJECT => $p['cable1_id'],
                    C__CMDB__GET__CATG   => defined_or_default('C__CATG__SMVIEW_CABLE_WIRE'),
                ]);
            $cable2Url = isys_helper_link::get_base() . isys_helper_link::create_url([
                    C__CMDB__GET__OBJECT => $p['cable2_id'],
                    C__CMDB__GET__CATG   => defined_or_default('C__CATG__SMVIEW_CABLE_WIRE'),
                ]);

            $message = '';
            if (empty($row['section_element1_id']) && empty($row['route_element1_id'])) {
                $message = $p['wire1_title'] . ' is not part of a section or route ';
            }
            if (empty($row['section_element2_id']) && empty($row['route_element2_id'])) {
                $message .= $p['wire2_title'] . ' is not part of a section or route ';
            }

            if ($removeInvalidPatches) {
                $this->_removePatch($p['pin1_out_id']);
                $message = 'removed invalid patch successfully';
            }

            $this->_errors[] = [
                'object'         => $p['obj_title'],
                'object_id'      => $p['obj_id'],
                'object_url'     => $objectUrl,
                'pin1_out_title' => $p['pin1_out_title'],
                'pin1_in_title'  => $p['pin1_in_title'],
                'wire1_title'    => $p['wire1_title'] ? $p['wire1_title'] : '?',
                'cable1_title'   => $p['cable1_title'] ? $p['cable1_title'] : '?',
                'cable1_url'     => $cable1Url,
                'pin2_out_title' => $p['pin2_out_title'],
                'pin2_in_title'  => $p['pin2_in_title'],
                'wire2_title'    => $p['wire2_title'] ? $p['wire2_title'] : '?',
                'cable2_title'   => $p['cable2_title'] ? $p['cable2_title'] : '?',
                'cable2_url'     => $cable2Url,
                'message'        => $message,
            ];
        }
        return $this->_errors;
    }

    /**
     * Fix invalid patches.
     *
     * @param  array $sib_pin
     * @param  array $sib_last_pin
     * @throws isys_exception_dao
     */
    private function _fixPatches($sib_pin, $sib_last_pin)
    {
        if (empty($sib_pin['id']) || empty($sib_last_pin['id'])) {
            return 'can not fix PINs because at least one PIN could not be found';
        }

//        $pin_dao = new isys_cmdb_dao_smview_cable_pin($this->_database);
        if (! empty($sib_pin['patch_pin_id'])) {
            // remove obsolete patch from other object
            $this->_removePatch($sib_pin['id']);
        }

        if (! empty($sib_last_pin['patch_pin_id'])) {
            // remove obsolete patch from other object
            $this->_removePatch($sib_last_pin['id']);
        }

        // patch both connections
        $this->_addPatch($sib_pin['id'], $sib_last_pin['id']);
        $this->_addPatch($sib_last_pin['id'], $sib_pin['id']);

        // set new patch
//        $pin_dao->patch_pins($sib_pin['id'], $sib_last_pin['id']);
        return 'patched PINs successfully';
    }

    private function _addPatch($pin_id, $patch_pin_id)
    {
        $pin_dao = new isys_cmdb_dao_smview_cable_pin($this->_database);
//        $pin = $pin_dao->get_object('id', $pin_id);
        $sql = "UPDATE isys_catg_smview_cable_pin_list SET isys_catg_smview_cable_pin_list__patch_id = " . $patch_pin_id
            . " WHERE isys_catg_smview_cable_pin_list__id = " . $pin_dao->convert_sql_id($pin_id);
        $pin_dao->update($sql) && $pin_dao->apply_update();
    }

    /**
     * Remove patch.
     *
     * @param  int $patchPinId
     * @throws isys_exception_dao
     */
    private function _removePatch($patchPinId)
    {
        $pin_dao = new isys_cmdb_dao_smview_cable_pin($this->_database);
        $sql = "UPDATE isys_catg_smview_cable_pin_list SET isys_catg_smview_cable_pin_list__patch_id = NULL"
            . " WHERE isys_catg_smview_cable_pin_list__id = " . $pin_dao->convert_sql_id($patchPinId);
        $pin_dao->update($sql) && $pin_dao->apply_update();
        return 'removed patch PIN successfully';
    }
}
