<?php

namespace idoit\Module\Cable;

use isys_application;
use isys_cmdb_dao_category_g_global;
use isys_cmdb_dao_category_g_smview_cable_section_element;
use isys_cmdb_dao_smview_cable_connection_view;
use isys_cmdb_dao_smview_cable_route_element;
use isys_cmdb_dao_smview_cable_section;
use isys_exception_database;

/**
 * Section Tracing class.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class SectionTracing
{
    
    /**
     * Object ID.
     *
     * @var int
     */
    private $_objId = null;
    
    /**
     * List of connections.
     *
     * @var array
     */
    private $_sections = [];
    
    /**
     * List of used sections.
     *
     * @var array
     */
    private $_usedSections = [];
    
    /**
     * Constructor.
     *
     * @param int $obj_id
     */
    public function __construct($obj_id)
    {
        $this->_objId = $obj_id;
        
        $dao_view = new isys_cmdb_dao_smview_cable_section(isys_application::instance()->database);
        $conn_in = $dao_view->filter_objects(array('start_obj_id' => $obj_id));
        $conn_out = $dao_view->filter_objects(array('end_obj_id' => $obj_id));
        $this->_sections = array_merge($conn_in, $conn_out);
        
        $this->_usedSections = $this->_get_used_sections();
    }
    
    /**
     * Get used sections.
     *
     * @return array
     */
    private function _get_used_sections()
    {

        /*
         * 2020-09-17: Anpassung des Zustandes "verwendet" an einer Strecke
         *
         * Eine Strecke gilt als "verwendet", wenn diese als Teil eines Netzweges verwendet wird.
         * Das bisherige Kriterium ob der Start- oder Endpunkt der Strecke gepatcht ist gilt nicht mehr.
         */

        $element_dao = isys_cmdb_dao_smview_cable_route_element::instance(isys_application::instance()->database);
        $usedSections = [];
        foreach ($this->_sections as $s) {
            $routeElements = $element_dao->filter_objects(array('section_id' => $s['id']));
            if (! empty($routeElements)) {
                $usedSections[$s['id']] = $s;
            }
        }
        return $usedSections;
    }

    /**
     * Get unpatched sections.
     *
     * @param  int|null $ignored_section_id         Will be ignored in section listing
     * @param  int|null $ignored_wire_id            Will be ignored in connection listing
     * @param  boolean $handleAssociatedRoute       Handles creation of an associated route
     * @param  boolean $considerPatchedConnections  Allow usage of patched connections
     * @return array
     * @throws isys_exception_database
     */
    public function get_unpatched_sections($ignored_section_id = null, $ignored_wire_id = null, $handleAssociatedRoute = false, $considerPatchedConnections = false)
    {
        $obj_dao = isys_cmdb_dao_category_g_global::instance(isys_application::instance()->database);
        $obj = $obj_dao->get_object_by_id($this->_objId)->get_row();

        // get next section by ID
        $associated_element = null;
        if ($handleAssociatedRoute && ! empty($ignored_section_id)) {
            $associated_element = $this->_get_next_section($ignored_section_id);
        }
        
        // get all unpatched sections of the object
        $objects = $this->_get_unused_sections($ignored_section_id, $ignored_wire_id, $considerPatchedConnections);
        return array(
            'element' => array(
                'id'    => $this->_objId,
                'title' => $obj['isys_obj__title'],
            ),
            'associated_element' => $associated_element,
            'objects' => array_values($objects),
        );
    }

    /**
     * Try to get next section by matching next start- and end-wire.
     *
     * @param  int $section_id
     * @return array|null
     * @throws isys_exception_database
     */
    private function _get_next_section($section_id)
    {
        $section_dao = isys_cmdb_dao_smview_cable_section::instance(isys_application::instance()->database);
        $element_dao = isys_cmdb_dao_category_g_smview_cable_section_element::instance(isys_application::instance()->database);
        $s = $section_dao->get_object('id', $section_id);

        // get first and last element
        $elements = $element_dao->get_elements_by_section_id($section_id, null, false, false);
        $firstElement = array_shift($elements);
        $lastElement = count($elements) > 0 ? array_pop($elements) : $firstElement;

        // try to get next wire numbers
        $firstWireNr = ((int) $firstElement['wire_title']) + 1;
        $nextFirstWireTitle = str_pad($firstWireNr, strlen($firstElement['wire_title']), '0', STR_PAD_LEFT);
        $lastWireNr = ((int) $lastElement['wire_title']) + 1;
        $nextLastWireTitle = str_pad($lastWireNr, strlen($lastElement['wire_title']), '0', STR_PAD_LEFT);

        $sections = $section_dao->filter_objects(['start_obj_id' => $s['start_obj_id'], 'end_obj_id' => $s['end_obj_id']]);
        foreach ($sections as $s) {
            // get first and last element
            $elements = $element_dao->get_elements_by_section_id($s['id'], null, false, false);
            $firstElement = array_shift($elements);
            $lastElement = count($elements) > 0 ? array_pop($elements) : $firstElement;

            if ($nextFirstWireTitle == $firstElement['wire_title'] && $nextLastWireTitle == $lastElement['wire_title']) {
                return $s;
            }
        }
        return null;
    }
    
    /**
     * Get unused sections.
     *
     * @param  int|null $ignored_section_id
     * @param  int|null $ignored_wire_id
     * @param  boolean $considerPatchedConnections
     * @return array
     */
    private function _get_unused_sections($ignored_section_id, $ignored_wire_id = null, $considerPatchedConnections = false)
    {
        $sections = [];
        
        // find cable type by ignored/selected wire id
        $cable_technology = $this->_get_technology_id($ignored_section_id, $ignored_wire_id);
        
        foreach ($this->_sections as $s) {
            $section_id = $s['id'];
            if ($section_id == $ignored_section_id) {
                // ignore this section
                continue;
            }
            
            if (! $considerPatchedConnections && isset($this->_usedSections[$section_id])) {
                // this section is already in use
                continue;
            }
            
            if (isset($sections[$section_id])) {
                // this section is already in list
                continue;
            }
            
            // get first and last element
            $section = self::get_section_endpoints($section_id, $s, $this->_objId);
            
            $tc = $section['technology_id'];
            if (isset($cable_technology) && $tc != $cable_technology) {
                continue;
            }
            
            $sections[$section_id] = $section;
        }
        
        uasort($sections, array($this, '_sortByTitle'));
        return $sections;
    }

    /**
     * Get cable technology by section.
     *
     * @param  int|null $selected_section_id
     * @param  int|null $selected_wire_id
     * @return int|null
     */
    private function _get_technology_id($selected_section_id, $selected_wire_id = null)
    {
        if (empty($selected_section_id) && empty($selected_wire_id)) {
            return null;
        }

        if (isset($selected_wire_id)) {
            // get wire
            $dao = isys_cmdb_dao_smview_cable_connection_view::instance(isys_application::instance()->database);
            $res = $dao->filter_objects(array('wire_id' => $selected_wire_id));

            if (empty($res)) {
                return null;
            }
            $wire = $res[0];
            return $wire['technology_id'];
        }
        
        foreach ($this->_sections as $s) {
            $section_id = $s['id'];
            if (isset($selected_section_id) && $section_id == $selected_section_id) {
                $section = self::get_section_endpoints($section_id, $s, $this->_objId);
                return $section['technology_id'];
            }
        }
    }

    /**
     * Get section endpoints.
     *
     * @param  int   $section_id
     * @param  array $section
     * @param  int   $obj_id
     * @return array
     */
    public static function get_section_endpoints($section_id, $section = null, $obj_id = null)
    {
        $obj_dao = isys_cmdb_dao_category_g_global::instance(isys_application::instance()->database);
        $element_dao = isys_cmdb_dao_category_g_smview_cable_section_element::instance(isys_application::instance()->database);
        
        // get first and last element
        $elements = $element_dao->get_elements_by_section_id($section_id, null, false, false);
        $firstElement = array_shift($elements);
        $lastElement = count($elements) > 0 ? array_pop($elements) : $firstElement;
        
        // object is first element
        $element_out = $firstElement;
        $element_in = $lastElement;
        $out = 'out';
        $in = 'in';
        if ($obj_id && $obj_id == $lastElement['obj_in_id']) {
            // object is last element
            $element_out = $lastElement;
            $element_in = $firstElement;
            $out = 'in';
            $in = 'out';
        }

        // get object ID of corresponding object
        if (! $section) {
            $section_dao = isys_cmdb_dao_smview_cable_section::instance(isys_application::instance()->database);
            $section = $section_dao->get_object('id', $section_id);
        }
        $obj = $obj_dao->get_object_by_id($section['obj_id'])->get_row();
        return array(
            'section_id'          => $section_id,
            'section_title'       => $obj['isys_obj__title'],
            'length'              => $section['length'],
            'technology_id'       => $element_in['technology_id'],
            'cable_free_wires'    => 0,
            'cable_patched_wires' => 1,
            'cable_max_capacity'  => 100,
            'obj_out_id'          => $element_out["obj_{$out}_id"],
            'obj_out_title'       => $element_out["obj_{$out}_title"],
            'pin_out_id'          => $element_out["pin_{$out}_id"],
            'pin_out_title'       => $element_out["pin_{$out}_title"],
            'pin_out_scope'       => $element_out["pin_{$out}_scope"],
            'wire_pin_out_title'  => $element_out["wire_pin_{$out}_title"],
            'wire_pin_out_state'  => $element_out["wire_pin_{$out}_state"],
            'obj_in_id'           => $element_in["obj_{$in}_id"],
            'obj_in_title'        => $element_in["obj_{$in}_title"],
            'pin_in_id'           => $element_in["pin_{$in}_id"],
            'pin_in_title'        => $element_in["pin_{$in}_title"],
            'pin_in_scope'        => $element_in["pin_{$in}_scope"],
            'wire_pin_in_title'   => $element_in["wire_pin_{$in}_title"],
            'wire_pin_in_state'   => $element_in["wire_pin_{$in}_state"],
        );
    }
    
    /**
     * Sort array by title of an object.
     *
     * @param  array $a
     * @param  array $b
     * @return int
     */
    private function _sortByTitle($a, $b)
    {
        $left = $a['title'];
        $right = $b['title'];
        return strnatcmp($left, $right);
    }
}
