<?php

namespace idoit\Module\Cable;

use idoit\Exception\Exception;
use idoit\Module\Cable\Algorithms\Dijkstra;
use idoit\Module\Cable\Algorithms\BreadthFirstSearch;
use isys_application;
use isys_cmdb_dao_smview_cable_free_connection_view;
use isys_cmdb_dao_smview_cable_free_section_view;
use isys_exception_database;
use PhpOrient\Exceptions\SocketException;

/**
 * Tracing class.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class Tracing
{

    /**
     * List of connected objects.
     *
     * @var array
     */
    private $_connected_objects = [];

    /**
     * List of connections.
     *
     * @var array
     */
    private $_connections = [];

    /**
     * List of free connections incoming.
     *
     * @var array
     */
    private $_free_connections_in = [];

    /**
     * List of free connections outgoing.
     *
     * @var array
     */
    private $_free_connections_out = [];

    /**
     * List of free sections incoming.
     *
     * @var array
     */
    private $_free_sections_in = [];

    /**
     * List of free sections outgoing.
     *
     * @var array
     */
    private $_free_sections_out = [];

    /**
     * Base object ID.
     *
     * @var int
     */
    private $_base_obj_id;

    /**
     * Include section data.
     *
     * @var boolean
     */
    private $_include_sections = true;

    /**
     * List of handled objects.
     *
     * @var array
     */
    private $_handled_objects = [];

    /**
     * OrientDB Tracing client if cache is available.
     *
     * @var OrientDB
     */
    private $_orientdb;

    /**
     * Tracing constructor.
     *
     * @param int     $objId
     * @param boolean $include_sections
     * @param boolean $use_cache
     */
    public function __construct($objId, $include_sections = true, $use_cache = true)
    {
        $this->_base_obj_id = $objId;
        $this->_include_sections = $include_sections;

        // check OrientDB availability
        if ($use_cache) {
            try {
                $this->_orientdb = new OrientDB();
                $ret = $this->_orientdb->get_last_update_info();
                if ($ret === []) {
                    // no valid configuration available
                    $this->_orientdb = null;
                } else {
                    return;
                }
            } catch (SocketException $e) {
                // no cache available, go ahead
                $this->_orientdb = null;
            } catch (Exception $e) {
                // no cache available, go ahead
                $this->_orientdb = null;
            }
        }

        $this->_connected_objects = [];
        $this->_connections = [];
        $this->_handled_objects = array(
            $objId => true,
        );

        $this->_initialize_free_connections();
        $this->_initialize_free_sections();

        // todo: validate these parameters
        ini_set('xdebug.max_nesting_level', 1200);
        ini_set('memory_limit', '1024M');

        $this->_add_object_connections($objId);
    }

    /**
     * Initialize free connections.
     */
    private function _initialize_free_connections()
    {
        $dao_view = new isys_cmdb_dao_smview_cable_free_connection_view(isys_application::instance()->database);
        $connections = $dao_view->get_objects();

        $connections_in = [];
        $connections_out = [];
        foreach ($connections as $c) {
            $obj_in = $c['obj_id_in'];
            $obj_out = $c['obj_id_out'];
            if (! isset($connections_in[$obj_in])) {
                $connections_in[$obj_in] = [];
            }
            $connections_in[$obj_in][] = $c;

            if (! isset($connections_out[$obj_out])) {
                $connections_out[$obj_out] = [];
            }
            $connections_out[$obj_out][] = $c;
        }

        $this->_free_connections_in = $connections_in;
        $this->_free_connections_out = $connections_out;
    }

    /**
     * Initialize free sections.
     */
    private function _initialize_free_sections()
    {
        $section_view = new isys_cmdb_dao_smview_cable_free_section_view(isys_application::instance()->database);
        $sections = $section_view->get_objects();

        $sections_in = [];
        $sections_out = [];
        foreach ($sections as $s) {
            $obj_in = $s['start_obj_id'];
            $obj_out = $s['end_obj_id'];
            if (! isset($sections_in[$obj_in])) {
                $sections_in[$obj_in] = [];
            }
            $sections_in[$obj_in][] = $s;

            if (! isset($sections_out[$obj_out])) {
                $sections_out[$obj_out] = [];
            }
            $sections_out[$obj_out][] = $s;
        }

        $this->_free_sections_in = $sections_in;
        $this->_free_sections_out = $sections_out;
    }

    /**
     * Get shortest path between two objects.
     *
     * @param  int $target_obj_id
     * @return array
     */
    public function get_shortest_path($target_obj_id)
    {
        if ($this->_orientdb) {
            try {
                return $this->_orientdb->get_shortest_path($this->_base_obj_id, $target_obj_id, $this->_include_sections);
            } catch (isys_exception_database $e) {
                return [];
            }
        }

        // get shortest cable between two objects
        $weighted_connections = [];
        $shortest_lengths = [];
        $infiniteLength = 999999999;
        foreach ($this->_connected_objects as $source => $targets) {
            foreach (array_keys($targets) as $id) {
                $wires = $this->_get_connections($source, $id);
                foreach ($wires as $wire) {
                    $length = (int) $wire['length'];
                    if (empty($length)) {
                        $length = $infiniteLength; //INF; // do not use INF constant to avoid calculation errors
                    }

                    if (isset($shortest_lengths[$source][$id])) {
                        $shortest_length = $shortest_lengths[$source][$id];
                        $length = $length < $shortest_length ? $length : $shortest_length;
                    }
                    $shortest_lengths[$source][$id] = $length;
                    $weighted_connections[] = array($source, $id, $length);
                }
            }
        }

        // calculate shortest path
        $algorithm = new Dijkstra($weighted_connections);
        $result = $algorithm->calculate($this->_base_obj_id, $target_obj_id);

        // get all shortest wires in this path
        $route = [];
        $prev = null;
        foreach ($result as $obj_id) {
            if (! $prev) {
                $prev = $obj_id;
                continue;
            }

            // get all wires matching shortest length
            $wires = $this->_get_connections($prev, $obj_id);
            $shortest_length = $shortest_lengths[$prev][$obj_id];
            $shortest_wires = [];
            foreach ($wires as $id => $wire) {
                $length = (int) $wire['length'];
                if ($length == $shortest_length || ($length == 0 && $shortest_length == $infiniteLength)) {
                    $shortest_wires[$id] = $wire;
                }
            }
            $route[] = array(
                'from'  => $prev,
                'to'    => $obj_id,
                'wires' => $shortest_wires,
            );
            $prev = $obj_id;
        }

        return $route;
    }

    /**
     * Get least number of hops between two objects.
     *
     * @param  int $target_obj_id
     * @return array
     */
    public function get_least_number_of_hops($target_obj_id)
    {
        if ($this->_orientdb) {
            try {
                return $this->_orientdb->get_least_number_of_hops($this->_base_obj_id, $target_obj_id, $this->_include_sections);
            } catch (isys_exception_database $e) {
                return [];
            }
        }

        $connections = [];
        foreach ($this->_connected_objects as $k => $v) {
            $connections[$k] = array_keys($v);
        }
        $algo = new BreadthFirstSearch($connections);
        $result = $algo->calculate($this->_base_obj_id, $target_obj_id);

        $route = [];
        $prev = null;
        foreach ($result as $obj_id) {
            if ($prev) {
                $route[] = array(
                    'from'  => $prev,
                    'to'    => $obj_id,
                    'wires' => $this->_get_connections($prev, $obj_id),
                );
            }
            $prev = $obj_id;
        }
        return $route;
    }

    /**
     * Get connections between two objects.
     *
     * @param  int $prev
     * @param  int $obj_id
     * @return array
     */
    private function _get_connections($prev, $obj_id)
    {
        $id1 = "$prev-$obj_id";
        $id2 = "$obj_id-$prev";
        $connections = ! empty($this->_connections[$id1]) ? $this->_connections[$id1] : $this->_connections[$id2];
        $wires = [];
        foreach ($connections as $id => $c) {
            if (isset($c['start_obj_id'])) {
                // this is a section
                $wires[$id] = array(
                    'section_id'    => $c['id'],
                    'section_title' => $c['title'],
                    'length'        => $c['length'],
                );
            } else {
                $wires[$id] = array(
                    'cable_id'    => $c['cable_id'],
                    'cable_title' => $c['cable_title'],
                    'length'      => $c['length'],
                    'wire_id'     => $c['wire_id'],
                    'wire_title'  => $c['wire_title'],
                    'wire_scope'  => $c['wire_scope'],
                );
            }
        }
        return $wires;
    }

    /**
     * Add connections from object.
     *
     * @param int $objId
     */
    private function _add_object_connections($objId)
    {
        $this->_handled_objects[$objId] = true;

        // get connections
        $connections_in = isset($this->_free_connections_in[$objId]) ? $this->_free_connections_in[$objId] : [];
        $connections_out = isset($this->_free_connections_out[$objId]) ? $this->_free_connections_out[$objId] : [];

        $this->_add_to_connection_list($objId, $connections_in, 'obj_id_out');
        $this->_add_to_connection_list($objId, $connections_out, 'obj_id_in');

        if (! $this->_include_sections) {
            return;
        }

        // get sections
        $sections_in = isset($this->_free_sections_in[$objId]) ? $this->_free_sections_in[$objId] : [];
        $sections_out = isset($this->_free_sections_out[$objId]) ? $this->_free_sections_out[$objId] : [];

        $this->_add_to_connection_list($objId, $sections_in, 'end_obj_id');
        $this->_add_to_connection_list($objId, $sections_out, 'start_obj_id');
    }

    /**
     * Add connections and sections to connection list.
     *
     * @param int $objId
     * @param array $connection_list
     * @param string $identifier
     */
    private function _add_to_connection_list($objId, $connection_list, $identifier)
    {
        foreach ($connection_list as $c) {
            $objId2 = $c[$identifier];
            $lowerId = $objId;
            $higherId = $objId2;
            if ($objId2 < $objId) {
                $lowerId = $objId2;
                $higherId = $objId;
            }
            $ident = "$lowerId-$higherId";
            $id = isset($c['start_obj_id']) ? 's_' . $c['id'] : 'w_' . $c['wire_id'];

            $this->_connections[$ident][$id] = $c;
            $this->_connected_objects[$lowerId][$higherId] = true;
            // save also reverse connection
            $this->_connected_objects[$higherId][$lowerId] = true;

            if (! isset($this->_handled_objects[$objId2])) {
                $this->_add_object_connections($objId2);
            }
        }
    }
}
