<?php

namespace idoit\Module\Cable;

use isys_application;
use isys_auth;
use isys_auth_cmdb;
use isys_cmdb_dao_category_g_smview_cable_route;
use isys_cmdb_dao_category_g_smview_cable_virtual_route;
use isys_cmdb_dao_category_g_smview_cable_virtual_route_element;
use isys_cmdb_dao_smview_cable_pin;
use isys_cmdb_dao_smview_cable_route;
use isys_cmdb_dao_smview_cable_route_element;
use isys_cmdb_dao_smview_cable_virtual_route;
use isys_cmdb_dao_smview_cable_virtual_route_element;
use isys_component_database;
use isys_exception_cmdb;
use isys_exception_dao;
use isys_exception_database;

/**
 * Virtual Route class.
 *
 * @package    i-doit
 * @subpackage smview_cable
 * @author     Steve Groeber <steve.groeber@shd-online.de>
 * @copyright  2022 SHD System-Haus-Dresden GmbH
 * @license    http://www.i-doit.com/license
 */
class VirtualRoute
{

    /**
     * Database.
     *
     * @var isys_component_database
     */
    private $_db;

    /**
     * Constructor.
     */
    public function __construct()
    {
        $this->_db = isys_application::instance()->database;
    }

    /**
     * Check if a virtual route is already valid to be created as a normal route.
     *
     * @param  int $obj_id
     * @return array
     * @throws isys_exception_database
     */
    public function check($obj_id)
    {
        $lang = isys_application::instance()->container->get('language');
        $canActivateVR = isys_auth_cmdb::instance()->is_allowed_to(isys_auth::EXECUTE, "CATEGORY/C__CATG__SMVIEW_CABLE_VIRTUAL_ROUTE");
        if (! $canActivateVR) {
            return [
                'is_valid'    => false,
                'message'     => $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__ACTIVATE_VIRTUAL_ROUTE_NO_PERMISSION'),
                'wired_wires' => [],
            ];
        }

        $virtual_route_element_dao = new isys_cmdb_dao_category_g_smview_cable_virtual_route_element($this->_db);
        $elements = $virtual_route_element_dao->get_elements_by_obj_id($obj_id);

        $is_valid = true;
        $wired_wires = [];
        $wire_dao = new \isys_cmdb_dao_category_g_smview_cable_wire(isys_application::instance()->database);
        foreach ($elements as $e) {
                $in_use = $wire_dao->get_in_use($e['wire_id']);
                if (! empty($in_use) && $in_use == $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__WIRE_IS_WIRED')) {
                    $is_valid = false;
                    $wired_wires[] = $e['cable_title'] .'/' . $e['wire_scope'] . '/' . $e['wire_title'];
                }
        }

        $ok = $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__CHECK_VIRTUAL_ROUTE_OK');
        $notOk = $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__CHECK_VIRTUAL_ROUTE_NOT_OK');

        return [
            'is_valid'    => $is_valid,
            'message'     => $is_valid ? $ok : $notOk,
            'wired_wires' => $wired_wires,
        ];
    }

    /**
     * Get a message, that the patch is already in use.
     *
     * @param  int $pinId
     * @param  array $pin_in
     * @return string
     * @throws isys_exception_database
     */
    private function _get_patch_in_use_message($pinId, $pin_in)
    {
        $virtual_route_element_dao = new isys_cmdb_dao_category_g_smview_cable_virtual_route_element($this->_db);
        $objIn = $virtual_route_element_dao->get_object($pin_in['obj_id'])->get_row();
        $message = "{$objIn['isys_obj__title']} - {$pin_in['description']} (ID: {$pin_in['id']})";

        // try to get route by pin and wire
        $wire_pin_dao = new \isys_cmdb_dao_smview_cable_wire_pin($this->_db);
        $wirePins = $wire_pin_dao->filter_objects(['connected_pin_id' => $pinId]);
        $wirePin = $wirePins[0];

        $route_element_dao = new \isys_cmdb_dao_smview_cable_route_element($this->_db);
        $route_elements = $route_element_dao->filter_objects(['wire_id' => $wirePin['wire_id']]);
        if (! empty($route_elements)) {
            $route_dao = new isys_cmdb_dao_category_g_smview_cable_route($this->_db);
            $route_element = $route_elements[0];
            $route_row = $route_dao->get_data($route_element['route_id'])->get_row();

            $lang = isys_application::instance()->container->get('language');
            $message .= ' ' . $lang->get('LC__CMDB__CATG__SMVIEW_CABLE__USED_IN') . ' ' . $route_row['isys_obj__title'];
        }
        return $message;
    }

    /**
     * Create a route from an existing virtual route.
     *
     * @param  int $obj_id
     * @param  boolean $deleteVirtualRoute
     * @return int|null
     * @throws isys_exception_cmdb
     * @throws isys_exception_dao
     * @throws isys_exception_database
     */
    public function create_route_from_vr($obj_id, $deleteVirtualRoute = false)
    {
        $res = $this->check($obj_id);
        if (! $res['is_valid']) {
            throw new isys_exception_cmdb($res['message']);
        }

        $route_dao = new isys_cmdb_dao_category_g_smview_cable_route($this->_db);
        $virtual_route_dao = new isys_cmdb_dao_smview_cable_virtual_route($this->_db);
        $virtual_route_element_dao = new isys_cmdb_dao_smview_cable_virtual_route_element($this->_db);

        // create new object
        $virtual_route_object = $route_dao->get_object($obj_id)->get_row();
        $objType = defined_or_default('C__SHD__ROUTE');
        $route_obj_id = $route_dao->create_object($virtual_route_object['isys_obj__title'], $objType);

        // create category
        $virtual_route = $virtual_route_dao->get_object('obj_id', $obj_id);
        $virtual_route['isys_obj__id'] = $route_obj_id;
        $virtual_route['description'] = $virtual_route_object['isys_obj__title'];

        // prepare route elements
        $elements = $virtual_route_element_dao->filter_objects(['route_id' => $virtual_route['id']]);
        $routeElements = [];
        foreach ($elements as $e) {
            $routeElements[] = [
                'wire_id'    => $e['wire_id'],
                'section_id' => $e['section_id'],
                'index'      => $e['index'],
                'obj_out_id' => $virtual_route['start_obj_id'],
                'obj_in_id'  => $virtual_route['end_obj_id'],
            ];
        }

        // save category
        $routeElements = json_encode($routeElements);
        $virtual_route['route_elements'] = $routeElements;
        $route_dao->create_data($virtual_route);

        if ($deleteVirtualRoute) {
            $virtual_route_dao->delete_object($virtual_route['id']);
            $route_dao->delete_object($obj_id);
        }
        return $route_obj_id;
    }
}
